#ifndef _AZPROF_DEBUG_HPP
#define _AZPROF_DEBUG_HPP

#include <stdarg.h>
#include <stdio.h>
#include <stdint.h>
#include <sys/types.h>

namespace azprof {

////////////////////////////////////////////////////////////////////////////////
// Macros
////////////////////////////////////////////////////////////////////////////////

#ifndef NDEBUG
#define AZPROF_DEBUG_ONLY(x) x
#else
#define AZPROF_DEBUG_ONLY(x)
#endif

#ifdef AZTEK
#define AZPROF_DEBUG_FUNCTION __PRETTY_FUNCTION__
#else
#define AZPROF_DEBUG_FUNCTION "<function name not available>"
#endif

#ifndef NDEBUG
#define azprof_assert(e) \
    if (!(e)) { \
        azprof::assert_failed(#e, AZPROF_DEBUG_FUNCTION, __FILE__, __LINE__); \
    }
#else
#define azprof_assert(e)
#endif

#define azprof_guarantee(e) \
    if (!(e)) { \
        azprof::guarantee_failed(#e, AZPROF_DEBUG_FUNCTION, __FILE__, __LINE__); \
    }

#define azprof_should_not_reach_here() \
    azprof::hit_should_not_reach_here(AZPROF_DEBUG_FUNCTION, __FILE__, __LINE__)

#define azprof_unimplemented() \
    azprof::hit_unimplemented(AZPROF_DEBUG_FUNCTION, __FILE__, __LINE__)

extern void assert_failed(
    const char *expr, const char *func, const char *file, int line
);

extern void guarantee_failed(
    const char *expr, const char *func, const char *file, int line
);

extern void hit_should_not_reach_here(
    const char *func, const char *file, int line
);

extern void hit_unimplemented(const char *func, const char *file, int line);

////////////////////////////////////////////////////////////////////////////////
// Debug flags
////////////////////////////////////////////////////////////////////////////////

class DebugFlag {
public:
    enum {
        HTTP_STREAM_LOG_NON_SENSITIVE_IO = 0x1,
        FILE_SERVLET_DONT_CACHE = 0x2
    };

    static uint64_t get();
    static void set(uint64_t);
    static bool is_set(int flag) {return (get() & flag) != 0;}

private:
    DebugFlag();
};

class DebugFlagMark {
public:
    DebugFlagMark(uint64_t flags) {
        _old = DebugFlag::get();
        DebugFlag::set(flags);
    }
    ~DebugFlagMark() {
        DebugFlag::set(_old);
    }
private:
    uint64_t _old;
};

////////////////////////////////////////////////////////////////////////////////
// Logging
////////////////////////////////////////////////////////////////////////////////

class Log {
public:
    enum LogLevel {
	AZPROF_LOG_NONE = 0,
	AZPROF_LOG_FATAL,
	AZPROF_LOG_ERROR,
	AZPROF_LOG_WARN,
	AZPROF_LOG_INFO,
	AZPROF_LOG_DEBUG
    };

    static void init();
    static Log* get();
    static void set(Log*);

    Log(const char *name, const char *path);
    ~Log();

    static void set_log_level(LogLevel level) { _log_level = level; }
    static void set_log_level(const char* level);
    static LogLevel get_log_level() { return _log_level; }
    static void vmsg(const char*, const char*, va_list);
    static void verr(const char*, va_list);
    static void err(const char*, ...);
    static void vwarn(const char*, va_list);
    static void warn(const char*, ...);
    static void vinfo(const char*, va_list);
    static void info(const char*, ...);
    static void debug(const char*, const char*, ...);

    static ssize_t vprint(const char*, va_list);
    static ssize_t print(const char*, ...);
    static ssize_t write(const char*, size_t);
    static int put(const char*);
    static int put(char);
    static int flush();

private:
    static Log *_stderr;
    static LogLevel _log_level;

    static FILE* file() {return get()->_file;}
    static const char* name() {return get()->_name;}

    char _name[64];
    FILE *_file;
};


/**
 * This poorly named class aids in debug logging by printing when the target
 * function has been entered and again when it exits. Typical use is thus:
 * void foo::bar() {
 *   FunctionLog fl("foo::bar");
 *   // do stuff
 * }
 * It only works when verbose debugging has been set with the ARTADebugFlags option 
 * is activated with the third bit (0x4) set (e.g. -PX:ARTADebugFlags=0x5).
 */
class FunctionLog {
public:
    FunctionLog(const char* fname) : _fname(fname) {
	// Log::debug(_fname, "       --->>");
    }

    ~FunctionLog() {
	// Log::debug(_fname, "<<---");
    }
    
    const char* _fname;
};

class LogMark {
public:
    LogMark(Log *log) {
        _old = Log::get();
        Log::set(log);
    }
    ~LogMark() {
        Log::set(_old);
    }
private:
    Log *_old;
};

} // namespace azprof

#endif // _AZPROF_DEBUG_HPP
