#ifndef _AZPROF_EVENT_H
#define _AZPROF_EVENT_H

#include <azprof/azprof_constants.h>
#include <proxy/proxy_network.h>
#include <stddef.h>
#include <sys/types.h>

#ifdef __cplusplus
#include <azprof/azprof_util.hpp>
#include <azprof/azprof_io.hpp>
#define AZPROF_EXTERN_C extern "C"
#else
#include <stdbool.h>
#define AZPROF_EXTERN_C extern
#endif

struct azprof_event_class_s;
struct azprof_operation_class_s;

////////////////////////////////////////////////////////////////////////////////
// Attributes
////////////////////////////////////////////////////////////////////////////////

typedef enum {
    AZPROF_EVENT_ATTR_BOOL_T,
    AZPROF_EVENT_ATTR_INT8_T,
    AZPROF_EVENT_ATTR_UINT8_T,
    AZPROF_EVENT_ATTR_INT16_T,
    AZPROF_EVENT_ATTR_UINT16_T,
    AZPROF_EVENT_ATTR_INT32_T,
    AZPROF_EVENT_ATTR_UINT32_T,
    AZPROF_EVENT_ATTR_INT64_T,
    AZPROF_EVENT_ATTR_UINT64_T,
    AZPROF_EVENT_ATTR_INTPTR_T,
    AZPROF_EVENT_ATTR_FLOAT_T,
    AZPROF_EVENT_ATTR_DOUBLE_T,
    AZPROF_EVENT_ATTR_STRING_T,
    AZPROF_EVENT_ATTR_IN_ADDR_T,
    AZPROF_EVENT_ATTR_SOCKADDR_T
} azprof_event_attr_type_t;

typedef enum {
    AZPROF_AF_UNSPEC = 0,
    AZPROF_AF_INET = 4,
    AZPROF_AF_INET6 = 16
} azprof_af_t;

typedef struct {
    union {
        uint8_t ipv4[4];
        uint16_t ipv6[8];
        char data[16];
    } in_addr;
    uint16_t family;
} azprof_in_addr_t;

AZPROF_EXTERN_C azprof_in_addr_t azprof_swizzle_in_addr(const proxy_in_addr_t*);

typedef struct {
    azprof_in_addr_t sin_addr;
    uint16_t sin_port;
} azprof_sockaddr_t;

AZPROF_EXTERN_C azprof_sockaddr_t azprof_swizzle_sockaddr(
    const proxy_sockaddr_t*
);

// The description of an attribute of a profiler event used for reflection. It's
// declared as a 'struct', not a 'class', so that we can initialize a 'static'
// instance without running code and relying on the order in which static
// initializers are run.
typedef struct azprof_event_attr_s {
#ifdef __cplusplus
    size_t id() const {return ea_id;}
    size_t offset() const {return ea_offset;}
    azprof_event_attr_type_t type() const {return ea_type;}
    const char* name() const {return ea_name;}

    bool has_fixed_machine_size() const;
    bool has_fixed_network_size() const;
    size_t fixed_machine_size() const;
    size_t network_size(const void *ptr) const;

    bool is_bool() const {return type() == AZPROF_EVENT_ATTR_BOOL_T;}
    bool is_int8() const {return type() == AZPROF_EVENT_ATTR_INT8_T;}
    bool is_uint8() const {return type() == AZPROF_EVENT_ATTR_UINT8_T;}
    bool is_int16() const {return type() == AZPROF_EVENT_ATTR_INT16_T;}
    bool is_uint16() const {return type() == AZPROF_EVENT_ATTR_UINT16_T;}
    bool is_int32() const {return type() == AZPROF_EVENT_ATTR_INT32_T;}
    bool is_uint32() const {return type() == AZPROF_EVENT_ATTR_UINT32_T;}
    bool is_int64() const {return type() == AZPROF_EVENT_ATTR_INT64_T;}
    bool is_uint64() const {return type() == AZPROF_EVENT_ATTR_UINT64_T;}
    bool is_intptr() const {return type() == AZPROF_EVENT_ATTR_INTPTR_T;}
    bool is_float() const {return type() == AZPROF_EVENT_ATTR_FLOAT_T;}
    bool is_double() const {return type() == AZPROF_EVENT_ATTR_DOUBLE_T;}
    bool is_string() const {return type() == AZPROF_EVENT_ATTR_STRING_T;}
    bool is_in_addr() const {return type() == AZPROF_EVENT_ATTR_IN_ADDR_T;}
    bool is_sockaddr() const {return type() == AZPROF_EVENT_ATTR_SOCKADDR_T;}

    ssize_t skip(azprof::StaticRingBuffer&) const;
    ssize_t read(
        azprof::StaticRingBuffer& buf, char *strbuf, size_t strbuflen, void *ptr
    ) const;
    ssize_t write(azprof::StaticRingBuffer& buf, const void *ptr) const;
    void print(azprof::PrintStream *stream, const void *ptr) const;
#endif // __cplusplus

    // These fields should be treated as private as matter of practice.
    size_t ea_id;
    size_t ea_offset;
    azprof_event_attr_type_t ea_type;
    const char *ea_name;
} azprof_event_attr_t;

////////////////////////////////////////////////////////////////////////////////
// Event classes
////////////////////////////////////////////////////////////////////////////////

#define AZPROF_EVENT_CLASS_ABSTRACT_FLAG            0x1
#define AZPROF_EVENT_CLASS_BEGIN_OPERATION_FLAG     0x2
#define AZPROF_EVENT_CLASS_END_OPERATION_FLAG       0x4
#define AZPROF_EVENT_CLASS_FIXED_NETWORK_SIZE_FLAG  0x8
#define AZPROF_EVENT_CLASS_FIXED_MACHINE_SIZE_FLAG  0x10
#define AZPROF_EVENT_CLASS_PACKED_FLAG              0x20
#define AZPROF_EVENT_CLASS_CAN_TRACE_TO_STDOUT_FLAG 0x40
#define AZPROF_EVENT_CLASS_CAN_TRACE_TO_STDERR_FLAG 0x80
#define AZPROF_EVENT_CLASS_CAN_TRACE_TO_FILE_FLAG   0x100
#define AZPROF_EVENT_CLASS_CAN_TRACE_TO_MEMORY_FLAG 0x200

AZPROF_EXTERN_C struct azprof_event_class_s *AZPROF_EVENT_CLASSES_BY_ID[];

// A description of a profiler event used for reflection. It's declared as a
// 'struct', not a 'class', so that we can initialize a 'static' instance
// without running code and relying on the order in which static initializers
// are run.
typedef struct azprof_event_class_s {
#ifdef __cplusplus
    enum ParseMode {
        PARSE_EVENTS_MODE,
        PARSE_OPERATIONS_MODE
    };

    static void init();

    static void parse_set(const char*, azprof::BitSet&, azprof::Result&);
    static void parse_set_impl(ParseMode, const char*, azprof::BitSet&,
        azprof::Result&);

    static const struct azprof_event_class_s* by_id(size_t id) {
        return (id < count()) ? AZPROF_EVENT_CLASSES_BY_ID[id] : NULL;
    }
    static const struct azprof_event_class_s* by_name(const char *name);
    static size_t count() {return AZPROF_EVENT_COUNT;}

    size_t id() const {return ec_id;}
    const char* name() const {return ec_name;}
    size_t fixed_machine_size() const {return ec_fixed_machine_size;}
    size_t fixed_network_size() const {return ec_fixed_network_size;}

    bool is_abstract() const {
        return ec_flags & AZPROF_EVENT_CLASS_ABSTRACT_FLAG;
    }
    bool is_operation() const {
        return ec_flags & (AZPROF_EVENT_CLASS_BEGIN_OPERATION_FLAG |
            AZPROF_EVENT_CLASS_END_OPERATION_FLAG);
    }
    bool is_begin_operation() const {
        return ec_flags & AZPROF_EVENT_CLASS_BEGIN_OPERATION_FLAG;
    }
    bool is_end_operation() const {
        return ec_flags & AZPROF_EVENT_CLASS_END_OPERATION_FLAG;
    }
    bool has_fixed_network_size() const {
        return ec_flags & AZPROF_EVENT_CLASS_FIXED_NETWORK_SIZE_FLAG;
    }
    bool has_fixed_machine_size() const {
        return ec_flags & AZPROF_EVENT_CLASS_FIXED_MACHINE_SIZE_FLAG;
    }
    bool is_packed() const {
        return ec_flags & AZPROF_EVENT_CLASS_PACKED_FLAG;
    }
    bool can_trace_to_stdout() const {
        return ec_flags & AZPROF_EVENT_CLASS_CAN_TRACE_TO_STDOUT_FLAG;
    }
    bool can_trace_to_stderr() const {
        return ec_flags & AZPROF_EVENT_CLASS_CAN_TRACE_TO_STDERR_FLAG;
    }
    bool can_trace_to_file() const {
        return ec_flags & AZPROF_EVENT_CLASS_CAN_TRACE_TO_FILE_FLAG;
    }
    bool can_trace_to_memory() const {
        return ec_flags & AZPROF_EVENT_CLASS_CAN_TRACE_TO_MEMORY_FLAG;
    }

    const struct azprof_operation_class_s *operation_class() const {
        return ec_operation_class;
    }
    size_t operation_class_id() const {return ec_operation_class_id;}
    const struct azprof_event_class_s* superclass() const {
        return ec_superclass;
    }
    size_t subclass_count() const {
        return ec_subclass_count;
    }
    const struct azprof_event_class_s* subclass_at(size_t i) const {
        azprof_assert(i < subclass_count());
        return ec_subclasses[i];
    }

    size_t attr_count() const {return ec_attr_count;}
    const azprof_event_attr_t* attr_by_id(size_t id) const {
        azprof_assert(id < attr_count());
        return ec_attrs[id];
    }
    const azprof_event_attr_t* attr_by_name(const char *name) const;
    bool has_attr(const azprof_event_attr_t *attr) const {
        return (attr->id() < attr_count()) &&
            (attr_by_id(attr->id()) == attr);
    }

    void nonabstract_subclass_set(azprof::BitSet*) const;
#endif // __cplusplus

    size_t ec_id;
    unsigned int ec_flags;
    const char *ec_name;
    size_t ec_fixed_machine_size;
    size_t ec_fixed_network_size;
    struct azprof_operation_class_s *ec_operation_class;
    size_t ec_operation_class_id; // cached from 'ec_operation_class'
    struct azprof_event_class_s *ec_superclass;
    struct azprof_event_class_s **ec_subclasses;
    size_t ec_subclass_count;
    azprof_event_attr_t **ec_attrs;
    size_t ec_attr_count;
} azprof_event_class_t;

////////////////////////////////////////////////////////////////////////////////
// Operation classes
////////////////////////////////////////////////////////////////////////////////

#define AZPROF_OPERATION_CLASS_ABSTRACT_FLAG            0x1
#define AZPROF_OPERATION_CLASS_CAN_TRACE_TO_STDOUT_FLAG 0x2
#define AZPROF_OPERATION_CLASS_CAN_TRACE_TO_STDERR_FLAG 0x4
#define AZPROF_OPERATION_CLASS_CAN_TRACE_TO_FILE_FLAG   0x8
#define AZPROF_OPERATION_CLASS_CAN_TRACE_TO_MEMORY_FLAG 0x10

AZPROF_EXTERN_C struct azprof_operation_class_s*
    AZPROF_OPERATION_CLASSES_BY_ID[];

// A description of a profiler operation used for reflection. It's declared as a
// 'struct', not a 'class', so that we can initialize a 'static' instance
// without running code and relying on the order in which static initializers
// are run.
typedef struct azprof_operation_class_s {
#ifdef __cplusplus
    // Parses an event class set expression.
    //
    // The rules for parsing the expression are the following.
    // 1. Extra white space doesn't matter.
    // 2. The empty expression is the universal set.
    // 3. A sequence of event class set operations can be specified separated by
    //    commas.
    // 4. The sequence of operations is processed left-to-right.
    // 5. Event class set operations are either additive or subtractive. An
    //    additive operation is a event class name while a subtractive operation
    //    is an event class name prefixed with a "!". They both add/subtract the
    //    event class name and all its non-abstract subclasses from the set
    //    being operated on.
    // 6. If the first operation is additive then the sequence is taken to be
    //    operating on the empty set.
    // 7. If the first operation is subtractive then the sequence is taken to be
    //    operating on the universal set.
    // 8. The empty and universal sets can be specified explicitly as "none" and
    //    "all" respectively.
    //
    // Here are some example filters:
    // 1. "File I/O, !read, !write": file I/O operations except "read" and
    //    "write".
    // 2. "!send, !recv": all events and operations except "send" and "recv".
    static void parse_set(const char*, azprof::BitSet&, azprof::Result&);

    static const struct azprof_operation_class_s* by_id(size_t id) {
        return (id < count()) ? AZPROF_OPERATION_CLASSES_BY_ID[id] : NULL;
    }
    static const struct azprof_operation_class_s* by_name(const char *name);
    static size_t count() {return AZPROF_OPERATION_COUNT;}

    size_t id() const {return oc_id;}
    const char* name() const {return oc_name;}

    bool is_abstract() const {
        return oc_flags & AZPROF_OPERATION_CLASS_ABSTRACT_FLAG;
    }

    const struct azprof_operation_class_s* superclass() const {
        return oc_superclass;
    }
    const struct azprof_operation_class_s* subclass_at(size_t i) const {
        azprof_assert(i < subclass_count());
        return oc_subclasses[i];
    }
    size_t subclass_count() const {return oc_subclass_count;}

    const azprof_event_class_t* begin_event_class() const {
        return oc_begin_event_class;
    }
    const azprof_event_class_t* end_event_class() const {
        return oc_end_event_class;
    }

    void nonabstract_subclass_set(azprof::BitSet*) const;
#endif // __cplusplus

    // These fields should be treated as private as a matter of practice.
    size_t oc_id;
    unsigned int oc_flags;
    const char *oc_name;
    struct azprof_operation_class_s *oc_superclass;
    struct azprof_operation_class_s **oc_subclasses;
    size_t oc_subclass_count;
    azprof_event_class_t *oc_begin_event_class;
    azprof_event_class_t *oc_end_event_class;
} azprof_operation_class_t;

////////////////////////////////////////////////////////////////////////////////
// Event instances
////////////////////////////////////////////////////////////////////////////////

// An instance of a profiler event.
typedef struct azprof_event_s {
#ifdef __cplusplus
    const azprof_event_class_t* clazz() const {return e_class;}
    const uint64_t begin_event_clock() const {return e_begin_event_clock;}
    uint64_t clock() const {return e_clock;}
    uint32_t cpu_id() const {return e_cpu_id;}
    uint64_t thread_id() const {return e_thread_id;}

    size_t network_size() const;

    void* attr_pointer(const azprof_event_attr_t *attr) const {
        return ((char*) this) + attr->offset();
    }

    bool bool_attr(const azprof_event_attr_t*) const;
    bool bool_by_id(size_t id) const;

    int8_t int8_attr(const azprof_event_attr_t*) const;
    int8_t int8_by_id(size_t id) const;

    uint8_t uint8_attr(const azprof_event_attr_t*) const;
    uint8_t uint8_by_id(size_t id) const;

    int16_t int16_attr(const azprof_event_attr_t*) const;
    int16_t int16_by_id(size_t id) const;

    uint16_t uint16_attr(const azprof_event_attr_t*) const;
    uint16_t uint16_by_id(size_t id) const;

    int32_t int32_attr(const azprof_event_attr_t*) const;
    int32_t int32_by_id(size_t id) const;

    uint32_t uint32_attr(const azprof_event_attr_t*) const;
    uint32_t uint32_by_id(size_t id) const;

    int64_t int64_attr(const azprof_event_attr_t*) const;
    int64_t int64_by_id(size_t id) const;

    uint64_t uint64_attr(const azprof_event_attr_t*) const;
    uint64_t uint64_by_id(size_t id) const;

    uint64_t intptr_attr(const azprof_event_attr_t*) const;
    uint64_t intptr_by_id(size_t id) const;

    float float_attr(const azprof_event_attr_t*) const;
    float float_by_id(size_t id) const;

    double double_attr(const azprof_event_attr_t*) const;
    double double_by_id(size_t id) const;

    const char* string_attr(const azprof_event_attr_t*) const;
    const char* string_by_id(size_t id) const;

    const azprof_in_addr_t* in_addr_attr(const azprof_event_attr_t*) const;
    const azprof_in_addr_t* in_addr_by_id(size_t id) const;

    const azprof_sockaddr_t* sockaddr_attr(const azprof_event_attr_t*) const;
    const azprof_sockaddr_t* sockaddr_by_id(size_t id) const;

    static ssize_t skip(azprof::StaticRingBuffer& buf);
    static struct azprof_event_s* read(
        azprof::StaticRingBuffer& inbuf, char *outbuf, size_t outbuflen
    );
    ssize_t write(azprof::StaticRingBuffer&) const;
#endif // __cplusplus

    const azprof_event_class_t *e_class;
    uint64_t e_begin_event_clock;
    uint64_t e_clock;
    uint32_t e_cpu_id;
    uint32_t e_thread_id;
} azprof_event_t;

#define AZPROF_EVENT_HEADER_SIZE offsetof(azprof_event_t, e_clock)
#define AZPROF_EVENT_ID 0
#define AZPROF_EVENT_ATTR_COUNT 3

AZPROF_EXTERN_C azprof_event_class_t AZPROF_EVENT_CLASS;
AZPROF_EXTERN_C azprof_event_class_t *AZPROF_EVENT_SUBCLASSES[];
AZPROF_EXTERN_C azprof_event_attr_t*
    AZPROF_EVENT_ATTRS[AZPROF_EVENT_ATTR_COUNT];
AZPROF_EXTERN_C azprof_event_attr_t AZPROF_EVENT_CLOCK_ATTR;
AZPROF_EXTERN_C azprof_event_attr_t AZPROF_EVENT_CPU_ID_ATTR;
AZPROF_EXTERN_C azprof_event_attr_t AZPROF_EVENT_THREAD_ID_ATTR;

////////////////////////////////////////////////////////////////////////////////
// Event recorders
////////////////////////////////////////////////////////////////////////////////

#define AZPROF_MAX_EVENT_RECORDER_COUNT 8

// A closure used to process all profiler events in a given set.
typedef void (*azprof_event_recorder_t)(const azprof_event_t *event, void *arg);

// The data associated with a an event recorder registered with a single event
// class.
typedef struct azprof_erb_s {
    azprof_event_recorder_t erb_rec;
    void *erb_arg;
} azprof_erb_t;

// A table of event class IDs to recorder bucket vectorss registered for that
// class. The first NULL entry if any in a bucket vector denotes the end of the
// registered event recorders for that event class. A compile-time maximum is
// imposed on the number of event recorders which can be registered for a given
// event class for look-up efficiency.
AZPROF_EXTERN_C azprof_erb_t AZPROF_EVENT_RECORDERS_BY_CLASS_ID
    [AZPROF_EVENT_COUNT][AZPROF_MAX_EVENT_RECORDER_COUNT];

// An event recorder which does nothing which is used to place a null element in
// the event recorder vector of an event class when an event recorder is
// removed. This allows us to not have to lock the event recorder vector to
// remove an event recorder or do a null check when call all the registered
// event recorders.
AZPROF_EXTERN_C void azprof_nop_event_recorder(const azprof_event_t*, void*);

#endif // _AZPROF_EVENT_H
