#ifndef _AZPROF_DATA_HPP
#define _AZPROF_DATA_HPP

#include "azprof_constants.h"
#include "azprof_event.h"
#include "azprof_io.hpp"
#include "azprof_memory.hpp"
#include "azprof_os.hpp"
#include "azprof_util.hpp"

#include <stdio.h>
#include <stdlib.h>

namespace azprof {

////////////////////////////////////////////////////////////////////////////////
// Event recorders
////////////////////////////////////////////////////////////////////////////////

class EventRecorders {
public:
    static void init();

    // Attempts to add/remove a recorder for a given set of profiler event
    // classes. A value of 'true' is returned if the recorder was not
    // successfully added/removed for one-or-more of the classes in the set.
    static bool add(
        const BitSet& set, azprof_event_recorder_t rec, void *arg
    ) {
        return _instance->add0(set, rec, arg);
    }
    static bool remove(azprof_event_recorder_t rec, void *arg) {
        return _instance->remove0(rec, arg);
    }

private:
    static EventRecorders *_instance;

    EventRecorders();
    bool add0(const BitSet&, azprof_event_recorder_t, void*);
    bool remove0(azprof_event_recorder_t, void*);

    Mutex _mutex;
    azprof_erb_t _bkts[AZPROF_MAX_EVENT_RECORDER_COUNT];
    size_t _bkt_cnt;
};

////////////////////////////////////////////////////////////////////////////////
// Event buffering
////////////////////////////////////////////////////////////////////////////////

// An in-memory ring buffer of events.
class EventBuffer {
public:
    static void record(const azprof_event_t*, void*);

    EventBuffer(bool init_now = true);

    Condition& condition() {return _condition;}

    bool init();
    void alloc(size_t cap);
    void alloc_copy(EventBuffer&);
    void clear();
    void free();

    bool empty() const {return _buf.len() == 0;}
    azprof_event_t* pop(char *buf, size_t len);

    size_t length() {return _buf.len();}
    size_t capacity() {return _buf.len();}
    uint64_t dropped() const {return _dropped;}
    uint64_t errors() const {return _errors;}

private:
    void record0(const azprof_event_t*);

    Condition _condition;
    StaticRingBuffer _buf;
    uint64_t _dropped;
    uint64_t _errors;
};

////////////////////////////////////////////////////////////////////////////////
// Event formatting
////////////////////////////////////////////////////////////////////////////////

// Writes events as plain-text to a stream.
class EventFormatter {
public:
    static void init();

    EventFormatter(PrintStream *stream);

    PrintStream* stream() { return _stream; }

    void format(EventBuffer& buffer);
    void format(const azprof_event_t*);
    int close();

private:
    static BitSet *_end_syscall_subclass_set;

    PrintStream *_stream;

    // Record the event type ID and originating thread of unfinished operations
    // so that we can determine whether another event occurred before the end of
    // this operation. This allows us to collapse back-to-back begin and end
    // events for an operation into a single line.
    size_t _unfinished_class_id;
    uint64_t _unfinished_thread_id;
};

////////////////////////////////////////////////////////////////////////////////
// Event I/O
////////////////////////////////////////////////////////////////////////////////

// Background thread which writes events as plain-text to a stream.
//
// Events are taken from a buffer. Threads record events into the buffer without
// blocking for I/O, and this thread formats the events to a stream in the
// background. Events are dropped if the event buffer is full because the
// background thread can't keep up.
//
// Separating the I/O into a background thread is important. Doing I/O to record
// I/O operations can lead to deadlocks which are difficult to resolve. Also, if
// the I/O for tracing is stuck then the whole application could be effected if
// this I/O wasn't backgrounded.
class EventWriter : public Thread {
public:
    static void init();

    EventWriter();

    void configure(EventBuffer *buffer, PrintStream *stream);
    void close();

private:
    static BitSet *_end_syscall_subclass_set;

    // State of the formatter to exchange between threads.
    class State {
    public:
        State(EventBuffer *__buffer, PrintStream *stream) :
            _buffer(__buffer),
            _formatter((stream != NULL) ? new EventFormatter(stream) : NULL) {}

        ~State() {
            delete _formatter;
        }

        bool is_closed() {return buffer() == NULL;}

        EventBuffer* buffer() { return _buffer; }
        EventFormatter* formatter() { return _formatter; }

        void close() {
            _buffer = NULL;
            if (_formatter != NULL) _formatter->close();
            _formatter = NULL;
        }

    private:
        EventBuffer *_buffer;
        EventFormatter *_formatter;
    };

    // Background thread loop implementation.
    virtual void run();

    Mutex _mutex;

    // New state which is posted for the background thread to pick-up.
    State *_new_state;

    // Current state in use by the background thread, or NULL if it's inactive.
    State *_current_state;
};

////////////////////////////////////////////////////////////////////////////////
// Event tracing
////////////////////////////////////////////////////////////////////////////////

// Configuration for recording events either to stdout, stderr, a specified
// file, or an in-memory ring buffer. A filter expression can specified for
// which events are recorded which is parsed and converted to an event class
// set.
class EventTracing {
public:
    class State {
    public:
        enum Mode {
            OFF_MODE,
            STDOUT_MODE,
            STDERR_MODE,
            FILE_MODE,
            MEMORY_MODE
        };

        enum {
            DEFAULT_MEMORY_SIZE = 16*Memory::KB,
            MAX_MEMORY_SIZE = Memory::MB
        };

        static void parse_flags(
            const char *mode, const char *filter, State& state, Result& result
        );
        static const char* mode_to_str(int);
        static int mode_to_event_class_flag(int);

        State();
        State(const State&);
        ~State();

        State& operator=(const State&);

        Mode mode() const {return _mode;}
        const char* mode_name() const {return mode_to_str(_mode);}
        const char* file_path() const {return _file_path;}
        size_t memory_size() const {return _memory_size;}
        const char* filter() const {return _filter;}

        bool append() const {return _append;}

        void set_mode(Mode);
        void set_file_path(const char*);
        void set_memory_size(size_t);
        void set_filter(const char*);
        void set_append(bool);
        void set_thread_ids(bool);
        void set_timestamps(bool);

    private:
        Mode _mode;
        char *_file_path;
        size_t _memory_size;
        char *_filter;
        bool _append;
    };

    static void init1();
    static bool init2(const char *mode, const char *filter);
    static void get_event_buffer(EventBuffer&);
    static void get_state(State&);
    static void set_state(const State&, Result&);
    static bool enabled_for_current_thread();
    static void set_enabled_for_current_thread(bool enabled);

private:
    static EventTracing *_instance;

    EventTracing();

    bool init2_impl(const char *mode, const char *filter);
    void get_event_buffer_impl(EventBuffer&);
    void get_state_impl(State&);
    void set_state_impl(const State&, Result&);
    bool enabled_for_current_thread_impl();
    void set_enabled_for_current_thread_impl(bool enabled);

    OS::ThreadKey _enabled_key;

    Mutex _mutex;
    State _state;

    EventBuffer _buffer;
    EventWriter _writer;
};

////////////////////////////////////////////////////////////////////////////////
// Operation profiles
////////////////////////////////////////////////////////////////////////////////

class PerThreadOperationProfile {
    friend class OperationProfile;
    friend class OperationProfileRow;
public:
    class ThreadProfile {
    public:
        ThreadProfile() : _os_thr(0), _prof(NULL) {}
        ThreadProfile(OS::Thread thr, class OperationProfile *prof) :
            _os_thr(thr), _prof(prof) {}

        OS::Thread os_thread() const {return _os_thr;}
        OperationProfile* profile() const {return _prof;}

    private:
        OS::Thread _os_thr;
        OperationProfile *_prof;
    };

    static void init();

    PerThreadOperationProfile();

    const BitSet& operation_class_set() const {return _set;}
    size_t rows() const {return _set.size();}
    size_t threads() const {return _thrv.size();}

    void init(const char *op_class_set_expr, Result&);
    void thread_list(Vector<OS::Thread>&);
    OperationProfile* thread_profile(OS::Thread);
    void summary(OperationProfile&);
    void reset();

private:
    // Thread key used to save a pointer to the per-thread 'OperationProfile'
    // instances.
    OS::ThreadKey _key;

    // Mapping of operation class IDs to indexes in the table. Keeping this
    // mapping allows us avoid probing or wasted space vs. using a hash table.
    // The table is built once and shared between all the 'OperationProfile'
    // instances associated with a 'PerThreadOperationProfile'.
    uint16_t *_map;

    // Set of operation class IDs traced in the profile.
    BitSet _set;

    // Global list of all the threads which have created 'OperationProfile'
    // instances.
    Mutex _thr_mutex;
    Vector<ThreadProfile> _thrv;
};

// A row in an operation profile table.
class OperationProfileRow {
    friend class OperationProfile;
public:
    static void init();

    OperationProfileRow();

    uint64_t num() const;
    uint64_t dur_sum_us() const;
    uint64_t dur_min_us() const;
    uint64_t dur_max_us() const;
    float dur_mean_us() const;
    float dur_stddev_us() const;
    float dur_1_min_avg_us() const;
    float dur_5_min_avg_us() const;
    float rate_1_min_avg_op_per_sec() const;
    float rate_5_min_avg_op_per_sec() const;

    void reset();
    void add(OperationProfileRow&);
    void to_xml(Response*);
    void to_csv(Response*);

private:
    static const float ONE_MIN_EXP_FACTOR = -1.6666666666666667e-08;
    static const float FIVE_MIN_EXP_FACTOR = -3.3333333333333334e-09;
    static uint64_t FIVE_SEC;

    // To avoid expensive ticks to seconds conversions, overflows in duration
    // squared calculations, and needing 64-bit values to store durations we
    // store them in cycles divided by a power of two. Tracking durations in
    // 1/10 of millis gives us 16 as our power of two assuming a clocks
    // frequency on the order of 1 GHz. This leads to an overflow in a 32-bit
    // representation of a duration after around 5 days which is acceptable.
    static const int DUR_SHIFT = 16;

    void update_if_necessary() {
        uint64_t t = OS::clock();
        if (t > _next_update) update(t);
    }

    void update(uint64_t t);
    static void exp_moving_avg(float dt, float x, float *one, float *five);

    ////////////////////////////////////////////////////////
    // Cache line of hot data accessed for each operation //
    ////////////////////////////////////////////////////////

    // Threshold for the next time which the cold data should be updated stored
    // in clocks. We update the data approximately every five seconds for good
    // moving average behavior and to avoid overflows in the 32-bit values we
    // use on the fast-path.
    uint64_t _next_update;

    // Number of operations which have occurred since the last cold data update
    // and the sum of their durations.
    uint32_t _num_since_update;
    uint32_t _sum_since_update;

    // The sum of the duration squared of all operations which have occurred
    // used for the duration standard deviation calculation.
    uint64_t _sum_sq;

    // The minimum and maximum duration of all operations which have occurred.
    uint32_t _min;
    uint32_t _max;

    //////////////////////////////////////////////////
    // Cache line of cold data updated periodically //
    //////////////////////////////////////////////////

    // The total number of operations which have occurred and the sum of their
    // durations. These aren't updated on the fast-path to avoid needing to
    // touch more than one cache line.
    uint64_t _num;
    uint64_t _sum;

    // One and five minute moving averages of the duration of all operations
    // which have occurred stored in microseconds.
    float _dur_1_min_avg;
    float _dur_5_min_avg;

    // One and five minute moving averages of the rate of all operations which
    // have occurred stored in operations per second.
    float _rate_1_min_avg;
    float _rate_5_min_avg;
};

// An operation profile table attached to a single thread or used in summing
// stats across all threads.
class OperationProfile : public ReferenceCounted {
    friend class OperationProfileRow;
    friend class PerThreadOperationProfile;
public:
    OperationProfile(PerThreadOperationProfile*);
    ~OperationProfile();

    bool alloc();
    void reset();
    void copy(OperationProfile&);

    const PerThreadOperationProfile* per_thread() const {return _per_thr;}
    size_t rows() const {return per_thread()->rows();}

    const OperationProfileRow& row_by_id(size_t id) const {
        azprof_assert(_per_thr->_set.contains(id));
        return row_at(_per_thr->_map[id]);
    }

    OperationProfileRow& row_by_id(size_t id) {
        azprof_assert(_per_thr->_set.contains(id));
        return row_at(_per_thr->_map[id]);
    }

    const OperationProfileRow& row_at(size_t i) const {
        azprof_assert(i < rows());
        return _tbl[i];
    }
    OperationProfileRow& row_at(size_t i) {
        azprof_assert(i < rows());
        return _tbl[i];
    }

private:
    static OperationProfile* const NIL;

    static void record(const azprof_event_t*, void*);
    static void destroy(void*);

    // Pointer back to the associated 'PerThreadOperationProfile'.
    PerThreadOperationProfile *_per_thr;

    // A spin lock for the table so that readers can atomically copy-out the
    // table in the face of writters. We use a spin lock and can get away with
    // using a lock at all since readers only come along when someone looks at
    // the table through the ARTA GUI and we want to avoid the extra layers
    // associated with a full-blown mutex.
    SpinLock _lock;

    // Vector of entries the length of the operation class set which is indexed
    // by the indexes stored in '_map'.
    OperationProfileRow *_tbl;
    void *_tbl_unaligned;
};

}; // namespace azprof

#endif // _AZPROF_DATA_HPP
