#ifndef _AZPROF_IO_HPP
#define _AZPROF_IO_HPP

#include "azprof_memory.hpp"
#include "azprof_security.hpp"
#include "azprof_util.hpp"

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#ifndef AZPROF_NO_JNI
#include <jni.h>
#endif

// Do not change these without checking the globalDefinitions_os.hpp file for consistency.
#define PARSE_ARTAPORT_ANY_STR                    "any"
#define BIND_ARTA_PORT_MIN                        1025
#define BIND_ARTA_PORT_MAX                        65535

namespace azprof {

////////////////////////////////////////////////////////////////////////////////
// HTTP/monitoring servers
////////////////////////////////////////////////////////////////////////////////

class Request;
class Response;
class Servlet;

class Server {
public:
    typedef void (*DetachCallback)(void*);

    enum Protocol {
        UNKNOWN,
        HTTP,
        HTTPS,
        MONITORING,
        JNI
    };

    static const char* protocol_to_str(Protocol);

    Server(Protocol, Servlet*);
    virtual ~Server();

    Protocol protocol() {return _protocol;}
    bool is_http() const {return _protocol == HTTP;}
    bool is_https() const {return _protocol == HTTPS;}
    bool is_monitoring() const {return _protocol == MONITORING;}
    bool is_jni() const {return _protocol == JNI;}
    const char *protocol_name() const;

    Servlet* servlet() {return _servlet;}

    // Used so that HotSpot can register a callback to do a JNI detach from a
    // server thread which is exiting.
    void set_detach_callback(DetachCallback, void*);

    virtual void to_xml(Response*) {azprof_unimplemented();}

protected:
    void detach();

private:
    Protocol _protocol;
    Servlet* _servlet;
    DetachCallback _detach_callback;
    void *_detach_arg;
};

// Base class for an HTTP or monoitoring request.
class Request {
public:
    static Request* parse(const Server*, const char *uri);
    static char* unescape(char*);
    static time_t ptime(const char*);
    static void unit_test();

    virtual ~Request();

    const Server* server() const {return _server;}

    void set_session(Session* session) { _session = session; }
    Session* session() {return _session;}
    const Session* session() const {return _session;}
    bool is_azul_engineer() const {
        return session()->privilege().has_azul_engineer_access();
    }

    Server::Protocol protocol() const {return _protocol;}
    bool is_http() const {return protocol() == Server::HTTP;}
    bool is_https() const {return protocol() == Server::HTTPS;}
    bool is_monitoring() const {return protocol() == Server::MONITORING;}
    const char* protocol_name() const;

    const char* raw_uri() const {return _raw_uri;}
    const char* raw_query() const {return _raw_query;}

    const char* path() const {return _path;}
    const char* base() const {return _base;}
    const char* fragment() const {return _fragment;}
    char* cookie_prefix() const;

    size_t parameter_count() const {return _param_cnt;}
    const char* parameter_name_at(size_t i) const;
    char* parameter_value_at(size_t i);
    const char* parameter_value_at(size_t i) const;
    bool bool_parameter_at(size_t i) const;
    int32_t int32_parameter_at(size_t i) const;
    uint32_t uint32_parameter_at(size_t i) const;
    int64_t int64_parameter_at(size_t i) const;
    uint64_t uint64_parameter_at(size_t i) const;
    double double_parameter_at(size_t i) const;

    char* parameter_by_name(const char *name);
    const char* parameter_by_name(const char *name) const;
    bool bool_parameter_by_name(const char *name) const;
    int32_t int32_parameter_by_name(const char *name) const;
    uint32_t uint32_parameter_by_name(const char *name) const;
    int64_t int64_parameter_by_name(const char *name) const;
    uint64_t uint64_parameter_by_name(const char *name) const;
    double double_parameter_by_name(const char *name) const;

    size_t header_count() const {return _header_cnt;}
    const char* header_name_at(size_t i) const;
    const char* header_value_at(size_t i) const;
    const char* header_by_name(const char*) const;
    time_t time_header_by_name(const char*) const;

    const char* acceptable_content_type(
        const char **content_types, size_t content_type_count
    );

protected:
    Request(const Server*, Server::Protocol);

    size_t buf_len() {return _buf_len;}
    size_t buf_cap() {return sizeof(_buf) - _buf_len;}
    char* buf_begin() {return _buf;}
    char* buf_end() {return buf_begin()+_buf_len;}
    void buf_inc(size_t k = 1) {_buf_len += MIN(k, buf_cap());}
    void buf_copy(const char *input);

    bool add_header(char *name, char *value);
    bool add_default_paramemter(char *name, char *value);

    bool parse_uri(char*);
    char* parse_path(char*);
    bool parse_query(char*);

private:
    static const size_t MAX_PARAM_CNT = 16;
    static const size_t MAX_HEADER_CNT = 16;

    const Server *_server;
    Session *_session;
    Server::Protocol _protocol;
    char _buf[64*1024];
    size_t _buf_len;
    char *_raw_uri;
    char *_raw_query;
    char _base[32];
    char *_path;
    struct param_s {
        char *name;
        char *value;
    } _params[MAX_PARAM_CNT];
    size_t _param_cnt;
    char *_fragment;
    struct header_s {
        char *name;
        char *value;
    } _headers[MAX_HEADER_CNT];
    size_t _header_cnt;
};

// Base class for the response to an HTTP or monoitoring response.
class Response {
public:
    enum Code {
        UNKNOWN = 0,
        CONTINUE = 100,
        SWITCHING_PROTOCOLS = 101,
        OK = 200,
        CREATED = 201,
        ACCEPTED = 202,
        NON_AUTHORITATIVE_INFORMATION = 203,
        NO_CONTENT = 204,
        RESET_CONTENT = 205,
        PARTIAL_CONTENT = 206,
        MULTIPLE_CHOICES = 300,
        MOVED_PERMANENTLY = 301,
        MOVED_TEMPORARILY = 302,
        FOUND = 302,
        SEE_OTHER = 303,
        NOT_MODIFIED = 304,
        USE_PROXY = 305,
        TEMPORARY_REDIRECT = 307,
        BAD_REQUEST = 400,
        UNAUTHORIZED = 401,
        PAYMENT_REQUIRED = 402,
        FORBIDDEN = 403,
        NOT_FOUND = 404,
        METHOD_NOT_ALLOWED = 405,
        NOT_ACCEPTABLE = 406,
        PROXY_AUTHENTICATION_REQUIRED = 407,
        REQUEST_TIMEOUT = 408,
        CONFLICT = 409,
        GONE = 410,
        LENGTH_REQUIRED = 411,
        PRECONDITION_FAILED = 412,
        REQUEST_ENTITY_TOO_LARGE = 413,
        REQUEST_URI_TOO_LONG = 414,
        UNSUPPORTED_MEDIA_TYPE = 415,
        REQUESTED_RANGE_NOT_SATISFIABLE = 416,
        EXPECTATION_FAILED = 417,
        INTERNAL_SERVER_ERROR = 500,
        NOT_IMPLEMENTED = 501,
        BAD_GATEWAY = 502,
        SERVICE_UNAVAILABLE = 503,
        GATEWAY_TIMEOUT = 504,
        HTTP_VERSION_NOT_SUPPORTED = 505
    };

    enum State {
        BEGIN,
        HEADER,
        BODY,
        END
    };

    static const char* code_to_str(int);
    static char* escape(const char *input, char *buf, size_t len);
    static size_t check_response_printable(const char *buf, size_t buflen);
    static const char* ftime(char *buf, size_t len, time_t t = 0);

    virtual ~Response();

    virtual bool is_non_blocking() const = 0;
    virtual void set_non_blocking(bool) = 0;

	virtual void send_error(Code k);
	virtual void send_error(Code k, const char* msg);
    virtual void ok(const char *content_type, ssize_t content_length) = 0;
    virtual void ok() = 0;
    virtual void bad_request() = 0;
    virtual void see_other(const char *fmt, ...) = 0;
    virtual void not_modified() = 0;
    virtual void unauthorized() = 0;
    virtual void forbidden() = 0;
    virtual void not_acceptable() = 0;
    virtual void not_found() = 0;
    virtual void precondition_failed() = 0;
    virtual void internal_server_error() = 0;
    virtual void not_implemented() = 0;

    void redirect_to_referer();

    bool save_parameter(const char *name);
    void clear_saved_parameters();

    virtual void vheaderf(const char *name, const char *fmt, va_list ap) = 0;
    virtual void headerf(const char *name, const char *fmt, ...) = 0;
    virtual void header(const char *name, const char *value) = 0;
    virtual void end_header() = 0;

    // Writes the contents of the buffer directly to the response stream.
    virtual ssize_t write(const char *buf, size_t len) = 0;

    // Writes the contents of the buffer to the response stream and escapes any
    // XML special characters.
    virtual ssize_t xwrite(const char *buf, size_t len);

    // Writes the contents of the buffer to the response stream and escapes any
    // URL special characters. Once all URL special characters have been escaped
    // no XML special characters will be present.
    virtual ssize_t uwrite(const char *buf, size_t len);

    // Prints formatted output directly to the response stream.
    virtual ssize_t vprint(const char *fmt, va_list ap) = 0;
    virtual ssize_t print(const char *fmt, ...);

    virtual void end() = 0;

    Request* request() {return _request;}
    Code code() const {return _code;}
    State state() const {return _state;}

    ssize_t indent();
    void inc_indention_level();
    void dec_indention_level();

protected:
    Response(Request*);

    void set_state(State);
    void set_code(Code __code) {_code = __code;}

    uint64_t _saved_params;

private:
    Request *_request;
    Code _code;
    State _state;
    size_t _indention_level;
};

// Takes a request and generates a response.
class Servlet {
public:
    enum {
        NO_LOGIN_REQUIRED = 0x1
    };
    static const int NO_FLAGS = 0;
    static const int ALL_FLAGS = NO_LOGIN_REQUIRED;

    Servlet(int flags, Privilege required_privilege);
    virtual ~Servlet();

    int flags() const {return _flags;}
    Privilege required_privilege() const {return _required_privilege;}

    bool requires_login() const {return (_flags & NO_LOGIN_REQUIRED) == 0;}
    bool requires_azul_engineer_access() const {
        return _required_privilege.has_azul_engineer_access();
    }

    virtual void pre_service(Request*, Response*) {}
    virtual void service(Request*, Response*) = 0;
    virtual void post_service(Request*, Response*) {}

    void service0(Request *req, Response *res) {
        pre_service(req, res);
        service(req, res);
        post_service(req, res);
    }

private:
    int _flags;
    Privilege _required_privilege;
};

////////////////////////////////////////////////////////////////////////////////
// Duplex streams
////////////////////////////////////////////////////////////////////////////////

class Stream {
public:
    virtual ~Stream();
    virtual bool is_closed() = 0;
    virtual ssize_t read(char *buf, size_t len) = 0;
    virtual ssize_t write(const char *buf, size_t len) = 0;
    virtual ssize_t flush(ssize_t n = -1);
    virtual int close() = 0;
};

class PrintStream : public Stream {
public:
    virtual ssize_t vprint(const char *fmt, va_list ap) = 0;
    virtual ssize_t print(const char *fmt, ...);
    virtual ssize_t flush(ssize_t n = -1) = 0;
};

class FdStream : public Stream {
public:
    FdStream(int fd);

    int fd() {return _fd;}

    virtual bool is_closed();
    virtual ssize_t read(char *buf, size_t len);
    virtual ssize_t write(const char *buf, size_t len);
    virtual int close();

private:
    int _fd;
};

#ifndef AZPROF_NO_OPENSSL
class SslStream : public Stream {
public:
    SslStream(SSL*);

    SSL* ssl() {return _ssl;}

    virtual bool is_closed();
    virtual ssize_t read(char *buf, size_t len);
    virtual ssize_t write(const char *buf, size_t len);
    virtual int close();

private:
    SSL* _ssl;
};
#endif // AZPROF_NO_OPENSSL

class FileStream : public PrintStream {
public:
    FileStream(FILE*);

    FILE* file() {return _file;}

    virtual bool is_closed();
    virtual ssize_t read(char *buf, size_t len);
    virtual ssize_t write(const char *buf, size_t len);
    virtual ssize_t vprint(const char *fmt, va_list ap);
    virtual ssize_t flush(ssize_t n = -1);
    virtual int close();

private:
    FILE* _file;
};

/**
 * This is a stream for writing to stdout or stderr only.
 * Why not just use a FileStream? It's not thread safe: stdout and stderr are
 * actually macros that resolve to an address on the stack, so if a thread
 * creates a FileStream that points to stdout, hands that FileStream to another
 * thread and quits running, you're in for a world of hurt.
 */
class StdStream : public PrintStream {
public:
    // Please forgive this type enum (as well as the poor choice of creating
    // yet another Stream class). It seemed like the least obtrusive way to
    // add this functionality in a hurry.
    enum stream_type { STDSTREAM_STDOUT, STDSTREAM_STDERR };
     StdStream(const stream_type st): _st(st), _is_closed(false) {}

    virtual bool is_closed() { return _is_closed; }
    virtual ssize_t read(char *buf, size_t len);
    virtual ssize_t write(const char *buf, size_t len);
    virtual ssize_t flush(ssize_t n = -1);
    virtual int close();
    virtual ssize_t vprint(const char *fmt, va_list ap);

private:
    const stream_type _st;
    bool _is_closed;
};

class ResponseStream : public PrintStream {
public:
    ResponseStream(Response*);

    Response* response() {return _res;}

    virtual bool is_closed();
    virtual ssize_t read(char *buf, size_t len);
    virtual ssize_t write(const char *buf, size_t len);
    ssize_t vprint(const char *fmt, va_list ap);
    virtual ssize_t flush(ssize_t n = -1);
    virtual int close();

private:
    Response *_res;
};

// A stream backed by ring buffers. For writes, it uses a combination of
// compacting, resizing, and draining to manage the amount of memory used by its
// underlying buffer. It can be configured as non-blocking so that a 'write' or
// 'print' call will never cause I/O to occur. If the contents of a 'write' or
// 'print' are truncated then it's recorded that the buffer overflowed.
class BufferedStream : public PrintStream {
public:
    enum {
        NON_BLOCKING_WRITES = 0x1,
        READ_EOF = 0x2,
        WRITE_OVERFLOWED = 0x4
    };

    BufferedStream(
        Stream *raw, size_t rdbuf_cap, size_t wrbuf_cap, size_t max_wrbuf_cap
    );
    virtual ~BufferedStream();

    bool uses_non_blocking_writes() const {return _flags & NON_BLOCKING_WRITES;}
    void set_non_blocking_writes(bool x) {
        if (x) _flags |= NON_BLOCKING_WRITES;
        else _flags &= ~NON_BLOCKING_WRITES;
    }
    bool was_eof_read() const {return _flags & READ_EOF;}
    bool has_write_overflowed() const {return _flags & WRITE_OVERFLOWED;}

    bool alloc();

    virtual bool is_closed();

    virtual ssize_t read(char *buf, size_t len);
    virtual ssize_t readline(char *buf, size_t len);
    virtual void copy_in(const char *buf, size_t len);
    virtual ssize_t write(const char *buf, size_t len);
    virtual ssize_t vprint(const char *fmt, va_list ap);

    virtual ssize_t flush(ssize_t n = -1);
    virtual int close();

private:
    int readc();
    ssize_t read_more();

protected:
    Stream *_raw;
    DynamicRingBuffer _rdbuf;
    DynamicRingBuffer _wrbuf;
    int _flags;
};

////////////////////////////////////////////////////////////////////////////////
// HTTP
////////////////////////////////////////////////////////////////////////////////

class HttpStream;

class HttpServer : public Server {
public:
    enum {
        MAX_BIND_ATTEMPTS = 30,
        BIND_ATTEMPT_DELAY = 1000, // milliseconds
        MIN_NUM_THREADS = 2,
        MAX_NUM_THREADS = 8,
        THREAD_IDLE_TIMEOUT = 3000 // milliseconds
    };

    HttpServer(Protocol, Servlet*);
    virtual ~HttpServer();

    void state(Result&);
    virtual void start(
        const char *interface, int port, Privilege, bool reattempt_bind, Result&
    );
    void stop(Result&);
    void abort(const char*);

    virtual void to_xml(Response*);

    // Provide the port number
    // Needed when the port number is selected during initialization for
    // ARTAPort=any
    int port() {return _port;}

protected:
    void start_http(bool reattempt_bind, Result&);
    int select0();
    virtual int select();
    void service(int fd, HttpStream*, Server::Protocol);

    Mutex _mutex;
    char _interface[32];
    int _port;
    int _listening_fd;
    Privilege _privilege;
    Result _result;

private:
    class BindAttemptThread : public Thread {
    public:
        BindAttemptThread(HttpServer*);

        Log* log() {return _log;}
        void set_log(Log *__log) {_log = __log;}

        virtual void run();

    private:
        HttpServer *_server;
        Log        *_log;
    };

    class WorkerThread : public Thread {
    public:
        WorkerThread(HttpServer*);
        virtual void run();
    private:
        HttpServer *_server;
    };

    friend class HttpServer::BindAttemptThread;
    friend class HttpServer::WorkerThread;

    bool bind(bool reattempt, Result&);
    void run();
    int close0();

#ifdef AZ_LINUX
    Poll _poll;
#else
    Selector _selector;
#endif
    BindAttemptThread _bind_attempt_thread;
    Mutex _thread_mutex;
    WorkerThread _thread_template;
    int _thread_cnt;
    Atomic::Int _idle_thread_cnt;
};

#ifndef AZPROF_NO_OPENSSL
class HttpsServer : public HttpServer {
public:
    HttpsServer(
        Servlet *servlet,
        const char *key_path,
        const char *cert_path,
        const char *dh512_path,
        const char *dh1024_path
    );

    virtual void start(
        const char *interface, int port, Privilege, bool reattempt_bind, Result&
    );
    virtual void start(
        const char *interface, int port, unsigned int session_timeout,
        bool reattempt_bind, Result&
    );

protected:
    void start_https(bool reattempt_bind, Result&);
    virtual int select();

private:
    static void openssl_lock_callback(int, int, const char*, int);
    static unsigned long openssl_id_callback();
    static DH* openssl_tmp_dh_callback(SSL*, int, int);

    static SSL_CTX *_ssl_ctx;
    static Mutex *_openssl_locks;
    static int _openssl_lock_cnt;
    static DH *_dh512;
    static DH *_dh1024;

    unsigned int _session_timeout;
    const char *_key_path;
    const char *_cert_path;
    const char *_dh512_path;
    const char *_dh1024_path;
};
#endif // AZPROF_NO_OPENSSL

// An extension of a BufferedStream to support the HTTP chunked transfer
// encoding.
class HttpStream : public BufferedStream {
public:
    enum {
        STATUS_CODE_GIVEN = 0x8,
        HEADER_DONE = 0x10,
        CHUNKED = 0x20,
        SENSITIVE_CONTENT = 0x40
    };

    HttpStream(
        HttpServer *server,
        Stream *raw,
        size_t rdbuf_cap = Memory::KB,
        size_t wrbuf_cap = Memory::MB,
        size_t max_wrbuf_cap = getARTAMaxResponseSizeBytes()
    );

    HttpServer* server() {return _server;}
    bool was_status_code_given() const {return _flags & STATUS_CODE_GIVEN;}
    bool is_header_done() const {return _flags & HEADER_DONE;}
    bool is_chunked() const {return _flags & CHUNKED;}
    void set_chunked() {_flags |= CHUNKED;}
    bool is_connected() const { return !_raw->is_closed(); }

    void begin_response(Response::Code, const char *msg);
    void begin_response(Response::Code);
    void headerf(const char *name, const char *fmt, ...);
    void vheaderf(const char *name, const char *fmt, va_list ap);
    void header(const char *name, const char *value);
    void end_header();
    void end_response();

    void begin_sensitive_content();
    void end_sensitive_content();

    virtual ssize_t read(char *buf, size_t len);
    virtual ssize_t readline(char *buf, size_t len);
    virtual void copy_in(const char *buf, size_t len);
    virtual ssize_t write(const char *buf, size_t len);
    virtual ssize_t vprint(const char *fmt, va_list ap);
    virtual ssize_t flush(ssize_t n = -1);

    // Work around for bug 25976
    static unsigned int getARTAMaxResponseSizeBytes();
    static unsigned int setARTAMaxResponseSizeMB(unsigned int i);
private:
    void log_non_sensitive_content(const char *buf, ssize_t len);
    void log_non_sensitive_content(ssize_t len, const char *fmt, va_list ap);

    HttpServer *_server;
    size_t _hdroff;

    static unsigned int ARTAMaxResponseSizeBytes;
};

class HttpRequest : public Request {
public:
    enum {
        _EOF = 0x1
    };

    HttpRequest(const HttpServer*, Server::Protocol, int);
    static void unit_test();

    bool eof() const {return _flags & _EOF;}

    int port() const {return _port;}
    const char* method() const {return _method;}
    const char* version() const {return _version;}

    bool read(HttpStream*);

private:
    static HttpRequest* test_parse_request_line(const char *data);

    bool parse_request_line();

    int _flags;
    int _port;
    char *_method;
    char *_version;
};

class HttpResponse : public Response {
public:
    HttpResponse(HttpRequest *request, HttpStream *stream);

    static const char* server();

    virtual bool is_non_blocking() const;
    virtual void set_non_blocking(bool);

    virtual void send_error(Code k, const char *msg);
    virtual void send_error(Code k);
    virtual void ok(const char *content_type, ssize_t content_length);
    virtual void ok();
    virtual void see_other(const char *fmt, ...);
    virtual void not_modified();
    virtual void bad_request();
    virtual void unauthorized();
    virtual void forbidden();
    virtual void not_acceptable();
    virtual void not_found();
    virtual void precondition_failed();
    virtual void internal_server_error();
    virtual void not_implemented();

    virtual void vheaderf(const char *name, const char *fmt, va_list ap);
    virtual void headerf(const char *name, const char *fmt, ...);
    virtual void header(const char *name, const char *value);
    virtual void end_header();

    virtual ssize_t write(const char *buf, size_t len);
    virtual ssize_t vprint(const char *fmt, va_list ap);
    virtual void end();

private:
    void begin(Code, const char *);
    void begin(Code);
    void begin_and_end_header(Code, const char *);
    void begin_and_end_header(Code);

    HttpStream *_stream;
};

////////////////////////////////////////////////////////////////////////////////
// Azul Monitoring
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// JNI
////////////////////////////////////////////////////////////////////////////////

#ifndef AZPROF_NO_JNI

class JniServer : public Server {
public:
    static JniServer* init(Servlet*);
    static JniServer* self() {return _self;}

    static void service(JNIEnv *env, jstring uri, jobject obj);

private:
    JniServer(Servlet*);

    static JniServer *_self;
};

class JniRequest : public Request {
public:
    static JniRequest* parse(const JniServer*, JNIEnv *env, jstring uri);
    ~JniRequest();
private:
    JniRequest(const JniServer*);
};

class JniResponse : public Response {
public:
    JniResponse(JniRequest *req, JNIEnv *env, jobject obj);
    virtual ~JniResponse();

    virtual bool is_non_blocking() const {return true;}
    virtual void set_non_blocking(bool x) {}

    virtual void ok(const char *content_type, ssize_t content_length);
    virtual void ok() {}
    virtual void see_other(const char *fmt, ...);
    virtual void not_modified();
    virtual void bad_request();
    virtual void unauthorized();
    virtual void forbidden();
    virtual void not_acceptable();
    virtual void not_found();
    virtual void precondition_failed();
    virtual void internal_server_error();
    virtual void not_implemented();

    virtual void vheaderf(const char *name, const char *fmt, va_list ap);
    virtual void headerf(const char *name, const char *fmt, ...);
    virtual void header(const char *name, const char *value);
    virtual void end_header();

    virtual ssize_t write(const char *buf, size_t len);
    virtual ssize_t vprint(const char *fmt, va_list ap);

    virtual void end();

private:
    void err(Code code);

    int grow(size_t n);
    int grow_if_necessary(size_t n);

    JNIEnv *_env;
    jobject _obj;
    char *_buf;
    size_t _len;
    size_t _cap;
};

#endif // ifndef AZPROF_NO_JNI

////////////////////////////////////////////////////////////////////////////////
// XML
////////////////////////////////////////////////////////////////////////////////

#define AZPROF_DECL_INT_XML_LEAF_FUNC(type, fmt, xfmt, expr) \
    static void leaf(Response*, const char*, type); \
    static void xleaf(Response*, const char*, type);

#define AZPROF_DECL_NONINT_XML_LEAF_FUNC(type, fmt, xfmt, expr) \
    static void leaf(Response*, const char*, type);

#define AZPROF_FOREACH_INT_XML_LEAF_TYPE(macro) \
    macro(char, "%d", "%#x", k) \
    macro(unsigned char, "%u", "%#x", k) \
    macro(short, "%d", "%#x", k) \
    macro(unsigned short, "%u", "%#x", k) \
    macro(int, "%d", "%#x", k) \
    macro(unsigned int, "%u", "%#x", k) \
    macro(long, "%ld", "%#x", k) \
    macro(unsigned long, "%lu", "%#x", k) \
    macro(long long, "%lld", "%#llx", k) \
    macro(unsigned long long, "%llu", "%#llx", k)

#define AZPROF_FOREACH_NONINT_XML_LEAF_TYPE(macro) \
    macro(bool, "%s", "%s", k ? "true" : "false") \
    macro(float, "%f", "%f", (double) k) \
    macro(double, "%f", "%f", k)

class Xml {
public:
    Xml(Response *res, const char *__name);
    ~Xml();

    static void begin(Response *res, const char *name);
    static void end(Response *res, const char *name);

    // Leaf node which has its text XML escaped and an optional attribute.
    static void leaf(Response *res, const char *name, const char *txt);

    // Leaf node which has its text XML escaped and an optional attribute.
    static void leaf_precise_double(Response *res, const char *name, double d);

    // Leaf node which has its text URI escaped and an optional attribute.
    static void uleaf(Response *res, const char *name, const char *txt);

    // Leaf node with unescaped text and an optional attribute.
    static void rleaf(Response *res, const char *name, const char *txt);

    // Leaf node with unescaped formatted text.
    static void leaff(Response *res, const char *name, const char *fmt, ...);
    static void vleaff(
        Response *res, const char *name, const char *fmt, va_list ap
    );

    // Convenience functions to avoid needing to specify a format.
    AZPROF_FOREACH_INT_XML_LEAF_TYPE(AZPROF_DECL_INT_XML_LEAF_FUNC)
    AZPROF_FOREACH_NONINT_XML_LEAF_TYPE(AZPROF_DECL_NONINT_XML_LEAF_FUNC)

private:
    Response *_res;
    const char *_name;
};

} // namespace azprof

#endif // _AZPROF_IO_HPP
