/* -*- mode: c++-mode; c-basic-offset: 4; indent-tabs-mode: nil -*- */
#ifndef _AZPROF_LEAK_HPP
#define _AZPROF_LEAK_HPP

/*
 * Copyright 2004-2011 Azul Systems, Inc.  All rights reserved.
 * AZUL PROPRIETARY/CONFIDENTIAL.  Use is subject to license terms.
 */

#include "azprof_cpu.hpp"
#include "azprof_debug.hpp"
#include "azprof_io.hpp"
#include "azprof_os.hpp"
#include <os/memory.h>

namespace azprof {

class LeakSite;
class LeakObj;

// Stats related to an site of a resource allocation or deallocation.
extern int RPC_CNT;

typedef uint64_t program_ctr_t;


class LeakSite {
    friend class Leak;
public:
    enum State {
        UNUSED,
        ALLOCATING
    };

    enum {
        MAX_RPC_CNT = 20
    };

    // Functions used for different filtering modes.
    typedef bool (*Filt)(const LeakSite*);
    static Filt filt_by_id(int id);
    static bool filt0(const LeakSite*);
    static bool filt1(const LeakSite*);
    static bool filt2(const LeakSite*);

    // Comparison value functions for all the different sorting modes.
    typedef uint64_t (*Cmp)(const LeakSite*);
    static Cmp cmp_by_id(int id);
    static uint64_t cmp0(const LeakSite*);
    static uint64_t cmp1(const LeakSite*);
    static uint64_t cmp2(const LeakSite*);
    static uint64_t cmp3(const LeakSite*);
    static uint64_t cmp4(const LeakSite*);
    static uint64_t cmp5(const LeakSite*);
    static uint64_t cmp6(const LeakSite*);
    static uint64_t cmp7(const LeakSite*);

    int state() const {return (int) _state;}

    const program_ctr_t* rpcs() const {return _rpcs;}
    uint64_t alloc_cnt() const {return _alloc_cnt;}
    uint64_t alloc_bytes() const {return _alloc_bytes;}
    uint64_t dealloc_cnt() const {return _dealloc_cnt;}
    uint64_t dealloc_bytes() const {return _dealloc_bytes;}
    uint64_t release_cnt() const {return _release_cnt;}
    uint64_t release_bytes() const {return _release_bytes;}

    inline uint64_t cmp_alloc_bytes() const;
    inline uint64_t cmp_dealloc_bytes() const;

    bool rpcs_eq(uint64_t *rpcs) {
        for (int i = 0; i < RPC_CNT; i++) {
            if (rpcs[i] != _rpcs[i]) return false;
        }
        return true;
    }

    void alloc(uint64_t bytes) {
        AZPROF_ATOMIC_INC(&_alloc_cnt);
        AZPROF_ATOMIC_ADD(bytes, &_alloc_bytes);
    }

    void dealloc(uint64_t bytes) {
        AZPROF_ATOMIC_INC(&_dealloc_cnt);
        AZPROF_ATOMIC_ADD(bytes, &_dealloc_bytes);
    }

    void release(uint64_t bytes) {
        AZPROF_ATOMIC_INC(&_release_cnt);
        AZPROF_ATOMIC_ADD(bytes, &_release_bytes);
    }

    void to_xml(Response*) const;
    void to_txt(Response*) const;

private:
    LeakSite();
    ~LeakSite();

    uint64_t _state;
    program_ctr_t _rpcs[MAX_RPC_CNT];
    uint64_t _alloc_cnt;
    uint64_t _alloc_bytes;
    uint64_t _dealloc_cnt;
    uint64_t _dealloc_bytes;
    uint64_t _release_cnt;
    uint64_t _release_bytes;
};

// Handle to an allocated object used to find its allocation site when it is
// freed.
class LeakObj {
    friend class Leak;
public:
    enum State {
        UNUSED,
        ALLOCATING,
        DEALLOCATING,
        DEALLOCATED
    };

    int state() const {return (int) _state;}
    uint64_t id() const {return _id;}
    LeakSite* site() {return _site;}
    uint64_t bytes() const {return _bytes;}

private:
    uint64_t _state;
    uint64_t _id;
    LeakSite *_site;
    uint64_t _bytes;
};

// Leak detection interface for memory and kernel resources.
class Leak {
public:
    enum Type {
        MALLOC = 4,
        FIRST_TYPE = MALLOC,
        MMAP_2M = MALLOC+1,
        MMAP_4K = MMAP_2M+1,
        LAST_TYPE = MMAP_4K,

        ALL_TYPES = LAST_TYPE+1,
        TYPE_CNT = LAST_TYPE - FIRST_TYPE + 1
    };

    static bool set_rpc_cnt(int cnt);
    static bool is_enabled() {return _enabled;}
    static void set_enabled(bool);
    static void register_hooks();
    static void unregister_hooks();
    static void force_stop();
    static bool hasStopped() {return _fstopped;}
    static void reset();

    static Leak* get() {
        if (_enabled) {
            Atomic::membar();
            return _leak;
        } else {
            return NULL;
        }
    }

    static bool is_valid_ptr(void *p) {return (((intptr_t) p) & 7) == 0;}
    static uint64_t ptr_to_id(void *p) {return (uint64_t) p;}
    static uint64_t large_pg_to_id(uint64_t p) {return p >> 8;}
    static uint64_t small_pg_to_id(uint64_t p) {return p >> 7;}

    static void malloc(size_t size, void* addr, intptr_t site_pc, intptr_t frame_ptr);
    static void free(void* addr, intptr_t site_pc, intptr_t frame_ptr);
    static void realloc(void* ptr, size_t size, void* addr, intptr_t site_pc, intptr_t frame_ptr);

    static int name_to_type(const char *name);
    static const char* type_to_name(int type);

    void sites_to_xml(
        Response *res, int type, int show, int sort, int start, int stride
    ) const;
    void sites_to_txt(Response *res, int type, int show, int sort) const;

    static LeakSite* site_by_id(int id) {
        return &_leak->_sites[id];
    }

    LeakSite* current_site(int type, intptr_t site_pc, intptr_t site_fp);
    void alloc(
        int type, uint64_t id, LeakSite *site, uint64_t bytes
    );
    void dealloc(int type, uint64_t id);
    void release(int type, uint64_t id);

private:
    // This is a good source of prime numbers.
    //     http://www.rsok.com/~jrm/printprimes.html
    // FIXME - figure out good values to handle stacks 15-deep
    static const uint32_t SITE_CAP = 1048573;
    static const uint64_t OBJ_CAP = 32452867;

    static SpinLock _lock;
    static volatile bool _enabled;
    static Leak* volatile _leak;
    static volatile bool _fstopped;

    LeakSite _sites[SITE_CAP];
    LeakObj _objs[OBJ_CAP];
    uint64_t _site_cnt;
    uint64_t _dealloc_miss_cnt[TYPE_CNT];
    uint64_t _release_miss_cnt[TYPE_CNT];
};

inline uint64_t LeakSite::cmp_alloc_bytes() const {
    return alloc_bytes();
}
inline uint64_t LeakSite::cmp_dealloc_bytes() const {
    return dealloc_bytes();
}

} // namespace azprof

#endif // _AZPROF_LEAK_HPP

