/* -*- mode: c++-mode; c-basic-offset: 4; indent-tabs-mode: nil -*- */
#ifndef _AZPROF_POSIX_OS_HPP
#define _AZPROF_POSIX_OS_HPP

/*
 * Copyright 2004-2011 Azul Systems, Inc.  All rights reserved.
 * AZUL PROPRIETARY/CONFIDENTIAL.  Use is subject to license terms.
 */

#include "azprof_debug.hpp"
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <sys/poll.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#if defined(AZ_LINUX)
#include <os/thread.h>
#endif // AZ_LINUX

typedef void* address_t;
typedef unsigned char* address;

extern "C" {
typedef void (*_ThreadKeyDestroyFunction)(void*);
void* azprof_Thread_run0(void* arg);
};

namespace azprof {

class OS {
public:
    static void init();

    ////////////////////////////////////////////////////////////////////////////
    // Time
    ////////////////////////////////////////////////////////////////////////////

    static double sec_per_clk() {return 1.0/CLOCKS_PER_SEC;}
    static uint64_t clock() {
        struct timeval tv;
        gettimeofday(&tv, NULL);
        return 1000000llu*tv.tv_sec + tv.tv_usec;
    }
    static uint64_t clocks_to_nanos(uint64_t t) {return 1000*t;}
    static uint64_t clocks_to_micros(uint64_t t) {return t;}
    static uint64_t clocks_to_millis(uint64_t t) {return t/1000;}
    static uint64_t nanos_to_clocks(uint64_t t) {return t/1000;}
    static uint64_t micros_to_clocks(uint64_t t) {return t;}
    static uint64_t millis_to_clocks(uint64_t t) {return 1000*t;}
    static time_t clocks_to_time(uint64_t t) {return (time_t) (t/1000000);}

    ////////////////////////////////////////////////////////////////////////////
    // Cpus
    ////////////////////////////////////////////////////////////////////////////

    static uint32_t current_cpu_id() {return 0;} // TODO

    ////////////////////////////////////////////////////////////////////////////
    // Threads
    ////////////////////////////////////////////////////////////////////////////

    typedef pthread_t Thread;
    typedef pthread_key_t ThreadKey;
    typedef _ThreadKeyDestroyFunction ThreadKeyDestroyFunction;

    static Thread current_thread() {return pthread_self();}

#ifdef AZ_LINUX
    static void sleep(uint64_t dt) {
        
        //Fix for bug #29411  
        //poll() can return -1 setting errno to EINTR. For EINTR, ignore and retry.
        //assert on remaining errors. 
        int rc;
         
        while (1) {
            rc = poll(NULL, 0, clocks_to_millis(dt));
            if ( rc == 0) {
                // Expected return code for timeout
                break;
            } else {
                // Ignore EINTR and retry 
                if (errno == EINTR) 
                    continue;
                else
                    // Reaching this place means error could be EFAULT / EINVAL / ENOMEM  
                    azprof_guarantee(0);
            }
        }
    }
#else
    static void sleep(uint64_t dt) {
        azprof_guarantee(poll(NULL, 0, clocks_to_millis(dt)) == 0);
    }
#endif

    static ThreadKey new_thread_key(ThreadKeyDestroyFunction func) {
        ThreadKey key;
        azprof_guarantee(pthread_key_create(&key, func) == 0);
        return key;
    }
    static void delete_thread_key(ThreadKey key) {
        azprof_guarantee(pthread_key_delete(key) == 0);
    }
    static void* get_thread_specific(ThreadKey key) {
        return pthread_getspecific(key);
    }
    static void set_thread_specific(ThreadKey key, void *val) {
        azprof_guarantee(pthread_setspecific(key, val) == 0);
    }

    ////////////////////////////////////////////////////////////////////////////
    // Misc.
    ////////////////////////////////////////////////////////////////////////////

    static void abort() {::abort();}
    static size_t errname(int errnum, char *buf, size_t len);
    static bool dladdr(const address_t& addr, char* buf, const size_t buflen);

    static int gethostname(char *buf, size_t len) {
        return ::gethostname(buf, len);
    }

private:
    OS();
};

class Mutex {
public:
    Mutex(bool init_now = true) : _owner(0) {
        if (init_now) azprof_guarantee(pthread_mutex_init(&_mutex, NULL) == 0);
    }
    ~Mutex() {azprof_guarantee(pthread_mutex_destroy(&_mutex) == 0);}
    bool init() {
        azprof_guarantee(pthread_mutex_init(&_mutex, NULL) == 0);
        return false;
    }

    bool is_mine() {return pthread_equal(pthread_self(), _owner) != 0;}
    void lock() {
        azprof_guarantee(pthread_mutex_lock(&_mutex) == 0);
        _owner = pthread_self();
    }
    void unlock() {
        _owner = 0;
        azprof_guarantee(pthread_mutex_unlock(&_mutex) == 0);
    }

protected:
    pthread_mutex_t _mutex;
    pthread_t       _owner;
};

typedef Mutex SpinLock;

class Thread {
public:
    enum {DEFAULT_STACK_SIZE_LIMIT = 1024*1024};

    Thread(size_t stack_size_limit = DEFAULT_STACK_SIZE_LIMIT) :
        _stack_size_limit(stack_size_limit) {}

    virtual ~Thread() {}

    bool start() {
      int err; 

      pthread_attr_t attr;
      err = pthread_attr_init(&attr);
      azprof_guarantee(err == 0);
      err = pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
      azprof_guarantee(err == 0);
      err = pthread_attr_setstacksize(&attr, _stack_size_limit);
      azprof_guarantee(err == 0);

      void *thread_stack = NULL;
      thread_stack = thread_stack_create();
      azprof_guarantee(thread_stack != NULL);
      azprof_guarantee(((uint64_t) thread_stack) >= __THREAD_STACK_REGION_START_ADDR__ &&
                       ((uint64_t) thread_stack)  < __THREAD_STACK_REGION_END_ADDR__ );
      start_args *th_args = (start_args*) malloc(sizeof (start_args));
      th_args->startFunc  = (Thread_Start_Function) azprof_Thread_run0;
      th_args->args       = (void *) this;
      th_args->stkBase    = (void *) thread_stack;

      err = pthread_create(&_pthread, &attr, (void* (*)(void*)) preRun, (void *) th_args) != 0;
      azprof_guarantee(err == 0);
      pthread_attr_destroy(&attr);
      return err;
    }

    virtual void run() = 0;

private:
    size_t    _stack_size_limit;
    pthread_t _pthread;
};

#if defined(AZ_LINUX)
// (# of zivision session fds:6) + (# of listen fd: 1) + (pipe fd:1)
#define AZPROF_MAX_POLLFD       8

// cmd to add, delete fd from poll set
#define AZPROF_POLLCMD_ADD_FD   1
#define AZPROF_POLLCMD_DEL_FD   2
#define AZPROF_POLLCMD_INTR     3

// cmd structure to add, delete fd from poll set
struct poll_cmd {
    int cmd;
    int fd;
};

class Poll {
public:

    int create() {

        int i;

        _max_fds = AZPROF_MAX_POLLFD;
        memset(&_poll_set[0], 0, sizeof(struct pollfd)*_max_fds);

        // Create a pipe to send commands to add FD.
        if (pipe(_pipe)) {
            Log::info("Poll::create() pipe failed: errno = %d.", errno);
            return -1;
        }
        if ( (_pipe[0] < 0) || (_pipe[1] < 0)) {
            Log::info("Poll::create() failed pipe[0]=%d pipe[1]=%d", _pipe[0], _pipe[1]);
            return -1;
        }

        // Initialize poll_set fds 
        for (i=0; i < _max_fds; i++) {
            _poll_set[i].fd = -1;
        }

        // Add _pipe[0] to handle input commands for add/delete fd.
        _poll_set[0].fd = _pipe[0];
        _poll_set[0].events = POLLIN;
        _poll_set[0].revents = 0;
        _num_fds = 1;

        // Initialize debug counters
        _stats_fd_ready = 0;
        _stats_fd_add_ok = 0;
        _stats_fd_del_ok = 0;
        _stats_fd_add_dup = 0;
        _stats_fd_add_full = 0;
        _stats_fd_del_notfound = 0;
        _stats_pipe_wr_err = 0;
        _stats_pipe_rd_err = 0;
        _stats_poll_timedout = 0;
        _stats_poll_failed = 0;
        _stats_cmd_invalid = 0;
        _stats_cmd_intr = 0;

        return 0;
    }

    int interrupt() {
        // Send INTR  command to the pipe

        struct poll_cmd pollcmd; 
        int wr_size = sizeof(pollcmd);
        int ret;

        pollcmd.cmd = AZPROF_POLLCMD_INTR;
        pollcmd.fd = -1; // don't care
        ret = write(_pipe[1], &pollcmd, wr_size );
        if (ret != wr_size) {
            _stats_pipe_wr_err++;
            // Log::info("Poll::add() write pipe failed errno=%d", errno);
            return -1;
        }

        return 0;
    }

    int add(int fd) {
        // Send ADD command with fd to the pipe

        struct poll_cmd pollcmd; 
        int wr_size = sizeof(pollcmd);
        int ret;

        pollcmd.cmd = AZPROF_POLLCMD_ADD_FD;
        pollcmd.fd = fd;
        ret = write(_pipe[1], &pollcmd, wr_size );
        if (ret != wr_size) {
            _stats_pipe_wr_err++;
            // Log::info("Poll::add() write pipe failed errno=%d", errno);
            return -1;
        }

        return 0;
    }

    int del(int fd) {
        // Send DEL command with fd to the pipe

        struct poll_cmd pollcmd; 
        int wr_size = sizeof(pollcmd);
        int ret;

        pollcmd.cmd = AZPROF_POLLCMD_DEL_FD;
        pollcmd.fd = fd;
        ret = write(_pipe[1], &pollcmd, wr_size );
        if (ret != wr_size) {
            _stats_pipe_wr_err++;
            // Log::info("Poll::del() write pipe failed errno=%d", errno);
            return -1;
        }

        return 0;
    }

    ssize_t wait(int *fdv, size_t nfds, int timeout) {
        while (true) {
            int n;
            int i=0;
            size_t j=0;

            // poll for events
            n = poll(_poll_set, _max_fds, timeout);
            if (n == 0) {
                _stats_poll_timedout++;
                return 0;
            }

            // poll failed (most likely EINTR)
            // caller will retry.
            if (n <  0) {
                _stats_poll_failed++;
                // Log::info("Poll::wait() poll failed errno:%d.", errno);
                return -1;
            }

            // Go thru poll set for ready fd.
            for (i=0; i < _max_fds; i++) {
                
                // skip deleted entry
                if (_poll_set[i].fd == -1) {
                    continue;
                }

                // Receive input event from cmd fd (pipe[0]) 
                if ((_poll_set[i].fd == _pipe[0]) && 
                    (_poll_set[i].revents & POLLIN)) {

                    struct poll_cmd pollcmd;
                    int rd_size = sizeof(pollcmd);
                    int ret;

                    // read pollcmd 
                    ret = read(_pipe[0], &pollcmd, rd_size);
                    if (ret != rd_size) {
                         _stats_pipe_rd_err++;
                         // Log::info("Poll::wait() read failed ret:%d errno:%d.", ret, errno);
                         return -1;
                    }

                    if (pollcmd.cmd == AZPROF_POLLCMD_ADD_FD) {

                        // Handle add fd to poll

                        int pos;
                        int dup_fd = 0;

                        // Check for duplicate add
                        for (pos = 0; pos < _max_fds; pos++) {
                            if (_poll_set[pos].fd == pollcmd.fd) {
                                dup_fd = 1;
                                break;
                            }
                        }
 
                        if (dup_fd) {
                            // add global debug count
                            _stats_fd_add_dup++;
                            // Log::info("Poll::wait() add duplicate fd:%d.", pollcmd.fd);
                            continue;
                        }

                        // Find available pollset entry and add fd
                        for (pos = 0; pos < _max_fds; pos++) {
                            if (_poll_set[pos].fd == -1) {
                                _poll_set[pos].fd = pollcmd.fd;
                                _poll_set[pos].events = POLLIN;
                                _poll_set[pos].revents = 0;
                                _num_fds++;

                                // add global debug count
                                _stats_fd_add_ok++;
                                // Log::info("Poll::wait() added fd:%d pos:%d.", pollcmd.fd, pos);
                                break;
                            }
                        }

                        // if pollset is full, close the fd to prevent leak
                        if (pos >= _max_fds) {
                            _stats_fd_add_full++;
                            // Log::info("Poll::wait() pollset full fd:%d.", pollcmd.fd);
                            ::close(pollcmd.fd);
                            continue;
                        }

                    } else if (pollcmd.cmd == AZPROF_POLLCMD_DEL_FD) {

                        // Handle delete fd from poll

                        int pos;

                        // Remove fd from pollset entry 
                        for (pos = 0; pos < _max_fds; pos++) {
                            if (_poll_set[pos].fd == pollcmd.fd) {
                                _poll_set[pos].fd = -1;
                                _poll_set[pos].events = 0;
                                _poll_set[pos].revents = 0;
                                _num_fds--;

                                _stats_fd_del_ok++;
                                // Log::info("Poll::wait() removed fd:%d.", pollcmd.fd);
                                break;
                            }
                        }

                        azprof_guarantee(pos < _max_fds);
                        if (pos >= _max_fds) {
                            _stats_fd_del_notfound++;
                            // Log::info("Poll::wait() can't delete fd:%d.", pollcmd.fd);
                            continue;
                        }

                    } else if (pollcmd.cmd == AZPROF_POLLCMD_INTR) {
                        // Get a poll.interrupt(), listen fd closed
                        // Caller will retry with a new listen fd.
                        _stats_cmd_intr++;
                        return -1;
                    } else {
                        _stats_cmd_invalid++;
                        Log::info("Poll::wait() bad cmd:%d.", pollcmd.cmd);
                    }

                    azprof_guarantee(pollcmd.cmd == AZPROF_POLLCMD_ADD_FD || 
                                     pollcmd.cmd == AZPROF_POLLCMD_DEL_FD);
                } else if (_poll_set[i].revents & POLLIN) {
                   // An FD is ready.
                   fdv[j++] = _poll_set[i].fd;

                   _stats_fd_ready++;
                   // Log::info("Poll::wait() found ready fd:%d.", _poll_set[i].fd);
                   // remove from pollset to match selector behavior.
                   _poll_set[i].fd = -1;
                   _poll_set[i].events = 0;
                   _poll_set[i].revents = 0;
                   _num_fds--;

                   // Stay within bound of fdv[] array.
                   if (j >= nfds) 
                       break;
                }
            }
            // Return # of fds that have input
            if (j > 0) 
                return ((ssize_t) j);

        }
    }

private:
    struct pollfd _poll_set[AZPROF_MAX_POLLFD];
    int _pipe[2];
    int _num_fds;
    int _max_fds;
    int _stats_fd_ready;
    int _stats_fd_add_ok;
    int _stats_fd_add_dup;
    int _stats_fd_add_full;
    int _stats_fd_del_ok;
    int _stats_fd_del_notfound;
    int _stats_pipe_wr_err;
    int _stats_pipe_rd_err;
    int _stats_poll_timedout;
    int _stats_poll_failed;
    int _stats_cmd_invalid;
    int _stats_cmd_intr;

};
#endif

class Selector {
public:
    Selector() : _maxfd(0) {FD_ZERO(&_fds);}

    int create() {
        // Create a pipe to use to send interrupts when FDs are added. It
        // doubles as a queue for these FDs by writing the FD to the pipe.
        if (pipe(_pipe)) {
            return -1;
        } else {
            if ( _pipe[0] < 0 || _pipe[0] > 1023) 
            {
                return -1;
                //Log::info("Selector::create() called with fd = %d.", _pipe[0]);
            }
            azprof_guarantee(0 <= _pipe[0] && _pipe[0] <= 1023);
            FD_SET(_pipe[0], &_fds);
            _maxfd = _pipe[0];
            return 0;
        }
    }

    int kick() {
        // Enqueue a junk FD for addition by writing it to the pipe.
        return add(-1);
    }

    int add(int fd) {
        // Enqueue the FD for addition by writing it to the pipe.
        int16_t k = (int16_t) fd;
        char buf[2] = {(char) ((k & 0xFF00) >> 8), (char) (k & 0xFF)};
        return (write(_pipe[1], buf, 2) == 2) ? 0 : -1;
    }

    ssize_t wait(int *fdv, size_t nfds, int timeout) {
        while (true) {
            struct timeval tv = {timeout / 1000, 1000 * (timeout % 1000)};
            fd_set rfds = _fds;
            int n = select(_maxfd + 1, &rfds, NULL, NULL, &tv);
            if (n > 0) {
                size_t i = 0;
                for (int fd = 0; (fd <= _maxfd) && (i < nfds); fd++) {
                    if (FD_ISSET(fd, &rfds)) {
                        if (fd == _pipe[0]) {
                            // An FD add has been enqueued by writing it to the
                            // pipe.
                            char buf[2];
                            if (read(_pipe[0], buf, 2) != 2) return -1;
                            int fd = (int) (((int16_t) buf[0]) << 8) | ((int16_t) buf[1]);
                            if (fd >= 0) 
                            {
                                if (fd > 1023) 
                                {
                                  return -1;
                                  // Log::info("Selector::wait() called with pipe's fd value = %d.", fd);
                                }
                                azprof_guarantee(fd <= 1023);
                                FD_SET(fd, &_fds);
                                if (fd > _maxfd) 
                                {  
                                  _maxfd = fd;
                                }
                                if (_maxfd < 0 || _maxfd > 1023) 
                                {
                                  return -1;
                                  // Log::info("Selector::wait() called with _maxfd value = %d.", _maxfd);
                                }
                                azprof_guarantee(0 <= _maxfd && _maxfd <= 1023);
                            }
                        } else {
                            // An FD is ready.
                            fdv[i++] = fd;
                            if (fd < 0|| fd > 1023) 
                            {
                              return -1;
                              // Log::info("Selector::wait() called with existing fd = %d.", fd);
                            }
                            azprof_guarantee(0 <= fd && fd <= 1023);
                            FD_CLR(fd, &_fds);
                        }
                    }
                }
                if (i > 0) {
                    return (ssize_t) i;
                }
            } else if (n == 0) {
                return 0;
            } else {
                return -1;
            }
        }
    }

private:
    fd_set _fds;
    int _pipe[2];
    int _maxfd;
};

}; // namespace azprof

#endif // _AZPROF_POSIX_OS_HPP
