/* A subsystem for authenticating a user of the web GUI
 *
 * --- The basics ---
 *
 * Authentication is either enabled or disabled. When authentication is enabled
 * SSL is used for all HTTP communications and a user name and password are
 * required to establish a session. When authentication is disabled plain-old
 * HTTP is used, a user name or password isn't required, and user sessions
 * aren't tracked (i.e. it's totally stateless).
 *
 * --- User databases ---
 *
 * Two user, group, and password databases are supported: the standard POSIX
 * /etc/passwd database and a local one at "etc/arta/passwd" in our JDK. The
 * entries in the local DB take precedence over those in the POSIX DB. We need
 * two DBs since users may not feel comfortable using their UNIX password to
 * login plus it allows us to add the special "azul" user described later.
 *
 * Using the POSIX DB gives us automatic LDAP integration. On platforms where
 * the POSIX DB is not supported we'll only support the local DB.
 *
 * --- Security configuration ---
 *
 * The file "etc/arta/security.conf" in our JDK controls authentication.
 *
 * To control whether authentication is enabled or disabled "enable
 * authentication" or "disable authentication" is placed in "security.conf".
 * There is an implied "disable authentication" if neither is specified.
 * If multiple "disable/enable authentication" entries exist then the first one
 * takes precedence. Once a VM instance has started with authentication enabled
 * or disabled this setting cannot be changed dynamically.
 *
 * When authentication is enabled a simple rules change is used to grant or
 * deny access to users and groups. The first rule that matches a user trying to
 * login is final and no rules farther down the chain are evaluated. Here's an
 * example.
 *
 * # Assume 'adam', 'bob', and 'chris' are in the group 'dev', 'david' is a
 * # security guru, and 'eric' is a lowly sysadmin
 * grant user adam level 3  # 'adam' is sketchy give him a lower access level
 *                          # than everyone else in 'dev'
 * grant group dev level 6  # give all the other 'dev' users level 6 access
 * grant user david level 9 # 'david' is god give him whatever he wants
 * grant user eric level 1  # let 'eric' poke around to monitor here-and-there
 * deny all                 # boot everyone else out
 *
 * There is an implied "deny all" at the end of the rules chain if it's not
 * specified explicitly.
 */

#ifndef _AZPROF_SECURITY_HPP
#define _AZPROF_SECURITY_HPP

#include "azprof_debug.hpp"
#include "azprof_util.hpp"

#include <string.h>

#ifndef AZPROF_NO_OPENSSL
#include <string>
#include <openssl/ssl.h>

#include <tr1/unordered_map>
using std::tr1::unordered_map;

#define AZPROF_OPENSSL_ONLY(x) x
#else
#define AZPROF_OPENSSL_ONLY(x)
#endif

namespace azprof {

#ifndef AZPROF_NO_IO
class Response;
#endif

// Description of a user from the POSIX or local DB.
class User {
public:
    static const size_t MAX_NAME_LEN = 15;
    static const size_t MAX_PASSWORD_LEN = 15;
    static const size_t ENCRYPTED_PASSWORD_LEN = 13;

    static bool is_valid_name(const char*);
    static bool is_valid_password(const char*);
    static bool is_valid_encrypted_password(const char*);
    static bool is_strong_password(const char *password, Result& result);
    static void encrypt_password(
        const char *input, const char *salt, char *output
    );

    User();
    User(const char *name, const char *encrypted_password);
    User(const User&);
    User& operator=(const User&);

    bool is_valid() const;
    bool has_shadow_password() const;
    const char* name() const {return _name;}
    const char* encrypted_password() const {return _encrypted_password;}

    bool is_password(const char*) const;
    void set_password(const char*);

private:
    char _name[MAX_NAME_LEN+1];
    char _encrypted_password[ENCRYPTED_PASSWORD_LEN+1];
};

// Description of a group from the POSIX or local DB.
class Group {
public:
    static const size_t MAX_NAME_LEN = 15;
    static bool is_valid_name(const char*);

    Group();
    Group(const char *name);
    Group(const Group&);
    Group& operator=(const Group&);

    bool is_valid() const;
    const char* name() const {return _name;}

    const User& member_at(size_t i) const {return _members[i];}
    size_t member_count() const {return _members.size();}
    bool contains(const char *username) const;
    User* add_member(const User&);
    bool delete_member(const User&);

private:
    char _name[MAX_NAME_LEN+1];
    Vector<User> _members;
};

// Database of users and groups.
class UserDb {
public:
    virtual ~UserDb();

    size_t user_count() const {return _users.size();}
    User& user_at(size_t i) {return _users[i];}
    size_t group_count() const {return _groups.size();}
    Group& group_at(size_t i) {return _groups[i];}
    User* add_user(const User&);
    bool delete_user(const char *name);
    Group* add_group(const Group&);
    bool delete_group(const char *name);
    User* add_user_to_group(const char *name, const User&);

    virtual User* current_user();
    virtual User* user_by_name(const char*);
    virtual Group* group_by_name(const char*);

private:
    Vector<User> _users;
    Vector<Group> _groups;

};

// Access to the POSIX user DB (/etc/passwd and /etc/group).
class PosixUserDb : public UserDb {
public:
    PosixUserDb();

    virtual User* current_user();
    virtual User* user_by_name(const char*);
    virtual Group* group_by_name(const char*);

private:
    static const size_t PWBUF_LEN = 4096;
    static const size_t GRBUF_LEN = 4096;
    User _current_user;
};

// List of users imported from '$JAVA_HOME/etc/arta/passwd'.
class LocalUserDb : public UserDb {
public:
    void read(const char *path, Result&);
    int write(const char *path);
};

// A list of multiple UserDb's acting as one.
class UserDbList : public UserDb {
public:
    UserDbList();
    ~UserDbList();

    size_t size() const {return _vec.size();}
    UserDb* at(size_t i) {return _vec[i];}
    void add(UserDb *db) {_vec.add(db);}
    void clear();

    virtual User* current_user();
    virtual User* user_by_name(const char*);
    virtual Group* group_by_name(const char*);

private:
    Vector<UserDb*> _vec;
};

// A description of the privilege level granted to a user or required to access
// certain information.
class Privilege {
public:
    enum {
        NO_ACCESS_LEVEL = 0,
        MIN_ACCESS_LEVEL = 1,
        NO_APP_DATA_ACCESS_LEVEL = MIN_ACCESS_LEVEL,
        APP_DATA_ACCESS_LEVEL = 6,
        MAX_ACCESS_LEVEL = 9,

        AZUL_ENGINEER_FLAG = 0x1,
        ALL_FLAGS = AZUL_ENGINEER_FLAG,
        NO_FLAGS = 0
    };

    static Privilege no_access() {return Privilege(NO_ACCESS_LEVEL);}
    static Privilege min_access() {return Privilege(MIN_ACCESS_LEVEL);}
    static Privilege min_azengr_access() {
        return Privilege(MIN_ACCESS_LEVEL, AZUL_ENGINEER_FLAG);
    }
    static Privilege no_app_data_access() {
        return Privilege(NO_APP_DATA_ACCESS_LEVEL);
    }
    static Privilege no_app_data_azengr_access() {
        return Privilege(NO_APP_DATA_ACCESS_LEVEL, AZUL_ENGINEER_FLAG);
    }
    static Privilege app_data_access() {
        return Privilege(APP_DATA_ACCESS_LEVEL);
    }
    static Privilege app_data_azengr_access() {
        return Privilege(APP_DATA_ACCESS_LEVEL, AZUL_ENGINEER_FLAG);
    }
    static Privilege max_access() {return Privilege(MAX_ACCESS_LEVEL);}
    static Privilege max_azengr_access() {
        return Privilege(MAX_ACCESS_LEVEL, AZUL_ENGINEER_FLAG);
    }

    Privilege(int level = NO_ACCESS_LEVEL, int flags = NO_FLAGS);

    static bool parse(const char*, Privilege&);

    int level() const {return _level;}
    int flags() const {return _flags;}

    bool has_access() const {return level() > NO_ACCESS_LEVEL;}
    bool has_app_data_access() const {return level() >= APP_DATA_ACCESS_LEVEL;}
    bool has_max_access() const {return level() >= MAX_ACCESS_LEVEL;}
    bool has_azul_engineer_access() const {return flags() & AZUL_ENGINEER_FLAG;}

    bool is_sufficient(Privilege) const;

private:
    int _level;
    int _flags;
};

// A rule in a chain used to filter access for users.
class SecurityRule {
public:
    enum Flag {
        FOR_OWNER = 0x1,
        FOR_A_USER = 0x2,
        FOR_A_GROUP = 0x4,
        FOR_ALL_USERS = 0x8
    };
    static const int TARGET_FLAGS =
        FOR_OWNER | FOR_A_USER | FOR_A_GROUP | FOR_ALL_USERS;

    enum ReturnCode {
        APPLIES = 0,
        ERROR = -1,
        DOESNT_APPLY = -2
    };

    SecurityRule();
    SecurityRule(int flags, Privilege privilege, const char *name);

    bool is_valid() const;
    int flags() const {return _flags;}
    bool is_for_owner() const {return _flags & FOR_OWNER;}
    bool is_for_a_user() const {return _flags & FOR_A_USER;}
    bool is_for_a_group() const {return _flags & FOR_A_GROUP;}
    bool is_for_everyone() const {return _flags & FOR_ALL_USERS;}
    const char* name() const {return _name;}
    Privilege privilege() const {return _privilege;}

    // Applies the rule to a given user and returns a result of 0 if access was
    // granted or a negative number otherwise.
    void apply(const char *username, UserDb *db, Result&) const;

private:
    int _flags;
    Privilege _privilege;
    char _name[16];
};

// A chain of rules which create a security policy which is usually pulled from
// '$JAVA_HOME/etc/arta/security.conf'.
class SecurityConf {
public:
    static const unsigned int DEFAULT_SESSION_TIMEOUT = 30*60;

    SecurityConf();

    void read(const char *path, Result&);
    void add(const SecurityRule&);
    void add(int flags, Privilege privilege, const char *name);
    void clear();

    size_t rules_count() const {return _rules.size();}
    const SecurityRule& rule_at(size_t i) const {return _rules[i];}

    bool is_authenticated() const {return _authenticated;}
    void set_authenticated(bool x) {_authenticated = x;}

    unsigned int session_timeout() const {return _session_timeout;}
    void set_session_timeout(unsigned int t) {_session_timeout = t;}

    void apply_rules(const char *username, UserDb*, Result&, Privilege&) const;

private:
    static const size_t MAX_WORD_CNT = 5;

    static ssize_t extract_flags_and_name(
        char **words, size_t word_num, int *flags, char **name
    );

    Vector<SecurityRule> _rules;
    bool _authenticated;
    unsigned int _session_timeout;
};

// Manages authentication of users.
class Authenticator {
public:

    Authenticator(
        const char *authentication,
        const char *authorization,
        const char *resource_dir
    );
    ~Authenticator();

    const char* authentication() const {return _authentication;}
    const char* authorization() const {return _authorization;}
    const char* resource_dir() const {return _resource_dir;}
    SecurityConf& security_conf() {return _security_conf;}
    UserDbList& user_db_lsit() {return _user_db_list;}

    void init(const char *authentication, const char *authorization, Result&);

    Privilege unauthenticated_privilege() const;

    void authenticate(const char *username, const char *password, Result&, Privilege&);

private:
    char *_authentication;
    char *_authorization;
    const char *_resource_dir;
    SecurityConf _security_conf;
    UserDbList _user_db_list;
};

class Session {
public:
    enum Flag {
        TEMPORARY = 0x1,
        LOGGED_IN = 0x2,
        AUTHENTICATED = 0x4
    };
    static const int ALL_FLAGS =
        TEMPORARY |
        LOGGED_IN |
        AUTHENTICATED;

#ifndef AZPROF_NO_OPENSSL
    static Session* get(const char* session_id);
#endif
    static Session* tmp(int flags, Privilege privilege) {
	return new Session(flags | TEMPORARY, "", privilege);
    }

    Session(int flags, const char *username, Privilege);

    bool is_temporary() const {return _flags & TEMPORARY;}
    bool is_logged_in() const {return _flags & LOGGED_IN;}
    bool is_authenticated() const {return _flags & AUTHENTICATED;}

    bool has_azul_engineer_access() const {
        return _privilege.has_azul_engineer_access();
    }

    int flags() const {return _flags;}
    void flags(const int __flags) {_flags = __flags;}

    const char* username() const;
    int username(const char*);

    Privilege privilege() const {return _privilege;}

    void authenticated(const char *username, Privilege privilege);
    void authenticated() { _flags |= AUTHENTICATED; }
    void logout();

#ifndef AZPROF_NO_OPENSSL
    const char* session_id() const { 
	return _session_id.empty() ? NULL : _session_id.c_str(); 
    }
#endif

#ifndef AZPROF_NO_IO
    void to_xml(Response *res);
#endif

private:

#ifndef AZPROF_NO_OPENSSL
    static const std::string generate_session_id();

    typedef unordered_map<std::string, Session*> session_map_t;
    static session_map_t session_map;

    const std::string _session_id;
#endif 

    int _flags;
    char _username[User::MAX_NAME_LEN+1];
    Privilege _privilege;
};

};

#endif // _AZPROF_SECURITY_HPP
