#ifndef _AZPROF_SERVLETS_HPP
#define _AZPROF_SERVLETS_HPP

#ifndef AZPROF_NO_EVENTS
#include "azprof_event.hpp"
#endif

#include "azprof_io.hpp"
#include "azprof_servlets.hpp"

namespace azprof {

////////////////////////////////////////////////////////////////////////////////
// ServletSet
////////////////////////////////////////////////////////////////////////////////

// A set of servlets hashed by their URI path.
class ServletSet : public Servlet {
public:
    static const size_t MAX_SERVLET_COUNT = 256;

    ServletSet();

    size_t size() const;
    Servlet* get(const char*) const;
    Servlet* add(const char*, Servlet*);

    virtual void service(Request*, Response*);

private:
    size_t find_entry_with_mutex(const char*) const;

    class Entry {
    public:
        Entry() : path(NULL), servlet(NULL) {}
        const char *path;
        Servlet *servlet;
    };

    Entry _entries[MAX_SERVLET_COUNT];
    size_t _size;
    Mutex _mutex;
};

////////////////////////////////////////////////////////////////////////////////
// XmlServlet
////////////////////////////////////////////////////////////////////////////////

// A servlet which sets the content-type and writes the header for dynamically
// generated XML.
class XmlServlet : public Servlet {
public:
    XmlServlet(Servlet*);
    Servlet* servlet() {return _servlet;}
    virtual void service(Request*, Response*);
private:
    Servlet *_servlet;
};


// Servlet which spits up a compile-time constant string.
class StringServlet : public Servlet {
public:
    StringServlet(const char *str, const char *content_type);
    virtual void service(Request*, Response*);
private:
    const char *_str;
    size_t _strlen;
    const char *_content_type;
};

////////////////////////////////////////////////////////////////////////////////
// StatelessServlet
////////////////////////////////////////////////////////////////////////////////

// A servlet which is implemented as a stateless callback function.
class StatelessServlet : public Servlet {
public:
    typedef void (*Function)(Request*, Response*);
    StatelessServlet(int flags, Privilege required_privilege, Function);
    virtual void service(Request*, Response*);
private:
    Function _callback;
};

////////////////////////////////////////////////////////////////////////////////
// Process config servlets
////////////////////////////////////////////////////////////////////////////////

class ProcessConfigurationServlet : public Servlet {
public:
    ProcessConfigurationServlet();
    virtual void service(Request*, Response*);
};

class EnvironmentVariablesServlet : public Servlet {
public:
    EnvironmentVariablesServlet();
    virtual void service(Request*, Response*);
};

////////////////////////////////////////////////////////////////////////////////
// Operation profile servlets
////////////////////////////////////////////////////////////////////////////////

#ifndef AZPROF_NO_EVENTS

class OperationProfileServlet : public Servlet {
public:
    OperationProfileServlet(int flags, PerThreadOperationProfile*);
    PerThreadOperationProfile* per_thread_operation_profile() {
        return _per_thread_operation_profile;
    }
private:
    PerThreadOperationProfile *_per_thread_operation_profile;
};

class OperationProfileThreadListServlet : public OperationProfileServlet {
public:
    OperationProfileThreadListServlet(
        const char *title, PerThreadOperationProfile*
    );
    const char* title() const {return _title;}
    virtual void service(Request*, Response*);
private:
    const char *_title;
};

class OperationProfileXmlServlet : public OperationProfileServlet {
public:
    OperationProfileXmlServlet(
        const char *title, const char *csv_href, const char *reset_href,
        PerThreadOperationProfile*
    );
    const char* title() const {return _title;}
    const char* csv_href() const {return _csv_href;}
    const char* reset_href() const {return _reset_href;}
    virtual void service(Request*, Response*);
private:
    const char *_title;
    const char *_csv_href;
    const char *_reset_href;
};

class OperationProfileCsvServlet : public OperationProfileServlet {
public:
    OperationProfileCsvServlet(PerThreadOperationProfile*);
    virtual void service(Request*, Response*);
};

class ResetOperationProfileServlet : public OperationProfileServlet {
public:
    ResetOperationProfileServlet(PerThreadOperationProfile*);
    virtual void service(Request*, Response*);
};

#endif // AZPROF_NO_EVENTS

////////////////////////////////////////////////////////////////////////////////
// Event tracing servlets
////////////////////////////////////////////////////////////////////////////////

#ifndef AZPROF_NO_EVENTS

class EventTraceSettingsServlet : public Servlet {
public:
    EventTraceSettingsServlet();
    virtual void pre_service(Request*, Response*);
    virtual void service(Request*, Response*);
};

class EventTraceBufferXmlServlet : public Servlet {
public:
    static const size_t MAX_TMP_EVENT_SIZE = 8*1024;
    EventTraceBufferXmlServlet();
    virtual void service(Request*, Response*);
};

class EventTraceBufferTxtServlet : public Servlet {
public:
    static const size_t MAX_TMP_EVENT_SIZE = 8*1024;
    EventTraceBufferTxtServlet();
    virtual void service(Request*, Response*);
};

#endif // AZPROF_NO_EVENTS

////////////////////////////////////////////////////////////////////////////////
// Leak detection sevlets
////////////////////////////////////////////////////////////////////////////////

#ifndef AZPROF_NO_LEAK_DETECTION

class EnableLeakServlet : public Servlet {
public:
    EnableLeakServlet();
    virtual void service(Request*, Response*);
};

class ResetLeakServlet : public Servlet {
public:
    ResetLeakServlet();
    virtual void service(Request*, Response*);
};

class LeakSitesXmlServlet : public Servlet {
public:
    LeakSitesXmlServlet();
    virtual void service(Request*, Response*);
};

class LeakSitesTxtServlet : public Servlet {
public:
    LeakSitesTxtServlet();
    virtual void service(Request*, Response*);
};

#endif // AZPROF_NO_LEAK_DETECTION

////////////////////////////////////////////////////////////////////////////////
// Abort
////////////////////////////////////////////////////////////////////////////////

class AbortServlet : public Servlet {
public:
    AbortServlet();
    virtual void service(Request*, Response*);
};

}; // namespace azprof

#endif // _AZPROF_SERVLETS_HPP
