#ifndef _AZPROF_STATS
#define _AZPROF_STATS

#include "azprof_debug.hpp"
#include "azprof_os.hpp"

namespace azprof {

template<class T> class StatSample {
public:
    StatSample(const T& __value, uint64_t __timestamp) :
        _value(__value), _timestamp(__timestamp) {}

    const T& value() const {return _value;}
    uint64_t timestamp() const {return _timestamp;}

private:
    T _value;
    uint64_t _timestamp;
};

// Ring buffer of periodic samples of a statistic. A lock guards between threads
// adding new samples and copying out all the current samples to analyze.
template<class T> class StatBuffer {
public:
    StatBuffer(size_t __capacity) :
        _samples(new StatSample<T>[__capacity]),
        _position(0),
        _length(0),
        _capacity(__capacity)
    {
        azprof_assert(is_power_of_2(__capacity));
    }
    ~StatBuffer() {delete _samples;}

    const StatSample<T>& sample_at(size_t i) const {
        azprof_assert(i < length());
        return _samples[(_position + i) & (_capacity-1)];
    }
    size_t length() const {return _length;}
    size_t capacity() const {return _capacity;}

    void add(const T& value) {
        WithSpinLock wsl(_lock);
        _samples[_position] = StatSample(value, OS::clock());
        _position = (_position + 1) & (_capacity - 1);
        if (_length < _capacity) ++_length;
    }

    void copy(StatBuffer& sb);
        WithSpinLock wsl(_lock);
        azprof_assert(sb._capacity == _capacity);
        memcpy(sb._samples, _samples, sizeof(StatSample<T>) * _capacity);
        sb._position = _position;
        sb._length = _length;
    }

private:
    SpinLock _lock;
    StatSample<T> *_samples;
    size_t _position;
    size_t _length;
    size_t _capacity;
};

class StatValue {
public:
    StatValue(double __value, uint64_t __timestamp) :
        _value(__value),
        _timestamp(__timestamp) {}

    double value() const {return _value;}
    uint64_t timestamp() const {return _timestamp;}

private:
    double _value;
    uint64_t _timestamp;
};

class StatPlotter {
public:
    virtual ~StatPlotter() {}
    virtual void plot(Vector<StatValue>&) = 0;
};

}; // namespace azprof

#endif // _AZPROF_STATS
