#ifndef _AZPROF_UTIL_HPP
#define _AZPROF_UTIL_HPP

#include "azprof_cpu.hpp"
#include "azprof_debug.hpp"
#include "azprof_memory.hpp"
#include "azprof_os.hpp"

#include "os/utilities.h"

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

namespace azprof {

////////////////////////////////////////////////////////////////////////////////
// Macros
////////////////////////////////////////////////////////////////////////////////

#define ARTA_NAME "Azul ZVision"
#define ARTA_ABBREVIATION "ARTA"

#ifndef MIN
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a, b) ((a) >= (b) ? (a) : (b))
#endif

#define AZPROF_CAS(val, dst, cmpval) \
    Atomic::cas((val), (volatile Atomic::Int*) (dst), (cmpval))

#define AZPROF_ATOMIC_ADD(k, p) Atomic::add(k, (volatile Atomic::Int*) (p))
#define AZPROF_ATOMIC_SUB(k, p) Atomic::sub(k, (volatile Atomic::Int*) (p))

#define AZPROF_ATOMIC_INC(p) Atomic::inc((volatile Atomic::Int*) (p))
#define AZPROF_ATOMIC_DEC(p) Atomic::dec((volatile Atomic::Int*) (p))

////////////////////////////////////////////////////////////////////////////////
// Sorting
////////////////////////////////////////////////////////////////////////////////

static inline int intcmpa(uint64_t a, uint64_t b) {
    return (a < b) ? -1 : ((a == b) ? 0 : 1);
}

static inline int intcmpd(uint64_t a, uint64_t b) {
    return (a > b) ? -1 : ((a == b) ? 0 : 1);
}

static inline int dblcmpa(double a, double b) {
    return (a < b) ? -1 : ((a == b) ? 0 : 1);
}

static inline int dblcmpd(double a, double b) {
    return (a > b) ? -1 : ((a == b) ? 0 : 1);
}

extern int strcmpa(const char *a, const char *b);
extern int strcmpd(const char *a, const char *b);

extern int sort_column(const char *param, int default_column = 0);
extern int sort_order(const char *param, int default_order = -1);

////////////////////////////////////////////////////////////////////////////////
// Bit magic
////////////////////////////////////////////////////////////////////////////////

extern uint32_t next_power_of_2(uint32_t);
extern bool population_count(uint32_t);

inline bool is_power_of_2(uint32_t n) {
    return (n & (n-1)) == 0;
}

static inline uint32_t nbits32(int n) {return ((1u << n) - 1);}
static inline uint64_t nbits64(int n) {return ((1llu << n) - 1);}

inline int32_t sext(int bits, uint32_t val) {
    return ((int32_t) val << (32 - bits)) >> (32 - bits);
}

inline void set_flag_value(int& flags, int flag, bool x) {
    if (x) {
        flags |= flag;
    } else {
        flags &= ~flag;
    }
}

////////////////////////////////////////////////////////////////////////////////
// Strings
////////////////////////////////////////////////////////////////////////////////

extern char* join_paths(const char *a, const char *b);
extern void cleanse(char *buf);
extern size_t xml_escape(
    char *dstbuf, size_t dstlen, const char *srcbuf, size_t srclen
);
inline bool is_str_end(char *s) {return s && (*s == '\0');}

extern int64_t parse_size(const char *buf, size_t len);
extern size_t format_size(char *buf, size_t len, uint64_t n);

extern bool matches_ip_addr(const char *s);

////////////////////////////////////////////////////////////////////////////////
// Vector
////////////////////////////////////////////////////////////////////////////////

template<class T> class Vector {
public:
    Vector(size_t __capacity = 8) :
        _capacity(next_power_of_2((uint32_t) __capacity)),
        _size(0)
    {
        _ptr = (T*) Memory::calloc(_capacity, sizeof(T));
    }
    ~Vector() {Memory::free(_ptr);}
    Vector(const Vector& v) :
        _ptr((T*) Memory::calloc(v._capacity, sizeof(T))),
        _capacity(v._capacity),
        _size(v._size)
    {
        memcpy(_ptr, v._ptr, sizeof(T)*v._size);
    }
    Vector& operator=(const Vector& v) {
        azprof_assert(Memory::is_trash(this, sizeof(*this)));
        _ptr = (T*) Memory::calloc(v._capacity, sizeof(T));
        memcpy(_ptr, v._ptr, sizeof(T)*v._size);
        _capacity = v._capacity;
        _size = v._size;
        return *this;
    }

    T* ptr() {return _ptr;}
    size_t size() const {return _size;}
    size_t capacity() const {return _capacity;}

    void resize(size_t n) {
        if (n > _capacity) realloc(n);
        _size = n;
    }

    const T& operator[](size_t i) const {
        azprof_assert(i < _size);
        return _ptr[i];
    }

    T& operator[](size_t i) {
        azprof_assert(i < _size);
        return _ptr[i];
    }

    void add(const T& x) {
        if (_size == _capacity) realloc(_size+1);
        _ptr[_size++] = x;
    }

    void remove(size_t i) {
        azprof_assert(i < _size);
        memmove(_ptr+i, _ptr+i+1, sizeof(T)*(_size-i-1));
        --_size;
    }

    void clear() {
        _size = 0;
    }

    void realloc(size_t cap) {
        size_t newcap = next_power_of_2((uint32_t) cap);
        _ptr = (T*) Memory::realloc(
            _ptr, sizeof(T)*_capacity, sizeof(T)*newcap
        );
        _capacity = newcap;
    }

private:
    T *_ptr;
    size_t _capacity;
    size_t _size;
};

////////////////////////////////////////////////////////////////////////////////
// Fixed-size bit sets
////////////////////////////////////////////////////////////////////////////////

class BitSet {
public:
    BitSet(size_t __capacity) : _capacity(__capacity), _size(0) {
        _element_count = (__capacity/ELEMENT_BITS)+1;
        _elements = (element_t*) Memory::calloc(_element_count, ELEMENT_SIZE);
        memset(_elements, 0, ELEMENT_SIZE*_element_count);
    }

    BitSet& operator=(const BitSet& set) {
        azprof_assert(set._capacity == _capacity);
        memcpy(_elements, set._elements, sizeof(element_t)*_element_count);
        return *this;
    }

    size_t capacity() const {return _capacity;}
    bool is_empty() const {return _size == 0;}
    size_t size() const {return _size;}

    bool contains(size_t k) const {
        azprof_assert(k < _capacity);
        return (
            _elements[k >> ELEMENT_SHIFT] & (1llu << (k & ELEMENT_MASK))) != 0;
    }

    void add(size_t k) {
        if (!contains(k)) {
            _elements[k >> ELEMENT_SHIFT] |= (1llu << (k & ELEMENT_MASK));
            ++_size;
        }
    }

    void add_all(const BitSet& set) {
        azprof_assert(set._capacity == _capacity);
        for (size_t k = 0; k < _capacity; k++) {
            if (set.contains(k)) add(k);
        }
    }

    void remove(size_t k) {
        if (contains(k)) {
            azprof_assert(size() > 0);
            _elements[k >> ELEMENT_SHIFT] &= ~(1llu << (k & ELEMENT_MASK));
            --_size;
        }
    }

    void remove_all(const BitSet& set) {
        azprof_assert(set._capacity == _capacity);
        for (size_t k = 0; k < _capacity; k++) {
            if (set.contains(k)) remove(k);
        }
    }

    void universal() {
        memset(_elements, 0xFF, ELEMENT_SIZE*_element_count);
        _size = _capacity;
    }

    void empty() {
        memset(_elements, 0, ELEMENT_SIZE*_element_count);
        _size = 0;
    }

private:
    typedef uint64_t element_t;
    static const size_t ELEMENT_SIZE = sizeof(element_t);
    static const size_t ELEMENT_BITS = 8*ELEMENT_SIZE;
    static const size_t ELEMENT_SHIFT = 6;
    static const size_t ELEMENT_MASK = ELEMENT_BITS-1;

    element_t *_elements;
    size_t _element_count;
    size_t _capacity;
    size_t _size;
};

////////////////////////////////////////////////////////////////////////////////
// RingBuffer
////////////////////////////////////////////////////////////////////////////////

// A statically sized ring buffer.
// TODO: Unify the shared code between StaticRingBuffer and DynamicRingBuffer.
class StaticRingBuffer {
public:
    StaticRingBuffer() : _ptr(NULL), _pos(0), _len(0), _cap(0) {}
    ~StaticRingBuffer() {Memory::free(ptr(), cap());}

    bool alloc(size_t cap);
    bool alloc_copy(const StaticRingBuffer&);
    void free();

    char* ptr() {return _ptr;}
    const char* ptr() const {return _ptr;}
    size_t pos() const {return _pos;}
    size_t len() const {return _len;}
    size_t cap() const {return _cap;}

    size_t avail() const {return cap() - len();}
    bool wrapped() const {return (pos() + len()) > cap();}

    char* head() {return ptr() + pos();}
    char* head(size_t off) {
        azprof_assert(off < len());
        return ptr() + ((pos() + off) & (cap() - 1));
    }
    size_t head_gap() const {return MIN(len(), cap() - pos());}

    char* tail(size_t off = 0) {return ptr() + tail_pos(off);}
    size_t tail_pos(size_t off = 0) const {
        azprof_assert((len() + off) <= cap());
        return ((pos() + off + len()) & (cap() - 1));
    }
    size_t tail_gap() const {return MIN(cap() - len(), cap() - tail_pos());}

    void inc(size_t n = 1) {
        azprof_assert((len() + n) <= cap());
        _len += n;
    }

    void dec(size_t n = 1) {
        azprof_assert(n <= len());
        _pos = (pos() + n) & (cap() - 1);
        _len -= n;
    }

    ssize_t read(char *buf, size_t len);
    ssize_t read8(uint8_t&);
    ssize_t read16(uint16_t&);
    ssize_t read32(uint32_t&);
    ssize_t read64(uint64_t&);
    ssize_t readstr(char *buf, size_t len);

    ssize_t write(const char *buf, size_t len);
    ssize_t write8(uint8_t);
    ssize_t write16(uint16_t);
    ssize_t write32(uint32_t);
    ssize_t write64(uint64_t);
    ssize_t writestr(const char*);

    void clear();

private:
    char *_ptr;
    size_t _pos;
    size_t _len;
    size_t _cap;
};

////////////////////////////////////////////////////////////////////////////////
// DynamicRingBuffer
////////////////////////////////////////////////////////////////////////////////

// A ring buffer which can dynamically grow in capacity if needed. A maximum
// capacity can be imposed though.
class DynamicRingBuffer {
public:
    DynamicRingBuffer(size_t cap, size_t max_cap);
    ~DynamicRingBuffer();

    char* ptr() {return _ptr;}
    const char* ptr() const {return _ptr;}
    size_t pos() const {return _pos;}
    size_t len() const {return _len;}
    size_t cap() const {return _cap;}
    size_t max_cap() const {return _max_cap;}
    size_t avail() const {return _cap - _len;}
    bool wrapped() const {return (_pos + _len) > _cap;}
    bool compacted() const {return tail_gap() == avail();}

    char* head() {return _ptr + _pos;}
    const char* head() const {return _ptr + _pos;}
    size_t head_pos() const {return _pos;}
    size_t head_gap() const {return MIN(_len, _cap - _pos);}

    char* tail() {return _ptr + tail_pos();}
    const char* tail() const {return _ptr + tail_pos();}
    size_t tail_pos() const {return (_pos + _len) & (_cap - 1);}
    size_t tail_gap() const {return MIN(_cap - _len, _cap - tail_pos());}
    void inc(size_t n = 1) {
        azprof_assert((_len + n) <= _cap);
        _len += n;
    }

    void dec(size_t n = 1) {
        azprof_assert(n <= _len);
        _pos = (_pos + n) & (_cap - 1);
        _len -= n;
        if (_len == 0) _pos = 0;
    }

    void write(const char *buf, size_t len);

    bool alloc();
    bool realloc(size_t);
    void compact();

private:
    char *_ptr;
    size_t _pos;
    size_t _len;
    size_t _cap;
    size_t _max_cap;
};

////////////////////////////////////////////////////////////////////////////////
// IORingBuffer
////////////////////////////////////////////////////////////////////////////////

// A ring buffer for characters that gets overwritten as new data arrives.
// Contrary to the other ring buffers there is no read position.
class IORingBuffer {
public:
    IORingBuffer(const char* name, size_t capacity);
    ~IORingBuffer();
    
    Mutex& mutex() { return _mutex; }
    
    const char* name() const { return _name; }
    size_t capacity() const { return _cap; }
    
    size_t read (      char* buf, size_t off, size_t len);
    size_t write(const char* buf, size_t off, size_t len);
    
    void read_fully (      char* buf, size_t len);
    void write_fully(const char* buf, size_t len);
    
    IORingBuffer* next() const { return _next; }
    
    // Global registration of IORingBuffers to be able to find them
    // and display them in ARTA.
    static void init();
    
    static void add   (IORingBuffer *iobuf);
    static void remove(IORingBuffer *iobuf);
    
    static Mutex&  list_mutex() { return *_list_mutex; }
    static IORingBuffer* list() { return _list; }
  
private:
    Mutex          _mutex;
    char*          _ptr;
    const char*    _name;
    size_t         _cap;
    size_t         _pos;
    IORingBuffer * _next;
    
    // Static fields to maintain the registration of IORingBuffers
    static Mutex*        _list_mutex;
    static IORingBuffer* _list;
};

////////////////////////////////////////////////////////////////////////////////
// Auto free/delete/close
////////////////////////////////////////////////////////////////////////////////

class WithPointer {
public:
    WithPointer(void *__x) : _x(__x) {}
    ~WithPointer() {Memory::free(_x);}
private:
    void *_x;
};

template<class T> class WithObject {
public:
    WithObject(T *__x) : _x(__x) {}
    ~WithObject() {delete _x;}
private:
    T *_x;
};

class WithOpenFD {
public:
    WithOpenFD(const char *path, int mode);
    ~WithOpenFD();

    operator int() {return _fd;}

    int read(char *buf, size_t len);
    int read_all(char *buf, size_t len);
    int read_all_at(char *buf, size_t pos, size_t len);

private:
    int _fd;
};

class WithOpenFile {
public:
    WithOpenFile(const char *path, const char *mode) : _f(fopen(path, mode)) {}
    ~WithOpenFile() {if (_f) fclose(_f);}
    operator FILE*() {return _f;}
private:
    FILE *_f;
};

////////////////////////////////////////////////////////////////////////////////
// Reference counting
////////////////////////////////////////////////////////////////////////////////

class ReferenceCounted {
public:
    ReferenceCounted() : _reference_count(0) {}
    virtual ~ReferenceCounted() {}

    Atomic::Int reference_count() const {return _reference_count;}

    void acquire() {
        Atomic::inc(&_reference_count);
    }

    void release() {
        Atomic::Int n = Atomic::dec(&_reference_count);
        azprof_assert(n >= 0);
        if (n == 0) delete this;
    }

private:
    Atomic::Int _reference_count;
};

class WithReferenceCount {
public:
    WithReferenceCount(ReferenceCounted *__refcnt) : _refcnt(__refcnt) {
        __refcnt->acquire();
    }
    ~WithReferenceCount() {_refcnt->release();}

private:
    ReferenceCounted *_refcnt;
};

////////////////////////////////////////////////////////////////////////////////
// Mutexes and spin locks
////////////////////////////////////////////////////////////////////////////////

class WithMutex {
public:
    WithMutex(Mutex& __mutex) : _mutex(__mutex) {_mutex.lock();}
    ~WithMutex() {_mutex.unlock();}
private:
    Mutex& _mutex;
};

class WithSpinLock {
public:
    WithSpinLock(SpinLock& __spin_lock) :
        _spin_lock(__spin_lock)
    {
        _spin_lock.lock();
    }
    ~WithSpinLock() {
        _spin_lock.unlock();
    }
private:
    SpinLock& _spin_lock;
};

////////////////////////////////////////////////////////////////////////////////
// Result
////////////////////////////////////////////////////////////////////////////////

class Result {
public:
    Result();
    Result(int value);
    Result(int value, const char *fmt, ...);
    Result(Result&);
    Result& operator=(const Result&);

    int value() const {return _value;}
    const char* message() const {return _message;}

    Result& set(int value);
    Result& vset(int value, const char *fmt, va_list);
    Result& set(int value, const char *fmt, ...);

private:
    int _value;
    char _message[256];
};

////////////////////////////////////////////////////////////////////////////////
// Thread locals
////////////////////////////////////////////////////////////////////////////////

// Thread local keys used within azprof and hung off a single libos key.
class ThreadLocal {
public:
    enum {
        LOG,
        DEBUG_FLAGS,
        COUNT
    };

    static void init();

    static void* get(int key);
    static void set(int key, void *val);

private:
    static void destroy(void*);
    static OS::ThreadKey _os_key;
};

}; // namespace azprof

#endif // _AZPROF_UTIL_HPP
