// Copyright 2003-2010 Azul Systems, Inc. All rights reserved.
// AZUL Systems PROPRIETARY/CONFIDENTIAL. Use is subject to license terms

// atomic.h - Azul OS system services atomic operations.

// Don't use this for now...
#define _AZUL_ATOMIC_H_ 1
#ifndef _AZUL_ATOMIC_H_

#ifdef __cplusplus
extern "C" {
#endif

// Inline versions of everything that it makes sense to inline:

inline static int64_t atomic_compare_and_exchange_64(int64_t            _exchange_value,
                                                     volatile int64_t*  _dest,
                                                     int64_t            _compare_value) __attribute__((always_inline));

inline static int64_t atomic_compare_and_exchange_u64(uint64_t           _exchange_value,
                                                      volatile uint64_t* _dest,
                                                      uint64_t           _compare_value) __attribute__((always_inline));

inline static int64_t atomic_exchange_64(int64_t _exchange_value, volatile int64_t* _dest)  __attribute__((always_inline));
inline static int64_t atomic_exchange_u64(uint64_t _exchange_value, volatile uint64_t* _dest)  __attribute__((always_inline));

inline static int64_t atomic_add_64(int64_t _delta, volatile int64_t* _dest) __attribute__((always_inline));
inline static int64_t atomic_add_u64(uint64_t _delta, volatile uint64_t* _dest) __attribute__((always_inline));

inline static int64_t atomic_exchange_and_add_64(int64_t _delta, volatile int64_t* _dest) __attribute__((always_inline));
inline static int64_t atomic_exchange_and_add_u64(uint64_t _delta, volatile uint64_t* _dest) __attribute__((always_inline));

inline static int64_t atomic_increment_64(volatile int64_t* _dest) __attribute__((always_inline));
inline static int64_t atomic_increment_u64(volatile uint64_t* _dest) __attribute__((always_inline));

inline static int64_t atomic_decrement_64(volatile int64_t* _dest) __attribute__((always_inline));
inline static int64_t atomic_decrement_u64(volatile uint64_t* _dest) __attribute__((always_inline));

inline static int64_t atomic_add_limit_64(volatile int64_t* ptr, int64_t inc, int64_t limit) __attribute__((always_inline));

// Atomically compare the value stored at dest with compare value, and if equal perform an exchange
// with exchange value.  Returns the previous value stored at dest, which will equal compare_value 
// if the exchange succeeds.

inline static int64_t atomic_compare_and_exchange_64(int64_t _exchange_value, volatile int64_t* _dest, int64_t _compare_value) 
{
    return __sync_val_compare_and_swap(_dest, _compare_value, _exchange_value);
}

inline static int64_t atomic_compare_and_exchange_u64(uint64_t _exchange_value, volatile uint64_t* _dest, uint64_t _compare_value) 
{
    return __sync_val_compare_and_swap(_dest, _compare_value, _exchange_value);
}

// Atomically exchange the value stored at dest with the exchange value.  Returns the previous value.

inline static int64_t atomic_exchange_64(int64_t _exchange_value, volatile int64_t* _dest) 
{
    return __sync_lock_test_and_set(_dest,_exchange_value);
}

inline static int64_t atomic_exchange_u64(uint64_t _exchange_value, volatile uint64_t* _dest) 
{
    return __sync_lock_test_and_set(_dest,_exchange_value);
}

// Atomically add delta and the value stored at dest and store the sum back at that address.  Returns the sum.

inline static int64_t atomic_add_64(int64_t _delta, volatile int64_t* _dest) 
{
    return __sync_add_and_fetch(_dest, _delta);
}

inline static int64_t atomic_add_u64(uint64_t _delta, volatile uint64_t* _dest) 
{
    return __sync_add_and_fetch(_dest, _delta);
}

// Atomically add delta and the value stored at dest and store the sum back at that address.  Returns the previous value.

inline static int64_t atomic_exchange_and_add_64(int64_t _delta, volatile int64_t* _dest) 
{
    return __sync_fetch_and_add(_dest, _delta);
}

inline static int64_t atomic_exchange_and_add_u64(uint64_t _delta, volatile uint64_t* _dest) 
{
    return __sync_fetch_and_add(_dest, _delta);
}

// Atomically increment the value at dest.  Returns the incremented value.

inline static int64_t atomic_increment_64(volatile int64_t* _dest) 
{
    return __sync_add_and_fetch(_dest, 1);
}

inline static int64_t atomic_increment_u64(volatile uint64_t* _dest) 
{
    return __sync_add_and_fetch(_dest, 1);
}

// Atomically decrement the value at dest.  Returns the decremented value.

inline static int64_t atomic_decrement_64(volatile int64_t* _dest) 
{
    return __sync_add_and_fetch(_dest, -1);
}

inline static int64_t atomic_decrement_u64(volatile uint64_t* _dest) 
{
    return __sync_add_and_fetch(_dest, -1);
}


/* If (old value exceeds limit), returns (limit) without doing the increment */
// convert to inline from macro in atomic.h
inline static int64_t atomic_add_limit_64(volatile int64_t* ptr, int64_t inc, int64_t limit)
{ 
    int64_t __old, __new, __res = *ptr;

    do 
    { 
		    __old = __res; 

        // current thread lost the race.
        if ((inc > 0 && (uint64_t)__old >= (uint64_t)limit) ||
            (inc < 0 &&  (uint64_t)__old <= (uint64_t)limit)) {
            return limit;
        }

        __new = __old + inc; 
        __res = atomic_compare_and_exchange_64(__new,  (ptr), __old); 

	  } while (__res != __old); 

    return __old;
}


// Atomically compare the value stored at dest with compare value, and if equal perform an exchange
// with exchange value.  Returns the previous value stored at dest, which will equal compare_value if 
// the exchange succeeds.

static inline int32_t atomic_compare_and_exchange_32(int32_t  _exchange_value, volatile int32_t* _dest, int32_t  _compare_value) 
{
    return __sync_val_compare_and_swap(_dest, _compare_value, _exchange_value);
}

// Atomically exchange the value stored at dest with the exchange value.
// Returns the previous value.

static inline int32_t atomic_exchange_32(int32_t  _exchange_value, volatile int32_t* _dest) 
{
    // This builtin, as described by Intel, is not a traditional test-and-set operation,
    // but rather an atomic exchange operation. It writes value into *ptr, and returns
    // the previous contents of *ptr.
    // Many targets have only minimal support for such locks, and do not support a full
    // exchange operation. In this case, a target may support reduced functionality here
    // by which the only valid value to store is the immediate constant 1. The exact value
    // actually stored in *ptr is implementation defined.
    // This builtin is not a full barrier, but rather an acquire barrier. This means that
    // references after the builtin cannot move to (or be speculated to) before the 
    // builtin, but previous memory stores may not be globally visible yet, and previous
    // memory loads may not yet be satisfied.
    return __sync_lock_test_and_set(_dest, _exchange_value);
}


// Atomically add delta and the value stored at dest and store the sum back at that address.  Returns the sum.
extern int32_t atomic_add_32(int32_t  _delta, volatile int32_t* _dest);

// Atomically increment the value at dest.  Returns the incremented value.
inline static int32_t atomic_increment_32(volatile int32_t* _dest) 
{
	return atomic_add_32(1L, _dest);
}

// Atomically decrement the value at dest.  Returns the decremented value.
inline static int32_t atomic_decrement_32(volatile int32_t* _dest) 
{
	return atomic_add_32(-1L, _dest);
}

// Entry points below are for temporary historical backwards-compatibility.

// Atomically exchange long word with another if the old value compares equal to compare_value.  Returns: old value
extern int64_t atomic_compare_and_exchange_long(int64_t  _exchange_value, volatile int64_t* _dest, int64_t  _compare_value);

// Atomically exchange (int) word with another if the old value compares equal to compare_value.  Returns: old value
extern int32_t atomic_compare_and_exchange_word(int32_t  _exchange_value, volatile int32_t* _dest, int32_t  _compare_value);

// Atomically exchange (int) word with another.  Returns: old value
extern int32_t atomic_exchange_word(int32_t  _exchange_value, volatile int32_t* _dest);

// Atomically add some delta to value at the given address and store the result back at that address.
// Returns: old value. Note: This differs from the equivalent (internal) HotSpot function semantics.
inline static int64_t atomic_add(volatile int64_t* _dest, int64_t _delta) 
{
    return atomic_exchange_and_add_64(_delta, _dest);
}

inline static int64_t atomic_increment(volatile int64_t* _dest) 
{
    return atomic_add_64(1LL, _dest);
}

inline static int64_t atomic_decrement(volatile int64_t* _dest) 
{
    return atomic_add_64(-1LL, _dest);
}

// Memory barrier
inline static void atomic_fence_all(void) __attribute__((always_inline));

inline static void atomic_fence_all(void) { __asm__ __volatile__ ("mfence":::"memory"); }

#define FENCEALL() atomic_fence_all()
#define LOCK_FENCE()
#define UNLOCK_FENCE()

// Prefetches
inline static void prefetch_shared(void *addr) 
{
  __asm__ __volatile__ ("prefetcht0 %0" : /* outputs */ : "m" (*(unsigned long*)addr));
}


#ifdef NOT_YET_IMPLEMENTED
// As yet unimplemented prefetch
inline static void prefetch_exclusive(void *addr __attribute__ ((unused))) 
{
	unimplemented("prefetch_exclusive");
}
#endif // NOT_YET_IMPLEMENTED

#define atomic_read(x) (*(x))
#define atomic_cas(a, b, c) CAS((a), (b), (c))

/*
 * compare_and_swap()
 */
inline static uint64_t CAS(int64_t* dst, int64_t old, int64_t newx)
{
    return __sync_val_compare_and_swap(dst, old, newx);
}

inline static uint64_t UCAS(volatile uint64_t* dst, uint64_t old, uint64_t newx)
{
    return __sync_val_compare_and_swap(dst, old, newx);
}

#define ATOMIC_64_OP(_dest, _val, _op) \
do \
{ \
    uint64_t oldv, ret; \
    do \
    { \
        oldv = *(_dest); \
        ret = atomic_compare_and_exchange_64((oldv _op (_val)), (_dest), oldv); \
    } while (ret != oldv); \
} while ( 0)

#define ATOMIC_64_OP_OR(_dest, _val) ATOMIC_64_OP(_dest, _val, |) 

#ifdef __cplusplus
} // extern "C"
#endif // __cplusplus

#endif // _AZUL_ATOMIC_H_
