// Copyright 2003 Azul Systems, Inc. All rights reserved.
// AZUL Systems PROPRIETARY/CONFIDENTIAL. Use is subject to license terms

// exceptions.h - Exception handling subsystem for generic Linux  OS Services

#ifndef _OS_EXCEPTIONS_H_
#define _OS_EXCEPTIONS_H_ 1

#include <signal.h>
#include <ucontext.h>

#include <os/types.h>


#ifdef __cplusplus
extern "C" {
#endif

// User-level chained signal handlers.
//
// Chained signal handlers can return one of three results:
//
//   AZUL_SIGNAL_NOT_HANDLED:      This handler didn't process the signal, invoke the next chained signal handler.
//   AZUL_SIGNAL_RESUME_EXECUTION: The handler processed the signal, resume at the point the signal occurred. 
//   AZUL_SIGNAL_SKIP_INSTRUCTION: The handler processed the signal, resume at the instruction following where the signal occurred.
//   AZUL_SIGNAL_JUMP_TO_ADDR:     The handler has satisfied the exception and has also stored the address of a 
//                                 function to execute (on the main stack) once the signal handler returns.
//   AZUL_SIGNAL_ABORTING:         The handler knows it's dying, call the unexpected signal handler instead.
//   AZUL_SIGNAL_ERROR:            The handler has encountered a pathological error, going futher is pointless, give up.
//   address_t:                    The handler processed the signal, resume at the returned address.
//
// TODO: AZUL_SIGNAL_SKIP_INSTRUCTION is not implemented: This would return at the instruction after the one that
//       triggered the signal.  To do this, I'd need to parse the x86 ASM to figure out the IP of the next
//       instruction along.
//
// Handler order:
//
//   Multiple handlers can be registered for a signal.  (Up to AZUL_MAX_CHAINED_HANDLERS)
//   In order to provide deterministic ordering of which handlers are called,
//   each handler is registered with an order #.  Handlers are invoked, going from
//   order 0 to AZUL_MAX_CHAINED_HANDLERS-1, until a handler returns something other
//   than AZUL_SIGNAL_NOT_HANDLED.
//
//   It is an error to register a handler with a particular order # if there is already a handler
//   registered with that order #.  exception_unregister_chained_handler() should be used to
//   unregister a handler before replacing it with a new handler.
//

// Signal Handler function types:
typedef address_t (*azul_chained_handler_t) (int, siginfo_t *, ucontext_t *);
typedef address_t (*azul_direct_handler_t)  (int, siginfo_t *, ucontext_t *);

// The Unexpected Signal Handler type is intentionally done as a different type signature
// so as to be incompatible with the chained or direct signal handlers.  This function
// should only be assigned using the register_unexpected_exception_handler().
typedef void (*azul_unexpected_exception_handler_t) ( ucontext_t*, siginfo_t*, int );

// The VMError Reporting Handler type yet another type of signal handler, called only
// when we've already encountered an unexpected signal handler and need to register
// the error reporting mechanism.
typedef void (*azul_vmerror_reporting_handler_t) ();

// Handler return values:
#define AZUL_SIGNAL_NOT_HANDLED      ((address_t)-1)
#define AZUL_SIGNAL_RESUME_EXECUTION ((address_t)-2)
#define AZUL_SIGNAL_SKIP_INSTRUCTION ((address_t)-3)
#define AZUL_SIGNAL_JUMP_TO_ADDR     ((address_t)-4)
#define AZUL_SIGNAL_ABORTING         ((address_t)-5)
#define AZUL_SIGNAL_ERROR            ((address_t)-6)

// Maximum number of handlers that may be registered for a particular signal.
// The maximum order # that can be specified when registering a handler
// is AZUL_MAX_CHAINED_HANDLERS-1.
#define AZUL_MAX_CHAINED_HANDLERS    (5)

// The exception handler order is defined a priori.  Do not refer to these
// defines directly.  Instead, use the signal specific macros defined below.
// Do not change the first entry in this list, nor assign it a value other
// than zero!!  Make sure each successive entry in the list is one greater
// than the next, and that the last entry equals AZUL_MAX_CHAINED_HANDLERS-1.
#define AZUL_EXCPT_HNDLR_FIRST       (0)
#define AZUL_EXCPT_HNDLR_SECOND      (AZUL_EXCPT_HNDLR_FIRST + 1)     
#define AZUL_EXCPT_HNDLR_THIRD       (AZUL_EXCPT_HNDLR_SECOND + 1)     
#define AZUL_EXCPT_HNDLR_FORTH       (AZUL_EXCPT_HNDLR_THIRD + 1)     
// This must be the last exception handler define.  Do not exceed 
// AZUL_MAX_CHAINED_HANDLERS-1, This define must equal it.
#define AZUL_EXCPT_HNDLR_LAST        (AZUL_EXCPT_HNDLR_FORTH + 1)

//*************************************************************************
// For each of the various signal types, define the order of handlers here,
// so that it is clear to anyone what gets called when.  Do not use the macro
// definitions above when calling the exception_register_chained_handler().
// If you decide to add another entry, make sure you follow the naming
// conventions shown here.
//*************************************************************************

// SIGSEGV chained signal handler - the order of these matters.
#define AZUL_SIGSEGV_STACK_EXCPT_HNDLR       (AZUL_EXCPT_HNDLR_FIRST)
#define AZUL_SIGSEGV_JVM_EXCPT_HNDLR         (AZUL_EXCPT_HNDLR_SECOND)
#define AZUL_SIGSEGV_NIO_EXCPT_HNDLR         (AZUL_EXCPT_HNDLR_THIRD)
#define AZUL_SIGSEGV_UNSAFE_EXCPT_HNDLR      (AZUL_EXCPT_HNDLR_FORTH)

// SIGBUS chained signal handler - right now there is only one. If
// you add more, following the naming convention.
#define AZUL_SIGBUS_JVM_EXCPT_HNDLR          (AZUL_EXCPT_HNDLR_FIRST)
#define AZUL_SIGBUS_UNSAFE_EXCPT_HNDLR       (AZUL_EXCPT_HNDLR_FORTH)

// SIGILL chained signal handler - right now there is only one. If
// you add more, following the naming convention.
#define AZUL_SIGILL_JVM_EXCPT_HNDLR          (AZUL_EXCPT_HNDLR_FIRST)

// Robb's special test handler.
#define AZUL_SIGUSR2_EXCPT_HNDLR             (AZUL_EXCPT_HNDLR_FIRST)

//
// User-level (chained) signal handlers register/unregister using these functions:
//
extern sys_return_t exception_register_chained_handler  (int signum, azul_chained_handler_t handler, int order);
extern sys_return_t exception_unregister_chained_handler(int signum, azul_chained_handler_t handler);

// The Signal Handler array has AZUL_MAX_CHAINED_HANDLERS + 1 entries, the
// last entry being reserved for the "direct" handler for the signal.  You
// may not have both a chained signal handler, and a direct signal handler
// simultaneously present.
#define AZUL_MAX_HANDLERS            (AZUL_MAX_CHAINED_HANDLERS + 1)
#define AZUL_DIRECT_HANDLER          (AZUL_MAX_CHAINED_HANDLERS)  // The last offset in the array.

//
// User-level (direct) signal handlers register/unregister using these functions:
//
extern sys_return_t exception_register_direct_handler  (int signum, azul_direct_handler_t handler);
extern sys_return_t exception_unregister_direct_handler(int signum, azul_direct_handler_t handler);

//
// Registration for the global unexpected signal handler.  This should be called
// at JVM init with a pointer to the application's unexpected signal handler function.
// This is what will be called regardless of signal.  It is not legal to place
// your unexpected signal handler function within the chained signal handler list,
// nor it it okay to register a previously registered chained signal handler as
// the unexpected signal handler.
//
void      register_unexpected_exception_handler (azul_unexpected_exception_handler_t  handler); 
void      register_jvm_vmerror_reporting_handler(azul_vmerror_reporting_handler_t     handler); 

#ifdef __cplusplus
} // extern "C"
#endif // __cplusplus
  
#endif // _OS_EXCEPTIONS_H_
