// Copyright 2006 Azul Systems, Inc. All rights reserved.
// AZUL Systems PROPRIETARY/CONFIDENTIAL. Use is subject to license terms
//
// memstats.hpp - Helper class for calculating memory statistics.

#ifndef _OS_MEMSTATS_HPP_
#define _OS_MEMSTATS_HPP_ 1

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <os/memory.h>
#include <aznix/az_allocid.h>

namespace libos {

//-------------------------------------------------------------------------
// Process Accounts
//-------------------------------------------------------------------------

class AccountStats {
public:
    AccountStats (const account_stats_t *account) {
        clear();
        calc(account);
    }

    void clear() {
        allocated_bytes = 0;
        funded_bytes = 0;
        overdraft_bytes = 0;
        footprint_bytes = 0;
        committed_bytes = 0;
        grant_bytes = 0;
        allocated_from_committed_bytes = 0;
    }

    void calc (const account_stats_t *account) {
        // General calculations.
        allocated_bytes = account->ac_allocated;
        funded_bytes = account->ac_balance + account->ac_allocated;
        if (account->ac_balance < 0) {
            overdraft_bytes = -account->ac_balance;
        }
        footprint_bytes = 
            (allocated_bytes > funded_bytes) ?
            allocated_bytes :
            funded_bytes;

        // Fund-specific calculations.
        if (account->ac_fund == AZMM_COMMITTED_FUND) {
            committed_bytes = funded_bytes;

            allocated_from_committed_bytes = 
                (allocated_bytes < committed_bytes) ?
                allocated_bytes :
                committed_bytes;
        }
        if (account->ac_overdraft_fund == AZMM_OVERDRAFT_FUND) {
            grant_bytes = overdraft_bytes;
        }
    }

    uint64_t getAllocatedBytes() const { return allocated_bytes; }
    uint64_t getFundedBytes() const { return funded_bytes; }
    uint64_t getOverdraftBytes() const { return overdraft_bytes; }
    uint64_t getFootprintBytes() const { return footprint_bytes; }
    uint64_t getCommittedBytes() const { return committed_bytes; }
    uint64_t getGrantBytes() const { return grant_bytes; }
    uint64_t getAllocatedFromCommittedBytes() const { return allocated_from_committed_bytes; }

private:
    uint64_t allocated_bytes;
    uint64_t funded_bytes;
    uint64_t overdraft_bytes;
    uint64_t footprint_bytes;
    uint64_t committed_bytes;
    uint64_t grant_bytes;
    uint64_t allocated_from_committed_bytes;
};

class AccountInfo {
public:
    AccountInfo () :
        // Purpose:  Create an accounting object.
        //
        _account_info_allocated (false),
        _account_info (0) {
        clear();
    }

    AccountInfo (const account_info_t *account_info) :
        // Purpose:  Calculate accounting statistics.
        // Params:   account_info -- Previously-collected memory_accounting
        //                           information.
        //                           The pointer is cached inside the object.
        //                           The data is not corrupted.
        //                           This pointer is never freed.
        _account_info_allocated (false),
        _account_info (const_cast<account_info_t*>(account_info)) {
        clear();
        calc(account_info);
    }

    ~AccountInfo() {
        deallocateAccountInfoIfNecessary();
    }

    sys_return_t inspectProcess (az_allocid_t allocid);

    sys_return_t inspectProcessSelf () {
        return inspectProcess (0);
    }

    void clear() {
        deallocateAccountInfoIfNecessary();

        allocated_bytes = 0;
        funded_bytes = 0;
        overdraft_bytes = 0;
        footprint_bytes = 0;

        committed_bytes = 0;
        grant_bytes = 0;
        allocated_from_committed_bytes = 0;

        default_allocated_bytes = 0;
        default_committed_bytes = 0;
        default_footprint_bytes = 0;
        default_grant_bytes = 0;

        heap_allocated_bytes = 0;
        heap_committed_bytes = 0;
        heap_footprint_bytes = 0;
        heap_grant_bytes = 0;
    }

    void calc (const account_info_t *account_info) {
        for (size_t i = 0; i < account_info->ac_count; i++) {
            const account_stats_t *account = &account_info->ac_array[i];
            const AccountStats as (account);

            // Collect process-wide information.
            allocated_bytes += as.getAllocatedBytes();
            funded_bytes += as.getFundedBytes();
            overdraft_bytes += as.getOverdraftBytes();
            footprint_bytes += as.getFootprintBytes();
            committed_bytes += as.getCommittedBytes();
            grant_bytes += as.getGrantBytes();
            allocated_from_committed_bytes += as.getAllocatedFromCommittedBytes();

            // Grab information about specific accounts.
            if (i == PROCESS_ACCOUNT_DEFAULT) {
                default_allocated_bytes = as.getAllocatedBytes();
                default_committed_bytes = as.getCommittedBytes();
                default_footprint_bytes = as.getFootprintBytes();
                default_grant_bytes = as.getGrantBytes();
            }
            else if (i == PROCESS_ACCOUNT_HEAP) {
                heap_allocated_bytes = as.getAllocatedBytes();
                heap_committed_bytes = as.getCommittedBytes();
                heap_footprint_bytes = as.getFootprintBytes();
                heap_grant_bytes = as.getGrantBytes();
            }
        }
    }

    const account_info_t *getAccountInfo() const { return _account_info; }

    uint64_t getAllocatedBytes() const { return allocated_bytes; }
    uint64_t getFundedBytes() const { return funded_bytes; }
    uint64_t getOverdraftBytes() const { return overdraft_bytes; }
    uint64_t getFootprintBytes() const { return footprint_bytes; }

    uint64_t getCommittedBytes() const { return committed_bytes; }
    uint64_t getGrantBytes() const { return grant_bytes; }
    uint64_t getAllocatedFromCommittedBytes() const { return allocated_from_committed_bytes; }

    uint64_t getDefaultAllocatedBytes() const { return default_allocated_bytes; }
    uint64_t getDefaultCommittedBytes() const { return default_committed_bytes; }
    uint64_t getDefaultFootprintBytes() const { return default_footprint_bytes; }
    uint64_t getDefaultGrantBytes() const { return default_grant_bytes; }

    uint64_t getHeapAllocatedBytes() const { return heap_allocated_bytes; }
    uint64_t getHeapCommittedBytes() const { return heap_committed_bytes; }
    uint64_t getHeapFootprintBytes() const { return heap_footprint_bytes; }
    uint64_t getHeapGrantBytes() const { return heap_grant_bytes; }

private:
    void deallocateAccountInfoIfNecessary() {
        if (_account_info_allocated) {
            free (_account_info);
            _account_info_allocated = false;
        }
        _account_info = 0;
    }

    bool _account_info_allocated;
    account_info_t *_account_info;

    // Process-wide, not fund specific.
    uint64_t allocated_bytes;
    uint64_t funded_bytes;
    uint64_t overdraft_bytes;
    uint64_t footprint_bytes;

    // Process-wide, fund specific.
    uint64_t committed_bytes;
    uint64_t grant_bytes;
    uint64_t allocated_from_committed_bytes;

    // Default account only.
    uint64_t default_allocated_bytes;
    uint64_t default_committed_bytes;
    uint64_t default_footprint_bytes;
    uint64_t default_grant_bytes;

    // Main heap account only.
    uint64_t heap_allocated_bytes;
    uint64_t heap_committed_bytes;
    uint64_t heap_footprint_bytes;
    uint64_t heap_grant_bytes;
};


//-------------------------------------------------------------------------
// System Funds
//-------------------------------------------------------------------------

class FundStats {
public:
    FundStats (const fund_stats_t *fund) {
        clear();
        calc(fund);
    }

    void clear() {
        balance = 0;
        maximum = 0;
        used = 0;
    }

    void calc (const fund_stats_t *fund) {
        // General calculations.
        balance = fund->fnd_balance;
        maximum = fund->fnd_maximum;
        used = maximum - balance;
    }

    uint64_t getBalance() const { return balance; }
    uint64_t getMaximum() const { return maximum; }
    uint64_t getUsed() const { return used; }

private:
    uint64_t balance;
    uint64_t maximum;
    uint64_t used;
};

class FundInfo {
public:
    FundInfo () :
        // Purpose:  Create an funding object.
        //
        _fund_info_allocated (false),
        _fund_info (0) {
        clear();
    }

    FundInfo (const funds_info_t *fund_info) :
        // Purpose:  Calculate funding statistics.
        // Params:   fund_info -- Previously-collected memory_funding
        //                        information.
        //                        The pointer is cached inside the object.
        //                        The data is not corrupted.
        //                        This pointer is never freed.
        _fund_info_allocated (false),
        _fund_info (const_cast<funds_info_t*>(fund_info)) {
        clear();
        calc(fund_info);
    }

    ~FundInfo() {
        deallocateFundInfoIfNecessary();
    }

    sys_return_t inspectSystem() {
        // Purpose:  Calculate funding statistics for a process.
        // Returns:  SYSERR_NONE if the operation succeeded.
        //           SYSERR_<something else> otherwise.
        // Notes:    A deep copy of the system's memory_funding stats
        //           is made, and freed when the object goes out of scope.

        deallocateFundInfoIfNecessary();

        size_t len = sizeof(funds_info_t);

        _fund_info = (funds_info_t *) malloc (len);
        if (!_fund_info) {
            return SYSERR_NO_MEMORY;
        }

        memset (_fund_info, 0, len);
        _fund_info_allocated = true;

        if (az_pmem_subsys_available() < 0) {
            // Return underlying error code.
            return errno;
        }
        sys_return_t ret = az_pmem_get_fund_stats(_fund_info, &len);
        if (ret) {
            return ret;
        }

        calc (_fund_info);

        return SYSERR_NONE;
    }

    void clear() {
        deallocateFundInfoIfNecessary();

        total_bytes = 0;
        committed_total_bytes = 0;
        committed_used_bytes = 0;
        grant_total_bytes = 0;
        grant_used_bytes = 0;
    }

    void calc (const funds_info_t *fund_info) {
        total_bytes = 0;

        for (size_t i = 0; i < fund_info->fi_count; i++) {
            const fund_stats_t *fund = &fund_info->fi_array[i];
            const FundStats fs (fund);

            total_bytes += fs.getMaximum();

            // Grab information about specific funds.
            if (i == AZMM_COMMITTED_FUND) {
                committed_total_bytes = fs.getMaximum();
                committed_used_bytes = fs.getUsed();
            }
            else if (i == AZMM_OVERDRAFT_FUND) {
                grant_total_bytes = fs.getMaximum();
                grant_used_bytes = fs.getUsed();
            }
        }
    }

    const funds_info_t *getFundInfo() const { return _fund_info; }

    uint64_t getTotalBytes() const { return total_bytes; }
    uint64_t getCommittedTotalBytes() const { return committed_total_bytes; }
    uint64_t getCommittedUsedBytes() const { return committed_used_bytes; }
    uint64_t getGrantTotalBytes() const { return grant_total_bytes; }
    uint64_t getGrantUsedBytes() const { return grant_used_bytes; }

private:
    void deallocateFundInfoIfNecessary() {
        if (_fund_info_allocated) {
            free (_fund_info);
            _fund_info_allocated = false;
        }
        _fund_info = 0;
    }

    bool _fund_info_allocated;
    funds_info_t *_fund_info;

    uint64_t total_bytes;
    uint64_t committed_total_bytes;
    uint64_t committed_used_bytes;
    uint64_t grant_total_bytes;
    uint64_t grant_used_bytes;
};

}

#endif
