// Copyright 2003 - 2010 Azul Systems, Inc. All rights reserved.
// AZUL Systems PROPRIETARY/CONFIDENTIAL. Use is subject to license terms

// thread.h - Shared Linux & (Linux + azul DLKM) thread implementation
//
#ifndef _OS_THREAD_H_
#define _OS_THREAD_H_ 1

#include <string.h>
#include <signal.h>         // needed for siginfo_t  definition
#include <ucontext.h>       // needed for ucontext_t definition
#include <os/errors.h>
#include <os/types.h>
#include <os/pagesize.h>
#include <os/azulmmap.h>
#include <os/utilities.h>

#include <pthread.h>

#ifdef __cplusplus
extern "C" {
#endif

// Defined for FDC
#define thread_key_create(key, func) (sys_return_t)pthread_key_create(key, func)
#define thread_key_destroy(key)      (sys_return_t)pthread_key_delete(key)
#define thread_set_specific(key, value)     pthread_setspecific(key, value)
#define thread_get_specific(key)      pthread_getspecific(key)
#define THR_SELF     pthread_self()
#define NULL_THREAD  0L

typedef unsigned char* address;

// Base of type-stable libos managed Thread memory
struct _ThreadBase {
  int64_t _user_vtable;     // User Memory.  Must be at StackBase+0
  int64_t _user_pending_ex; // User Memory.  Must be at StackBase+8
  uint64_t _sig_jump_to_addr;         // AZUL_SIGNAL_JUMP_TO_ADDR PC address to execute when we 
                                      // return from the signal handler and resume on the main stack.
  int64_t _libos_eva, _libos_epc; // Tell user-mode SEGV crash dumps where we crashed
  uint64_t _event_profiling_buffer; // Event Profiling Tick Buffer address of current buffer.
  int32_t _local_thread_timer_ticks;  // Used to calc TTSP freq from Tick freq.

  // These next counts are the number of mapped VM_SMALL_PAGEs.
  // For the JEX stack, pages are mapped in from thread_map_jex_stack up to 
  // the _jex_stk_ext_cnt; the _jex_stk_ext_cnt pagenum is the 1st page NOT 
  // mapped in.
  int16_t _jex_stk_ext_cnt; // Number of stack extensions for the interpreter expression stack

  // Type-stable HotSpot bits.
  int8_t _is_Complete_Java_thread:1, // If the rest of the JavaThread* is initialized
    _is_Complete_Thread:1;     // If the rest of the Thread* is initialized

  // Record SIGSEGV access type for nio user-stack, delayed signal handling...
  // ( expected values are: 0 - instruction fetch, 1 - store,  2- load )
  int8_t _libos_nio_sigsegv_access_type;   

  // Signal Handling Fields.
  int8_t _nested_segv_cnt;                // Number of recursive (nested) SIGSEGV's we've seen.
  int8_t _sighandler_aborting;            // Flag that indicates we are dying from a SIGSEGV. 

  int8_t _tick_signal_handler;            // Tick signal-handler flag. {i.e. handler_active, slow_timer, etc.}

  uint64_t _self;         // thread self pointer

  intptr_t *       _jexstk_top; // Top of the expression stack, the last live value is at _jexstk_top[-1]
  unsigned char*   _stack_base;
  size_t           _stack_size;
  // State of the stack guard pages for this thread.
  enum StackGuardState {
    stack_guard_unused,          // not needed
    stack_guard_yellow_disabled, // disabled (temporarily) after stack overflow
    jex_stack_guard_yellow_disabled, // disabled jex stack yellow zone
    stack_guard_enabled          // enabled
  } _stack_guard_state;

  // Unused LIBOS Fields DO NOT TOUCH OR REORDER, the first field is used to insure
  // that the alignment of this structure matches its correllary in exceptions.hpp
  int8_t _libos06;
  int8_t _libos07;
  int8_t _libos08;

  // Further memory is reserved for user programs.
};
typedef struct _ThreadBase ThreadBase;

// Thread Stacks!
// Stacks memory map is defined in azulmmap.h
// Stacks start at  __THREAD_STACK_REGION_START_ADDR__
// Stacks end   at  __THREAD_STACK_REGION_END_ADDR__
//    __THREAD_STACK_REGION_START_ADDR__ is equal to (AZUL_THREAD_STACKS_SHIFTED_BASE<<AZUL_THREAD_STACKS_BASE_SHIFT)

extern uint64_t _thread_stack_size;
extern unsigned long _exec_stack_yellow_pages;
extern unsigned long _exec_stack_red_pages;
extern unsigned long _jex_stack_yellow_pages;
extern unsigned long _jex_stack_red_pages;

// Each individual stack is broken into these regions, and each region is
// restricted to being a multiple of small OS pages to allow for
// hardware protection.
enum {
  // User local data.  Holds the HotSpot JavaThread* structure for HotSpot threads.
  thread_size_local_data    = 1*VM_SMALL_PAGE_SIZE, // Holds for HotSpot the JavaThread* structure
  // Thread-local-data overflow detection.
  // Alt-signal-stack overflow detection.
  // Either is a fatal error.
  thread_size_deadzone1     = 4*VM_SMALL_PAGE_SIZE, 
  // Alternate Signal Stack: the stack used when processing SEGVs.  Since we
  // do our own stack overflow management, we take SEGVs any time a stack
  // pointer overflows its current space, and we extend the stack (map in some
  // more storage) or kill the whole process (fatal red zone touch).  However
  // the SEGV handling requires stack space precisely when we have none, hence
  // the need for an alternate signal stack.
  thread_size_alt_sig_stack = 8*VM_SMALL_PAGE_SIZE, // Alternate stack for SEGV handling
  // The siginfo_t & ucontext_t structures.  Since we want to be able to handle
  // up to 10 recursive SIGSEGV's, we also want to capture the siginfo_t and
  // ucontext_t structure data for each of these instances.  Therefore we keep
  // an array of max_recursive_exceptions entries, determined by the avilable
  // space within this segment, that contains both of the structures.  If you exceed
  // the recursive exceptions limit we disregard additional entries and abort.
  thread_size_siginfo_ucontext = 4*VM_SMALL_PAGE_SIZE, // Cache ucontext for SEGV handling
  // Thread-local buffer for collecting hi-res profiling ticks
  thread_size_ttsp_buf      = 1*VM_SMALL_PAGE_SIZE, // The initial and intermediate tick buffers for this thread.
  max_recursive_exceptions  = thread_size_siginfo_ucontext / (sizeof(siginfo_t) + sizeof(ucontext_t) + sizeof(void*))
};

// Stack Layout, replicated for every stack at 2Meg offsets
enum {
  thread_map_local_data    = 0,
  thread_map_deadzone1     = thread_map_local_data   + thread_size_local_data,
  thread_map_alt_sig_stack = thread_map_deadzone1    + thread_size_deadzone1,
  thread_map_siginfo_ucontext = thread_map_alt_sig_stack    + thread_size_alt_sig_stack,
  thread_map_ttsp_buf         = thread_map_siginfo_ucontext + thread_size_siginfo_ucontext,
  thread_map_jex_stack     = thread_map_ttsp_buf     + thread_size_ttsp_buf
};

inline static void* thread_current();

inline static void set_signal_jump_to_addr(uint64_t addr)
{
    ThreadBase*        thr = (ThreadBase*) thread_current();
    thr->_sig_jump_to_addr = addr;
}

// Because we want to be able to capture the siginfo_ & ucontext_t for
// multiple, recursirve SIGSEGVs, we've created an area above the alternate
// stack space but below the Tick Profiling buffers, and then provide
// accessor routines to make sure that data is always aligned properly.
// Which array element you get is based on the 
inline static siginfo_t* altsigstack_siginfo_addr_get(int segv_cnt)
{ 
    ThreadBase *thr = (ThreadBase*) thread_current();

    if (thr->_nested_segv_cnt >= max_recursive_exceptions)
    {
        // display_signum_and_errmsg(SIGSEGV, segv_exceeded_space_str);
        abort();
    }

    if (segv_cnt > thr->_nested_segv_cnt)
    {
        // display_signum_and_errmsg(SIGSEGV, segv_exceeded_count_str);
        abort();
    }

    uint64_t addr = ((uint64_t) thr) + thread_map_siginfo_ucontext;
    addr += ((segv_cnt) * (sizeof(siginfo_t) + sizeof(ucontext_t) + sizeof(void*)));
    addr&= ~(sizeof(void*) - 1);            
    return (siginfo_t*) addr;
}

// Returns the ucontext_t start address, word aligned (rounded up) from the
// siginfo_t structure at the bottom of the alt stack, given a stack pointer.
inline static ucontext_t* altsigstack_ucontext_addr_get(int segv_cnt)
{ 
    uint64_t addr = (uint64_t) altsigstack_siginfo_addr_get(segv_cnt);
    addr+= sizeof(siginfo_t);
    addr+= sizeof(void*);             // Round up to next word.
    addr&= ~(sizeof(void*) - 1);
    return (ucontext_t*) addr;
}

// Copies the siginfo_t data into the alt stack.  
inline static void altsigstack_siginfo_set(int segv_cnt, siginfo_t* pSigInfo)
{
    void*  addr =  altsigstack_siginfo_addr_get(segv_cnt);
    memcpy(addr, pSigInfo, sizeof(siginfo_t));
}

// Copies the ucontext_t data, word aligned (rounded up from the
// siginfo_t structure) into the alt stack.
inline static void altsigstack_ucontext_set(int segv_cnt, ucontext_t* pUcontext)
{
    void*  addr = altsigstack_ucontext_addr_get(segv_cnt);
    memcpy(addr, pUcontext, sizeof(ucontext_t));
}

// To make the code simpler and more readable, we have created a wrapper layer
// which hides differences between az_mmap/az_mreserve/az_munmap and mmap/munmap.  
// NOTE.  These routines are not generalized for hotspot and the runtime.  These
// are used ONLY for thread stack memory, is allocated in azmemory for all proxied
// VMs and in linux memory for non-proxied (Espresso).
void* thrdStk_mmap_wrapper(void *start, size_t size, int prot , int flags, int fd, off_t offset);
void* thrdStk_mreserve_wrapper(void *start, size_t size, int prot , int flags, int fd, off_t offset);
int   thrdStk_munmap_wrapper(void *start, size_t size);
void  thread_jexstack_yellow_zone_reguard(void* thread);

// The start function prototype passed to the pThreads library when we
// create a new thread.
typedef void* (*Thread_Start_Function)(void *);

typedef struct {
  Thread_Start_Function startFunc;
  void *args;
  void *stkBase;
} start_args;

extern pid_t thread_gettid();

// The thread stack callback function type.
typedef void (*threadStackCallbackFuncPtr_t) ( intptr_t );

// Set a call-back function, called after any new thread is created and after
// its stack has moved from the initial pThreads stack to the normal large
// libos managed stack.
extern threadStackCallbackFuncPtr_t _thread_start_callback;
// Set a call-back function, called when any thread is about to have its
// normal large libos managed stack be reclaimed.
extern threadStackCallbackFuncPtr_t _thread_death_callback;
static inline void thread_start_callback_register(threadStackCallbackFuncPtr_t f) { _thread_start_callback=f; }
static inline void thread_death_callback_register(threadStackCallbackFuncPtr_t f) { _thread_death_callback=f; }

//
//  These functions manage thread stacks that conform to Hotspot's particular alignment and layout requirements.
//
extern void  init_all_thread_stacks ();
extern void  init_stack_arguments(unsigned long thread_stack_size, unsigned long jex_yellow_pages, unsigned long jex_red_pages, unsigned long stack_yellow_pages, unsigned long stack_red_pages);
extern void* thread_stack_create    ();

extern sys_return_t thread_set_priority(pthread_t tid, int64_t priority);
extern sys_return_t thread_get_priority(pthread_t tid, int64_t* priority_ptr);

inline static sys_return_t thread_self(thread_t *t) {
    *t = pthread_self();
    return SYSERR_NONE;
}

inline static sys_return_t thread_terminate(thread_t thr __attribute__ ((unused))) {
    // TODO, need to figure out if pthread_cancel can be used
    return SYSERR_UNIMPLEMENTED;
}

inline static uint16_t thread_txu_id() {
    uint64_t id = 0;
    // TODO: Need to figure out how to do this
    return (uint16_t) id;
}

inline static uint64_t thread_get_identifier_self() {
    return (uint64_t) pthread_self();
}

extern sys_return_t thread_sleep(uint64_t sleep_nanos);

// TODO: For now define these, will need to revisit
typedef struct thread_perf2_context {
  ureg_t          perfcntctl0;
  ureg_t          perfcntctl1;
  int32_t         perfcnt0;
  reg_t           perfcnt1;
  int32_t         tlbperfcnt0;
} thread_perf2_context_t;


// --- Current thread's ThreadBase address
inline static void* thread_current() {
  void *ret = 0;
  asm("movq %%gs:%c1,%0" : "=r"(ret) : "g"(__builtin_offsetof(struct _ThreadBase, _self)) );
  return ret;
}

extern void* get_thread_current_from_key(); 

extern int set_pthread_key_data_for_current_thread(void *val);

// guarded_raw_thread_local_address returns current thread pointer or NULL.
extern void* guarded_raw_thread_local_address();

inline static size_t  stack_yellow_zone_size() { return _exec_stack_yellow_pages * VM_SMALL_PAGE_SIZE; }
inline static size_t stack_red_zone_size() { return _exec_stack_red_pages * VM_SMALL_PAGE_SIZE; }

inline static address stack_yellow_zone_base(ThreadBase* thread) { 
  return (address)(thread->_stack_base - (thread->_stack_size - (stack_red_zone_size() + stack_yellow_zone_size()))); 
}

inline static address stack_red_zone_base(ThreadBase* thread) { 
  return (address)(thread->_stack_base - (thread->_stack_size - stack_red_zone_size())); 
}

inline static bool in_stack_red_zone(ThreadBase* thread, address a) {
 return (a <= stack_red_zone_base(thread)) && (a > (address)((intptr_t)thread->_stack_base - thread->_stack_size)); 
}

inline static bool in_stack_yellow_zone(ThreadBase* thread, address a) {
  return (a <= stack_yellow_zone_base(thread) && (a > stack_red_zone_base(thread)));
}


inline static size_t  thread_stack_size() { return _thread_stack_size; }
inline static address jex_stack_base(ThreadBase* thread) { return (address)thread; }

inline static size_t  jex_stack_yellow_zone_size() { return _jex_stack_yellow_pages * VM_SMALL_PAGE_SIZE; }
inline static size_t  jex_stack_red_zone_size() { return _jex_stack_red_pages * VM_SMALL_PAGE_SIZE; }

inline static address jex_stack_yellow_zone_base(ThreadBase* thread) {
  return (address)(jex_stack_base(thread)  + thread_stack_size() - jex_stack_yellow_zone_size() - jex_stack_red_zone_size());
}

inline static address jex_stack_red_zone_base(ThreadBase* thread) { 
  return (address)(jex_stack_base(thread)  + thread_stack_size() - jex_stack_red_zone_size()); 
}

// JEX stack grows upwards, yellow_zone_addr < red_zone_addr
inline static bool in_jex_stack_yellow_zone(ThreadBase* thread, address a) {
  return (a >= jex_stack_yellow_zone_base(thread)) && (a < jex_stack_red_zone_base(thread));
}

inline  static bool in_jex_stack_red_zone(ThreadBase* thread, address a) {
  return (a >= jex_stack_red_zone_base(thread)) && (a < (jex_stack_base(thread)+thread_stack_size() ));
}

extern int init_main_thread_stack();

typedef int (*pthread_create_handle_azsys_fn)(pthread_t *thread, const pthread_attr_t *attr, void *(*start_routine) (void *), void *arg);
extern pthread_create_handle_azsys_fn pthread_create_handle_azsys;

extern void* preRun(void *args);

#ifdef __cplusplus
} // extern "C"
#endif

#endif // _OS_THREAD_H_
