// Copyright 2003-2010 Azul Systems, Inc. All rights reserved.
// AZUL Systems PROPRIETARY/CONFIDENTIAL. Use is subject to license terms

// utilities.h - Remaining bits and pieces of OS Abstraction Layer for OS Services

#ifndef _OS_UTILITIES_H_
#define _OS_UTILITIES_H_ 1

#ifdef __cplusplus
extern "C" {
#endif

#include <stdarg.h>
#include <stdbool.h>
#include <signal.h>
#include <os/types.h>
#include <os/memory.h>

#include <arpa/inet.h>

typedef struct {
  int num_options;
  char** options;
} LauncherArgs;

extern void init_azsys(void);

extern void init_main_source(const char *source);

extern int find_default_license(
    char *default_license_file,
    size_t buff_size);

extern int validate_license(
    const char *license_file,
    char *failure_buff,
    size_t failure_buff_size);

// Process exit related interface
extern void exit_out_of_memory(void);
extern void exit_abort(uint64_t code);
extern void set_abort_on_out_of_memory(int flag/* boolean value */);
extern void set_oom_callback_with_return(bool (*f) (void));
extern void set_os_backtrace_callback(void (*f) (void));

// Interface for a non-os backtrace callback, typically used for profiling.
typedef void (*backtrace_callback_t) (intptr_t pc, intptr_t sp, intptr_t fp, const uint32_t rpc_count, 
                                      uint64_t* rpcs);
backtrace_callback_t get_backtrace_callback();
extern void set_backtrace_callback(backtrace_callback_t f);

// How to make as-yet-unspecified syscalls
extern sys_return_t systemcall(int64_t syscall_num,
                               int64_t arg1,
                               int64_t arg2,
                               int64_t arg3,
                               int64_t arg4,
                               int64_t arg5);

// Miscellaneous convenience functions

// Formatting into a string buffer
#define os_snprintf  snprintf
#define os_vsnprintf vsnprintf

// Safe (syscall) print-to-console function
extern int uprintf(const char *fmt, unsigned long long arg);

// Assertions, etc.
inline static void _os_breakpoint(void)
{
	__asm__ __volatile__("int	$0x3");
}

// Always-on assertion
#define os_guarantee(e, s)   ((e) ? (void)0 : _os_abort("guarantee failed", __FUNCTION__, __FILE__, __LINE__, (s), #e))
// Check system call
#define os_syscallok(x, s)   (((x) == SYSERR_NONE) ? (void)0 : _os_syscallok(x, __FUNCTION__, __FILE__, __LINE__, (s)))

extern void _os_syscallok(sys_return_t rc,
                          const char *_function,
                          const char *_file,
                          int         _line,
                          const char *_msg);

// Debug-only assertions
#ifdef DEBUG
#define os_assert(e, s)      ((e) ? (void)0 : _os_abort("assertion failed",     __FUNCTION__, __FILE__, __LINE__, (s), #e))
// Precondition
#define os_require(e, s)     ((e) ? (void)0 : _os_abort("precondition failed",  __FUNCTION__, __FILE__, __LINE__, (s), #e))
// Postcondition
#define os_ensure(e, s)      ((e) ? (void)0 : _os_abort("postcondition failed", __FUNCTION__, __FILE__, __LINE__, (s), #e))
#else
#define os_assert(e, s)      ((void)0)
#define os_require(e, s)     ((void)0)
#define os_ensure(e, s)      ((void)0)
#endif

// Used internally for noting implementation state/progress
#define os_unreachable()     os_guarantee(0, "should not reach here")
#define os_unimplemented()   os_guarantee(0, "unimplemented")
#define os_untested()        os_guarantee(0, "untested")

// Helper function
extern void _os_abort(const char *_type,
                      const char *_function,
		      const char *_file,
		      int         _line,
		      const char *_msg,
		      const char *_expression);

// For explicit debugging
extern void os_debug(void);

// Print a backtrace
extern void os_backtrace(void);

// TODO: Macro to retrieve the value of a control register into a variable
// src: name of control register (in quotes, e.g., "perfcnt1")
// dst: int64_t variable that wil receive the value
#define  os_get_control_register(src,dst) dst=0

// TODO: Macro to set the value of a control register.
// src: int64_t variable with the value to set
// dst: name of control register (in quotes, e.g., "perfcnt1")
#define os_set_control_register(src,dst)

extern void       os_disable_azmem();
extern int        os_should_use_azmem();
extern void       os_disable_codecache_in_azmem();
extern void       os_enable_codecache_in_azmem();
extern void       os_disable_azsched();
extern void       os_enable_azsched();
extern int        os_should_use_azsched();
extern void       os_disable_azpgroups();
extern int        os_should_use_azpgroups();
extern void       os_set_start_suspended(int ss);
extern int        os_should_start_suspended();
extern void       os_get_memory_accounts(account_info_t* acct_info);
extern int        os_get_memory_accounts_usage(account_usage_info_t* acct_usage_info);
extern int        os_has_account_usage_support();
extern int        os_has_segv_probe();
extern void       os_set_mem_partition(int mem_partition);
extern int        os_get_mem_partition();

typedef int (*is_current_thread_detached_fn)(void *);
extern void set_is_current_thread_detached_fn(is_current_thread_detached_fn callback);

typedef int (*check_and_call_user_signal_handler_fn)(int sig, siginfo_t* siginfo, void* ucontext);
extern void set_check_and_call_user_signal_handler_fn(check_and_call_user_signal_handler_fn callback);

typedef void (*thread_stack_disable_yellow_zone_callback_fn)(intptr_t);
extern void set_thread_stack_disable_yellow_zone_callback_fn(thread_stack_disable_yellow_zone_callback_fn callback);

extern void os_setup_avm_launch_environment();

extern void set_azul_vmstructs_attributes();
extern int get_azul_vmstructs_version();

extern ssize_t fullRead (int fd, char *buf, size_t count);

extern size_t strlcat(char *dst, const char *src, size_t siz);
extern size_t strlcpy(char *dst, const char *src, size_t siz);

extern void signal_chain_handler(int signum, siginfo_t* pSiginfo, ucontext_t* pUcontext);

# if __BYTE_ORDER == __BIG_ENDIAN
/* The host byte order is the same as network byte order,
   so these functions are all just identity.  */
# define ntohll(x) (x)
# define htonll(x) (x)
# else
#  if __BYTE_ORDER == __LITTLE_ENDIAN
#   define ntohll(x) __bswap_64 (x)
#   define htonll(x) __bswap_64 (x)
#  endif
# endif

#ifdef __cplusplus
} // extern "C"
#endif

#endif // _OS_UTILITIES_H_
