/*
 * Copyright 2012 Azul Systems, Inc.  All rights reserved.
 * AZUL PROPRIETARY/CONFIDENTIAL.  Use is subject to license terms.
 */
#ifndef XLATE_BITFLAGS_HPP
#define XLATE_BITFLAGS_HPP

#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>

/**
 ** 
 ** class XlateBits - Container class for the bit translation.
 **
 **/
class XlateBits
{
  friend class XlateFlag;
  friend class XlateMode;
  friend class XlateProt;

  public:

    // An array of bitflag/namestring pairs, used by xlate() in
    // translating various parameters.
    typedef struct _bitflag_element_t
    {
        const char*   name;
        int           flag;
      
    } BITFLAG_ELEMENT_T;

    XlateBits()  { exit(-1); }
    ~XlateBits() { exit(-1); }

    /**
     ** xlate() - Translate a set of bitflags into character strings, based on bit/string tables
     ** provided by the caller.  This is especially useful for describing the various OS flags we'll be using.
     ** The table provided by the user is a pairing of bit assignments and their corresponding string value.
     ** This code is re-entrant, but care should be taken when translating mutiple flags within the same operation
     ** as the user supplied result string buffer will be over-written if used sequentially.
     ** 
     ** RETURNS: A pointer to a string buffer containing the string translation of the bit flags, or an error string,
     ** or the string "none" if there bitflag was zero.
     **/
    static const char* xlate(int flag, char* result, size_t res_size, BITFLAG_ELEMENT_T* bitflags, size_t elements)
    {
        const char* xlate_error  = "xlate() error";
        const char* no_flags     = "no flags";
  
        size_t pos   = 0;
        size_t len   = 0;
  
        if (result      == NULL)    { return xlate_error; }  // Missing return string buffer.
        if (bitflags    == NULL)    { return xlate_error; }  // Missing translation array.
        if (elements    <  1)       { return xlate_error; }  // Missing translation array.
        if (flag        == 0)       { return no_flags;    }  // No flags set.
  
        *result = 0;
        --elements;
  
        // Make sure the last entry is null terminated.
        if (bitflags[elements].name != NULL ) { return xlate_error; }  // Wrong array termination.
        if (bitflags[elements].flag != ~0   ) { return xlate_error; }  // Wrong array termination.
  
        // Loop through the rest of the items looking for bit matches,
        // for each match, write the namestring into the result buffer,
        // with a plus sign between them.
        for (size_t index = 0; index < elements; index++)
        {
            if (bitflags[index].name == 0) { return xlate_error; }   // Missing name string.
            if (bitflags[index].flag == 0) { return xlate_error; }   // Missing bit mask.
  
            if (flag & bitflags[index].flag)
            {
                len = strlen(bitflags[index].name);
  
                if (len <= (0))              { return xlate_error; }   // Empty name string.
                if (len >  (res_size - pos)) { return xlate_error; }   // Not enough space.
  
                memcpy((void*) &result[pos], (void*) &bitflags[index].name[0], len);
                pos+= len;
                result[pos++] = '+';
            }
        }
  
        if (pos == 0) { return "none"; }  // No matching flags set.
  
        result[pos - 1] = 0; // Null terminate the string, chop off the trailing "+" of the last item.
  
        if (strlen(result) >= res_size) { return xlate_error; } // Pathological.
  
        return result;  // Found 1 or more flags.

    }  // end of xlate().
      
};  // end of class XlateBits.
  
/**
 ** For flag parameters within the following calls...
 **    lseek()
 **    mmap()
 **    open()
 **/
class XlateFlag : public XlateBits
{
  public:

    static const char* lseek(int flag, char* result, size_t result_size)
    {
        BITFLAG_ELEMENT_T bitflags[] = 
        {
            { "SEEK_SET",       SEEK_SET,      },
            { "SEEK_CUR",       SEEK_CUR,      },
            { "SEEK_END",       SEEK_END,      },
            { NULL,             ~0,            },
        };
  
        return xlate(flag, result, result_size, bitflags, sizeof(bitflags)/sizeof(BITFLAG_ELEMENT_T));
  
    } // end of lseek().
  
  
    static const char* mmap(int flag, char* result, size_t result_size)
    {
        BITFLAG_ELEMENT_T bitflags[] = 
        {
            { "MAP_FIXED",       MAP_FIXED,      },
            { "MAP_SHARED",      MAP_SHARED,     },
            { "MAP_PRIVATE",     MAP_PRIVATE,    },
            { "MAP_DENYWRITE",   MAP_DENYWRITE,  },
            { "MAP_NORESERVE",   MAP_NORESERVE,  },
            { "MAP_LOCKED",      MAP_LOCKED,     },
            { "MAP_GROWSDOWN",   MAP_GROWSDOWN,  },
            { "MAP_ANONYMOUS",   MAP_ANONYMOUS,  },
            { "MAP_32BIT",       MAP_32BIT,      },
            { "MAP_POPULATE",    MAP_POPULATE,   },
            { "MAP_NONBLOCK",    MAP_NONBLOCK,   },
            { NULL,              ~0,             },
        };
  
        return xlate(flag, result, result_size, bitflags, sizeof(bitflags)/sizeof(BITFLAG_ELEMENT_T));
  
    } // end of mmap().
  
    static const char* open(int flag, char* result, size_t result_size)
    {
  
        BITFLAG_ELEMENT_T bitflags[] = 
        { 
            { "O_APPEND",        O_APPEND,     },
            { "O_ASYNC",         O_ASYNC,      },
            { "O_RDONLY",        O_RDONLY,     },
            { "O_WRONLY",        O_WRONLY,     },
            { "O_RDWR",          O_RDWR,       },
            { "O_CREAT",         O_CREAT,      },
            { "O_DIRECT",        O_DIRECT,     },
            { "O_DIRECTORY",     O_DIRECTORY,  },
            { "O_EXCL",          O_EXCL,       },
            { "O_LARGEFILE",     O_LARGEFILE,  },
            { "O_NOATIME",       O_NOATIME,    },
            { "O_NOFOLLOW",      O_NOFOLLOW,   },
            { "O_NONBLOCK",      O_NONBLOCK,   },
            { "O_NDELAY",        O_NDELAY,     },
            { "O_SYNC",          O_SYNC,       },
            { "O_TRUNC",         O_TRUNC,      },
            { NULL,              ~0,           },
        };
  
        return xlate(flag, result, result_size, bitflags, sizeof(bitflags)/sizeof(BITFLAG_ELEMENT_T));
  
    } // end of open().
  
};  // end of class XlateFlag.
  
/**
 ** For the "prot" parameter within the following calls...
 **    mmap()
 **/
class XlateProt : public XlateBits
{
  public:

    static const char* mmap(int flag, char* result, size_t result_size)
    {
        BITFLAG_ELEMENT_T bitflags[] = 
        {
            { "PROT_EXEC",    PROT_EXEC,  },
            { "PROT_READ",    PROT_READ,  },
            { "PROT_WRITE",   PROT_WRITE, },
            { NULL,           ~0,         },
        };
  
        return xlate(flag, result, result_size, bitflags, sizeof(bitflags)/sizeof(BITFLAG_ELEMENT_T));
  
    }  // end of mmap().
  
};  // end of class XlateProt.
  
/**
 ** For the "mode" parameter within the following calls...
 **    open()
 **/
class XlateMode : public XlateBits
{
  public:

    static const char* open(int flag, char* result, size_t result_size)
    {
        BITFLAG_ELEMENT_T bitflags[] = 
        { 
            { "S_IRWXU",         00700,        },
            { "S_IRUSR",         00400,        },
            { "S_IWUSR",         00200,        },
            { "S_IXUSR",         00100,        },
            { "S_IRWXG",         00070,        },
            { "S_IRGRP",         00040,        },
            { "S_IWGRP",         00020,        },
            { "S_IXGRP",         00010,        },
            { "S_IRWXO",         00007,        },
            { "S_IROTH",         00004,        },
            { "S_IWOTH",         00002,        },
            { "S_IXOTH",         00001,        },
            { NULL,              ~0,           },
        };
  
        return xlate(flag, result, result_size, bitflags, sizeof(bitflags)/sizeof(BITFLAG_ELEMENT_T));
  
    } // end of open().
      
};  // end of class XlateMode.

#endif // XLATE_BITFLAGS_HPP

// end of file bitflags.hpp.

