/*
 * Copyright 2012 Azul Systems, Inc.  All rights reserved.
 * AZUL PROPRIETARY/CONFIDENTIAL.  Use is subject to license terms.
 */
#ifndef ZVISION_CSL_HPP
#define ZVISION_CSL_HPP

#include "zvision.hpp"

 // Used for Unit Tests: Comma Separated List files allow us to preload shared memory files.
#define CSL_BRACKET_WIDTH                     7
#define CSL_ELEMENT_DELIMITER                 ','
#define CSL_START_BRACKET_DELIMITER           '['
#define CSL_END_BRACKET_DELIMITER             ']'
#define CSL_REPEAT_DELIMITER                  'R'
#define CSL_REPEAT_WIDTH                      5
#define CSL_MAX_REPEATS                       9999

/**
 ** class ZVM_API - class for the Azul JVM specific ZVision API.  Only methods called by the ZVM go here.
 **/
class ZVisionCSL : public ZVision
{
  private:


  public:

    // Utility conversion noop/fail versions, used where applicable for CSL_METHOD_T casts.
    static bool noop_csl_to_bin(char* addr, size_t len, int fd, const char* name, size_t* count) { return true;  }
    static bool fail_csl_to_bin(char* addr, size_t len, int fd, const char* name, size_t* count) { return false; }

    /**
     ** Replicate a block of memory that you've just loaded into memory via a comma separated list file entry.
     ** This is called by the convert_csl_to_bin(), whenever a user creates a line beginning with the CSL_REPEAT_DELIMITER
     ** followed by the count of how many times to repeat the data within the shared memory file.
     **/
    static bool repeat_block_copy(char*   repeat_block_start_addr, 
                                  char*   repeat_block_end_addr, 
                                  size_t  repeat_block_count,
                                  char**  shared_mem_addr,
                                  char*   last_valid_addr)
    {
        ZV_ASSERT(repeat_block_end_addr != NULL);
        ZV_ASSERT(repeat_block_end_addr > repeat_block_start_addr);
        ZV_ASSERT(repeat_block_count > 0);
        ZV_ASSERT(repeat_block_count <= CSL_MAX_REPEATS);
        ZV_ASSERT(*shared_mem_addr == repeat_block_end_addr);

        size_t size = (repeat_block_end_addr >= repeat_block_start_addr) ? (repeat_block_end_addr - repeat_block_start_addr) : 0;

        // TODO: Add a check on the address range and repeat count.
        while (repeat_block_count-- > 0)
        {
            if (*shared_mem_addr + 1 > last_valid_addr)
            {
                PATHOLOGICAL("Exceed available memory during CSL conversion repeat block.\n");
                return false;
            }

            ::memcpy((void*) *shared_mem_addr, (const void*) repeat_block_start_addr, size);
            *shared_mem_addr += size;
        }

        return true;
    }


    /**
     ** Each line begins with a bracketed value, followed by the data itself.  Values within the brackets are 
     ** fixed size and represent the type of data, where character strings length values, and hexidecimal data
     ** has values withn the bracket that begin with 'Nx' (or 'NX') where N is the binary data width (which can
     ** be only 1, 2, 4 or 8) in bytes, and the remaining digits within the bracket indicate the number of times
     ** to repeat the same data packet (which immediately follows the brackets) in memory.
     **
     ** shared_mem_addr is the shared memory file address where we want to load the data.  This value may, or may NOT
     ** be the starting address in memory of the file, as some file types (Request/Response Buffers) may have several
     ** other fields before the response data.
     **/
    static bool convert_csl_to_bin(char* shared_mem_addr, size_t len, SOCKET_FD_T fd, const char* name, size_t* count)
    {
        ssize_t          bytes_read        = 0;            // Number of bytes returned from ::read() call.
        size_t           offset            = 0;            // Characters read from the input file, including newlines.
        size_t           line              = 0;            // Currently processed input file line number.
        size_t           repeat_count      = 0;            // How many times to repeat the block of information.

        char*            repeat_start_addr = NULL;         // The starting address of the already written data.
        char*            repeat_end_addr   = NULL;         // The ending address of the already written data.
        char*            last_valid_addr   = NULL;         // The last valid address we can write to.

        char             buf[MAX_ZVISION_PATHNAME_LEN];

        DIAGOUT(4, "Converting CSL file \"%s\" into binary.\n", name);

        if (shared_mem_addr == NULL)
        {
            PATHOLOGICAL("shared_mem_addr is NULL.\n");
            return false;
        }

        if (len > MAX_RESERVE_MEMORY)
        {
            PATHOLOGICAL("maximum shared memory size %lu exceeds MAX_RESERVE_MEMORY (%ld).\n", len, MAX_RESERVE_MEMORY);
            return false;
        }

        // Calculate the last valid address we can write to.
        last_valid_addr = shared_mem_addr + len;

        // Parse the csl (comma separated list) of input values.
        while (true)
        {
            size_t size = 0;

            // Clear the entire temporary buffer, just to be safe.
            memset(&buf[0], 0, sizeof(buf));

            bytes_read = ::read(fd, &buf[0], 1);

            // If there is nothing there, we're probably done.
            if (bytes_read != 1) { break; } 

            // First character of the line is either a bracket, or the prior newline/carriage return from
            // the previous line, in the latter case we just throw them away.  Newlines are not required.
            if (buf[0] == '\n' || buf[0] == '\r') 
            {
                offset++;
                continue;
            }

            // We should be at a start of line bracket.
            line++;

            // If the user wants to repeat a block of data, entering an 'R' followed by the number of times
            // to repeat the block, followed by a comma, will copy the next block of data (up to the next
            // CSL_REPEAT_DELIMITER or the EOL) and copy that at successive points in the memory.  This is
            // useful when you are trying to test that you can move data, and want to sanity test the data
            // content, but don't care too much about the diversity of the data. 
            if (buf[0] == CSL_REPEAT_DELIMITER)
            {
                size_t loop;

                if (repeat_start_addr == NULL)
                {
                    ZV_ASSERT(repeat_end_addr == NULL);
                    ZV_ASSERT(repeat_count == 0);
                }
                else
                {
                    if (repeat_block_copy(repeat_start_addr, 
                                          repeat_end_addr, 
                                          repeat_count, 
                                          &shared_mem_addr, 
                                          last_valid_addr) != true)
                    {
                        PATHOLOGICAL("REPEAT block exceeded memory at line %lu in file \"%s\" [fd %d] .\n", line, name, fd);
                        return false;
                    }
                }

                repeat_start_addr = shared_mem_addr;
                repeat_end_addr   = shared_mem_addr;
                repeat_count      = 0;

                offset++;

                // Grab the rest of the new repeat clause.
                bytes_read = ::read(fd, &buf[0], CSL_REPEAT_WIDTH);

                if (bytes_read != CSL_REPEAT_WIDTH) 
                { 
                    PATHOLOGICAL("Insufficient data after REPEAT clause (\'%c\') at line %lu in file \"%s\" [fd %d] .\n", 
                        CSL_REPEAT_DELIMITER, line, name, fd);
                    return false;
                }

                // Determine the value.
                for (loop = 0; loop < CSL_REPEAT_WIDTH; loop++)
                {
                    if (buf[loop] < '0' || buf[loop] > '9') { break; }

                    repeat_count = (repeat_count * 10) + (buf[loop] - '0');
                    offset++;
                }

                offset++;

                if (buf[loop] != CSL_ELEMENT_DELIMITER)
                {
                    PATHOLOGICAL("Missing CSL_REPEAT_DELIMITER \"%c\" got \"%c\" instead at line %lu in file \"%s\" [fd %d] .\n", 
                        CSL_ELEMENT_DELIMITER, buf[loop], line, name, fd);
                    return false;
                }

                buf[loop] = 0;  

                if (sscanf(&buf[0], "%ld", &repeat_count) != 1)
                {
                    PATHOLOGICAL("Encountered invalid REPEAT clause value \"[%s]\" at line %lu in file \"%s\" (fd %d) .\n", 
                        &buf[1], line, name, fd);
                    return false;
                }

                if (repeat_count == 0)
                {
                    // The user has terminated a repeat block.  We've already block copied it, so
                    // we reset everything back to zero.
                    repeat_start_addr = NULL;
                    repeat_end_addr   = NULL;
                }

                continue;
            }

            if (buf[0] != CSL_START_BRACKET_DELIMITER)
            {
                PATHOLOGICAL("Expected start of line bracket \"%c\" got \"%c\" instead at line %lu in file \"%s\" [fd %d] .\n", 
                    CSL_START_BRACKET_DELIMITER, buf[0], line, name, fd);
                return false;
            }

            // We parse the contents within the brackets either as the length of the string following 
            // the bracket (up to the delimiter) or as an indicator of the binary data width of the 
            // following hexidecimal value (up to the delimiter).  If there is an optional repetition
            // character within the bracket, repeat the binary data that many times within the shared
            // memory file.

            // Read the rest of the bracket in.  We've already overwritten the starting delimiter, so
            // we need to grab N-1 characters instead of the full bracket width.
            bytes_read = ::read(fd, &buf[0], CSL_BRACKET_WIDTH - 1);

            if (bytes_read != (CSL_BRACKET_WIDTH - 1))
            {
                PATHOLOGICAL("Expected %d characters after '%c', got %lu instead, at line %lu in file \"%s\" [fd %d] .\n", 
                    (CSL_BRACKET_WIDTH - 1), CSL_START_BRACKET_DELIMITER, bytes_read, line, name, fd);
                return false;
            }

            // Make sure that the last charcter is an end delimiter.  Since we've already overwritten
            // the start character, we'll need to look at position N-2 within the buffer.
            if (buf[(CSL_BRACKET_WIDTH - 2)] != CSL_END_BRACKET_DELIMITER)
            {
                buf[(CSL_BRACKET_WIDTH - 2)] = 0;
                PATHOLOGICAL("Missing end delimiter (contained '%c' but expecting '%c') at line %lu in file \"%s\" (fd %d) .\n", 
                    buf[(CSL_BRACKET_WIDTH - 2)], CSL_END_BRACKET_DELIMITER, line, name, fd);
                return false;
            }

            buf[(CSL_BRACKET_WIDTH - 2)] = 0;   // overwrite the trailing bracket.
            offset  += CSL_BRACKET_WIDTH;       // account for the entire bracket's length in the input file.

            // If we have a numeric value in this location, it must be a character string.
            if (buf[1] >= '0' && buf[1] <= '9')
            {
                if (sscanf(&buf[1], "%ld", &size) != 1 || size < 1 || size >= MAX_ZVISION_PATHNAME_LEN - CSL_BRACKET_WIDTH)
                {
                    PATHOLOGICAL("Encountered invalid string size \"[%s]\" at line %lu in file \"%s\" (fd %d) .\n", 
                        &buf[1], line, name, fd);
                    return false;
                }

                bytes_read = ::read(fd, &buf[0], (size + 1));

                if (bytes_read != ((ssize_t) (size + 1))) 
                { 
                    PATHOLOGICAL("Unexpected EOF (read %lu of %lu characters) at line %lu in file \"%s\" (fd %d) .\n", 
                        bytes_read, size, line, name, fd);
                    return false;
                }

                if (buf[size] != CSL_ELEMENT_DELIMITER)
                {
                    PATHOLOGICAL("Missing delimiter for string starting at line %lu in file \"%s\" (fd %d) .\n", 
                        line, name, fd);
                    return false;
                }

                if (shared_mem_addr + size > last_valid_addr)
                {
                    PATHOLOGICAL("Exceed available memory during CSL conversion at line %lu in file \"%s\" (fd %d) .\n", 
                        line, name, fd);
                    return false;
                }

                buf[size] = 0;                                              // overwrite the trailing comma.
                ::memcpy((void*) shared_mem_addr, (const void*) buf, size); // copy the data into memory.
                shared_mem_addr   += size;                                  // bump the memory pointer.
                offset += (size + 1);                                       // increment the bytes read counter.

                // If we already were in a repeat block, bump the repeat block end pointer along.
                if (repeat_end_addr != NULL)
                {
                    repeat_end_addr += size;
                }
                    
                continue;
            }

            // If the second character after the brackets is an x or an X, then treat the data after the
            // brackets, up to the delimiter, as a hexidecimal value.  The value before the X is the byte width
            // within the shared memory file of the hexidecimal value.  Valid data widths values are 1, 2,
            // 4, & 8.  We write the unsigned value described in the input file into the address within memory.
            // The digits following the 'X' are treated as a repition value, indicating how many times to repeat
            // the hexidecimal data value (after the brackets) in memory.
            if (buf[1] == 'x' || buf[1] == 'X')
            {
                if (buf[0] != '1' && buf[0] != '2' && buf[0] != '4' && buf[0] != '8')
                {
                    PATHOLOGICAL("Invalid value \"[%s]\" for binary width at line %lu in file \"%s\" (fd %d) .\n", 
                        &buf[0], line, name, fd);
                    return false;
                }

                size_t repeat = 0;

                if (sscanf(&buf[2], "%ld", &repeat) != 1 || repeat > 999)
                {
                    PATHOLOGICAL("Invalid repetition value \"[%s]\" at line %lu in file \"%s\" (fd %d) .\n", 
                        &buf[1], line, name, fd);
                    return false;
                }

                size_t   width = (buf[0] - '0');
                uint64_t data  = 0;

                switch (width)
                {
                case 1:
                case 2:
                case 4:
                case 8:
                    break;
                default:
                    PATHOLOGICAL("Data width \"[%s]\" converted to illegal value (%ld) at line %lu in file \"%s\" (fd %d) .\n", 
                        &buf[0], width, line, name, fd);
                    return false;
                    break;
                }

                // Read the hexstring (two character for each hex byte) and one more for the trailing delimiter.
                bytes_read = ::read(fd, &buf[0], (width * 2) + 1);

                if (bytes_read != ((ssize_t) (width * 2) + 1)) 
                {
                    PATHOLOGICAL("Unexpected EOF (read %lu of %lu hex characters) at line %lu in file \"%s\" (fd %d) .\n", 
                        bytes_read, width, line, name, fd);
                    return false;
                }

                if (buf[(width * 2)] != CSL_ELEMENT_DELIMITER)
                {
                    PATHOLOGICAL("Missing delimiter for hexstring starting at line %lu in file \"%s\" (fd %d) .\n", 
                        line, name, fd);
                    return false;
                }

                buf[(width * 2)] = 0; // overwrite the trailing delimiter.
                    
                // Convert the data into an unsigned long long.
                errno = 0;
                data = ::strtoull(&buf[0], NULL, 16);

                if (errno != 0)
                {
                    PATHOLOGICAL("Invalid value \"[%s]\" for binary data at line %lu in file \"%s\" (fd %d) .\n", 
                        buf, line, name, fd);
                    return false;
                }

                switch (width)
                {
                case 1:
                    {
                         uint8_t  value = (uint8_t) (data & 0xFF);
                         for (size_t loop = 0; loop <= repeat; loop++)  
                         {
                             if (shared_mem_addr + 1 > last_valid_addr)
                             {
                                 PATHOLOGICAL("Exceed available memory during CSL conversion at "
                                              "line %lu in file \"%s\" (fd %d) .\n", 
                                     line, name, fd);
                                 return false;
                             }

                             ::memcpy((void*) shared_mem_addr, (const void*) &value, 1); 
                             shared_mem_addr += 1; 
                        }
                    }
                    break;

                case 2:
                    {
                        uint16_t value = (uint16_t) (data & 0xFFFF);
                        for (size_t loop = 0; loop <= repeat; loop++)  
                        { 
                             if (shared_mem_addr + 2 > last_valid_addr)
                             {
                                 PATHOLOGICAL("Exceed available memory during CSL conversion at "
                                              "line %lu in file \"%s\" (fd %d) .\n", 
                                     line, name, fd);
                                 return false;
                             }

                            ::memcpy((void*) shared_mem_addr, (const void*) &value, 2); 
                            shared_mem_addr += 2; 
                        }
                    }
                    break;

                case 4:
                    {
                        uint32_t value = (uint32_t) (data & 0xFFFFFFFF);
                        for (size_t loop = 0; loop <= repeat; loop++)  
                        { 
                             if (shared_mem_addr + 4 > last_valid_addr)
                             {
                                 PATHOLOGICAL("Exceed available memory during CSL conversion at "
                                              "line %lu in file \"%s\" (fd %d) .\n", 
                                     line, name, fd);
                                 return false;
                             }

                            ::memcpy((void*) shared_mem_addr, (const void*) &value, 4); 
                            shared_mem_addr += 4; 
                        }
                    }
                    break;

                case 8:
                    {
                        uint64_t value = (uint64_t) (data & 0xFFFFFFFFFFFFFFFF);
                        for (size_t loop = 0; loop <= repeat; loop++)  
                        {  
                             if (shared_mem_addr + 8 > last_valid_addr)
                             {
                                 PATHOLOGICAL("Exceed available memory during CSL conversion at "
                                              "line %lu in file \"%s\" (fd %d) .\n", 
                                     line, name, fd);
                                 return false;
                             }

                            ::memcpy((void*) shared_mem_addr, (const void*) &value, 8); 
                            shared_mem_addr += 8; 
                        }
                    }
                    break;

                default:
                    PATHOLOGICAL("Illegal width value (%lu) while processing line %lu in file \"%s\" (fd %d) .\n", 
                        width, line, name, fd);
                    return false;
                    break;

                }  // end of switch (width).

                offset += (width * 2);   // Hexidecimal characters read.
                offset++;                // trailing delimiter.

                // If we are in a repeat block, bump the repeat block end pointer along for whatever the 
                // value width is, and then again as many times that the repeat parameter calls for.
                if (repeat_start_addr != NULL && repeat_end_addr != NULL)
                {
                    repeat_end_addr += (width + (width * repeat));
                }
                    
                continue;

            } // end of if (buf[1] == 'x' || buf[1] == 'X')

            PATHOLOGICAL("Unexpected bracket contents \"[%s]\" at line %lu in file \"%s\" (fd %d) .\n", buf, line, name, fd);
            return false;

        }  // end of while (true)
        
        // Check that we weren't already in a repeat block, if so, then replicate the data
        // that we've already written into memory that many times more.
        if (repeat_start_addr != NULL)
        {
            if (repeat_block_copy(repeat_start_addr, 
                                  repeat_end_addr, 
                                  repeat_count, 
                                  &shared_mem_addr, 
                                  last_valid_addr) != true)
                {
                    PATHOLOGICAL("REPEAT block exceeded memory at line %lu in file \"%s\" [fd %d] .\n", line, name, fd);
                    return false;
                }
        }

        // If the caller supplied a pointer to a size_t, fill in the number of bytes written into shared memory.
        if (count != NULL) {  *count = (size_t) offset; }
           
        DIAGOUT(3, "Finished reading \"%s\" from fd %d.\n", name, fd);
        DIAGOUT(3, "read %lu lines [total characters = %lu].\n", line, offset);

        return true;

    }  // end of convert_csl_to_bin().




    /**
     ** 
     **/
    static bool convert_bin_to_csl(SOCKET_FD_T csl_fd, SOCKET_FD_T bin_fd, size_t* lines_written)
    {

        uint64_t bin_word;                             // Internal representation of binary word of input data.
        uint8_t  bin_byte;                             // Internal representation of binary byte of input data.

        uint8_t*  bin_byte_addr;                       // Address of a byte of data in the bin_data buffer.

#define CSL_WORD_LEN       25                          // Characters required to describe 1 CSL (64bit) word.
#define CSL_BYTE_LEN       11                          // Characters required to describe 1 CSL (8bit) word.
#define BIN_DATA_LEN       1024                        // Read input in chunks.  Must be multiple of word size.
#define WORDS_PER_LINE     4                           // Can reasonably be between 1 and WORDS_PER_LINE_MAX.
#define WORDS_PER_LINE_MAX 5                           // Limit on number of words we'll display in a line.

        size_t   word_size         = sizeof(bin_word); // We process a "unit64_t" as a word.

        size_t   line_cnt          = 0;                // Currently processed input file line number.
        size_t   index             = 0;                // Subscript into a unit64_t word of bin_data array.
        size_t   words_written     = 0;                // Number of CSL words written per line.

        ssize_t  bytes_read        = 0;                // Bytes actually read in (per attempt) from binary file.

        char     csl_string[CSL_WORD_LEN];             // Buffer to create CSL string output.
        char     bin_data[BIN_DATA_LEN];               // Buffer to read in chunks of binary input.

        ZV_ASSERT(CSL_WORD_LEN >= (sizeof(bin_word) * 2) + CSL_BRACKET_WIDTH + 1);                     
        ZV_ASSERT(CSL_WORD_LEN >= (sizeof(bin_byte) * 2) + CSL_BRACKET_WIDTH + 1);                     

        ZV_ASSERT(BIN_DATA_LEN % word_size == 0);
        ZV_ASSERT(BIN_DATA_LEN == sizeof(bin_data));

        ZV_ASSERT(WORDS_PER_LINE > 0);
        ZV_ASSERT(WORDS_PER_LINE < WORDS_PER_LINE_MAX);

        // Parse the binary file input until we run out of data.
        while (true)
        {
            // Clear the entire temporary buffer, just to be safe.
            memset((void*) &bin_data[0], 0, BIN_DATA_LEN);

            bytes_read = ::read(bin_fd, &bin_data[0], BIN_DATA_LEN);

            if (bytes_read < 0) 
            { 
                ERRNO_RESULT_T   failure = errno;
                
                DIAGOUT(0, "Error reading binary file, error was \"%s\" (%d)",
                    strerror(failure), failure);
                return false; 
            }

            if (bytes_read == 0) { break; }

            // Process a full block of data, N words at a time.
            if (bytes_read == BIN_DATA_LEN) 
            { 
                for (index = 0; index < BIN_DATA_LEN; index+= (word_size * WORDS_PER_LINE))
                {
                    for (words_written = 0; words_written < WORDS_PER_LINE; words_written++)
                    {
                        bin_byte_addr = (uint8_t*) &bin_data[(index + (word_size * words_written))];

                        ::sprintf(csl_string, "[8X0000]%02X%02X%02X%02X%02X%02X%02X%02X,", 
                            bin_byte_addr[0], bin_byte_addr[1], bin_byte_addr[2], bin_byte_addr[3],
                            bin_byte_addr[4], bin_byte_addr[5], bin_byte_addr[6], bin_byte_addr[7]); 
                        ::write(csl_fd, (void*) csl_string, CSL_WORD_LEN);
                    }

                    ::write(csl_fd, "\n", 1); 
                    line_cnt++;
                }

                // Go grab another block.
                continue;
            }

            // If we have something more than one words worth of data in the buffer,
            // then handle them one word at a time.  But don't put more than WORDS_PER_LINE
            // of words on a line of output.
            if (bytes_read >= ((ssize_t) word_size)) 
            { 
                for (index = words_written = 0; index <= (bytes_read - word_size); index += word_size)
                {
                    bin_byte_addr = (uint8_t*) &bin_data[(index + (word_size * words_written))];

                    ::sprintf(csl_string, "[8X0000]%02X%02X%02X%02X%02X%02X%02X%02X,", 
                        bin_byte_addr[0], bin_byte_addr[1], bin_byte_addr[2], bin_byte_addr[3],
                        bin_byte_addr[4], bin_byte_addr[5], bin_byte_addr[6], bin_byte_addr[7]); 
                    ::write(csl_fd, (void*) csl_string, CSL_WORD_LEN);

                    if (++words_written >= WORDS_PER_LINE)  
                    {
                        ::write(csl_fd, "\n", 1); 
                        line_cnt++;
                        words_written = 0;
                    }
                }

                // The loop should have dropped us out before we overshot
                // the number of "wordsize" words.
                ZV_ASSERT(bytes_read >= ((ssize_t) index));

                // This should reduce the byte_read count by the number of words
                // we've written, which should be less than a word_size in value.
                bytes_read -= index;
            }

            ZV_ASSERT(bytes_read < ((ssize_t) word_size));

            // Write out the remaining few bytes as individual 
            // CSL instructions, one per byte.
            while (bytes_read > 0)
            {
                bin_byte_addr = (uint8_t*) &bin_data[index];
                bin_byte      = (uint8_t) *bin_byte_addr;

                ::sprintf(csl_string, "[1X0000]%02X,", bin_byte); 
                ::write(csl_fd, (void*) csl_string, CSL_BYTE_LEN);

                // Although we aren't writing 8 byte words, the concept is the same,
                // so use the same mechanism to force the last remaining CSL byte 
                // descriptions to not overflow the line.
                if (++words_written >= WORDS_PER_LINE)  
                {
                    ::write(csl_fd, "\n", 1); 
                    line_cnt++;
                    words_written = 0;
                }
            }

            // If we get here, we're done.
            break;

        }  // end of while (true)
        
        // If the caller supplied a pointer to a size_t, fill in the number of bytes written into shared memory.
        if (lines_written != NULL) {  *lines_written = line_cnt; }
           
        DIAGOUT(3, "Finished reading from fd %d into fd %d\n  created %ld lines of CSL data.\n", bin_fd, csl_fd, line_cnt);

        return true;

    }  // end of convert_bin_to_csl().

}; // end of class ZVisionCSL.

#endif // ZVISION_CSL_HPP

