/*
 * Copyright 2013 Azul Systems, Inc.  All rights reserved.
 * AZUL PROPRIETARY/CONFIDENTIAL.  Use is subject to license terms.
 */
#ifndef ZVISION_DIAG_HPP
#define ZVISION_DIAG_HPP

#include <errno.h>
#include <stdio.h>

#define __FILENAME__ (strrchr(__FILE__, '/')+1)

#define PATHOLOGICAL(format, ...) \
{ \
    fprintf(stderr, "%*.*s=>[% 6d %-20.20s]:  ", \
        ZVisionDiag::_prefix_len, ZVisionDiag::MAX_PREFIX_STRLEN, (const char*) ZVisionDiag::_prefix, __LINE__, __FILENAME__); \
    fprintf(stderr, format, ##__VA_ARGS__); \
    ZVisionDiag::abend(__LINE__, __FILE__, ::getpid()); \
}


#ifndef ShouldNotReachHere
#define ShouldNotReachHere() \
{ \
    fprintf(stderr, "ERROR: thread %d should not have reached line %d in file \"%s\".\n", ::getpid(), __LINE__, __FILE__); \
    while (1) {} \
}
#endif

// TODO: Replace with real thing.
#define ZV_ASSERT(...) \
{ \
    if (__VA_ARGS__) \
    {} \
    else \
    { \
        PATHOLOGICAL("ZV_ASSERT() failed at line %d in file \"%s\".\n", __LINE__, __FILE__); \
    } \
}

// Debugging output.
#define DIAGOUT(level, format, ...) \
{ \
    if ((int)level <= (int)ZVisionDiag::_verbosity) \
    { \
        fprintf(stderr, "%*.*s %lu[% 6d %-20.20s]:  ", \
            ZVisionDiag::_prefix_len, ZVisionDiag::MAX_PREFIX_STRLEN, (const char*) ZVisionDiag::_prefix, \
            ((size_t) level), __LINE__, __FILENAME__); \
        fprintf(stderr, format, ##__VA_ARGS__); \
    } \
}



// Debugging without the prefix.
#define DIAGOUTNOPREFIX(level, format, ...) \
{ \
    if ((int)level <= (int)ZVisionDiag::_verbosity) \
    { \
        fprintf(stderr, format, ##__VA_ARGS__); \
    } \
}

#define DIAGBUFFEROUT(level, addr, size)  \
{ \
    if ((int)level <= (int)ZVisionDiag::_verbosity) \
    { \
        ZVisionDiag::print_buffer_contents(level, addr, size); \
    } \
}


/**
 ** class ZVisionDiag - ZVision Diagnostics.
 **/
class ZVisionDiag
{
  public:

    /**
     ** Enums for this class.
     **/
    enum
    {
        MAX_PREFIX_STRLEN              = 30,
        MAX_PATHOLOGICAL_ERRORS        = 10,
        MAX_PRINTABLE_BUFFER_CONTENTS  = 1000000,
        LINE_OUTPUT_LEN                = 66, // ::strlen("00000   0x0000000000000000  0x0000000000000000   ................")
    };

    static const char*        _prefix[MAX_PREFIX_STRLEN];     // Abbreviation string placed in front of error message.
    static int                _prefix_len;                    // Length of the prefix string.
    static volatile bool      _triggered;                     // Has an error already been recorded?
    static volatile size_t    _index;                         // Offset into the line/file array of the error order.
    static int                _pid[MAX_PATHOLOGICAL_ERRORS];  // The Unix process id of the thread in error.
    static size_t             _line[MAX_PATHOLOGICAL_ERRORS]; // At what line number did the macro get invoked?
    static char*              _file[MAX_PATHOLOGICAL_ERRORS]; // In what source file did the error occur?
    static size_t             _verbosity;                     // Level of debugging output.

    /**
     ** Set the Debugging Output prefix string, which will be included before each output line.
     **/
    static void set_prefix(const char* prefix) 
    { 
        if (prefix == NULL) { return; }

        size_t len = ::strnlen(prefix, (MAX_PREFIX_STRLEN-1));

        ::memset(_prefix, 0, MAX_PREFIX_STRLEN);
        ::memcpy(_prefix, prefix, len);

        _prefix_len = len;

    }  // end of set_prefix().


    /**
     ** Abnormal End - The process has encountered an error that can not be resolved.  This
     ** probably means we are shutting down, and that there will be several additional errors
     ** that follow on (or result from) this situation.  Record the details of the pathological
     ** error in the order in which they occur, saving the line number and file name of the
     ** source code line in which the error was encountered..
     **/
    static void abend(size_t line, const char* filename, int pid)
    {
        volatile bool current_state = _triggered;

        while (current_state != true)
        {
            if (__sync_bool_compare_and_swap(&_triggered, current_state, true) == true)
            {
                _line[0]  = line;
                _file[0]  = (char*) filename;
                _pid[0]   = pid;
                return;
            }

            current_state = _triggered;
        }

        volatile size_t current_index = _index;
        volatile size_t next_index    = current_index + 1;

        if (next_index >= MAX_PATHOLOGICAL_ERRORS) { return; }

        while (__sync_val_compare_and_swap(&_index, current_index, next_index) != current_index)
        {
            current_index = _index;
            next_index    = current_index + 1;
        }

        _line[next_index]  = line;
        _file[next_index]  = (char*) filename;
        _pid[next_index]   = pid;

        return;

    } // end of abend().


    /**
     ** Display the pathological error stack, if there were errors, print to standard error.
     **/
    static void print_error_stack(const char* title) 
    { 
        if (_triggered == true)
        {
            volatile size_t count = _index;

            fprintf(stderr, "%*.*s  [% 6d %-20.20s]:  ", \
                ZVisionDiag::_prefix_len, ZVisionDiag::MAX_PREFIX_STRLEN, \
                (const char*) ZVisionDiag::_prefix, __LINE__, __FILENAME__);
            fprintf(stderr, "Recorded Error Stack");

            if (_prefix != NULL)
            {
                fprintf(stderr, " for %*.*s", _prefix_len, (MAX_PREFIX_STRLEN-1), (const char*) title);
            }

            fprintf(stderr, ":\n");

            for (size_t index = 0; index <= _index; index++)
            {
                fprintf(stderr, "%*.*s  [% 6d %-20.20s]:  ", \
                    ZVisionDiag::_prefix_len, ZVisionDiag::MAX_PREFIX_STRLEN, \
                    (const char*) ZVisionDiag::_prefix, __LINE__, __FILENAME__);
                fprintf(stderr, "#%u pid: %d line: %u file: \"%s\".\n", \
                    (unsigned int) index, _pid[index], (unsigned int) _line[index], _file[index]);
            }

            if (count >= MAX_PATHOLOGICAL_ERRORS) 
            { 
                fprintf(stderr, "%*.*s  [% 6d %-20.20s]:  ", \
                    ZVisionDiag::_prefix_len, ZVisionDiag::MAX_PREFIX_STRLEN, \
                    (const char*) ZVisionDiag::_prefix, __LINE__, __FILENAME__);
                fprintf(stderr, "NOTE: Exceeded stack limit (%d), addtional errors disregarded :\n", MAX_PATHOLOGICAL_ERRORS);
            }
        }

        return;

    } // end of print_error_stack().


    /**
     **  Print the contents of the buffer to stderr, using the DIAGOUT macros.
     **/
    static void print_buffer_contents(size_t verbosity, volatile char* data, size_t size)
    {
        if (data == NULL)
        {
            PATHOLOGICAL("Data was NULL.\n");
            return;
        }

        if (size <= 0)
        {
            PATHOLOGICAL("Illegal buffer size (%u) parameter value.\n", (unsigned int) size);
            return;
        }

        if (size > MAX_PRINTABLE_BUFFER_CONTENTS)
        {
            DIAGOUT(verbosity, "buffer size (%u) exceeds limit (%u).\n", \
                (unsigned int) size, (unsigned int) MAX_PRINTABLE_BUFFER_CONTENTS);
            size = MAX_PRINTABLE_BUFFER_CONTENTS;
        }

        size_t null_lines = 0;
        size_t index      = 0;

        while (index < size)
        {
            char  buf[LINE_OUTPUT_LEN];       // ::strlen("00000   0x0000000000000000  0x0000000000000000   ................")
            bool  all_null = true;

            sprintf(&buf[0], "%05u   0x", (unsigned int) index);

            for (size_t loop = 0; loop < 8; loop++)
            {
                sprintf(&buf[10 + (loop*2)], "%02X", (uint8_t) data[index + loop]);

                if (data[index + loop] != 0) { all_null = false; }
            }

            index+=8;

            sprintf(&buf[26], "  0x");
            for (size_t loop = 0; loop < 8; loop++)
            {
                sprintf(&buf[30 + (loop*2)], "%02X", (uint8_t) data[index + loop]);

                if (data[index + loop] != 0) { all_null = false; }
            }

            index+=8;

            sprintf(&buf[46], "   ");

            for (size_t loop = 0; loop < 16 ; loop++)
            {
                size_t offset = (index - 16) + loop;
                char   letter = data[offset];

                sprintf(&buf[49 + loop], "%c", isprint(letter) ? letter : '.');
            }

            if (all_null == false)
            {
                if (null_lines > 0)
                { 
                      DIAGOUT(verbosity, "        %u %s all zeros.\n", 
                          (unsigned int) null_lines, null_lines == 1 ? "line was" : "lines were"); 
                }
                null_lines = 0;
                DIAGOUT(verbosity, "%s\n", buf);
            }
            else
            {
                null_lines++;
            }
        }

        if (null_lines > 0) 
        {   
            DIAGOUT(verbosity, "        %u %s all zeros.\n", 
                (unsigned int) null_lines, (null_lines == 1) ? "line was" : "lines were"); 
        }

        DIAGOUT(verbosity, "%05u   --END--\n", (unsigned int) index);


        return;

    } // end of print_buffer_contents().


    /**
     ** Debugging output level of the JVM.  0 = show nothing, 4 = show everything.
     **/
    static inline size_t verbosity()                 { return _verbosity;  }
    static inline void   set_verbosity(size_t level) { _verbosity = level; }
    static inline void   clr_verbosity()             { _verbosity = 0;     }


};  // end of class ZVisionDiag

#endif // ZVISION_DIAG_HPP

// end of file zvision_diag.hpp.

