/*
 * Copyright 2013 Azul Systems, Inc.  All rights reserved.
 * AZUL PROPRIETARY/CONFIDENTIAL.  Use is subject to license terms.
 */
#ifndef ZVISION_HA_HPP
#define ZVISION_HA_HPP

#include "zvision.hpp"

#define ZVisionHostAgentFailIfNoMatchingVersion  true

// This helper class is here to allow us to arbitrate between the Host Agent and the ZVision Service Thread.
// It allows us to create locks around a given section type, so that if the underlying shared memory needs to
// change dynamically, we can avoid a SIGSEGV.
class ZVisionHostAgentSectionLock : public ZVision
{
  public:

    // Depending on when they were created, the Host Agent's (and JVM's) notion of how many section types there
    // are may differ.  We choose the smaller of the two, and ignore any requests or notifications that exceed it.
    static SECTION_TYPE_E    last_jvm_type;
    static SECTION_TYPE_E    last_ha_type;

  private:

    SECTION_TYPE_E         _type;            // The section type (most probably tick_profiling) this lock governs
    SECTION_LOCK_T         _section_lock;    // The section lock value in the local object.  If zero, no unlock is required.
    FILE_DATA_T*           _file_data;       // The type_stable memory from which we obtain the section type's file_data.

  public:

    inline ZVisionHostAgentSectionLock(FILE_DATA_T* file_data, bool enable, SECTION_TYPE_E limit)
    {
      SECTION_LOCK_T compare_value;
      SECTION_LOCK_T exchange_value;

      // Set this now.  If the destructor is called, and finds a zero
      // in this field, it knows not to do anything but return.
      _section_lock = 0;

      // Only arbitrate for the lock if we need it.
      if (!enable) { return; }

      if (file_data == NULL || file_data->cookie != COOKIE_FILE_DATA)
      {
        // TODO: Need to signal an error has occurred!
        ZV_ASSERT(false);
        return;
      }
      
      // Ignore requests that we don't support.
      if (file_data->type >= limit) { return; }

      // Arbitrate for the lock.  Each section-type has a lock that
      // contains the "reset iteration" count (number of times we've
      // changed the memory pointers to that section) which is only
      // ever incremented by the EventTickBuffer::resize_memory()
      // method.  If we need to touch that memory, we'll grab the
      // lock first (we set the high order bit, but not increment the
      // "reset iteration" value) and then (when the destructor runs)
      // we release the lock by clearing the high order bit.
      do
      {
        compare_value = (SECTION_LOCK_T) file_data->section_lock;

        // If the lock is zero, then something is really wrong.
        if (compare_value == 0)
        { 
          // TODO: Need to signal an error has occurred!
          ZV_ASSERT(false);
          return;
        }

        // If its in use by the JVM, then we can't get it, keep trying...
        // TODO: Need to limit the spins.
        if ((compare_value & ZVISION_JVM_SECTION_LOCK_MASK) != 0) { continue; }

        // If the section is already locked by the Host Agent, then another
        // Host Agent thread has it, so try again...
        // TODO: Need to limit the spins.
        if ((compare_value & ZVISION_HA_SECTION_LOCK_MASK) != 0) { continue; }

        // Set the lock bit.
        exchange_value = compare_value | ZVISION_HA_SECTION_LOCK_MASK;
        exchange_value = __sync_val_compare_and_swap(&file_data->section_lock, compare_value, exchange_value);

      } while (exchange_value != compare_value);

      // Record the section_lock details, so that the destructor knows what to do.
      _section_lock = file_data->section_lock;
      _file_data    = file_data;

    } // end of ZVisionHostAgentSectionLock().


    // The destructor will free the lock.  We should be the one holding
    // the lock (i.e. the high order bit should be set, and the object's
    // local value should match the section_lock field), if it doesn't, 
    // we've got an error.  Otherwise, clear the local field and free
    // the section_lock.
    inline ~ZVisionHostAgentSectionLock()
    {
      // If no local lock was set, then there is nothing for us to do.
      if (_section_lock == 0) { return; }

      if (_file_data == NULL || _file_data->cookie != COOKIE_FILE_DATA)
      {  
        // TODO: Need to signal an error has occurred!
        ZV_ASSERT(false);
        return; 
      }

      // Therefore no one should have updated the lock while we had possession.
      if (_section_lock != _file_data->section_lock)
      {  
        // TODO: Need to signal an error has occurred!
        ZV_ASSERT(false);
        return; 
      }

      // The lock value should have the bit set to indicate it belongs to us.
      if ((_section_lock & ZVISION_HA_SECTION_LOCK_MASK) == 0)
      {
        // TODO: Need to signal an error has occurred!
        ZV_ASSERT(false);
        return; 
      }

      // Now release the lock.
      _file_data->section_lock ^= ZVISION_HA_SECTION_LOCK_MASK;

    } // end of ~ZVisionHostAgentSectionLock().

};  // end of class ZVisionHostAgentSectionLock


/**
 **/
class ZVisionHostAgent : public ZVision
{
  public:

    /**
     ** This enum describes the connection status of the associated JVM, from the Host Agent's perspective.
     ** You can also look at the ipc_state(), to see what the JVM thinks is happening.
     **/
    enum JVMConnectStatus         
    {
        JVM_hello      = 111,  //  The JVM contacted the Host Agent on its well-known port.
        HA_start       = 222,  //  The Host Agent has initialized and sent a start_token request to the JVM.
        JVM_connect    = 333,  //  The JVM has sent a start_token response back, and the connection is established.
        HA_end         = 444,  //  The HA has sent a shutdown request.
        JVM_end        = 555,  //  The JVM has sent a shutdown notification back, and is exiting.
        JVM_exit       = 666,  //  The JVM is exiting unexpectedly.
        HA_error       = 888,  //  The Host Agent encountered an error.
        JVM_error      = 999,  //  The JVM encouhntered an error and died. 
    };


  private:
    
    // Each JVM, as part of the connection process, creates it own object within the Host Agent's ZVision API.  The 
    // fields within the object are specific only to that particular JVM, which may be at any given revision, and
    // therefore have differing properties (such as the number of section types supported) than other JVMs with which
    // the Host Agent has established a connection.
    
    TYPE_STABLE_LAYOUT_T*    _ts;                             // (shared) memory-mapped type_stable file addr created by the JVM.
    char                     _name[MAX_IPC_SOCKET_NAME_SIZE]; // Fully qualified type_stable file name for this JVM.
    JVMConnectStatus         _status;                         // Current connection status from the Host Agent's perspective.
    size_t                   _jvm_seq_num;                    // Sequence number assigned by the Host Agent for cross-checking.
    JVM_KEY_T                _jvm_key;                        // Security key generated by the Zing JVM.
    PID_T                    _jvm_pid;                        // The JVM's Unix Process id.
    HA_HANDLE_T              _handle;                         // Host Agent's indentifying handle for the JVM, assigned by HA.
    SECTION_TYPE_E           _last_jvm_type;                  // The last valid JVM section type, may differ depending on release.
    SECTION_TYPE_E           _last_ha_type;                   // The highest type this connection will support, min of JVM & HA. 

  public:

    /**
     ** ZVisionHostAgent constructor method - Initialize the Host Agent ZVision API.  This call creates an object
     ** for each connection with a Zing Virtual Machine.   The JVM must be on the same machine (same virtual address
     ** space) as the Host Agent, so that they can both (simultaneously) access the shared (memory-mapped) ZRC data.
     **
     ** This method is called by the ZVisionHostAgent::jvm_hello() method, when the Host Agent receives a JVM "hello"
     ** message.  The Host Agent polls on a well-know port it established, and upon receiving the message from the JVM,
     ** instantiates a new Host Agent object.  The object establishes a connection with the JVM using information passed
     ** to it in the hello message.
     **/
    ZVisionHostAgent(const char* basename, JVM_KEY_T jvm_key, PID_T jvm_pid, SOCKET_FD_T output_fd = -1)
    {
        DIAGOUT(1, ">>> ZVisionHostAgent(basename: \"%s\", key: 0x%lX, pid: %d, output fd %d) request for new JVM connection!!\n", 
            basename, jvm_key, jvm_pid, output_fd);
    
        SOCKET_FD_T                    ts_fd            = -1;
        size_t                         basename_len     = ::strnlen(basename, MAX_ZVISION_PATHNAME_LEN);
        const CONSTANTS_T*             ts_const         = constants_by_type(type_stable);

        ERRNO_RESULT_T     failure                   = 0;

       // This value is assigned after the object is created, by the Host Agent unit test logic.
       // It is not needed, and used only as a cross checking value.
        _jvm_seq_num = ~0;

        // If we haven't already, call ZVision common initialization routines.
        if (zvision_init() != true)
        {
            PATHOLOGICAL("zvision_init() failed.\n");
            _status = HA_error;
            return; 
        }
    
        // The JVM has announced it is alive, and this object got created as a result of that.
        _status = JVM_hello;

        // Establish the type_stable file basename.
        if (basename == NULL)
        {
            PATHOLOGICAL("Type Stable shared memory basename is NULL.\n");
            _status = HA_error;
            return;
        }
    
        // Make sure the type stable constants are valid.
        if (ts_const == NULL)
        {
            PATHOLOGICAL("type_stable_constants pointer was NULL.\n");
            _status = HA_error;
            return; 
        }

        // Verify that the user passed us a reasonable type_stable basename.
        if (basename_len >= MAX_IPC_SOCKET_NAME_SIZE)
        {
            PATHOLOGICAL("Constructed type stable file name len (%ld) exceeds available namespace size (%ld).\n", 
                basename_len, MAX_IPC_SOCKET_NAME_SIZE); 
            _status = HA_error;
            return;
        }
        
        DIAGOUT(3, "Issuing open(\"%s\", O_RDWR | O_NONBLOCK)...\n", basename);

        // We open this file read/write, but because this is the Host Agent, we don't
        // want to create it, nor can we wait on it.
        ts_fd = ::open(basename, O_RDWR | O_NONBLOCK);

        if (ts_fd < 0)
        {
            failure = errno;
            PATHOLOGICAL("open(name=\"%s\") failed with errno set to %d (%s).\n", basename, failure, ::strerror(failure)); 
            _status = HA_error;
            return;
        }
    
        // The JVM has already created the file, we are simply mapping it into memory using the
        // file descriptor already established when we opened the file to read its header contents.
        // We map it as shared; Because the type_stable file is fixed size, and little accessed during
        // the normal operation of the JVM, we treat it differently than other files, and are willing
        // to pay the cost of a file system backed entity.
        void* addr = mmap_file(ts_fd,                                    // File descriptor of the type_stable file.
                               sizeof(TYPE_STABLE_LAYOUT_T),             // Size of the type_stable file contents.
                               (char*) xlate_section_type(type_stable)); // The name of the file type.

        if (addr == MAP_FAILED)
        {
            PATHOLOGICAL("Host Agent's mmap_file() returned MAP_FAILED for type_stable file.\n");
            if (::close(ts_fd) != 0) { PATHOLOGICAL("::close(%d) failed.\n", ts_fd); }
            _status = HA_error;
            return; 
        }

        // Initialize the local ZVisionHostAgent object with the information we
        // received from the JVM's hello call.
        _ts         = (TYPE_STABLE_LAYOUT_T*) addr;
        _jvm_key    = jvm_key;
        _jvm_pid    = jvm_pid;
        ::memcpy((void*) &_name[0], basename, MAX_IPC_SOCKET_NAME_SIZE);

        // Depending on when they were created, the Host Agent's (and JVM's) notion of how many section types there
        // are may differ.  We choose the smaller of the two, and ignore any requests or notifications that exceed it.
        _last_jvm_type = (SECTION_TYPE_E) (((size_t) start_token) + _ts->jvm_section_cnt - 1);
        _last_ha_type  = (end_token <  _last_jvm_type) ? end_token : _last_jvm_type;

        DIAGOUT(2, "Higest JVM section type is %d, HA is %d.\n", _last_jvm_type, _last_ha_type);
        DIAGOUT(3, "Accessing the JVM-created section entry files based on the type_stable's section entry data...\n");

        for (size_t loop = (size_t) start_token; loop < (size_t) _last_ha_type; loop++)
        {
            SECTION_TYPE_E     type          = (SECTION_TYPE_E) loop;
            SECTION_ENTRY_T*   entry         = section_entry_by_type(_ts, type);
            const CONSTANTS_T* constants     = constants_by_type(type);

            if (constants == NULL)
            {
                PATHOLOGICAL("Constants pointer for \"%s\" entry was NULL.\n", xlate_section_type(type));
                if (::close(ts_fd) != 0) { PATHOLOGICAL("::close(%d) failed.\n", ts_fd); }
                _status = HA_error;
                return; 
            }

            if (constants->cookie != COOKIE_CONSTANTS)
            {
                PATHOLOGICAL("Corrupted cookie (within constants struct) in \"%s\" entry, expected 0x%lX, got 0x%lX instead.\n", 
                    xlate_section_type(type), COOKIE_CONSTANTS, constants->cookie);
                if (::close(ts_fd) != 0) { PATHOLOGICAL( "::close(%d) failed.\n", ts_fd); }
                _status = HA_error;
                return; 
            }

            if (entry->cookie != COOKIE_SECTION_ENTRY)
            {
                PATHOLOGICAL("Corrupted SECTION_ENTRY_T in \"%s\" entry.\nExpected 0x%lX but found 0x%lX.\n",
                    xlate_section_type(type), COOKIE_SECTION_ENTRY, entry->cookie); 
                if (::close(ts_fd) != 0) { PATHOLOGICAL( "::close(%d) failed.\n", ts_fd); }
                _status = HA_error;
                return;
            }
       
            if (entry->file.cookie != COOKIE_FILE_DATA)
            {
                PATHOLOGICAL("Corrupted FILE_DATA_T in \"%s\" entry.\nExpected 0x%lX but found 0x%lX.\n",
                    xlate_section_type(type), COOKIE_FILE_DATA, entry->file.cookie); 
                if (::close(ts_fd) != 0) { PATHOLOGICAL("::close(%d) failed.\n", ts_fd); }
                _status = HA_error;
                return;
            }
      
            if (entry->jvm_funcs.cookie != COOKIE_JVM_FUNCS)
            {
                PATHOLOGICAL("Corrupted JVM FUNC_DATA_T in \"%s\" entry.\nExpected 0x%lX but found 0x%lX.\n",
                    xlate_section_type(type), COOKIE_JVM_FUNCS, entry->jvm_funcs.cookie); 
                if (::close(ts_fd) != 0) { PATHOLOGICAL("::close(%d) failed.\n", ts_fd); }
                _status = HA_error;
                return;
            }

            if (entry->ha_funcs.cookie != COOKIE_HA_FUNCS)
            {
                PATHOLOGICAL("Corrupted HA FUNC_DATA_T in \"%s\" entry.\nExpected 0x%lX but found 0x%lX.\n",
                    xlate_section_type(type), COOKIE_HA_FUNCS, entry->ha_funcs.cookie); 
                if (::close(ts_fd) != 0) { PATHOLOGICAL("::close(%d) failed.\n", ts_fd); }
                _status = HA_error;
                return;
            }

            // If it is the type_stable entry, then we've already processed this file above, 
            // just fill in the details in the section entry, mostly for completeness.
            if (loop == (size_t) type_stable) 
            { 
                entry->file.ha_mmap_fd   = ts_fd;
                entry->file.ha_mmap_addr = addr;
                continue; 
            } 

            // Only open the files that have been created by the JVM.
            if (entry->file.allocated != true || entry->file.active != true)
            {
                DIAGOUT(3, "Skipping Cross Reference initalization for section \"%s\".\n", xlate_section_type(type));
                continue;
            }

            char buf[RESULT_SIZE] = { 0 };

            xlate_version_struct((VERSION_INFO_T*) &(_ts->version), buf);

            DIAGOUT(3, "Looking up Cross Reference for section \"%s\", version %s.\n", xlate_section_type(type), buf);

            // Now get the appropriate cross reference method funcs structure address for this type & version,
            // which may (or may not) match what the JVM used.   Since it is possible that the JVM's ZVision API
            // rev is out-of-date, or uses a version we don't yet support, we first check to see if there is a
            // matching version.
            const CROSS_REF_ENTRY_T* xref = get_method_info(type, (VERSION_INFO_T*) constants->version);
    
            if (xref == NULL)
            { 
                if (ZVisionHostAgentFailIfNoMatchingVersion)
                {
                    DIAGOUT(0, "ZVision API for type \"%s\" (%d), version \"%s\" is unsupported.\n",
                        xlate_section_type(type), type, xlate_version_struct((VERSION_INFO_T*) &(_ts->version), buf));

                    DIAGOUT(0, "Latest supported  Host Agent ZVision version is \"%s\".\n",
                        xlate_version_struct((VERSION_INFO_T*) current_sw_ver(), buf));

                    DIAGOUT(0, "Earliest supported Host Agent ZVision version is \"%s\".\n",
                        xlate_version_struct((VERSION_INFO_T*) earliest_sw_ver(), buf));

                    if (::close(ts_fd) != 0) { PATHOLOGICAL("::close(%d) failed.\n", ts_fd); }
                    _status = HA_error;
                    return; 
                }

                entry->ha_funcs.compress_buf_func = NULL;
                DIAGOUT(0, "WARNING!!  No matching version support for section type \"%s\" (%d), forgoing future requests.\n", 
                    xlate_section_type(type), type);
                set_forgo_data(_ts, type);
                continue;
            }
            else
            {
                if (xref->ha_funcs.compress_buf_func == NULL && ZVisionHostAgentFailIfNoMatchingVersion)
                {
                }

                // We now have the cross reference data, so fill the Host Agent function pointers within 
                // the cross reference section entry data.
                entry->ha_funcs.compress_buf_func = xref->ha_funcs.compress_buf_func;
            }
    
            // If we are using SYSV shared memory (i.e. ram that is not backed by disk) then open
            // this file using the alternative SYSV methods, otherwise default to the BSD open call.
            if (constants->use_dev_shm == true)
            {
                DIAGOUT(3, "Issuing shm_open(name=\"%s\", O_RDWR, 0x%X) in \%s\".\n", 
                    entry->file.name, constants->open_mode, SHMEM_DIR_STR);
        
                entry->file.ha_mmap_fd = ::shm_open(entry->file.name, O_RDWR, constants->open_mode);
    
                if (entry->file.ha_mmap_fd < 0)
                {
                    failure = errno;
                    PATHOLOGICAL("shm_open(name=\"%s\") failed with errno set to %d (%s).\n", 
                      entry->file.name, failure, ::strerror(failure)); 
                    _status = HA_error;
                    return;
                }
            }
            else
            {
                DIAGOUT(3, "Issuing open(name=\"%s\", O_RDWR, 0x%X)...\n", 
                    entry->file.name, constants->open_mode);
        
                entry->file.ha_mmap_fd = ::open(entry->file.name, O_RDWR, constants->open_mode);
    
                if (entry->file.ha_mmap_fd < 0)
                {
                    failure = errno;
                    PATHOLOGICAL("open(name=\"%s\") failed with errno set to %d (%s).\n", 
                      entry->file.name, failure, ::strerror(failure)); 
                    _status = HA_error;
                    return;
                }
            }

            // The file is open (either through a SYSV shm_open or standard BSD open system call), now map 
            // in the memory, and update the section entry within the type stable file.
            entry->file.ha_mmap_addr = mmap_file(entry->file.ha_mmap_fd, 
                                                 entry->file.max_size,
                                                 (char*) xlate_section_type(type));

            if (entry->file.ha_mmap_addr == MAP_FAILED) 
            {
                PATHOLOGICAL("mmap_file() returned MAP_FAILED for \"%s\" (%d) file.\n", xlate_section_type(type), type);
                _status = HA_error;
                return; 
            }

            entry->file.ha_mmap_len = entry->file.max_size;
            entry->file.ha_mmap_gen = 1;

            // If this is a request buffer, then initialize the header information.
            if (constants->file_type == ZVision::request_response_file)
            {
                ZVision::REQUEST_BUFFER_T* req_buf = (ZVision::REQUEST_BUFFER_T*) entry->file.ha_mmap_addr;

                req_buf->hdr.cntl.request_id    = ZVision::host_agent_pid(_ts) == 0 ? ::getpid() : ZVision::host_agent_pid(_ts);
                req_buf->hdr.cntl.response_id   = ZVision::jvm_pid(_ts);
                req_buf->hdr.cntl.cookie        = ZVision::COOKIE_REQUEST_BUF;
             }

            // Make sure the compress_buf_func points at something.
            if (entry->ha_funcs.compress_buf_func == NULL)
            { 
                PATHOLOGICAL("\"%s\" entry Get function was NULL.\n", xlate_section_type(type));
                _status = HA_error;
                return; 
            }

            entry->file.ha_output_cnt = 0;
            entry->file.ha_output_fd  = output_fd;

            DIAGOUT(3, "%s \"%s\" initialized.\n", xlate_file_type(constants->file_type), xlate_section_type(type));

        }  // end of for (size_t loop = (size_t) start_token; loop <= (size_t) last_ha_type; loop++)
    
        // This connects to the port that should have already been established by the JVM.
        // It also initiates the JVM's state change to ipc_state_connection_est.
        char*               ipc_name     = _ts->ipc_domain_id;
        size_t              ipc_name_len = (ipc_name != NULL) ? ::strnlen(ipc_name, MAX_IPC_SOCKET_NAME_SIZE) : 0;
        struct sockaddr_un  socket_info;
        struct sockaddr_un* socket_info_addr = &socket_info;

        if (ipc_name_len <= 0)
        {
            PATHOLOGICAL("Unix Domain Socket Filename length is zero.\n");
            _status = HA_error;
            return; 
        }

        if (ipc_name_len >= sizeof(socket_info.sun_path))
        {
            PATHOLOGICAL("Unix Domain Socket Filename length (%lu) exceeds \"sun_path\" size (%lu).\n", 
                ipc_name_len, sizeof(socket_info.sun_path));
            _status = HA_error;
            return; 
        }

        DIAGOUT(3, ">>>>  Connecting to JVM using \"%s\".  <<<<<\n", ipc_name);

        ::memset(&socket_info, 0, sizeof(struct sockaddr_un));
        socket_info.sun_family = AF_UNIX;
        ::memcpy(socket_info.sun_path, _ts->ipc_domain_id, ipc_name_len);

        DIAGOUT(3, "calling socket(AF_UNIX, SOCK_SEQPACKET, 0).\n");

        _ts->ha_ipc_fd = ::socket(AF_UNIX, SOCK_SEQPACKET, 0);

        if (_ts->ha_ipc_fd <= 0)
        {
            failure = errno;
            PATHOLOGICAL("Could not open send fd %d.  Error is %d \"%s\".\nExiting prematurely.\n", 
                _ts->ha_ipc_fd, failure, ::strerror(failure));
            _status = HA_error;
            return; 
        }

        DIAGOUT(3, "socket(AF_UNIX, SOCK_SEQPACKET, 0) returned fd %d.\n", _ts->ha_ipc_fd);

        DIAGOUT(3, "calling connect(fd %d, sock_info %p, size %lu).\n", 
            _ts->ha_ipc_fd, socket_info_addr, sizeof(struct sockaddr_un));

        if (::connect(_ts->ha_ipc_fd, (struct sockaddr*) socket_info_addr, sizeof(struct sockaddr_un)) != 0)
        {
            failure = errno;

            PATHOLOGICAL("::connect(%d, %p, %lu) failed for send fd.  Error is %d \"%s\"\nExiting prematurely.\n", 
                _ts->ha_ipc_fd, socket_info_addr, sizeof(struct sockaddr_un), failure, ::strerror(failure));
            _status = HA_error;
            return; 
        }
          
        DIAGOUT(2, "Connected to JVM on fd %d, file is: \"%s\".\n", _ts->ha_ipc_fd, socket_info_addr->sun_path);

        // Double check that the jvm state hasn't been corrupted.
        if (ZVision::jvm_is_ready(_ts) != true)
        {
            PATHOLOGICAL("host_agent_zvision_init() completed but JVM (PID %d) is now not ready.\n", ZVision::jvm_pid(_ts));
            _status = JVM_error;
            return; 
        }

        if (ZVision::host_agent_is_ready(_ts) == true)
        {
            PATHOLOGICAL("Host Agent (PID %d) was ready (non-zero) during initialization.\n", ZVision::host_agent_pid(_ts));
            _status = HA_error;
            return; 
        }

        FILE_DATA_T* file_data = file_data_by_type(_ts, start_token);

        if (file_data == NULL)
        {
            PATHOLOGICAL("file_data for start_token (%d) was NULL.\n", start_token);
            _status = HA_error;
            return; 
        }

        volatile REQUEST_BUFFER_T* req_buf = (volatile REQUEST_BUFFER_T*) file_data->ha_mmap_addr;

        if (req_buf == NULL)
        {
            PATHOLOGICAL("req_buf for start_token (%d) was NULL.\n", start_token);
            _status = HA_error;
            return; 
        }

        // The handle is an indentifying token that the Host Agent will use to recognize one JVM from another when it
        // receives future notification messages.   Using the ZVisionHostAgent object address should make the value 
        // unique between JVMs.
        _handle = (HA_HANDLE_T) this;

        char*  param_buf_addr = (char*) &req_buf->params[0];
        size_t param_buf_len  = MAX_REQUEST_PARAMS_SIZE;

        ZV_ASSERT(param_buf_len == sizeof(req_buf->params));

        // Build a JVM start message, this contains the basename, jvm_key and process id sent to us
        // by the JVM when it started up.
        JVM_HELLO_MSG_T   hello_msg;
        JVM_HELLO_MSG_T*  hello_msg_addr = &hello_msg;

        size_t  hello_msg_len = sizeof(JVM_HELLO_MSG_T);
        size_t  copy_len       = (hello_msg_len < param_buf_len) ? hello_msg_len : param_buf_len;

        ::memset((void*) hello_msg_addr, 0, hello_msg_len);

        hello_msg.jvm_key = jvm_key;
        hello_msg.jvm_pid = jvm_pid;
        hello_msg.handle  = _handle;
        hello_msg.cookie  = COOKIE_HELLO_MSG;

        ::memcpy((void*) &hello_msg.name[0], basename, MAX_IPC_SOCKET_NAME_SIZE);

        // Fill in the start_token request buffer with the contents of the JVM_HELLO_MSG_T notification.
        req_buf->hdr.cntl.type           = start_token;
        req_buf->hdr.cntl.params_len     = hello_msg_len;
        req_buf->hdr.cntl.params_cnt     = 1;
        req_buf->hdr.cntl.request_id     = ::getpid();
        req_buf->hdr.cntl.request_size   = param_buf_len;
        req_buf->hdr.cntl.request_token  = 1;
        req_buf->hdr.cntl.response_id    = _ts->jvm_pid;   // The prior ::memset() reset this value.
        req_buf->hdr.cntl.cookie         = COOKIE_REQUEST_BUF;

        ::memset((void*) param_buf_addr, 0, param_buf_len);
        ::memcpy((void*) param_buf_addr, (const void*) hello_msg_addr, copy_len);

        DIAGOUT(1, "Host Agent sending \"%s\" to JVM...\n", xlate_section_type(start_token));
        DIAGOUT(1, "Handle is 0x%lu, key is 0x%lX, pid %d, state  is \"%s\" (0x%X).\n", 
            _handle, jvm_key, jvm_pid, xlate_jvm_connect_status(_status), _status);

        // Send a start_token "request" to the JVM.  The request buffer contains the JVM key, a handle that the
        // JVM will use on all future requests, JVM's Unix process id, and type_stable filename that the JVM
        // passed back to us when it sent an unsolicited message to the Host Agent's well-known socket.
        if (send_request(start_token, 1, (const char*) req_buf) != true)
        {
            PATHOLOGICAL("Could not send a start message to the JVM. Exiting prematurely.\n");
            _status = HA_error;
            return; 
        }

        DIAGOUT(2, "Send succesful, changing state  to \"%s\" (0x%X).\n", xlate_jvm_connect_status(HA_start), HA_start);

        // This should always be the last thing we do in the constructor, and if the state isn't set to this,
        // the invoking logic will know that something went wrong during object creation.
        _status = HA_start;

        return;
    
    }  // end of ZVisionHostAgent().


    /**
     ** Shut down the connection with a JVM, or reset it.
     **/
    ~ZVisionHostAgent()
    {
        ZVisionDiag::print_error_stack("ZVision Host Agent");
 
        DIAGOUT(1, "Deleting ZVisionHostAgentObject: handle 0x%lu, JVM status \"%s\", fd %d, basename is \"%s\".\n", 
            _handle, xlate_jvm_connect_status(_status), _ts->ha_ipc_fd, _name);

        for (size_t loop = (size_t) start_token; loop <= (size_t) _last_ha_type; loop++)
        {
            SECTION_TYPE_E     type          = (SECTION_TYPE_E) loop;
            SECTION_ENTRY_T*   entry         = section_entry_by_type(_ts, type);

            // We'll close this one at the end of the loop, it is self-referential
            // and there are fields within it we may access before we finish shutting
            // down the rest of the JVM connection.
            if (type == type_stable) { continue; }

            if (entry->file.ha_mmap_fd >= 0)
            {
                DIAGOUT(3, "Closing Host Agent fd %d, \"%s\" of JVM %lu, handle 0x%lu.\n", 
                    entry->file.ha_mmap_fd, xlate_section_type(type), _jvm_seq_num, _handle);
                ::close(entry->file.ha_mmap_fd);
                entry->file.ha_mmap_fd = -1;
            }

        }  // end of for (size_t loop = (size_t) start_token; loop <= (size_t) _last_ha_type; loop++)

        // Close the communications link to the JVM.
        if (_ts->ha_ipc_fd >= 0)
        {
            DIAGOUT(3, "Closing Host Agent fd %d, \"%s\", JVM %lu, handle 0x%lu.\n", 
                _ts->ha_ipc_fd, _ts->ipc_domain_id, _jvm_seq_num, _handle);
            ::close(_ts->ha_ipc_fd);
            _ts->ha_ipc_fd = -1;
        }

        SECTION_ENTRY_T* entry = section_entry_by_type(_ts, type_stable);

        // Close the type_stable file.
        if (entry->file.ha_mmap_fd >= 0)
        {
            DIAGOUT(3, "Closing Host Agent fd %d, \"%s\" JVM %lu, handle 0x%lu.\n", 
                entry->file.ha_mmap_fd, xlate_section_type(type_stable), _jvm_seq_num, _handle);
            ::close(entry->file.ha_mmap_fd);
            entry->file.ha_mmap_fd = -1;
        }

        return;

    }  // end of ~ZVisionHostAgent().


    /**
     ** This is an wrapper method for jvm_hello(JVM_HELLO_MSG_T* hello_msg, size_t timeout_ms), used
     ** for unit testing and should not be unnecessary for production code.  Mostly it is here so that we
     ** can test the ZVision API without any knowledge of the production Host Agent code.
     **
     ** This method is called by the Host Agent Unit Test (HAUT) when it has received a connect request (from a JVM)
     ** on its well-known port.   It passes in the file descriptor generated by the ::connect(), and returns an object
     ** handle.  It then establishes its own connection (on a new, separate fd) with the JVM.
     **
     ** It is called by the HAUT logic when a JVM sends a message to the Host Agent announcing its presence.
     ** The Host Agent will have been pended on a well-known socket listening for just such a message.
     ** Once recevived, the Host Agent calls this method, passing it the file descriptor on which it was polling.
     **
     ** This method reads that file descriptor, and calls the ZVisionHostAgent constructor method, which then
     ** establishes contact with the calling JVM, sets up the commnucation paths, and memory maps in the
     ** shared files created by the JVM.  Any future responses or notifications from the JVM will then be handled by
     ** the methods associated with these structures.
     ** 
     ** If the caller specifies a timeout equal to zero, the function will not wait at all for the JVM to
     ** establish a connection back.  In that case, the caller should invoke the wait_for_JVM_start_token() method
     ** (with a zero timeout) in their own non-blocking processing loop.  A timeout value less than zero will
     ** pend indefinitely.
     **
     ** Note: If the caller does not wait, take care should be taken NOT to set the Host Agent state
     ** to ready until the connection is actually established.
     **
     ** Returns either a handle to the the new ZVisionHostAgent object, or on failure, a NULL.
     **/
    static ZVisionHostAgent* parse_jvm_connect(SOCKET_FD_T connection_fd, int timeout_ms, SOCKET_FD_T output_fd = -1)
    {
        JVM_HELLO_MSG_T     hello_msg;
        ERRNO_RESULT_T       failure = 0;
        ssize_t              bytes_read;

        bytes_read = read_file(connection_fd, (void*) &hello_msg, sizeof(JVM_HELLO_MSG_T), &failure);

        if (bytes_read != sizeof(JVM_HELLO_MSG_T))
        {
            PATHOLOGICAL("read %lu bytes (but expected %lu) for JVM_HELLO_MSG_T on fd %d .\n", 
                bytes_read, sizeof(JVM_HELLO_MSG_T), connection_fd);
            return NULL;
        }

        if (failure != 0)
        {
            PATHOLOGICAL("read_file() returned \"%s\" (%d), when reading JVM_HELLO_MSG_T on fd %d.\n",
                ::strerror(failure), failure, connection_fd);
            return NULL;
        }

        ZVisionHostAgent* ha_obj = jvm_hello(&hello_msg, timeout_ms, output_fd);

        // No matter whether we could establish a connection or not, we're not using the connection_fd after
        // the hello message came in on, so close it.
        if (connection_fd >= 0)
        {
            ::close(connection_fd);
        }

        if (ha_obj != NULL) { return ha_obj; }

        DIAGOUT(0, "parse_jvm_connect(connection fd %d, %dms, output fd %d) failed.\n", connection_fd, timeout_ms, output_fd);
        return NULL;

    }  // end of parse_jvm_connect().



    /**
     ** This method is used to (ultimately) create a ZVisionHostAgent object, which associates a JVM with a handle that
     ** can be used by the Host Agent whevever a notification (or request response) message arrives.  
     **
     ** It is called by the Host Agent logic when a JVM sends a message to the Host Agent announcing its presence.
     ** The Host Agent will have been pended on a well-known socket listening for just such a message.
     **
     ** Once recevived, the Host Agent will read the file desriptor generated by the ::connect() method, obtaining
     ** the JVM start message details, including the JVM generated security key.  It then calls the ZVisionHostAgent 
     ** constructor method, which establishes contact with the calling JVM, sets up the commnucation paths, and 
     ** memory maps in the shared files created by the JVM.
     ** 
     ** Any future responses or notifications from the JVM will then be handled by the methods associated with these
     ** structures.
     ** 
     ** If the caller specifies a timeout equal to zero, the function will not wait at all for the JVM to
     ** establish a connection back.  In that case, the caller should invoke the wait_for_JVM_start_token() method
     ** (with a zero timeout) in their own non-blocking processing loop.  A timeout value less than zero will
     ** pend indefinitely.
     **
     ** Note: If the caller does not wait, take care should be taken not to set the Host Agent state
     ** to ready until the connection is actually established.
     **
     ** The output_fd is an already open (and available for writing) file descriptor, to which the transport callback
     ** may (or may not) send logging and/or a copy of the data being written.  This allows the Host Agent to record
     ** the information being transmitten locally as well.   Generally, this fd is expected to be used for general-
     ** purpose logging, as it will be shared by the entire JVM (i.e. all other types will write to the same fd). If
     ** you want to record only a specific type of data (e.g. config_details) the Host Agent must create a separate
     ** O_RDWR file descriptor, and set "output_fd" in the type_stable section, using ZVisionHostAgent::set_output_fd().
     **
     ** 
     ** Returns either a handle to the the new ZVisionHostAgent object, or a NULL on error.
     **
     **/
    static ZVisionHostAgent* jvm_hello(JVM_HELLO_MSG_T* hello_msg, int timeout_ms, SOCKET_FD_T output_fd)
    {
        ZVisionHostAgent*    ha_obj  = NULL;
        JVMConnectStatus     status;

        if (hello_msg->cookie != COOKIE_HELLO_MSG)
        {
            PATHOLOGICAL("got corrupted cookie 0x%lX (expected 0x%lX) when reading JVM_HELLO_MSG_T %p.\n", 
                hello_msg->cookie, COOKIE_HELLO_MSG, hello_msg);
            goto bad_exit;
        }

        DIAGOUT(1, "Got JVM_HELLO_MSG_T %p, key is 0x%lX, JVM pid is %d, type_stable name is \"%s\".\n", 
            hello_msg, hello_msg->jvm_key, hello_msg->jvm_pid, hello_msg->name);

        // Create an object specific to the JVM that just connected to us.
        ha_obj = new ZVisionHostAgent(&hello_msg->name[0], hello_msg->jvm_key, hello_msg->jvm_pid, output_fd);

        // Establish the object's handle.
        ha_obj->_handle = (HA_HANDLE_T) ha_obj;

        // Technically, these two fields should be identical, but there isn't any reason they have to be.
        DIAGOUT(1, "Created new ZVisionHostAgent object %p, setting handle to 0x%lu.\n", ha_obj, ha_obj->_handle);

        // Check its status.
        status = ha_obj->_status;

        // HA_start means that the Host Agent has initialized, and sent a start_token request to the JVM.
        // If anything went wrong during the object creation, the state will be other than this.
        if (status != HA_start)
        {
            PATHOLOGICAL("Invalid state \"%s\" (%d), expecting \"%s\" (%d).\n",
                xlate_jvm_connect_status(status), status, xlate_jvm_connect_status(HA_start), HA_start);
            goto bad_exit;
        }

        // The caller may not want to wait for the JVM to respond.  If so, they can call the
        // wait_for_JVM_start_token() method directly, with a zero timeout, in whatever processing
        // loop is necessary for their application.
        if (timeout_ms == 0) 
        { 
            DIAGOUT(3, "Created ZVisionHostAgent object %p, not waiting for start_token response.\n", ha_obj);
            return ha_obj; 
        }
        
        // Otherwise, wait (with a timeout) for some amount of time for the JVM to send back a start_token response.
        // Failing to do so here doesn't necessarily mean a pathological condition.  The JVM could have died for unrelated
        // reasons, or been terminated by the user.
        if (ha_obj->wait_for_JVM_start_token(timeout_ms) != true)
        {
            DIAGOUT(0, "Could not establish connetion with JVM (ZVisionHostAgent object %p) within %d milliseconds.\n", 
                ha_obj, timeout_ms);
            return NULL;
        }

        DIAGOUT(3, "Created ZVisionHostAgent object %p.\n", ha_obj);
        return ha_obj;

bad_exit:
        PATHOLOGICAL("jvm_hello() failed.\n");
        if (ha_obj != NULL) { delete(ha_obj); }
        return NULL;
         
    }  // end of jvm_hello()


    /**  
     **  Wait for the return start message from the JVM connection request.
     **  NOTE: A negative timeout_ms value means we will poll indefinitely.
     **/
    bool wait_for_JVM_start_token(int timeout_ms)
    {
        DIAGOUT(3, "wait_for_JVM_start_token(%p, timeout %dms).\n", this, timeout_ms);

        ssize_t            bytes_read;
        JVM_NOTIFY_MSG_T   jvm_notify_msg;
        ERRNO_RESULT_T     failure = 0;

        // The memory mapped type_stable file created by the JVM.
        TYPE_STABLE_LAYOUT_T* ts = _ts;

        if (type_stable_valid(ts) != true)
        {
            PATHOLOGICAL("handle points to invalid object.\n");
            return false;
        }

        // Get the IPC socket established by the JVM, from the type_stable file.
        SOCKET_FD_T      fd                 = ts->ha_ipc_fd;

        if (fd < 0)
        {
            PATHOLOGICAL("ZVisionHostAgent object %p has ha_ipc_fd (%d) less than zero.\n", this, fd);
            return false;
        }

        // If the caller used a value less than zero for the timeout_ms parameter, then
        // pend indefinitely during the ::poll() call.  If they used a value of zero,
        // then don't pend at all.  Otherwise pend for no more than <timeout_ms> seconds
        // during the ::poll() call.
        if (timeout_ms >= 0)
        {
            DIAGOUT(1, "Waiting %d milliseconds while polling for JVM start_token on fd %d...\n", timeout_ms, fd);
        }
        else
        {
            DIAGOUT(1, "Polling indefinitely for JVM start_token on fd %d...\n", fd);
        }

        // We should be waiting for a response, and nothing else.
        if (_status != HA_start)
        {
            // If it wasn't an error, and it wasn't a shutdown, then its an error.
            if (_status == HA_error || _status == JVM_error || _status != JVM_exit) 
            { 
                // TODO: figure out what to do here.
                PATHOLOGICAL("Host Agent status is now \"%s\" (0x%X), exiting prematurely.\n", 
                    xlate_jvm_connect_status(_status), _status);

                return false;
            }

            PATHOLOGICAL("Host Agent %p (fd %d) status \"%s\" invalid, expected \"%s\" while waiting for start_token.\n",
                this, fd, xlate_jvm_connect_status(_status), xlate_jvm_connect_status(HA_start));
            _status = HA_error;
            return false;
        }

//TODO: mark file as non-blocking read, Put in loop that cycles until timeout.

        // Loop continously, ignoring asynchronous messages that might have arrived out of order.
        while (true)
        {
            // Read the response buffer.  It should contain the JVM's response back from the our start_token request.
            bytes_read = read_file(fd, (void*) &jvm_notify_msg, sizeof(jvm_notify_msg), &failure);
           
            if (bytes_read != sizeof(jvm_notify_msg))
            {
                DIAGOUT(0, "Read %lu bytes of start_token response, expected %lu, for ZVisionHostAgent %p.\n",
                    bytes_read, sizeof(jvm_notify_msg), this);
                _status = HA_error;
                return false;
            }

            if (jvm_notify_msg.cookie != COOKIE_NOTIFY_MSG)
            {
                PATHOLOGICAL("JVM message corrupted; expecting cookie 0x%lX, got 0x%lX instead; handle 0x%lu, type \"%s\" (%d).\n",
                    COOKIE_NOTIFY_MSG, jvm_notify_msg.cookie, jvm_notify_msg.handle, 
                    xlate_section_type(jvm_notify_msg.type), jvm_notify_msg.type);
                _status = JVM_error;
                return false;
            }

            if (jvm_notify_msg.type == tick_profiling || jvm_notify_msg.type == method_names) { continue; }
        }

        if (jvm_notify_msg.type != start_token)
        {
            PATHOLOGICAL("Got unexpected msg type \"%s\" (%d), was expecting \"%s\" (%d) from JVM.\n", 
                xlate_section_type(jvm_notify_msg.type), jvm_notify_msg.type, xlate_section_type(start_token), start_token);
            _status = JVM_error;
            return false;
        }

        // The handle should match what we sent initially.
        if (jvm_notify_msg.handle != _handle)
        {
            PATHOLOGICAL("Got invalid handle in \"%s\" message; was expecting 0x%lu, got 0x%lu instead.\n", 
                xlate_section_type(start_token), jvm_notify_msg.handle, _handle);
            _status = JVM_error;
            return false;
        }

        // Okay, everything matches, we have now established a full connection with the JVM.
        _handle = jvm_notify_msg.handle;
        _status = JVM_connect;

        DIAGOUT(3, "Connected to JVM, socket \"%s\", handle 0x%lu.\n", xlate_section_type(start_token), _handle);
        DIAGOUT(3, "JVM connetion status is \"%s\" (0x%X).\n", xlate_jvm_connect_status(_status), _status);

        set_host_agent_ready(ts);

        DIAGOUT(2, "Host Agent is now ready.\n");

        return true;

    }  // end of wait_for_JVM_start_token().


    /**
     **  Called only by the Host Agent (or Host Agent Unit Test) to process data when made available.
     **  This could be either indexed buffer or request/response notifications.  The supplied xport_method
     **  parameter will be passed as the function to be invoked. 
     **
     **  Whenever the host agent is notified via a Unix Domain socket message from the JVM that a
     **  type section shared memory buffer is now available, it should then call this function, which 
     **  will then call the "compress" function associated with that type, passing the compressed data
     **  (i.e. Google Protocol Buffer) to the transport layer callback, which then moves that section's
     **  block of data out across the network.
     **/
    bool process_notify_message(SOCKET_FD_T fd, HA_XPORT_BUF_METHOD_T xport_method)
    {
        ZVision::JVM_NOTIFY_MSG_T   jvm_notify_msg;
        ssize_t                     jvm_notify_msg_len = sizeof(jvm_notify_msg);
        ERRNO_RESULT_T              failure = 0;
    
        DIAGOUT(4, "process_notify_message(fd %d, callback %p).\n", fd, xport_method);

        // Read the response buffer.  
        ssize_t bytes_read = ZVision::read_file(fd, (void*) &jvm_notify_msg, jvm_notify_msg_len, &failure);
       
        if (bytes_read != jvm_notify_msg_len)
        {
            // This might not actually be a pathological case (perhaps the JVM died while writing?), so we 
            // can't just abort if it happens, but we should mention the fact that it did.
            DIAGOUT(1, "Read %lu bytes of JVM notify message on fd %d, expected %lu.\n",
                bytes_read, fd, jvm_notify_msg_len);
            return false;
        }
    
        if (xport_method == NULL)
        {
            PATHOLOGICAL("transport method callback was NULL.\n");
            return false;
        }

        if (jvm_notify_msg.cookie != COOKIE_NOTIFY_MSG)
        {
            PATHOLOGICAL("Corrupted JVM notify message of type \"%s\".  Expected 0x%lX cookie, got 0x%lX instead.\n", 
                xlate_section_type(jvm_notify_msg.type), COOKIE_NOTIFY_MSG, jvm_notify_msg.cookie);
            return false;
        }

        TYPE_STABLE_LAYOUT_T* ts = (TYPE_STABLE_LAYOUT_T*) _ts;
        
        if (type_stable_valid(ts) != true)
        {
            PATHOLOGICAL("JVM message handle points to invalid type_stable structure.\n");
            return false;
        }
        
        SECTION_TYPE_E type = jvm_notify_msg.type;

        if (type < start_token || type >= _last_ha_type)
        {
            PATHOLOGICAL("JVM notify message type \"%s\" (%d) invalid.  Must be value between %d and %d.\n", 
                xlate_section_type(type), type, start_token, _last_ha_type);
            return false;
        }

        // If the JVM sends us a notification (indexed buffer) or request that happens to be newer than what 
        // this Host Agent supports, then we ignore it.
        if (type >= _last_ha_type)
        {
            DIAGOUT(1, "Got type request (%d) not supported by this revision, ignoring.\n", type);
            return true;
        }

        if (type == shutdown) 
        {
          DIAGOUT(1, "Connection shutdown request received from JVM on fd %d, setting connect status to \"%s\".\n", 
            fd, xlate_jvm_connect_status(JVM_end)); 
          set_jvm_connect_status(JVM_end);
          clr_host_agent_ready(ts);
          return true;
        }

        SECTION_ENTRY_T* entry = section_entry_by_type(ts, type);

        if (entry == NULL)
        {
            PATHOLOGICAL("section entry for type \"%s\" (%d) was NULL.\n", xlate_section_type(type), type);
            return false; 
        }

        FILE_DATA_T* file_data = &entry->file;

        if (_status == JVM_end || _status == HA_end)
        {
            DIAGOUT(2, "Connection shutdown in progress, state is \"%s\", ignoring message.\n", xlate_jvm_connect_status(JVM_end)); 
            return true;
        }

        // If the JVM is already shutting down, then ignore the request.
        if (_status != JVM_connect)
        {
            DIAGOUT(2, "Host Agent object %p, JVM %lu (fd %d) received \"%s\" response while connections state was \"%s\".\n",
                this, _jvm_seq_num, ts->ha_ipc_fd, xlate_section_type(type), xlate_jvm_connect_status(_status));
            return true;
        }

        // TODO: Make this more efficient, we don't need to constantly look this up.
        const CONSTANTS_T* constants = constants_by_type(type);

        //----------------------------------------------------------------------------------------------------//
        // Process the notify message type for content, compress it, then pass it on to the transport layer.  //
        // We are gauranteed that the underpinning memory will not go away until we exit this routine.        //
        //----------------------------------------------------------------------------------------------------//

        DIAGOUT(2, "Got \"%s\" msg from JVM [handle 0x%lu, fd %d].\n", xlate_section_type(type), jvm_notify_msg.handle, fd);

        COMPRESS_BUF_METHOD_T  compress_buf_func;

        char*   raw_data_addr   = (char*) file_data->ha_mmap_addr;
        char*   final_data_addr = (char*) file_data->ha_mmap_addr;

        size_t  raw_data_size   = file_data->unit_size;
        size_t  final_data_size = file_data->unit_size;

        if (raw_data_addr == NULL)
        {
            PATHOLOGICAL("Host Agent object %p, JVM %lu \"%s\" (fd %d) raw data addr is NULL.\n",
              this, _jvm_seq_num, xlate_section_type(type), file_data->ha_mmap_fd);
            return false;
        }

        if (raw_data_size <= 0)
        {
            PATHOLOGICAL("Host Agent object %p, JVM %lu \"%s\" (fd %d) raw data size (%ld) invalid.\n",
              this, _jvm_seq_num, xlate_section_type(type), file_data->ha_mmap_fd, raw_data_size);
            return false;
        }

        if (constants->file_type == indexed_buffer_file)
        {
            // It is possible (should the Tick Profiling Mechanism, or Method Names Mechanism, or some
            // other indexed_buffer_file that can be dynamically resized by the JVM), to have exceeded
            // its available buffers such  that the ZVision Service Thread (or some other entity) will
            // shutdown the file recording subsystem, free the buffer memory related to that file type
            // and reallocate new memory.  To that end, we need to arbitrate how that happens, so that
            // we don't encounter the case where this code tries to access memory that has just disappeared
            // out from under it.  This lock prevents the ZST from freeing memory until we're done copying
            // it.  The lock disappears when we return from this scope.  The lock is ignored if we aren't
            // using dynamic resizing.
            ZVisionHostAgentSectionLock   lock(file_data, true, _last_ha_type);

            // Grab the memory sizing "generation" of the message we just received.  If this doesn't match
            // the memory size generation of the current data section, then we'll need to resize.
            // This is only applicable to indexed_buffer_file types and is ignored otherwise.
            size_t msg_mmap_gen = file_data->section_lock & (ZVISION_SECTION_LOCK_MAX_ITERATIONS - 1);

            // Okay, we should now be the only thread that can manipulate the memory, and the iteration
            // count within the lock field should not change.  Read that iteration count, and compare it
            // against what we think is our current iteration for this section type.  If they don't match
            // then the JVM has resized the underpinning memory for that section type (so we need to unmap
            // the current memory, remap to the JVM's new memory size, and disregard this message) or it
            // is a stale message from the prior iteration of memory, in which case we just discard it.
            if (msg_mmap_gen < file_data->ha_mmap_gen)
            {
              DIAGOUT(0, "Discarding stale (gen %ld) \"%s\" message, current memory generation is %ld.\n", 
                msg_mmap_gen, xlate_section_type(type), file_data->ha_mmap_gen);
              // TODO: add a counter.
              return true;
            }
            else if (msg_mmap_gen > file_data->ha_mmap_gen)
            {
              // We get here becaues the Host Agent's notion of memory is out of date and we need to
              // resize it.  Call the function to do so, but discard the current entry.
              if (resize_ha_section_memory(file_data) != true)
              {
                PATHOLOGICAL("Host Agent object %p, JVM %lu error resizing \"%s\" (fd %d) memory, current gen %ld, msg gen %ld.\n",
                    this, _jvm_seq_num, xlate_section_type(type), file_data->ha_mmap_fd, file_data->ha_mmap_gen, msg_mmap_gen); 
                return false;
              }
            }

            // Double check that the memory maps between Host Agent and JVM now match.
            if (msg_mmap_gen != file_data->ha_mmap_gen)
            {
                PATHOLOGICAL("Host Agent object %p, JVM %lu \"%s\" (fd %d) memory gen mismatch, current gen %ld, msg gen %ld.\n",
                    this, _jvm_seq_num, xlate_section_type(type), file_data->ha_mmap_fd, file_data->ha_mmap_gen, msg_mmap_gen); 
                return false;
            }

            // Set the buffer pointer to the index within the memory mapped file.  We use
            // "raw data" to mean, in the case of indexed_buffer_files, the starting point
            // of the data memory space (i.e. the start of the tick_buffer address space, if
            // we are dealing with a tick buffer), and "final data" to mean the actual offset to the
            // specific buffer within that space.
            final_data_addr += (jvm_notify_msg.index * raw_data_size);

#ifdef RKANEDONT
            // If the Host Agent has (during initialization) created a ring buffer for this file type, then update
            // the ring buffer pointers appropriately.  This allows consumer tasks to apply filters to the data
            // (and/or a copy of the data), and access it (in the order it was delivered) via ring buffers.
            if (file_data->ha_ring_obj != NULL) 
            {
                ZVisionFilterTicks* ft = (ZVisionFilterTicks*) file_data->ha_ring_obj;
                ft->put_buffer(jvm_notify_msg.index);
            }
#endif // RKANEDONT
       }           
       else if (constants->file_type == request_response_file)
       {
            if (type == heartbeat)
            {
                REQUEST_BUFFER_T* req = (REQUEST_BUFFER_T*) file_data->ha_mmap_addr;

                if (req == NULL) 
                { 
                    PATHOLOGICAL("NULL ha_map_addr value.\n");
                    return false;
                }


// Defining this will dump out the various counters within the Tick Profiling mechanism.
#undef SHOW_TICK_BUFFER_INTERNALS
#ifdef SHOW_TICK_BUFFER_INTERNALS

                uint64_t* addr = (uint64_t*) &req->response_data_start; 

                uint64_t in_use      = *addr++;
                uint64_t recycled    = *addr++;
                uint64_t last_bufno  = *addr++;
                uint64_t list_len    = *addr++;
                uint64_t hint_idx    = *addr++;
                uint64_t created     = *addr++;
                uint64_t deleted     = *addr++;
                uint64_t margin      = *addr++;

                uint64_t active      = created - deleted;
                uint64_t max_list    = in_use - active;

                DIAGOUT(0, "---------------------------------\n");
                DIAGOUT(0, "  Buffers In Use: %lu\n", in_use);
                DIAGOUT(0, "Recycled Buffers: %lu\n", recycled);
                DIAGOUT(0, "      Last Bufno: %lu\n", last_bufno);
                DIAGOUT(0, "Recycle List Len: %lu\n", list_len);
                DIAGOUT(0, "    Max List Len: %lu\n", max_list);
                DIAGOUT(0, "     Hint Offset: %lu\n", hint_idx);
                DIAGOUT(0, " Threads Created: %lu\n", created);
                DIAGOUT(0, " Threads Deleted: %lu\n", deleted);
                DIAGOUT(0, "  Active Threads: %lu\n", active);
                DIAGOUT(0, "   Buffer Margin: %lu\n", margin);

#endif // SHOW_TICK_BUFFER_INTERNALS

            }

            compress_buf_func = entry->ha_funcs.compress_buf_func;

            if (compress_buf_func == NULL)
            {
                PATHOLOGICAL("compress buffer function was NULL for ZVisionHostAgent object %p, JVM %lu.\n", this, _jvm_seq_num);
                return false;
            }

            // Calculate the address of where to place the compressed data.  It should
            // begin 1 unit_size worth of bytes after the start of the raw data.
            final_data_addr = raw_data_addr + offsetof(REQUEST_BUFFER_T, response_data_start);
 
            // If there is no compress_buf_func method, then skip this step and send the data out as-is.
            if (compress_buf_func != noop_buffer_compress)
            {
                final_data_addr = raw_data_addr + raw_data_size;
    
                // Call the function which knows how to compress the binary data (into a Google Protocol Buffer?).
                // The function expects the buffer space allocated for the request type to be large enough to 
                // also contain the transformed data, placed just after the raw data.  The function will return the new
                // address of the transformed data, or zero if it encounters an error.
                final_data_size = compress_buf_func(file_data, raw_data_addr, raw_data_size, final_data_addr);

                if (final_data_size < 0)
                {
                    PATHOLOGICAL("compress_buf_func(%p, %p, %ld, %p) failed, ZVisionHostAgent %p, \"%s\", JVM %lu.\n", 
                        file_data, raw_data_addr, raw_data_size, final_data_addr, this, xlate_section_type(type), _jvm_seq_num);
                    return false;
                }
            }
        }
        else
        {
            PATHOLOGICAL("Invalid type \"%s\" in notify msg, index %lu, handle %p, JVM %lu, size %lu bytes.\n", 
               xlate_section_type(type), jvm_notify_msg.index, this, _jvm_seq_num, raw_data_size);
            return false;
        }

        ssize_t bytes_written = xport_method((const void*) final_data_addr, (size_t) final_data_size, (void*) file_data, &failure);

        if (failure != 0)
        {
            PATHOLOGICAL("xport_method(%p, %ld, %p, %p) failed for ZVisionHostAgent object %p, JVM %lu; \"%s\" (%d).\n", 
                raw_data_addr, final_data_size, file_data, &failure, this, _jvm_seq_num, strerror(failure), failure);
            return false;
        }
  
        if (bytes_written != (ssize_t) final_data_size)
        {
            PATHOLOGICAL("xport_method(%p, %ld, %p, %p) failed for ZVisionHostAgent object %p, JVM %lu. "
              "wrote %ld bytes but data size was %ld.\n", 
              raw_data_addr, final_data_size, file_data, &failure, this, _jvm_seq_num, bytes_written, final_data_size);
            return false;
        }
  
        DIAGOUT(3, "Processed \"%s\" msg on fd %d, index %lu, handle %p, JVM %lu, data len %lu (%lu compressed), wrote %lu.\n", 
           xlate_section_type(type), fd, jvm_notify_msg.index, this, _jvm_seq_num, raw_data_size, final_data_size, bytes_written);

        file_data->ha_output_cnt += bytes_written;

        return true;

    }  // end of process_notify_message().


    /**
     ** Send a request message from the Host Agent to the JVM based on type: The JVM fills in the shared memory response
     ** buffer, then sends back (via IPC socket) a message.  The response from the JVM, if received, should then be 
     ** processed asynchronously, via a call to the process_notify_message() method, to call the transport layer repsonse
     ** associated with the type.
     **
     ** If the params_cnt is zero, and/or the params pointer is NULL, then use whatever is already in the request_buffer
     ** for this type.
     **/
    bool send_request(SECTION_TYPE_E type, size_t params_cnt=0, const char* params=NULL)
    {
        if (type_stable_valid(_ts) != true)
        {
            PATHOLOGICAL("type_stable pointer (%p) invalid.\n", _ts);
            return false;
        }

        if (type == type_stable || type < start_token)
        {
            PATHOLOGICAL("Illegal request, type may not be \"%s\" (%d).\n", xlate_section_type(type), type);
            return false;
        }

        // If the type we are sending isn't supported by the rev of the JVM, then don't bother.
        if (type >= _last_jvm_type)
        {
            DIAGOUT(2, "Ignoring section type request (#%d), not supported by the revision of the connected JVM.\n", type);
            return true;
        }

        // If the user previously asked to ignore (forgo) requests of this type, just return true.
        if (forgo_data(_ts, type) == true)
        {
            DIAGOUT(2, "Forgoing requests for section type \"%s\" (%d).\n", xlate_section_type(type), type);
            return true;
        }

        // Calculate the length of the parameters string.
        size_t len = (params != NULL) ? ::strnlen(params, MAX_REQUEST_PARAMS_SIZE) : 0;

        ZV_ASSERT(len <= MAX_REQUEST_PARAMS_SIZE);

        // Grab the type_stable file data for this section type.
        FILE_DATA_T* file_data = file_data_by_type(_ts, type);

        if (file_data == NULL) 
        {
            PATHOLOGICAL("file_data for type \"%s\" was NULL.\n", xlate_section_type(type));
            return false; 
        }

        // Find the memory mapped address for this section.
        volatile REQUEST_BUFFER_T* req_buf = (volatile REQUEST_BUFFER_T*) file_data->ha_mmap_addr;

        if (req_buf == NULL)
        {
            PATHOLOGICAL("req_buf for type \"%s\" was NULL.\n", xlate_section_type(type));
            return false; 
        }

        // This field should never change once initialized.
        if (req_buf->hdr.cntl.cookie != COOKIE_REQUEST_BUF)
        {
            PATHOLOGICAL("req_buf corrupted, expecting cookie (0x%lX), got 0x%lX instead.\n",
                COOKIE_REQUEST_BUF, req_buf->hdr.cntl.cookie);
            return false;
        }

        // This field should never change once initialized.
        if (req_buf->hdr.cntl.request_id != ::getpid()) // host_agent_pid(_ts))
        {
            PATHOLOGICAL("req_buf corrupted, expecting Host Agent pid (%d), got %d instead.\n",
                ::getpid(), /* host_agent_pid(_ts) */   req_buf->hdr.cntl.request_id);
            return false;
        }

        // This field should never change once initialized, but during shutdown the JVM pid field
        // may already be zeroed.
        if (req_buf->hdr.cntl.response_id != ZVision::jvm_pid(_ts) && _status == JVM_connect)
        {
            PATHOLOGICAL("send_request() \"%s\" request buffer invalid, expecting JVM pid (%d), got %d instead.\n",
                xlate_section_type(type), ZVision::jvm_pid(_ts), req_buf->hdr.cntl.response_id);
            return false;
        }

        if (len > sizeof(req_buf->params)) 
        { 
            ZV_ASSERT(len <= sizeof(req_buf->params));
            len = sizeof(req_buf->params); 
        }

        // If the user cleared both of these fields, then just use whatever is already
        // in the request buffer.  This allows callers to treat the parameter string as
        // a binary structure instead.
        if (params != NULL && params_cnt > 0)
        {
            ::memset((void*) &req_buf->params[0], 0, sizeof(req_buf->params));
            ::memcpy((void*) &req_buf->params[0], params, len);
            req_buf->hdr.cntl.params_cnt  = params_cnt;
        }

        req_buf->hdr.cntl.type        = type;
        req_buf->hdr.cntl.params_len  = len;

        if (req_buf->hdr.cntl.params_len >= MAX_REQUEST_PARAMS_SIZE)
        {
            PATHOLOGICAL("user_req_buf size (%lu) exceeds request message parameter string (%lu).\n",
                req_buf->hdr.cntl.params_len, MAX_REQUEST_PARAMS_SIZE);
            return false;
        }

        SOCKET_FD_T fd = _ts->ha_ipc_fd;

        if (fd <= 0)
        {
            PATHOLOGICAL("Unix Domain Socket (%d) invalid.\n", fd);
            return false;
        }

        // Set the ID fields.
        __sync_fetch_and_add(&req_buf->hdr.cntl.request_token, 1);     // Atomically increment the token id.

        // Update the counters.
        __sync_fetch_and_add(&req_buf->hdr.cntl.request_cnt, 1);       // Atomically increment the request count.

        HA_REQUEST_MSG_T request_msg;

        request_msg.handle  = _handle;
        request_msg.type    = type;
        request_msg.cookie  = COOKIE_REQUEST_MSG;

        DIAGOUT(3, "Host Agent (pid %d) sending a request for type \"%s\" (handle 0x%lX) on fd %d, to the JVM (pid %d).\n", 
            req_buf->hdr.cntl.request_id, xlate_section_type(req_buf->hdr.cntl.type), _handle, fd, req_buf->hdr.cntl.response_id);
        DIAGOUT(3, "request message parameter count = %lu, parameter string length = %lu, cookie = 0x%lX\n",
            req_buf->hdr.cntl.params_cnt, req_buf->hdr.cntl.params_len, req_buf->hdr.cntl.cookie);
        DIAGOUT(3, "Request Buffer Header: size %lu bytes.\n", sizeof(REQUEST_BUFFER_HEADER_T));
        DIAGBUFFEROUT(3, (volatile char*) &req_buf->hdr, sizeof(REQUEST_BUFFER_HEADER_T));
        DIAGOUT(3, "Request Buffer Parameters: size %lu bytes.\n", MAX_REQUEST_PARAMS_SIZE); 
        DIAGBUFFEROUT(3, (volatile char*) &req_buf->params, MAX_REQUEST_PARAMS_SIZE);
        
        ERRNO_RESULT_T       failure = 0;

        // Now that the parameters have been loaded into the shared memory, send only the request type (1 word) to the JVM.
        ssize_t bytes_written = write_file(fd, (const void*) &request_msg, sizeof(HA_REQUEST_MSG_T), &failure);

        if (bytes_written == sizeof(HA_REQUEST_MSG_T)) 
        { 
            DIAGOUT(4, "Successfully wrote %ld bytes to fd %d.\n", bytes_written, fd);

            return true; 
        }

        // The "failure" variable was set when we made the write_file() call.  If we're here, the call failed.  If this 
        // isn't because the write might block, it is a pathological error.
        if (failure != EAGAIN)
        {
            PATHOLOGICAL( "send_request(): wrote %ld bytes, was expecting to write %lu bytes.\n", 
                bytes_written, sizeof(HA_REQUEST_MSG_T));
            PATHOLOGICAL("failed with errno %d, \"%s\".\n", failure, ::strerror(failure));
            return false;
        }

        // If this is because the write "might" block, but we wrote a partial amount (instead
        // of zero) then it is also a pathological error.
        if (bytes_written != 0)
        {
            PATHOLOGICAL("send_request(): Wrote only %lu bytes when EAGAIN.\n", bytes_written);
            PATHOLOGICAL("failed with errno %d, \"%s\".\n", failure, ::strerror(failure));
            return false;
        }

        // Okay, the write would have blocked, we didn't write anything, consider 
        // this a dropped message.  Atomically increment the dropped count.
        __sync_fetch_and_add(&req_buf->hdr.cntl.dropped, 1);

        // Mark the size as less than zero, indicating the request failed.
        req_buf->hdr.cntl.response_size  = -1;
        
        DIAGOUT(3, "(\"%s\") write would have blocked, message dropped.\n", xlate_section_type(req_buf->hdr.cntl.type));

        return true;

    }  // end of send_request().


    /**
     ** Returns the file descriptor used by the Host Agent to connect to the JVM associated with this object.
     ** This method should be used only by the Host Agent to associate the handle with an fd.  Do not write or
     ** read to the fd directly, use the send_request() & process_notify_message() calls, respectively, if you
     ** need to move data between the Host Agent and a JVM.
     **
     ** On error, returns -1, calls PATHOLOGICAL(), sets communciation status field to HA_error.
     **/
    inline SOCKET_FD_T jvm_fd() 
    { 
        if (type_stable_valid(_ts) != true)
        {
            PATHOLOGICAL("type_stable ptr %p invalid, for JVM handle 0x%lX, \"%s\".\n", _ts, _handle, _name);
            _status = HA_error;
            return -1;
        }
     
     return _ts->ha_ipc_fd; 

     }  // end of jvm_fd().


    /**
     ** Return the Host Agent assigned JVM sequence number.
     **/
    inline size_t jvm_seq_num() { return _jvm_seq_num; }


    /**
     ** Return the Host Agent assigned JVM pid.
     **/
    inline PID_T jvm_pid() { return _jvm_pid; }


    /**
     ** Return the Host Agent assigned JVM key.
     **/
    inline JVM_KEY_T jvm_key() { return _jvm_key; }


    /**
     ** Assign JVM sequence number.  This is a unique value assigned to every JVM the Host Agent deals with.
     ** The only invalid value is ~0, which is used to indicate an as yet unassigned object.
     **/
    inline bool jvm_seq_num(size_t jvm_seq_num) 
    { 
        ZV_ASSERT(_jvm_seq_num == ((size_t) ~0));

        if (jvm_seq_num == ((size_t) ~0))
        {
            PATHOLOGICAL("Can not set jvm_seq_num to 0x%ld for Host Agent object %p.\n", jvm_seq_num, this);
            return false;
        }

        _jvm_seq_num = jvm_seq_num; 

        return true;

        }


    /**
     ** Return the Host Agent generated handle.
     **/
    inline HA_HANDLE_T handle() { return _handle; }


    /**
     ** Return the basename for this object.
     **/
    inline const char* basename() { return &_name[0]; }


    /**
     ** Return the current connection status (from the Host Agent's perspective) of the JVM.
     **/
    JVMConnectStatus jvm_connect_status() { return _status; }

    /**
     ** Set the current connection status (from the Host Agent's perspective) of the JVM.
     **/
    inline void set_jvm_connect_status(JVMConnectStatus status) { _status = status; }


    /**
     ** Set the output file descriptor for the JVM's section type.  Normally every output_fd within
     ** the JVM would all go to the same output fd, to be used as a logging mechanism.  But it is also
     ** possible to use it for other purposes, for instance, to save off data to a local disk (this
     ** would allow us to record the data of a particular section type) when unit testing (to capture
     ** a local copy of the data being sent, for comparison against a "golden data" file.
     **
     ** The output_fd is only usable by the Host Agent side, it should not be modified by the JVM.
     **/
    inline bool set_output_fd(SECTION_TYPE_E type, SOCKET_FD_T output_fd)
    {
        ZV_ASSERT(_handle == (HA_HANDLE_T) this);

        if (type_stable_valid(_ts) != true)
        {
            PATHOLOGICAL("_ts ptr (%p) invalid for Host Agent object 0x%lX.\n", _ts, _handle);
            return false;
        }

        SECTION_ENTRY_T* entry = section_entry_by_type(_ts, type);

        if (entry == NULL)
        {
            PATHOLOGICAL("section_entry_by_type(%p, \"%s\") failed for ZVisionHostAgent object 0x%lX.\n", 
                _ts, xlate_section_type(type), _handle);
            return false;
        }

        entry->file.ha_output_fd = output_fd;

        DIAGOUT(3, "Set \"%s\" output_fd to %d for ZVisionHostAgent object 0x%lX.\n", 
            xlate_section_type(type), output_fd, _handle);

        return true;

    }  // end of set_output_fd().


    /**
     ** resize_ha_section_memory(FILE_DATA_T* file_data) - 
     **
     ** Unmaps the existing memory for a given (indexed_buffer_file) section type, then re-mmap()'s 
     ** memory for the new size.  Called each time we iterate through a new version of the underpinning
     ** section type memory, but only from within a routine that holds the ZVISION_HA_SECTION_LOCK_MASK
     ** section lock.  The corresponding JVM memory has already been resized.  We use the unit_cnt field
     ** times the unit_size field to determine the new size, and we use the ha_mmap_len field to determine
     ** how much memory had been in play before the resize request occurred.
     **
     ** Returns true if successful, false otherwise.
     **/
    bool resize_ha_section_memory(FILE_DATA_T* file_data)
    {
      int rc = 0;

      if (file_data == NULL) 
      {
        PATHOLOGICAL("FILE_DATA_T* was NULL.\n"); 
        ZV_ASSERT(false);
        return false; 
      }

      if (file_data->cookie != COOKIE_FILE_DATA)
      {
        PATHOLOGICAL("FILE_DATA_T* %p has corrupt cookie, expecting 0x%lX, found 0x%lX.\n",
          file_data, COOKIE_FILE_DATA, file_data->cookie); 
        return false; 
      }

      int    ha_mmap_fd   = file_data->ha_mmap_fd;
      size_t ha_mmap_gen  = file_data->ha_mmap_gen;
      size_t ha_mmap_len  = file_data->ha_mmap_len;
      void*  ha_mmap_addr = file_data->ha_mmap_addr;

      if (ha_mmap_fd < 0)
      {
        PATHOLOGICAL("File: \"%s\" (%p) has illegal ha_mmap_fd (%d).\n", 
          xlate_section_type(file_data->type), file_data, ha_mmap_fd);
        return false; 
      }

      if (ha_mmap_addr == NULL)
      {
        PATHOLOGICAL("File: \"%s\" (%p), ha_mmap_fd %d, has NULL ha_mmap_addr.\n", 
          xlate_section_type(file_data->type), file_data, ha_mmap_fd);
        return false; 
      }

      if (ha_mmap_len <= 0)
      {
        PATHOLOGICAL("File: \"%s\" (%p), ha_mmap_fd %d, has invalid ha_mmap_len (%ld).\n", 
          xlate_section_type(file_data->type), file_data, ha_mmap_fd, ha_mmap_len);
        return false; 
      }

      DIAGOUT(0, "resize_ha_section_memory(%p) called for \"%s\" section: fd %d, gen %lu, len %lu, addr %p.\n", 
        file_data, xlate_section_type(file_data->type), ha_mmap_fd, ha_mmap_gen, ha_mmap_len, ha_mmap_addr);

      // Clear these fields, in case we encounter an error along the way.
      file_data->ha_mmap_addr = NULL;
      file_data->ha_mmap_len = 0;

      DIAGOUT(0, "Issuing ::munmap(%p, %ld) for file \"%s\", fd %d.\n", 
            ha_mmap_addr, ha_mmap_len, xlate_section_type(file_data->type), ha_mmap_fd); 

      // Release the current virtual memory map.
      rc = ::munmap(ha_mmap_addr, ha_mmap_len);

      while (rc != 0)
      {
        ERRNO_RESULT_T failure = errno;

        switch (failure)
        {
        case EINTR:
          continue;
          break;

        default:
          PATHOLOGICAL("::munmap(%p, %ld) for \"%s\" file (%p), ha_mmap_fd %d failed, error \"%s\".\n",
            ha_mmap_addr, ha_mmap_len, xlate_section_type(file_data->type), file_data, ha_mmap_fd, ::strerror(failure)); 
          return false;
          break;
        }

        rc = ::munmap(ha_mmap_addr, ha_mmap_len);
      }

      off_t new_size = file_data->max_size;

      DIAGOUT(0, "Issuing ::truncate(%d, %ld) for file \"%s\".\n", ha_mmap_fd, new_size, xlate_section_type(file_data->type)); 

      // Adjust the file descriptor's notion of how big the file is.  This
      // value can increase or decrease from the current size.
      rc = ::ftruncate(ha_mmap_fd, new_size);

      while (rc != 0)
      {
        ERRNO_RESULT_T failure = errno;

        switch (failure)
        {
        case EINTR:
          continue;
          break;

        default:
          PATHOLOGICAL("::truncate(%d, %ld) for \"%s\" file (%p), failed, error \"%s\".\n",
            ha_mmap_fd, new_size, xlate_section_type(file_data->type), file_data, ::strerror(failure)); 
          return false;
          break;
        }

        rc = ::ftruncate(ha_mmap_fd, new_size);
      }

      DIAGOUT(0, "Issuing ::mmap(%d, %ld) for file \"%s\".\n", ha_mmap_fd, new_size, xlate_section_type(file_data->type)); 

      file_data->ha_mmap_addr = mmap_file(ha_mmap_fd, new_size, (char*) xlate_section_type(file_data->type));

      if (file_data->ha_mmap_addr == MAP_FAILED)
      {
        PATHOLOGICAL("mmap_file() returned MAP_FAILED for \"%s\" (%p), ha_mmap_fd %d failed.\n",
          xlate_section_type(file_data->type), file_data, ha_mmap_fd);
        return false; 
      }

      file_data->ha_mmap_len = new_size;
      file_data->ha_mmap_gen++;

      DIAGOUT(0, "Re-mmap()'ed the Host Agent \"%s\" memory fd %d, gen %lu, len %lu, addr %p.\n", 
        xlate_section_type(file_data->type), file_data->ha_mmap_fd, file_data->ha_mmap_gen, 
        file_data->ha_mmap_len, file_data->ha_mmap_addr);

      return true;
      
    }  // end of resize_ha_section_memory().


    /**
     ** Convert the JVM connection status to a printable string.
     **/
    static const char* xlate_jvm_connect_status(JVMConnectStatus status)
    {
        switch (status)
        {
        case JVM_hello:
            return "JVM_hello";
            break;
        case HA_start:
            return "HA_start";
            break;
        case JVM_connect:
            return "JVM_connect";
            break;
        case HA_end:
            return "HA_end";
            break;
        case JVM_end:
            return "JVM_end";
            break;
        case JVM_exit:
            return "JVM_exit";
            break;
        case HA_error:
            return "HA_error";
            break;
        case JVM_error:
            return "JVM_error";
            break;
        default:
            return "unknown";
            break;
        }

    } // end of xlate_jvm_connect_status().

    /**
     ** ts - Return the (shared) type_stable file address memory-mapped into the Host Agent's address space.
     **/
    inline TYPE_STABLE_LAYOUT_T* ts() { return _ts; }

    /**
     ** Sets and clears the host agent Process Id field, in the type_stable shared memory space.  This allows the JVM
     ** to know when the Host Agent is ready, or not.
     **/
    static inline void set_host_agent_ready(TYPE_STABLE_LAYOUT_T* ts) { set_pid(ts, user_is_host_agent, ::getpid()); }
    static inline void clr_host_agent_ready(TYPE_STABLE_LAYOUT_T* ts) { set_pid(ts, user_is_host_agent, 0);          }

};  // end of class ZVisionHostAgent 



#endif // ZVISION_HA_HPP

// end of file zvision_ha.hpp.

