/*
 * Copyright 2013 Azul Systems, Inc.  All rights reserved.
 * AZUL PROPRIETARY/CONFIDENTIAL.  Use is subject to license terms.
 */
#ifndef ZVISION_JVM_HPP
#define ZVISION_JVM_HPP

#include "arta.hpp"
#include "zvision.hpp"
#include "zvision_csl.hpp"

/**
 ** Datestamp formating info used when creating the base filename.
 **/
#define ZVISION_DATESTAMP_FORMAT              "%Y%m%d_%H%M%S_"
#define ZVISION_DATESTAMP_LEN                 (::strlen("YYYYMMDD_HHMMSS_"))
#define MAX_HOST_AGENT_CONNECTS               2

/**
 ** class ZVisionJVM - class for the Azul JVM ZVision API calls.  Only methods called by the JVM go here, see
 ** the ZVisionHostAgent class for Host Agent speicifc API calls.
 **/
class ZVisionJVM: public ZVision
{
  private:

    char     _basename_prefix[MAX_ZVISION_PATHNAME_LEN];  // type_stable filename prefix.
    char     _date_time_procid[MAX_ZVISION_PATHNAME_LEN]; // type_stable filename suffix (default is date/time/procid).
    char     _basename[MAX_ZVISION_PATHNAME_LEN];         // Fully qualified type_stable file name (prefix & suffix combined).
 
    ZVision::TYPE_STABLE_LAYOUT_T* _ts;                   // JVM's (shared) memory-mapped type_stable file addr.
    ZVision::JVM_KEY_T             _jvm_key;              // JVM generated security key.
    HA_HANDLE_T                    _handle;               // Host Agent generated handle that identifies this JVM.

  public:

    typedef size_t (*IDX_BUF_SIZE_METHOD_T)();
    typedef size_t (*IDX_BUF_UNIT_SIZE_METHOD_T)();
    typedef void   (*IDX_BUF_INIT_METHOD_T)(void*);

    typedef struct indexed_buffer_callback_t
    {
        SECTION_TYPE_E                 type;
        IDX_BUF_SIZE_METHOD_T          size_callback;      // The JVM func that returns the total size of the file.
        IDX_BUF_UNIT_SIZE_METHOD_T     unit_size_callback; // The JVM func that returns the unit size of a buffer within the file.
        IDX_BUF_INIT_METHOD_T          init_callback;      // The JVM func that initializes the file and any assocated items.
        indexed_buffer_callback_t*     next;
      
    } IDX_BUF_CALLBACK_ENTRY_T;

    static IDX_BUF_CALLBACK_ENTRY_T*  _cb_list_head;    // Pointer to linked list of indexed buffer initialization callbacks.

    /**
     ** Return the callbacks structure entry that matches the type given, or NULL if no matching type.
     **/
    static IDX_BUF_CALLBACK_ENTRY_T* find_indexed_buffer_entry(SECTION_TYPE_E type)
    {
        IDX_BUF_CALLBACK_ENTRY_T* current = ZVisionJVM::_cb_list_head;

        while (current != NULL) 
        { 
            if (current->type == type) { return current; }
            current = current->next; 
        }

        return NULL;

    } // end of find_indexed_buffer_entry().


    /**
     ** Returns the callback method registered by the JVM subsystem associated with the particular indexed buffer file when
     ** determining the amount of memory required for the indexed buffer file itself.
     **
     ** For example, the Tick Profiling subsystem would use the EventTickBuffer::allocate_buffer_pool() method to determine
     ** the amount of memory it needs for the tick buffers.  This callback would have been registered during initialization
     ** (see the create_vm() method within thread.cpp) and it would be called before the EventTickBuffer::init() method had
     ** a chance to run.  The ZVisionServiceThread::init() method would call open() and mmap() on its behalf, establishing
     ** the memory.
     **/
    static IDX_BUF_SIZE_METHOD_T find_indexed_buffer_size_callback(SECTION_TYPE_E type)
    {
        IDX_BUF_CALLBACK_ENTRY_T* entry = find_indexed_buffer_entry(type);

        if (entry == NULL) { return NULL; }
          
        return entry->size_callback;
    }


    static IDX_BUF_UNIT_SIZE_METHOD_T find_indexed_buffer_unit_size_callback(SECTION_TYPE_E type)
    {
        IDX_BUF_CALLBACK_ENTRY_T* entry = find_indexed_buffer_entry(type);

        if (entry == NULL) { return NULL; }
          
        return entry->unit_size_callback;
    }


    static IDX_BUF_INIT_METHOD_T find_indexed_buffer_init_callback(SECTION_TYPE_E type)
    {
        IDX_BUF_CALLBACK_ENTRY_T* entry = find_indexed_buffer_entry(type);

        if (entry == NULL) { return NULL; }
          
        return entry->init_callback;
    }


    /**
     ** Allows a non-ZVision task to register a callback that establishes the size of a file, when we need 
     ** to call an initialization routine for that (indexed buffer type) file during ZVision init.
     **/
    static void register_indexed_buffer_callbacks(SECTION_TYPE_E              type, 
                                                  IDX_BUF_SIZE_METHOD_T       size_callback,
                                                  IDX_BUF_UNIT_SIZE_METHOD_T  unit_size_callback,
                                                  IDX_BUF_INIT_METHOD_T       init_callback)
    {
        IDX_BUF_CALLBACK_ENTRY_T* current = _cb_list_head;
        IDX_BUF_CALLBACK_ENTRY_T** last   = &_cb_list_head;

        while (current != NULL)
        {  
            last    = &current->next;  
            current = current->next; 
        }

        current = (IDX_BUF_CALLBACK_ENTRY_T*) malloc(sizeof(IDX_BUF_CALLBACK_ENTRY_T));

        if (current == NULL)
        {
            PATHOLOGICAL("Could not allocate %lu bytes for callback registration.\n", sizeof(IDX_BUF_CALLBACK_ENTRY_T));
            return;
        }

        *last                       = current;
        current->type               = type;
        current->size_callback      = size_callback;
        current->unit_size_callback = unit_size_callback;
        current->init_callback      = init_callback;
        current->next               = NULL;

        return;

    }  // end of register_indexed_buffer_callbacks().


    /**
     ** Free all registered callbacks.
     **/
    static void free_indexed_buffer_callbacks()
    {
        IDX_BUF_CALLBACK_ENTRY_T* current = _cb_list_head;

        while (current != NULL) 
        { 
            IDX_BUF_CALLBACK_ENTRY_T* next = current->next;
            free(current);
            current = next; 
        }

    } // end of free_indexed_buffer_callbacks().


    /**
TODO: blocks when there are no Host Agent requests to process.
     **
     ** ZVisionJVM.   Called only once per life cycle of the JVM.  It is passed whatever overridden default strings
     ** the user specified at startup.  It creates the IPC socket, sets up the connection socket with the Host Agent, and
     ** creates all of the memory mapped (shared) files.
     **
     ** ON ERROR:  If the constructor completes and the IPC state isn't ipc_state_no_connection, then something has failed,
     ** otherwise the IPC state will be set to ipc_state_no_connection.
     **
     ** NOTES: 
     **   - Called by JVM startup logic only, must be during the single threaded create_vm() method.
     **   - This method expects to instantiate and mmap() those files that will be shared by the JVM & Host Agent.
     **   - Because some types of files (such as the indexed buffer Tick Profiling buffers) are also created as part
     **     of the standard (non-ZVision) startup, it is necessary that those routines pre-register their initializaiton
     **     routines with ZVision.  When this function executes, it will invoke these callbacks, then allocate and
     **     mmap() the memory so that the non-ZVision methods can take advantage of the initialization, or, if ZVision
     **     is turned off, realize that they should do it themselves.
     **
     **/
    ZVisionJVM(const char*  zvision_dir_str, 
               const char*  basename_prefix_str, 
               const char*  basename_str, 
               const char*  ipc_name_suffix_str, 
               JVM_KEY_T    jvm_key,
               FILTER_KEY_T filter_key)
    {
        DIAGOUT(1, "ZVisionJVM():  dir \"%s\", prefix \"%s\", basename \"%s\", ipc \"%s\", jvm key 0x%lu, filter key 0x%lu.\n",
             zvision_dir_str, basename_prefix_str, basename_str, ipc_name_suffix_str, jvm_key, filter_key);
            
        // TODO: Check for duplicate invocations!  

        // If the user supplied a key, use it, otherwise default to the Unix Process Id.
        if (jvm_key == 0) { jvm_key = ::getpid(); }

        set_jvm_key(jvm_key);

        char buf[RESULT_SIZE] = { 0 };

        if (create_basename_prefix(basename_prefix_str) != true)
        {
            PATHOLOGICAL("create_basename_prefix() failed.\n");
            set_ipc_state(ipc_state_pathological);
            return; 
        }
    
        if (create_date_time_procid() != true)
        {
            PATHOLOGICAL("create_date_time_procid() failed.\n");
            set_ipc_state(ipc_state_pathological);
            return; 
        }
    
        if (create_basename(basename_str) != true)
        {
            PATHOLOGICAL("create_basename() failed.\n");
            set_ipc_state(ipc_state_pathological);
            return; 
        }
    
        if (zvision_init() != true)
        {
            PATHOLOGICAL("zvision_init() failed.\n");
            set_ipc_state(ipc_state_pathological);
            return; 
        }
    
        // Some sanity checking here to make sure that changes to the structures haven't exceeded our assumptions.
        // Compare the theoretical size of the type stable header with the actual structure size.
        size_t layout_header_size = (
                   (sizeof(MAGIC_COOKIE_T) * 1)                     +  // Cookie.
                   (sizeof(VERSION_INFO_T) * 1)                     +  // Version structure.
                   (sizeof(size_t)         * 2)                     +  // Section Count & Active Count.
                   (sizeof(SOCKET_FD_T)    * 4)                     +  // IPC fds for JVM & Host Agent, JVM Server fd, & unused fd.
                   (sizeof(int)            * 2)                     +  // Host Agent PID & JVM PID.
                   (sizeof(FORGO_DATA_T)   * FORGO_DATA_WORD_COUNT) +  // Data Available array
                   (sizeof(IPC_STATE_E)    * 1)                     +  // JVM State.
                   MAX_IPC_SOCKET_NAME_SIZE);                          // Socket Filename Buffer.

        if ((layout_header_size + sizeof(SECTION_ENTRY_T)) != sizeof(TYPE_STABLE_HEADER_T))
        {
            PATHOLOGICAL("size of header (%lu) doesn't match size of TYPE_STABLE_HEADER_T (%lu) .\n",
                (layout_header_size + sizeof(SECTION_ENTRY_T)), sizeof(TYPE_STABLE_HEADER_T));
            set_ipc_state(ipc_state_pathological);
            return; 
        }

        size_t ts_file_size  = layout_header_size + (sizeof(SECTION_ENTRY_T) * SECTION_TYPES_COUNT);
    
        if ((layout_header_size + (sizeof(SECTION_ENTRY_T) * SECTION_TYPES_COUNT)) != sizeof(TYPE_STABLE_LAYOUT_T))
        {
            PATHOLOGICAL("Type Stable file size (%lu) doesn't sum to components sizes (%lu).\n",
                sizeof(TYPE_STABLE_LAYOUT_T), ts_file_size);
            set_ipc_state(ipc_state_pathological);
            return;
        }
    
        DIAGOUT(3, "JVM Type Stable Component Sizes:\n");
        DIAGOUT(3, "  Header:               %lu\n",   layout_header_size);
        DIAGOUT(3, "  Section Entry:        %lu\n",   sizeof(SECTION_ENTRY_T));
        DIAGOUT(3, "  Entry Cnt:            %d\n",    SECTION_TYPES_COUNT);
        DIAGOUT(3, "  Total File Size:      %lu\n",   sizeof(TYPE_STABLE_LAYOUT_T));
    
    
        // For other files, the allocate() is called before the create_file(), to set up the default (zero'ed) fields, but
        // for the type_stable file we need to make that happen ahead of time.  Use the _earliest_sw_ver as the
        // version for this entry, which must match the _type_version_table entry for the (only) type_stable entry there.
        const CROSS_REF_ENTRY_T* xref = get_method_info(type_stable, (VERSION_INFO_T*) earliest_sw_ver());

        if (xref == NULL)
        { 
            xlate_version_struct((VERSION_INFO_T*) earliest_sw_ver(), buf);

            PATHOLOGICAL("Method for type \"%s\", version \"%s\" was NULL.\n", xlate_section_type(type_stable), buf);
            set_ipc_state(ipc_state_pathological);
            return;
        }

        if (xref->cookie != COOKIE_CROSS_REF)
        {
            xlate_version_struct((VERSION_INFO_T*) earliest_sw_ver(), buf);

            PATHOLOGICAL("Invalid cross reference cookie, expected %lu, got %lu instead, for \"%s\", version \"%s\".\n", 
                COOKIE_CROSS_REF, xref->cookie, xlate_section_type(type_stable), buf);
            set_ipc_state(ipc_state_pathological);
            return;
        }

        if (xref->max_size != sizeof(TYPE_STABLE_LAYOUT_T))
        {
            xlate_version_struct((VERSION_INFO_T*) earliest_sw_ver(), buf);

            // PATHOLOGICAL("Invalid cross reference max size, expected %lu, got %lu instead, for \"%s\", version \"%s\".\n", 
            //     sizeof(TYPE_STABLE_LAYOUT_T), xref->max_size, xlate_section_type(type_stable), buf);
            // return;
        }

        if (xref->unit_size != sizeof(TYPE_STABLE_LAYOUT_T))
        {
            xlate_version_struct((VERSION_INFO_T*) earliest_sw_ver(), buf);

            // PATHOLOGICAL("Invalid cross reference unit size, expected %lu, got %lu instead, for \"%s\", version \"%s\".\n", 
            //     sizeof(TYPE_STABLE_LAYOUT_T), xref->unit_size, xlate_section_type(type_stable), buf);
            // set_ipc_state(ipc_state_pathological);
            // return;
        }

        if (xref->unit_cnt != 1)
        {
            xlate_version_struct((VERSION_INFO_T*) earliest_sw_ver(), buf);

            PATHOLOGICAL("Invalid cross reference unit count, expected 1, got %lu instead, for \"%s\", version \"%s\".\n", 
                xref->unit_cnt, xlate_section_type(type_stable), buf);
            set_ipc_state(ipc_state_pathological);
            return;
        }

        if (xref->notify_lmt != 0)
        {
            xlate_version_struct((VERSION_INFO_T*) earliest_sw_ver(), buf);

            PATHOLOGICAL("Invalid cross reference notify limit, expected 0, got %lu instead, for \"%s\", version \"%s\".\n", 
                xref->notify_lmt, xlate_section_type(type_stable), buf);
            set_ipc_state(ipc_state_pathological);
            return;
        }

        // Now create the temporary FILE_DATA_T and CONSTANTS_T structures we will use to
        // call the create_file() and mmap_file() methods.
        const CONSTANTS_T*    ts_constants = constants_by_type(type_stable);

        // NOTE: This is a local structure, which we fill in and pass to the create_file() method.
        FILE_DATA_T           ts_file_data;                                
                    
        // Hardcoded constants and/or data that must be in place a priori.
        ts_file_data.cookie            = COOKIE_FILE_DATA;
        ts_file_data.max_size          = ts_file_size;
        ts_file_data.type              = type_stable;
        ts_file_data.allocated         = true;
        ts_file_data.active            = true;

        // This call fills in the file name in the local type_stable's FILE_DATA_T structure, as well as
        // the jvm_mmap_fd.  We've filled in the cookie, size & type above.
        bool rc = create_file(&ts_file_data, ts_constants, zvision_dir_str, basename_prefix_str, basename_str);
       
        if (rc != true)
        {
            PATHOLOGICAL("create_file() call failed for JVM type_stable file.\n");
            set_ipc_state(ipc_state_pathological);
            return;
        }

        // Memory map in the file; then we'll fill it in.
        ts_file_data.jvm_mmap_addr = 
            mmap_file(ts_file_data.jvm_mmap_fd,     // Filled in during create_file() method.
                      ts_file_data.max_size,        // Filled in by call to EventTickBuffer::allocate_buffer_pool.
                      ts_file_data.name);           // The file name of the type stable file.

        // This just shouldn't fail.
        if (ts_file_data.jvm_mmap_addr == MAP_FAILED)
        {
            PATHOLOGICAL("mmap_file() returned MAP_FAILED for JVM type_stable file.\n");
            set_ipc_state(ipc_state_pathological);
            return;
        }

        // Once "_ts" is established (i.e. we've memory mapped in the type_stable file), we can reference its contents
        // directly in the shared memory through the pointer.  At this point we should no longer refer to the locally
        // defined (temporary) type_stable FILE_DATA_T stucture (ts_file_data) above, and use the pointer (_ts) instead.
        _ts = (TYPE_STABLE_LAYOUT_T*) ts_file_data.jvm_mmap_addr;

        // No reason for this to ever happen.
        if (_ts == NULL)
        {
            PATHOLOGICAL("NULL address for JVM type_stable file pointer.\n");
            set_ipc_state(ipc_state_pathological);
            return;
        }

        // Okay, we now fill in the shared memory file contents, from data that is known at init time.
        _ts->cookie                                    = COOKIE_TYPE_STABLE;
        _ts->version.release                           = current_sw_ver()->release;
        _ts->version.major                             = current_sw_ver()->major;
        _ts->version.minor                             = current_sw_ver()->minor;
        _ts->version.revision                          = current_sw_ver()->revision;
        _ts->jvm_section_cnt                           = SECTION_TYPES_COUNT;
        _ts->active_cnt                                = 0;
        _ts->ipc_state                                 = ipc_state_initializing;
        _ts->jvm_pid                                   = 0;   // We don't fill this in until completely initiailized.
        _ts->host_agent_pid                            = 0;   // At this point the Host agent is still unitialized.
        
        for (int loop = 0; loop < ((int) FORGO_DATA_WORD_COUNT); loop++)
        {
            _ts->forgo_data[loop]                      = 0;   // There shouldn' be any transactions yet.
        }

        _ts->jvm_server_fd                             = -1;  // We haven't yet created the JVM server fd.
        _ts->jvm_ipc_fd                                = -1;  // We haven't yet created the JVM send fd.
        _ts->ha_ipc_fd                                 = -1;  // We haven't yet created the Host Agent send fd.
        _ts->ipc_domain_id[0]                          = 0;   // We haven't yet created the socket base filename.

        int index = type_index(type_stable);

        // This next chunk is a bit obscure and/or self-referential, so a little explanation is probably in order.
        // The type_stable file has an array of section entries, including a self-referential "type_stable" section entry.
        // For completness (because it is a shared-with-the-host-agent file) we need to fill in the fields of that structure.
        // We refer to the type_stable section of the type_stable file as the ts_section; confusing, but it can't be avoided.
        SECTION_ENTRY_T* ts_section = &_ts->sections[index];

        // These have already been set by the create_file() or mmap_file() methods, and then stored in the 
        // local FILE_DATA_T structure (ts_file_data) above, so must now be copied over.
        ts_section->file.jvm_mmap_addr          = ts_file_data.jvm_mmap_addr;
        ts_section->file.ha_mmap_addr           = NULL;
        ts_section->file.jvm_mmap_fd            = ts_file_data.jvm_mmap_fd;
        ts_section->file.name_len               = ts_file_data.name_len;

        ZV_ASSERT(ts_section->file.name_len > 0);

        // Do not copy over more than the size of the name array.
        ::memcpy((void*) ts_section->file.name, 
                 (const void*) ts_file_data.name, 
                 (ts_file_data.name_len <  MAX_ZVISION_PATHNAME_LEN) ?  ts_file_data.name_len : MAX_ZVISION_PATHNAME_LEN);

        ZV_ASSERT(::strlen((char*) &ts_section->file.name[0]) > 0);

        // This next bit effectively replaces what we do for all the other section entries. [i.e.  initialize_section_type()]
        ts_section->file.max_size               = ts_file_size;
        ts_section->file.type                   = type_stable;
        ts_section->file.version.release        = earliest_sw_ver()->release;
        ts_section->file.version.major          = earliest_sw_ver()->major;
        ts_section->file.version.minor          = earliest_sw_ver()->minor;
        ts_section->file.version.revision       = earliest_sw_ver()->revision;
        ts_section->file.ha_mmap_len            = ts_file_size;
        ts_section->file.ha_mmap_fd             = -1;
        ts_section->file.ha_mmap_gen            = 1;  
        ts_section->file.ha_output_fd           = -1;
        ts_section->file.ha_output_cnt          = 0;
        ts_section->file.allocated              = true;
        ts_section->file.active                 = true;
        ts_section->file.section_lock           = 0;  
        ts_section->file.cookie                 = COOKIE_FILE_DATA;
        ts_section->cookie                      = COOKIE_SECTION_ENTRY;
        
        // These come from the cross reference object.
        ts_section->file.alignment              = xref->alignment;
        ts_section->file.notify_lmt             = xref->notify_lmt;          
        ts_section->file.notify_cntdwn          = xref->notify_lmt;  // Countdown of writes without signaling HA.
        ts_section->file.unit_size              = xref->unit_size;            
        ts_section->file.unit_cnt               = xref->unit_cnt;           
        ts_section->jvm_funcs.init_buf_func     = xref->jvm_funcs.init_buf_func;
        ts_section->jvm_funcs.preload_buf_func  = xref->jvm_funcs.preload_buf_func;
        ts_section->jvm_funcs.cookie            = COOKIE_JVM_FUNCS;

        // The Host Agent function references are not meaning at this moment, mark it as unitialized.
        ts_section->ha_funcs.compress_buf_func  = NULL;
        ts_section->ha_funcs.cookie             = COOKIE_HA_FUNCS;

        // We've now created the type_stable file...
        DIAGOUT(3, "Created JVM type_stable shared-memory file \"%s\" (%lu bytes), mapped into address %p.\n", 
            ts_section->file.name, ts_section->file.max_size, ts_section->file.jvm_mmap_addr); 

        // Next go through the list of type sections and find the appropriate methods based on the version.
        for (size_t loop = (size_t) start_token; loop < (size_t) end_token; loop++)
        {
            DIAGOUT(2, "######################################\n");

            SECTION_TYPE_E type = (SECTION_TYPE_E) loop;
       
            // The type_stable section's associated file has already been created, initialized and memory mapped.
            // Therefore we can skip this section type.
            if (type == type_stable) { continue; }
        
            // Allocate the appropriate section information.  This fills in the type_stable
            // (shared memory) file SECTION_ENTRY_T entry for the appropriate type.
            if (initialize_section_type(_ts, type) != true) 
            {
                PATHOLOGICAL("initialize_section_type(section = \"%s\" (%d)) function call failed.\n", 
                    xlate_section_type(type), type);
                set_ipc_state(ipc_state_pathological);
                return;
            }
    
            const CONSTANTS_T* constants = constants_by_type(type);

            if (constants == NULL)
            {
                PATHOLOGICAL("constants_by_type(section = \"%s\" (%d)) function returned NULL.\n", 
                    xlate_section_type(type), type);
                set_ipc_state(ipc_state_pathological);
                return;
            }

            FILE_DATA_T*        file_data  = file_data_by_type(_ts, type);

            if (file_data == NULL)
            {
                PATHOLOGICAL("file_data_by_type(section = \"%s\" (%d)) function returned NULL.\n", 
                    xlate_section_type(type), type);
                set_ipc_state(ipc_state_pathological);
                return;
            }

            //--------------------------------------------------------------------------------------------------------//
            //                                                                                                        //
            // If this is an indexed_buffer_file (i.e., a file that contains data asynchronously generated by the JVM //
            // e.g. an Event Tick Buffer file) then we will want to allocate it here (before the file's subsystem     //
            // initialization logic has executed) so we'll need to call that appropriate (type specific) logic in     //
            // order to determine its size, instead of using the value defined in the constants structure.            //
            //                                                                                                        //
            //--------------------------------------------------------------------------------------------------------//
            if (constants->file_type == indexed_buffer_file)
            {

                // Indexed buffer files within the JVM should only derive their size through their
                // own (JVM specific) subsystem calls below, and never be defined within ZVision itself,
                // as it is possible for them to be any size chosen by the user at runtime.  That means 
                // that any indexed buffer file needs to provide some means of initialization that can
                // be called before it would otherwise be invoked during a non-ZVision run.  So, we need
                // to know what size it should be, and then after we open() and mmap() the file, we need
                // to set the subsystem's internal address to the memory we've allocated on its behalf.
                // e.g. For Tick Profiling, we call the init_buffer_pool_size() method to obtain its size,
                // and then the init() method to allocate the memory.
                if (file_data->type == tick_profiling || file_data->type == method_names)
                {
                    IDX_BUF_SIZE_METHOD_T size_callback = find_indexed_buffer_size_callback(file_data->type);

                    if (size_callback == NULL)
                    {
                        PATHOLOGICAL("File type \"%s\" missing indexed buffer size callback.\n", xlate_section_type(type));
                        set_ipc_state(ipc_state_pathological);
                        return;
                    }

                    file_data->max_size = (size_callback)();

                    // Check that the size of the allocated file still makes sense.
                    if (file_data->max_size <= 0)
                    {
                        PATHOLOGICAL("File type \"%s\" has invalid size %lu.\n", xlate_section_type(type), file_data->max_size);
                        set_ipc_state(ipc_state_pathological);
                        return;
                    }

                    file_data->ha_mmap_len = file_data->max_size;

                    // Grab the unit size of the entries.
                    IDX_BUF_UNIT_SIZE_METHOD_T unit_size_callback = find_indexed_buffer_unit_size_callback(file_data->type);

                    if (unit_size_callback == NULL)
                    {
                        PATHOLOGICAL("File type \"%s\" missing indexed buffer unit-size callback.\n", xlate_section_type(type));
                        set_ipc_state(ipc_state_pathological);
                        return;
                    }

                    file_data->unit_size = (unit_size_callback)();
                    file_data->unit_cnt  = (file_data->unit_size > 0) ? (file_data->max_size / file_data->unit_size) : 1;

                    DIAGOUT(0, "File type \"%s\" (indexed buffer file) unit size is %ld, unit cnt is %ld.\n", 
                        xlate_section_type(type), file_data->unit_size, file_data->unit_cnt);

                    // Complain about this in non-production builds.  Cope with it otherwise.
                    if (file_data->unit_size > file_data->max_size)
                    {
                        PATHOLOGICAL("File type \"%s\" has invalid unit size (%lu) greater than maximum size %lu.\n", 
                            xlate_section_type(type), file_data->unit_size, file_data->max_size);
                        file_data->unit_size = file_data->max_size;
                        set_ipc_state(ipc_state_pathological);
                    }
                    else if (file_data->unit_size == 0)
                    {
                        DIAGOUT(0, "File type \"%s\" has zero unit size.\n", xlate_section_type(type));
                    }
                }

            }
      
            bool rc;
                    
            // Most types are expected to be stored as SYSV (POSIX) /dev/shm/... shared memory objects.
            // Unless the user has specifically requested not to (which creates a disk backed file) we
            // will place everything except proprietary types in SYSV shared memory.  This means there is
            // no file system backed image of the file, and therefore eliminates pressure on either the OS 
            // cache and swap pages.
            if (constants->use_dev_shm == true)
            {
                // Create the shared memory (without disk backing) file for this type.
                rc = create_shared_memory_object(file_data, constants, basename_str);
           
                if (rc != true)
                {
                    PATHOLOGICAL("create_file() function call failed for type \"%s\" (%d).\n", 
                        xlate_section_type(type), type);
                    set_ipc_state(ipc_state_pathological);
                    return;
                }
            }
            else
            {
                // Create the shared memory (but disk backed) file for this type.
                rc = create_file(file_data, constants, zvision_dir_str, basename_prefix_str, basename_str);
           
                if (rc != true)
                {
                    PATHOLOGICAL("create_file() function call failed for type \"%s\" (%d).\n", 
                        xlate_section_type(type), type);
                    set_ipc_state(ipc_state_pathological);
                    return;
                }
            }

            // This is a shared memory file, memory map it in as a shared memory address space.
            file_data->jvm_mmap_addr = mmap_file(file_data->jvm_mmap_fd,     // file descriptor.
                                                 file_data->max_size,        // file size.
                                                 file_data->name);           // file name string.

            if (file_data->jvm_mmap_addr == MAP_FAILED) 
            {
                PATHOLOGICAL("mmap_file() returned MAP_FAILED, section type is \"%s\" (%d).\n", xlate_section_type(type), type);
                set_ipc_state(ipc_state_pathological);
                return;
            }
    
            // TODO:  We can probably eliminate the initialization function & therefore the reference to entry.
            SECTION_ENTRY_T* entry = section_entry_by_type(_ts, type);

            if (entry == NULL)
            {
                PATHOLOGICAL("section_entry_by_type(section = \"%s\" (%d)) function returned NULL.\n", 
                    xlate_section_type(type), type);
                set_ipc_state(ipc_state_pathological);
                return;
            }

            if (entry->jvm_funcs.init_buf_func == NULL)
            {
                PATHOLOGICAL("Buffer initialization function for \"%s\" section (%d) was NULL.\n", 
                    xlate_section_type(type), type);
                set_ipc_state(ipc_state_pathological);
                return;
            }

            // Call whatever initialization method is appropriate based the file type.
            if (entry->jvm_funcs.init_buf_func != noop_buffer_init)
            { 
               // Call the appropriate initialization method based on what kind of file type 
               // it is (request_response_file, indexed_buffer_file, etc.) as defined in the 
               // CONSTANTS_T structure for this type.
                if (entry->jvm_funcs.init_buf_func(_ts, type, file_data->jvm_mmap_addr) != true)
                { 
                    PATHOLOGICAL("User specific init(%p, \"%s\" (%d), %p) method failed.\n", 
                        _ts, xlate_section_type(type), type, file_data->jvm_mmap_addr);
                    set_ipc_state(ipc_state_pathological);
                    return;
                }
            }
            else if (constants->file_type == indexed_buffer_file)
            {
                //--------------------------------------------------------------------------------------------------------//
                //                             Indexed Buffer File shared memory file creation                            //
                //                                                                                                        //
                //                                                                                                        //
                // For each Indexed Buffer File, check its corresponding global to see if the user wants to create a      //
                // shared memory space ring buffer, such that whenever an indexed buffer entry is updated, the ring       //
                // buffer is updated as well.   This allows tasks to poll on the ring buffer and pull out data as it      //
                // is added.   A separate process can then use the ZVisionFilterTick APIs to pull out only those entries  //
                // it is intereset in.  For instance, in the case of tick buffers, whenever a new tick buffer is added    //
                // to the recycle list, a ring buffer entry is added that points to the newest Event Tick Buffer.  The    //
                // calling process can then use the API to pull out only tick entries that meet whatever criteria it has  //
                // previously set.                                                                                        //
                //                                                                                                        //
                //--------------------------------------------------------------------------------------------------------//

//TBD: This should apply to every indexed buffer file, not just tick profiling!!

                // If we are going to collect Event Tick Buffer entries for processing by a separate task, then create a 
                // shared memory file that will contain a constantly updating ring buffer of tick data.  We use the
                // ZVisionFilterTicks object API to do so, and a separate (customer's) process can then read (by polling
                // the ring buffer for new Event Tick Buffer) data.
                if (file_data->type == tick_profiling && filter_key != 0)
                {
                    size_t elem_size     = ZVisionFilterTicks::tick_element_size;
                    size_t hdr_size      = ZVisionFilterTicks::tick_header_size;
                    size_t alloc_offset  = ZVisionFilterTicks::tick_alloc_offset;

                    // Create the indexed buffer file's shared-memory ring buffer file, that we'll use to process ticks.
                    ZVisionFilterTicks* ft_obj = new ZVisionFilterTicks(file_data, filter_key, elem_size, hdr_size, alloc_offset);

                    // For the "Producer" task, the object should always intialize properly and
                    // be in state "active".  If the object failed to initialize, it is an error.
                    if (ft_obj->state() != ZVisionFilterTicks:: active)
                    {
                        PATHOLOGICAL("Could not create ZVisionFilterTicks object.\n");
                        delete ft_obj;
                        set_ipc_state(ipc_state_pathological);
                        return; 
                    }

                    file_data->jvm_filter_obj = (ZVisionFilter*) ft_obj;
                    
                    ZV_ASSERT(::strncmp(file_data->name, ft_obj->buf_filename(), file_data->name_len) == 0);

                    DIAGOUT(1, "Created ZVisionFilterTicks object %p for data file \"%s\".\n", ft_obj, ft_obj->buf_filename());
                }
             
                // Call the initialization logic (or re-initialization logic, if we happen to be resizing
                // the memory) for the indexed_buffer_file.  Files like this may dynamically resize their
                // address space.
                IDX_BUF_INIT_METHOD_T init_callback = find_indexed_buffer_init_callback(file_data->type);

                if (init_callback == NULL)
                {
                    PATHOLOGICAL("File type \"%s\" missing indexed buffer init callback.\n", xlate_section_type(type));
                    set_ipc_state(ipc_state_pathological);
                    return;
                }

                (init_callback)(file_data->jvm_mmap_addr);

                continue;
            }
            else if (constants->file_type == request_response_file)
            {
                // All request/response buffers should use a generic initialization routine,
                // which zeros out counters, clears buffers, etc.
                if (request_buffer_init(file_data->jvm_mmap_addr) != true)
                {
                    PATHOLOGICAL("request_buffer_init(%p) for type \"%s\" (%d) returned false.\n", 
                        file_data->jvm_mmap_addr, xlate_section_type(type), type);
                    set_ipc_state(ipc_state_pathological);
                    return;
                }

                continue;
            }
            else if (constants->file_type == proprietary_file)
            {
                // Proprietary files require no special initialization. Nothing to see here folks, move along...
                DIAGOUT(3, "Ignoring initialization method for \"%s\" file of type \"%s\" (%d).\n",
                    xlate_file_type(constants->file_type), xlate_section_type(type), type);

                continue;
            }
            else
            {
                PATHOLOGICAL("Constants file type \"%s\" (%d) invalid.\n", 
                    xlate_file_type(constants->file_type), constants->file_type);
                set_ipc_state(ipc_state_pathological);
                return;
            }

        }   // end of for (size_t loop = (size_t) start_token; loop < (size_t) end_token; loop++)
       
        // Okay, now build the socket file name that we'll use to receive the requests,
        // and on which we'll send the asynchronous notifications & responses.
        if (ipc_name_create(ipc_name_suffix_str, basename_str) != true)
        {
            PATHOLOGICAL("ipc_name_create() failed.\n");
            set_ipc_state(ipc_state_pathological);
            return;
        }

        // Also set the PID value to indicate that the JVM is ready and available.
        set_jvm_ready(_ts);

        // List the prohibited types, nothing should be restricted this point.
        show_forgo_data(_ts, 3);

        // Now create the JVM port for connection with the Host Agent.  
        if (_ts->jvm_server_fd >= 0)
        {
            PATHOLOGICAL("Unix Domain Server fd (%d) already exists.\n", _ts->jvm_server_fd);
            set_ipc_state(ipc_state_pathological);
            return; 
        }

        if (_ts->jvm_ipc_fd >= 0)
        {
            PATHOLOGICAL("Unix Domain IPC fd (%d) already exists.\n", _ts->jvm_ipc_fd);
            set_ipc_state(ipc_state_pathological);
            return; 
        }

        ERRNO_RESULT_T      failure       = 0;
        const char*         ipc_name      = (const char*) _ts->ipc_domain_id;
        size_t              ipc_name_len  = ipc_name == NULL ? 0 : ::strnlen(ipc_name, MAX_IPC_SOCKET_NAME_SIZE);

        struct sockaddr_un  socket_info;
        struct sockaddr_un* socket_info_addr = &socket_info;
        socklen_t           socket_info_len  = sizeof(struct sockaddr_un);

        // clear the structure & create a Unix Domain Named socket.
        ::memset(socket_info_addr, 0, sizeof(struct sockaddr_un));
        ::memcpy((void*) &socket_info.sun_path[0], (const void*) ipc_name, ipc_name_len);
        socket_info.sun_family = AF_UNIX;

        if (ipc_name_len >=  MAX_IPC_SOCKET_NAME_SIZE          || 
            ipc_name_len >= sizeof(socket_info_addr->sun_path) ||
            ipc_name_len <= 1)
        {
            PATHOLOGICAL("Unix Domain Socket filename length (%lu) invalid; \"sun_path\" size (%lu), max len (%lu).\n", 
                ipc_name_len, sizeof(socket_info_addr->sun_path), MAX_IPC_SOCKET_NAME_SIZE);
            set_ipc_state(ipc_state_pathological);
            return; 
        }

        // Create the socket, bind and listen on it.
        if (ipc_state(_ts) != ipc_state_initializing)
        {
            PATHOLOGICAL("State \"%s\" when establishing jvm socket for \"%s\", should have been \"%s\".\n",
                xlate_ipc_state(ipc_state(_ts)), ipc_name, xlate_ipc_state(ipc_state_initializing));
            set_ipc_state(ipc_state_pathological);
            return;
        }

        _ts->jvm_server_fd = ::socket(AF_UNIX, SOCK_SEQPACKET, 0);

        if (_ts->jvm_server_fd <= 0)
        {
            failure = errno;
            PATHOLOGICAL("Call to ::socket(AF_UNIX, SOCK_SEQPACKET, 0) for file \"%s\" failed, error is \"%s\" (%d).\n", 
                ipc_name, ::strerror(failure), failure);
            set_ipc_state(ipc_state_pathological);
            return; 
        }

        DIAGOUT(2, "Called ::socket(AF_UNIX, SOCK_SEQPACKET, 0) which returned fd %d.\n", _ts->jvm_server_fd);

        // Mark the file as blocking.
        if (mark_file_blocking(_ts->jvm_server_fd) != true)
        {
            PATHOLOGICAL("mark_file_blocking(%d) failed for file \"%s\".\n", _ts->jvm_server_fd, ipc_name);
            set_ipc_state(ipc_state_pathological);
            return; 
        }

        DIAGOUT(2, "Calling ::bind(oob fd %d, socket info addr %p, socklen_t %u) for file \"%s\"...\n", 
            _ts->jvm_server_fd, (const struct sockaddr*) socket_info_addr, socket_info_len, ipc_name);

        // Force errno to zero, in case we get back something we weren't expecting.
        errno = 0;

        if (::bind(_ts->jvm_server_fd, (struct sockaddr*) socket_info_addr, socket_info_len) != 0)
        {
            failure = errno;
            PATHOLOGICAL("Could not bind to \"%s\" (JVM IPC fd %d).  Error is %d \"%s\".\n", 
                ipc_name, _ts->jvm_server_fd, failure, ::strerror(failure));
            set_ipc_state(ipc_state_pathological);
            return; 
        }
          
        DIAGOUT(2, "Calling ::listen(oob fd %d, max outstanding requests = %u) for file \"%s\"...\n", 
            _ts->jvm_server_fd, MAX_HOST_AGENT_CONNECTS, ipc_name);

        // Force errno to zero, in case we get back something we weren't expecting.
        errno = 0;

        if (::listen(_ts->jvm_server_fd, MAX_HOST_AGENT_CONNECTS) != 0)
        {
            failure = errno;
            PATHOLOGICAL("Could not listen to \"%s\" (IPC fd %d), error is %d \"%s\".\n", 
                ipc_name, _ts->jvm_server_fd, failure, ::strerror(failure));
            set_ipc_state(ipc_state_pathological);
            return; 
        }

        DIAGOUT(2, "Created the JVM Server fd (%d) using file \"%s\"\n", _ts->jvm_server_fd, ipc_name);

        //--------------------------------------------------------------------------------------------------------//
        //                                   Completed Object Initialization                                      //
        //                                                                                                        //
        // Let the world know that we are now up and running, but haven't connected with a Host Agent as yet.  If //
        // the constructor completes and the IPC state isn't ipc_state_no_connection, then something has failed.  //
        //--------------------------------------------------------------------------------------------------------//

        set_ipc_state(ipc_state_no_connection);

        DIAGOUT(2, "ZVisionJVM() constructor initialization complete.\n");
        DIAGOUT(2, "\n");

        return;
    
    }  // end of ZVisionJVM().


    /**
     ** Shutdown the connection with the Host Agent.
     **/
    ~ZVisionJVM() 
    { 
        DIAGOUT(1, "Cleaning up ZVision Object (%p): handle 0x%lX, key 0x%lX, basename \"%s\".\n", 
            this, _handle, _jvm_key, _basename);
        return;
    }


    /**
     ** Returns the address of the memory mapped type stable memory for this ZVisionJVM.
     **/
    inline TYPE_STABLE_LAYOUT_T* type_stable_ptr() { return _ts; }

    /**
     ** Returns the address of the memory mapped type stable memory for this ZVisionJVM.
     **/
    inline HA_HANDLE_T handle() { return _handle; }

    /**
     ** Returns the special security key generated by the JVM at initialization.
     **/
    inline JVM_KEY_T jvm_key() { return _jvm_key; }

    /**
     ** Sets the special security key generated by the JVM at initialization.
     **/
    inline void      set_jvm_key(JVM_KEY_T key) { _jvm_key = key; }

    /**
     ** Returns the file descriptor the JVM should use to write messages to the Host Agent,
     ** or an illegal file descriptor if the type_stable file pointer is NULL.
     **/
    inline SOCKET_FD_T jvm_fd() { return (_ts != NULL) ? _ts->jvm_ipc_fd : -1; }

    /**
     ** Returns the file descriptor the JVM should use to accept connection from the Host Agent,
     ** or an illegal file descriptor if the type_stable file pointer is NULL.
     **/
    inline SOCKET_FD_T jvm_server_fd() { return (_ts != NULL) ? _ts->jvm_server_fd : -1; }

    /**
     ** Checks that the JVM ZVision state machine state transitions are valid.  We need a valid
     ** type_stable file, and the transition must also be permitted.  Any state change transition 
     ** that isn't sets us into ipc_state_pathological.
     **/
    inline void set_ipc_state(IPC_STATE_E new_state)
    {
        TYPE_STABLE_LAYOUT_T* ts = _ts;  // JVM's (shared) memory-mapped type_stable file addr.

        if (ts == NULL)                     
        { 
            PATHOLOGICAL("_ts was NULL while changing state to \"%s\" (0x%lX), setting to \"%s\" (0x%lX).\n",
                xlate_ipc_state(new_state),              new_state,
                xlate_ipc_state(ipc_state_pathological), ipc_state_pathological);
            ts->ipc_state = ipc_state_pathological; 
            return;
        }

        if (type_stable_valid(ts) == false)
        { 
            PATHOLOGICAL("ts invalid while changing state to \"%s\" (0x%lX), setting to \"%s\" (0x%lX).\n",
                xlate_ipc_state(new_state),              new_state,
                xlate_ipc_state(ipc_state_pathological), ipc_state_pathological);
            ts->ipc_state = ipc_state_pathological; 
            return;
        }

        IPC_STATE_E current_state = ipc_state(ts);

        switch (new_state)
        {
        case ipc_state_uninitialized:
        case ipc_state_initializing:
            if (true)                                         { goto error_exit; }
            break;
        case ipc_state_no_connection:         
            if (current_state != ipc_state_initializing &&
                current_state != ipc_state_connection_est)    { goto error_exit; }
            break;
        case ipc_state_ignoring_host_agent:         
            if (current_state != ipc_state_no_connection)     { goto error_exit; }
            break;
        case ipc_state_got_connect:         
            if (current_state != ipc_state_no_connection)     { goto error_exit; }
            break;
        case ipc_state_got_start:         
            if (current_state != ipc_state_got_connect)       { goto error_exit; }
            break;
        case ipc_state_connection_est:         
            if (current_state != ipc_state_got_start)         { goto error_exit; }
            break;
        case ipc_state_exiting:         
            if (current_state != ipc_state_initializing     &&
                current_state != ipc_state_no_connection    && 
                current_state != ipc_state_got_connect      && 
                current_state != ipc_state_got_start        && 
                current_state != ipc_state_connection_est)    { goto error_exit; }
            break;
        case ipc_state_exited:         
            if (current_state != ipc_state_exiting)           { goto error_exit; }
            break;
        case ipc_state_pathological:         
            if (current_state == ipc_state_pathological)      
            { 
                PATHOLOGICAL("Current state was \"%s\" (0x%lX), state change to same ignored.\n",
                    xlate_ipc_state(current_state), current_state);
                return;
            }
            break;
        default:   
            if (true)                                         { goto error_exit; }
            break;

        } // end of switch (new_state).

        DIAGOUT(3, "Changing state from \"%s\" (0x%lX) to \"%s\" (0x%lX).\n",
            xlate_ipc_state(current_state), current_state, xlate_ipc_state(new_state), new_state);
        _ts->ipc_state = new_state; 
        return;

error_exit:
        PATHOLOGICAL("state change error: state was \"%s\" (0x%lX), setting to \"%s\" (0x%lX).\n",
            xlate_ipc_state(current_state), current_state,
            xlate_ipc_state(ipc_state_pathological), ipc_state_pathological);
        _ts->ipc_state = ipc_state_pathological; 
        return;

    } // end of set_ipc_state().


    /**
     ** Writes the JVM's Unix process id value (in the type_stable file's JVM pid field), indicating that the JVM
     ** has initialized and is now ready to accept requests from the Host Agent.
     **/
    static inline void set_jvm_ready(TYPE_STABLE_LAYOUT_T* ts) { set_pid(ts, user_is_jvm, ::getpid()); }

    /**
     ** Clears the JVM's Unix process id value (in the type_stable file's JVM pid field), indicating that the JVM
     ** is no longer operations or able to accept requests from the Host Agent.
     **/
    static inline void clr_jvm_ready(TYPE_STABLE_LAYOUT_T* ts) { set_pid(ts, user_is_jvm, 0);          }

    /**
     ** initialize_section_type(type) - Fill in the ZVision type_stable shared memory "file_data" for
     ** a section entry [of <type>/<version>], making sure that the object begins at a word boundry appropriate
     ** to an object of its alignment, based  on the cross referrence information.
     ** 
     ** This function should return true, because the calling funtions should always supply the appropriate
     ** parameters during initialization, and once initialization is complete this function should never be
     ** called again.
     **
     ** RETURNS: true, or false on PATHOLOGICAL error.
     **
     ** ON ERROR:  If an inappropriate or invalid value is supplied, the ZVision Diagnostics may not work.
     **
     ** NOTES:
     **
     ** CALLED BY: JVM only.
     **
     **/
    bool initialize_section_type(TYPE_STABLE_LAYOUT_T* ts, SECTION_TYPE_E type)        
    {
        char buf[RESULT_SIZE] = { 0 };

        if (type == type_stable)
        {
            PATHOLOGICAL("initialize_section_type() can not be called for type_stable.\n");
            return false; 
        }

        if (check_type(type) != true)
        { 
            PATHOLOGICAL("check_type(%d) failed.\n", type);
            return false; 
        }

        if (type_stable_valid(ts) != true)
        {
            PATHOLOGICAL("type_stable_valid(%p) failed.\n", ts);
            return false; 
        }

        if (section_init_check(ts, type) != true)
        { 
            PATHOLOGICAL("section_init_check(%d) returned false.\n", type);
            return false; 
        }

        const CONSTANTS_T*       constants = constants_by_type(type);   
              SECTION_ENTRY_T*   section   = section_entry_by_type(ts, type);  
        const VERSION_INFO_T*    mmap_ver  = constants->version;

        // TODO: Atomically set the "allocated" & "active" flags for the segment, we should only allocate/activate a segment once.
        if (section->file.allocated != false)
        {
            PATHOLOGICAL("Duplicated \"%s\" type (%d) initialization.\n", xlate_section_type(type), type_index(type));
            return false; 
        }

        // Find the cross reference method data appropriate to the type & version, as defined by 
        // the constants.  This usually returns the last entry in the array (the newest) but it 
        // is possible to hardcode a version in constants instead, to force the code to use an
        // earlier rev.
        if (mmap_ver == NULL) { mmap_ver = current_sw_ver(); }
        
        const CROSS_REF_ENTRY_T*   xref = get_method_info(type, (VERSION_INFO_T*) mmap_ver);

        if (xref == NULL)
        { 
            //TODO: We need to back out?  section->allocated = false;
            PATHOLOGICAL("Method for type \"%s\", version \"%s\" was NULL.\n", 
                xlate_section_type(type), xlate_version_struct((VERSION_INFO_T*) mmap_ver, buf));
            return false; 
        }

        // Now construct a version based on the verno value in the cross reference.
        VERSION_INFO_T    version;

        if (convert_verno(xref->verno, &version) != true)
        { 
            PATHOLOGICAL("convert_verno() failed for type \"%s\" %d) failed.\n", xlate_section_type(type), type);
            return false; 
        }
        
        DIAGOUT(3, "Allocating section \"%s\" (%d), version \"%s\".\n", 
            xlate_section_type(type), type, xlate_version_struct((VERSION_INFO_T*) &version, buf));

        // Set these fields now, they will be used later on...
        section->file.type                  = type;
        section->file.version.release       = version.release;
        section->file.version.major         = version.major;
        section->file.version.minor         = version.minor;
        section->file.version.revision      = version.revision;
        section->file.alignment             = xref->alignment;
        section->file.max_size              = xref->max_size; 
        section->file.notify_lmt            = xref->notify_lmt;          
        section->file.notify_cntdwn         = xref->notify_lmt;  // This is a countdown, starting from the notify_lmt.
        section->file.unit_size             = xref->unit_size;            
        section->file.unit_cnt              = xref->unit_cnt;           
        section->jvm_funcs.init_buf_func    = xref->jvm_funcs.init_buf_func;
        section->jvm_funcs.preload_buf_func = xref->jvm_funcs.preload_buf_func;
        section->ha_funcs.compress_buf_func = NULL;   // This isn't meaningful to the JVM, only the Host Agent.

        // Magic cookies that should be the correct types.
        section->cookie                 = COOKIE_SECTION_ENTRY;
        section->file.cookie            = COOKIE_FILE_DATA;           
        section->jvm_funcs.cookie       = COOKIE_JVM_FUNCS;
        section->ha_funcs.cookie        = COOKIE_HA_FUNCS;

        // These are set by the create_file() function.
        section->file.jvm_mmap_fd       = -1;
        section->file.jvm_mmap_addr     = NULL;
        section->file.ha_mmap_fd        = -1;
        section->file.ha_mmap_addr      = NULL;
        section->file.ha_mmap_gen       = 0;
        section->file.ha_mmap_len       = 0;
        section->file.ha_output_fd      = -1;
        section->file.ha_output_cnt     = 0;
        section->file.name[0]           = (char) 0;
        section->file.name_len          = 0;
        section->file.allocated         = true;
        section->file.active            = true;

        // We only use the section_lock for "indexed_buffer_file" filetypes
        // i.e. for files whose memory layout might change dynamically, such as
        // the tick_profiling memory (see EventTickBuffer::resize_memory() method)
        // This value represents the number of times we've allocated memory for
        // the section type.  It should always be zero for all other file types.
        section->file.section_lock      = (constants->file_type == indexed_buffer_file) ? 1 : 0;  

        // We use ha_mmap_gen for indexed_buffer_files that can dynamically resize,
        // it should always be zero for all other types.  It helps inform the Host Agent
        // when a file has been resized, and therefore needs to be freed and re-mmap()'ed.
        section->file.ha_mmap_gen       = 1;  

        // We use ha_mmap_len for indexed_buffer_files that can dynamically resize,
        // it should always be zero for all other types.  It helps inform the Host Agent,
        // when resizing a file, how large a section to ::munmap() before remapping the
        // new memory.
        section->file.ha_mmap_gen       = 1;  

        return true;

    }  // end of initialize_section_type().


    
    /**
     ** create_shared_memory_object() - 
     **
     ** RETURNS: true, or false on PATHOLOGICAL error.
     **
     ** ON ERROR:  If an inappropriate or invalid value is supplied, the ZVision Diagnostics may not work.
     **
     ** NOTES: 
     **   - May not be called if type is type_stable.
     **
     **
     ** CALLED BY: JVM only.
     **
     **/
    bool create_shared_memory_object(FILE_DATA_T* file_data, const CONSTANTS_T* constants, const char* basename_str)
    {
        char result1[RESULT_SIZE]   = { 0 };         // Automatiic stack variable for flagbit translation.
        char result2[RESULT_SIZE]   = { 0 };         // Automatiic stack variable for flagbit translation.

        char*  section_type_str     = NULL;

        size_t section_type_str_len = 0;
        size_t basename_str_len     = 0;
        size_t filename_str_len     = 0;

        
        ERRNO_RESULT_T failure      = 0;

        if (file_data == NULL)
        {
            PATHOLOGICAL("File data pointer was NULL.\n");
            return false;
        }

        section_type_str = (char*) xlate_section_type(file_data->type);

        if (section_type_str == NULL)
        {
            PATHOLOGICAL("Filename unavailable for section type.\n");
            return false;
        }

        section_type_str_len = section_type_str != NULL ? ::strnlen(section_type_str, SECTION_TYPE_STR_LEN) : 0;

        if (section_type_str_len < 1)
        {
            PATHOLOGICAL("Section type name is less than 1 character in length.\n");
            return false;
        }

        if (constants == NULL)
        {
            PATHOLOGICAL("Constants pointer was NULL.\n");
            return false;
        }

        // If they didn't give a basename, use the predefined _date_time_procid coupled with the section type string.
        // The prefix will either be the user specified one, or the default if left unchanged.
        if (basename_str == NULL) 
        { 
            basename_str = _basename;
        }

        basename_str_len = ::strnlen(basename_str, MAX_POSIX_SHARED_MEMORY_NAME_SIZE);

        // Ensure that there are no slashes within the basename.
        if (::strchr(&basename_str[0], '/') != NULL)
        {
            PATHOLOGICAL("Basename string of shared memory object contains an illegal slash.\n");
            return false;
        }

        // Make sure the leading slash, the basename, the slash separator and the section type string will fit.
        if (1 + basename_str_len + 1 + section_type_str_len >= MAX_POSIX_SHARED_MEMORY_NAME_SIZE)
        {
            PATHOLOGICAL("Basename length (%lu) plus Section Type length (%lu) plus exceeds shared memory name space (%ld).\n",
                basename_str_len, section_type_str_len, MAX_POSIX_SHARED_MEMORY_NAME_SIZE); 
            return false;
        }

        // Now create the shared memory object name in the file->name buffer.  It must start with a slash, but
        // the file name doesn't contain the leading "/dev/shm".
        file_data->name[0] = '/';
        filename_str_len  = 1;

        // Copy in the basename.
        ::memcpy((void*) &file_data->name[filename_str_len], (const void*) basename_str, basename_str_len);
        filename_str_len += basename_str_len;

        // Add a separator character.
        file_data->name[filename_str_len] = '_';
        filename_str_len++;

        // Copy the section type onto the end of the basename.
        ::memcpy((void*) &file_data->name[filename_str_len], (const void*) section_type_str, section_type_str_len);
        filename_str_len += section_type_str_len;

        // One final check.  Also redundant.
        if (filename_str_len < 1 || filename_str_len >= MAX_POSIX_SHARED_MEMORY_NAME_SIZE)
        {
            PATHOLOGICAL("Shared memory file name length (%lu) exceeds maximum shared memory name size (%lu).\n",
                filename_str_len, MAX_POSIX_SHARED_MEMORY_NAME_SIZE); 
            return false;
        }

        DIAGOUT(2, "Creating POSIX shared memory object \"%s\".\n", file_data->name);
        DIAGOUT(3, "    located in \"%s\", flags=0x%X \"%s\", mode=0x%X \"%s\"...\n", SHMEM_DIR_STR,
                constants->open_flags, XlateFlag::open(constants->open_flags, result1, sizeof(result1)/sizeof(char)), 
                constants->open_mode,  XlateMode::open(constants->open_mode,  result2, sizeof(result2)/sizeof(char)));

        file_data->jvm_mmap_fd      = -1;
        file_data->allocated        = true;
        file_data->active           = true;
        file_data->mmap_flags       = 0;
        file_data->cookie           = COOKIE_FILE_DATA;
        file_data->name_len         = basename_str_len;

        // Force errno to zero, in case we get back something we weren't expecting.
        errno = 0;

        // Now create the new POSIX shared memory object for this type, and create an entry of the appropriate size.
        SOCKET_FD_T fd = ::shm_open((const char*) file_data->name, (int) constants->open_flags, (mode_t) constants->open_mode);

        if (fd < 0)
        {
            failure = errno;
            PATHOLOGICAL("::shm_open(name=\"%s\") failed with errno set to %d (%s).\n", 
              file_data->name, failure, ::strerror(failure)); 
            return false;
        }

        // Force errno to zero, in case we get back something we weren't expecting.
        errno = 0;

        // Seek to the end of the file, this positions the pointer.
        if (::ftruncate(fd, file_data->max_size) != 0)
        {
            failure = errno;

            PATHOLOGICAL("::ftruncate(fd=%d, size=%ld) of file \"%s\" failed with errno set to %d (%s).\n", 
                fd, file_data->max_size, file_data->name, failure, ::strerror(failure));

            if (::close(fd) != 0) { DIAGOUT(1, "::close(fd %d) failed.\n", fd); }

            return false;
        }

        // The file descriptor is stored back into the FILE_DATA_T structure.
        file_data->jvm_mmap_fd = fd;

        return true;

    } // end of create_shared_memory_object().


    /**
     ** create_file() - 
     **
     ** RETURNS: true, or false on PATHOLOGICAL error.
     **
     ** ON ERROR:  If an inappropriate or invalid value is supplied, the ZVision Diagnostics may not work.
     **
     ** NOTES: 
     **   - If the file section type is type_stable, and the basename subdirectory does not exist already,
     **     then we create the basename subdirectory.
     **
     ** If we run out of buffer space, complain and...???   TODO: Figure out what.
     **
     ** CALLED BY: JVM only.
     **
     **/
    bool create_file(FILE_DATA_T*         file_data,            // In ram section type entry.
                     const CONSTANTS_T*   constants,            // Hard coded constants for this section type.
                     const char*          zvision_dir_str,      // Shared memory directory path name string
                     const char*          basename_prefix_str,  // if NON-null, user assigned basename prefix string
                     const char*          basename_str)         // if NON-null, user assigned basename string
    {
        char result1[RESULT_SIZE]   = { 0 };         // Automatiic stack variable for flagbit translation.
        char result2[RESULT_SIZE]   = { 0 };         // Automatiic stack variable for flagbit translation.

        char*  section_type_str     = NULL;

        size_t section_type_str_len    = 0;
        size_t basename_str_len        = 0;
        size_t zvision_dir_str_len     = 0;
        size_t filename_str_len        = 0;
        
        ERRNO_RESULT_T failure      = 0;

        if (file_data == NULL)
        {
            PATHOLOGICAL("File data pointer was NULL.\n");
            return false;
        }

        section_type_str = (char*) xlate_section_type(file_data->type);

        if (section_type_str == NULL)
        {
            PATHOLOGICAL("Filename unavailable for section type.\n");
            return false;
        }

        section_type_str_len = section_type_str != NULL ? ::strnlen(section_type_str, SECTION_TYPE_STR_LEN) : 0;

        if (section_type_str_len <= 0)
        {
            PATHOLOGICAL("Section type name is less than 1 character in length.\n");
            return false;
        }

        if (constants == NULL)
        {
            PATHOLOGICAL("Constants pointer was NULL.\n");
            return false;
        }

        // If they didn't give a basename, use the default. 
        if (basename_str == NULL) 
        { 
            basename_str = &_basename[0]; 
        }

        basename_str_len = ::strnlen(basename_str, MAX_ZVISION_PATHNAME_LEN);

        if (basename_str_len <= 0)
        {
            PATHOLOGICAL("Basename length is less than 1.\n");
            return false;
        }

        if (zvision_dir_str == NULL) { zvision_dir_str = ZVISION_DIR_STR; }

        // Clear the file name field.
        ::memset((void*) &file_data->name[0], 0, MAX_ZVISION_PATHNAME_LEN);

        // Now build  the fully qualified "file name" which, for example, might consist of "/tmp/zvision/" followed
        // by the basename (i.e. the prefix, the date and time stamps, the proc id) then a slash, and then the section type.
        zvision_dir_str_len = ::strnlen(zvision_dir_str, MAX_ZVISION_PATHNAME_LEN);

        if (zvision_dir_str_len <= 0 || zvision_dir_str_len >= MAX_ZVISION_PATHNAME_LEN) 
        {
            PATHOLOGICAL("ZVision path name length (%ld) \"%*s\"[truncated] exceeds available name size (%ld).\n", 
                zvision_dir_str_len, (int) MAX_ZVISION_PATHNAME_LEN, zvision_dir_str, MAX_ZVISION_PATHNAME_LEN); 
            return false;
        }

        // Start by copying the ZVision directory path into the filename.
        ::memcpy((void*) &file_data->name[0], (const void*) zvision_dir_str, zvision_dir_str_len);
        filename_str_len += zvision_dir_str_len;

        // Check that there is room for adding the basename.
        if (basename_str_len >= (MAX_ZVISION_PATHNAME_LEN - filename_str_len))
        {
            PATHOLOGICAL("basename length (%ld) \"%*s\"[truncated] exceeds remaining available name size (%ld).\n", 
                basename_str_len, (int) MAX_ZVISION_PATHNAME_LEN, basename_str, MAX_ZVISION_PATHNAME_LEN - filename_str_len); 
            return false;
        }

        // Next copy the basename (prefix + date/time/procid) into the filename.
        ::memcpy((void*) &file_data->name[filename_str_len], (const void*) basename_str, basename_str_len);
        filename_str_len += basename_str_len;

        // Check that there is room for adding the separator.
        if (filename_str_len >= MAX_ZVISION_PATHNAME_LEN)
        {
            PATHOLOGICAL("Adding 1 character separator exceeds remaining available space.\n"); 
            return false;
        }

        // Add a separator character.
        file_data->name[filename_str_len] = '/';
        filename_str_len += 1;

        // If this is the type_stable file, then also create the directory for it.
        if (file_data->type == type_stable)
        {
            errno = 0;

            int rc = ::mkdir(file_data->name, constants->open_mode);

            if (rc != 0)
            {
                ERRNO_RESULT_T failure = errno;
    
                if (failure != EEXIST)
                {
                    PATHOLOGICAL("mkdir(\"%s\") failed with errno set to %d (%s)\n.", 
                      file_data->name, failure, ::strerror(failure)); 
                    return false;
                }
            }
        }

        // Make sure the section type will fit onto the end of the basename.
        if (section_type_str_len >= (MAX_ZVISION_PATHNAME_LEN - filename_str_len))
        {
            PATHOLOGICAL("section type length (%ld) \"%*s\"[truncated] exceeds remaining available basename size (%ld).\n", 
                section_type_str_len, (int) MAX_ZVISION_PATHNAME_LEN, section_type_str, \
                MAX_ZVISION_PATHNAME_LEN - filename_str_len); 
            return false;
        }

        // Copy the section type onto the end of the basename.
        ::memcpy((void*) &file_data->name[filename_str_len], (const void*) section_type_str, section_type_str_len);
        filename_str_len += section_type_str_len;

        // One last redundant check.
        if (filename_str_len >= MAX_ZVISION_PATHNAME_LEN)
        {
            PATHOLOGICAL("Filename length (%ld) exceeds maximum name size (%ld).\n",
                filename_str_len, MAX_ZVISION_PATHNAME_LEN); 
            return false;
        }

        file_data->name_len = filename_str_len;

        DIAGOUT(2, "Creating file \"%s\".\n", file_data->name);
        DIAGOUT(3, "    flags=0x%X \"%s\", mode=0x%X \"%s\"...\n", 
                constants->open_flags, XlateFlag::open(constants->open_flags, result1, sizeof(result1)/sizeof(char)), 
                constants->open_mode,  XlateMode::open(constants->open_mode,  result2, sizeof(result2)/sizeof(char)));

        errno = 0;
        // Now open the new shared memory file for this type, and create an entry of the appropriate size.
        SOCKET_FD_T fd = ::open(file_data->name, constants->open_flags, constants->open_mode);

        if (fd < 0)
        {
            failure = errno;

            PATHOLOGICAL("::open(name=\"%s\") failed with errno set to %d (%s).\n", 
              file_data->name, failure, ::strerror(failure)); 
            return false;
        }

        // Seek to the end of the file, this positions the pointer.
        if (lseek_file(fd, file_data->max_size, constants->lseek_flags) != true)
        {
            failure = errno;
            if (::close(fd) != 0)
            {
                DIAGOUT(1, "::close(%d) failed.\n", fd);
            }
            PATHOLOGICAL("lseek_file(fd=%d, %ld, 0x%X=\"%s\") of file \"%s\" failed with errno set to %d (%s).\n", 
                fd, file_data->max_size, constants->lseek_flags, 
                XlateFlag::lseek(constants->lseek_flags, result1, sizeof(result1)/sizeof(char)), 
                file_data->name, failure, ::strerror(failure));
            return false;
        }

        // Write nothing.  This fixes the size of the file.
        if (::write(fd, "", 1) != 1)
        {
            failure = errno;
            if (::close(fd) != 0)
            {
                DIAGOUT(1, "::close(%d) failed.\n", fd);
            }
            PATHOLOGICAL("::write(fd=%d) failed with errno set to %d (%s).\n", 
              fd, failure, ::strerror(failure)); 
            return false;
        }

        // The file descriptor is stored back into the FILE_DATA_T structure.
        file_data->jvm_mmap_fd = fd;
        file_data->ha_mmap_len = file_data->max_size;

        return true;

    } // end of create_file().
    
  

//-------------------------------------------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------------------------------------
//
//             Unix Domain Socket Functions.
//
//-------------------------------------------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------------------------------------


    /**
     ** lost_connection_to_host_agent() - Called by the JVM Service Thread whenever it appears that the TCP connection
     ** has dropped.
     **/
    void lost_connection_to_host_agent()
    {
         DIAGOUT(1, "Host Agent connection lost.\n");

         set_ipc_state(ipc_state_no_connection);

         if (_ts != NULL) 
         { 
             _ts->jvm_ipc_fd = -1; 
             _ts->host_agent_pid = 0;
         }

         // RKANE: This doesn't work, figure out a better way to zero this!
         // _handle = 0;

    }
 
    /**
     ** establish_host_agent_connection() - connect to the well-known port that should have already been established
     ** by the HA, and send a hello message.  Initiates the JVM's state change to ipc_state_got_connect.
     **
     **
     **
     **/
    inline bool establish_host_agent_connection(int* countdown, int timeout)
    {
        char*               ipc_name         = (char*) ZVISION_JVM_HELLO_PORT;
        size_t              ipc_name_len     = (ipc_name != NULL) ? ::strnlen(ipc_name, MAX_IPC_SOCKET_NAME_SIZE) : 0;

        struct sockaddr_un  socket_info;
        struct sockaddr_un* socket_info_addr = &socket_info;
        socklen_t           socket_info_len  = sizeof(struct sockaddr_un);

        size_t              connect_retries  = 20;
        size_t              connect_attempts = 0;
        int                 max_secs         = *countdown;
        int                 result;

        bool                connected        = false;
        ERRNO_RESULT_T      failure          = 0;

        SOCKET_FD_T         jvm_server_fd    = -1;
        SOCKET_FD_T         jvm_ipc_fd       = -1;

        if (countdown == NULL)
        { 
            PATHOLOGICAL("NULL countdown timer.\n");
            return false; 
        }

        if (ipc_name_len <= 0)
        {
            PATHOLOGICAL("Host Agent's Unix Domain Socket Filename length (%lu) too small.\n", ipc_name_len);
            return false;
        }
        
        if (ipc_name_len >= sizeof(socket_info.sun_path))
        {
            PATHOLOGICAL("Host Agent's Unix Domain Socket Filename length (%lu) exceeds \"sun_path\" size (%lu).\n", 
                ipc_name_len, sizeof(socket_info.sun_path));
            return false;
        }

        SOCKET_FD_T fd = ::socket(AF_UNIX, SOCK_SEQPACKET, 0);

        if (fd < 0)
        {
            failure = errno;
            PATHOLOGICAL("::socket() method failed for Host Agent's well-known port (fd %d), error is %d \"%s\".\n", 
                fd, failure, ::strerror(failure));
            return false;
        }

        DIAGOUT(3, "Called ::socket(AF_UNIX, SOCK_SEQPACKET, 0), got fd %d.\n", fd);

        //----------------------------------------------------------------------------//
        // Once we've got the socket fd, we don't want to return without closing it,  //
        // so from this point on, only leave this routine via the "exit" label.       //
        //----------------------------------------------------------------------------//

        DIAGOUT(2, "Waiting no more than %d seconds on fd %d for Host Agent connection.\n", *countdown, fd);

        // Mark the file as nonblocking, as we can't afford a JVM pause.
        if (mark_file_nonblocking(fd) != true)
        {
            PATHOLOGICAL("mark_file_blocking(%d) failed.\n", fd);
            return false;
        }

        // TODO: Make retries a parameter.
        while (connect_attempts++ < connect_retries)
        {
            ::memset(&socket_info, 0, sizeof(struct sockaddr_un));
            socket_info.sun_family = AF_UNIX;
            ::memcpy((void*) &socket_info.sun_path[0], (const void*) ipc_name, ipc_name_len);

            DIAGOUT(2, "Calling ::connect(fd %d, sock_info %p, size %lu), IPC name is \"%s\".\n", 
                fd, socket_info_addr, sizeof(struct sockaddr_un), socket_info.sun_path);
    
            result = ::connect(fd, (struct sockaddr*) socket_info_addr, sizeof(struct sockaddr_un));
            failure = errno;

            if (result == 0) { break; }

            if (*countdown == 0)
            {
                // If it failed because the port was unavailable/not created, then we just wait
                // for the Host Agent to create one.  Otherwise, it is a pathological error.
                if (failure != ENOENT && failure != ECONNREFUSED)
                {
                  PATHOLOGICAL("Got unexpected ::connect() error \"%s\".\n", ::strerror(failure));
                  return false;
                }
                else
                {
                  DIAGOUT(1, "Host Agent hasn't yet established its well-known port (\"%s\"), retrying...\n", 
                    ZVISION_JVM_HELLO_PORT);
                }
                goto exit; 
            }

            if (*countdown > 0) { (*countdown)--; }

            ::sleep(1);
        } 

        // If we got here with a non-zero result it is because we exceeded the connect_retries limit.
        if (result != 0)
        {
            PATHOLOGICAL("::connect(%d, %p, %lu) failed to connect to Host Agent after %lu attempts, error is %d \"%s\".\n", 
                fd, socket_info_addr, sizeof(struct sockaddr_un), connect_attempts, failure, ::strerror(failure));
            goto exit; 
        }

        // Send a hello message to the Host Agent, which should result in the Host Agent sending
        // a start_token request, which we should recieve within some reasonable period of time.
        if (send_host_agent_hello(fd, countdown) != true)
        {
            DIAGOUT(1, "Failed sending Host Agent (file \"%s\") a hello message.\n", ZVISION_JVM_HELLO_PORT);
            goto exit; 
        }

        DIAGOUT(2, "JVM has connected to Host Agent's well-known port (\"%s\"), and sent a hello msg on fd %d.\n", ipc_name, fd);
        DIAGOUT(2, "Waiting no more than %d seconds for a Host Agent start request.\n", *countdown);

        if (type_stable_valid(_ts) != true)
        {
            PATHOLOGICAL("Invalid type_stable pointer establishing connection.\n");
            goto exit; 
        }

        jvm_server_fd = _ts->jvm_server_fd;
        jvm_ipc_fd    = _ts->jvm_ipc_fd;

        // At this point we have established the JVM IPC fd, and connected to the Host Agent's well-known port,
        // and sent a "hello" message to said port, and are now waiting for a response (in the form of a
        // start_token request) back from the Host Agent.
        while (true)
        {
            DIAGOUT(2,  "calling ::accept(fd %d,  socket_info addr %p, len addr %p) called.\n", 
                jvm_server_fd, &socket_info, &socket_info_len); 

            // The ::accept() call has no internal timeout mechanism, so if we wish to pend for only a limited
            // amount of time, we first set the file to NON-BLOCKING, and then check the return code from the
            // accept to see if it came back with EAGAIN.  This will block if timeout is less than zero... 
            jvm_ipc_fd = ::accept(jvm_server_fd, (sockaddr*) &socket_info, &socket_info_len); 
    
            if (jvm_ipc_fd >= 0)
            {
                DIAGOUT(2, "::accept(fd = %d, socket_info_addr %p, len addr %p) returned fd %d, peer length contains %u.\n",
                    jvm_server_fd, socket_info_addr, &socket_info_len, jvm_ipc_fd, socket_info_len);

                if (_ts != NULL) { _ts->jvm_ipc_fd = jvm_ipc_fd; }

                set_ipc_state(ipc_state_got_connect);

                DIAGOUT(2, "JVM got connect from Host Agent on fd %d.\n", jvm_ipc_fd);
                DIAGOUT(2, "State is now \"%s\" (0x%lX).\n", xlate_ipc_state(ipc_state(_ts)), ipc_state(_ts));

                // This is the normal (non-pathological) exit from the while loop.
                break;

            } // end of if (fd >= 0)

            failure = errno;

            // We should only get here if the socket is marked NON-blocking, and we are willing to wait.
            if (failure == EAGAIN || failure == EWOULDBLOCK)
            {
                if (*countdown < 0 )
                { 
                    DIAGOUTNOPREFIX(5, ".");
                    ::sleep(1);
                    continue;
                } 

                if (*countdown == 0)
                { 
                    DIAGOUT(1, "The accept() system call was interrupted before a valid connection arrived.\n");
                    DIAGOUT(2, "Timed out waiting for Host Agent connection, error was %s.\n", ::strerror(failure));
                    goto exit; 
                }

                DIAGOUTNOPREFIX(5, ".");
                ::sleep(1);
                (*countdown)--;
                continue;                  
            }

            if (failure == EINTR)
            {
                if (*countdown < 0 )
                { 
                    DIAGOUTNOPREFIX(5, ".");
                    ::sleep(1);
                    continue;
                } 

                if (*countdown == 0)
                { 
                    DIAGOUT(1, "The accept() system call was interrupted before a valid connection arrived.\n");
                    DIAGOUT(2, "Timed out waiting for Host Agent connection, error was %s.\n", ::strerror(failure));
                    goto exit; 
                }

                DIAGOUTNOPREFIX(5, ".");
                ::sleep(1);
                (*countdown)--;
                continue;                  
            }

            // Anything else is just wrong.
            switch (failure)
            {
                break;
            case EBADF:
                DIAGOUT(1, "File Descriptor (%d) specified is invalid.\n", fd);
                break;
            case ECONNABORTED:
                DIAGOUT(1, "The connection has been aborted.\n");
                break;
            case EFAULT:
                DIAGOUT(1, "Buffer address parameter (%p) specified is outside accessible address space.\n", 
                    socket_info_addr);
                break;
            case EINVAL:
                DIAGOUT(1, "Socket not listening, or the addrlen is invalid.\n");
                break;
            case EMFILE:
                DIAGOUT(1, "Per process limit of open file descriptors has been reached.\n");
                break;
            case ENFILE:
                DIAGOUT(1, "System limit of open file descriptors has been reached.\n");
                break;
            case ENOMEM:
                DIAGOUT(1, "Not enough free memory.\n");
                break;
            case ENOTSOCK:
                DIAGOUT(1, "File descriptor (%d) is a file, not a socket.\n", fd);
                break;
            case EOPNOTSUPP:
                DIAGOUT(1, "Referenced socket (%d) type is not supported.\n", fd);
                break;
            case EPROTO:
                DIAGOUT(1, "Protocol error.\n");
                break;
            case EPERM:
                DIAGOUT(1, "Firewall rules forbid connection.\n");
                break;
            default:
                DIAGOUT(1, "Unexpected error (0x%X) returned.\n", failure);
                break;
            }

            PATHOLOGICAL("::accept(fd %d, struct addr %p, len %d [addr %p]) failed, errno was 0x%X \"%s\".\n", 
                fd, socket_info_addr, socket_info_len, &socket_info_len, failure, ::strerror(failure));

            goto exit; 

        } // end of  while (true)

        // Ask the Host Agent to process the first request, which must be a start_token.  Anything else would be an error.
        // We automatically return a start request with a start response, which should cause the Host Agent to become ready.
        if (process_host_agent_start_message( (timeout > 0) ? timeout : 1) != true)
        {
            DIAGOUT(1, "Did not receive Host Agent start message within %d milliseconds.\n", timeout);
            goto exit; 
        }

        // Wait for the Host Agent to become ready (by setting its pid).
        if (wait_for_host_agent_ready(*countdown) != true)
        {
            DIAGOUT(1, "Host Agent did not come ready. Timed out after %d seconds.\n", max_secs);
            goto exit; 
        }       
        
        DIAGOUT(2, "Connected to Host Agent after %d seconds.\n", 
            (*countdown > 0 && max_secs > 0) ? (max_secs - *countdown) : max_secs);
        connected = true;

exit:

        DIAGOUT(3, "Closing connection to the Host Agent's well-known port, fd %d.\n", fd);
        close(fd);

        DIAGOUT(3, "establish_host_agent_connection() returns %s.\n", connected == true ? "true" : "false");
        return true; 

    } // end of establish_host_agent_connection().



    /**
     ** Send a hello msg to the Host Agent's well-know port.  This IPC socket should have been established before the 
     ** JVM initialized, and it should already be marked non-blocking.
     **/
    bool send_host_agent_hello(SOCKET_FD_T fd, int* timeout)
    {
        // Build a hello message, this contains the basename, key and process id of the JVM.
        JVM_HELLO_MSG_T         hello_msg;

        ::memset((void*) &hello_msg, 0, sizeof(JVM_HELLO_MSG_T));

        hello_msg.jvm_key = _jvm_key;
        hello_msg.jvm_pid = ::getpid();
        hello_msg.cookie  = COOKIE_HELLO_MSG;

        // Check that this object hasn't become corrupted.
        if (type_stable_valid(_ts) != true)
        {
            PATHOLOGICAL("Invalid type_stable pointer when building JVM hello message.\n");
            return false; 
        }

        SECTION_ENTRY_T* entry = section_entry_by_type(_ts, type_stable);

        if (entry == NULL || entry->cookie != COOKIE_SECTION_ENTRY)
        {
            PATHOLOGICAL("Invalid type_stable section entry when building JVM hello message.\n");
            return false; 
        }

        const char*      ts_name     = &entry->file.name[0];
        size_t           ts_name_len = ::strnlen(ts_name, MAX_IPC_SOCKET_NAME_SIZE);

        if (ts_name_len < 1)
        {
            PATHOLOGICAL("The type_stable file \"%s\" name length (%lu) was too small.\n", 
                &entry->file.name[0], ts_name_len);
            return false; 
        }

        if (ts_name_len >= sizeof(hello_msg.name))
        {
            PATHOLOGICAL("The type_stable file \"%s\" name length (%lu) execeed hello_msg.name (%lu).\n", 
                &entry->file.name[0], ts_name_len, sizeof(hello_msg.name));
            return false; 
        }

        ::memcpy((void*) &hello_msg.name[0], (const void*) ts_name, ts_name_len);

        ERRNO_RESULT_T failure = 0;

        DIAGOUT(2, "Sending a JVM_HELLO_MSG_T on fd %d.\n", fd);
          
        // Now that the parameters have been loaded into the shared memory, send the hello msg to the JVM.
        ssize_t bytes_written = write_file(fd, (const void*) &hello_msg, sizeof(hello_msg), &failure);

        if (bytes_written != sizeof(hello_msg)) 
        { 
            PATHOLOGICAL("Send on fd %d of JVM_HELLO_MSG_T to Host Agent failed, sent only %lu bytes, size was %lu.\n", 
                fd, bytes_written, sizeof(hello_msg));
            return false; 
        }

        return true;

    } // end of send_host_agent_hello().



    /**
     ** Create the base filename of the Unix Domain socket we'll use to send & receive between the JVM  
     ** and the Host Agent.  We fill in the type_stable header area with the manufactured unix domain
     ** socket filename.  
     **/
    bool ipc_name_create(const char* ipc_name_suffix_str, const char* basename_str)
    {
        if (_ts == NULL)
        {
              PATHOLOGICAL("_ts is NULL.\n");
              return false;
        }

        if (_ts->cookie != COOKIE_TYPE_STABLE)
        {
              PATHOLOGICAL("_type_stable cookie corrupted.\n");
              return false;
        }

        // If the caller didn't supply one, use the default suffix.
        if (ipc_name_suffix_str == NULL)
        {
            ipc_name_suffix_str = IPC_NAME_SUFFIX_STR;
        }

        // If the caller didn't supply one, use the default basename (prefix/date/time/procid).
        if (basename_str == NULL)
        {
            basename_str = _basename;
        }

        size_t ipc_domain_str_len      = 0;
        size_t ipc_prefix_dir_str_len  = ::strnlen(IPC_PREFIX_DIR_STR,  MAX_IPC_SOCKET_NAME_SIZE);
        size_t basename_str_len        = ::strnlen(basename_str,        MAX_IPC_SOCKET_NAME_SIZE);
        size_t ipc_name_suffix_str_len = ::strnlen(ipc_name_suffix_str, MAX_IPC_SOCKET_NAME_SIZE);

        ::memset(_ts->ipc_domain_id, 0, MAX_IPC_SOCKET_NAME_SIZE);  

        if (ipc_prefix_dir_str_len >= MAX_IPC_SOCKET_NAME_SIZE)
        {
            PATHOLOGICAL("IPC_PREFIX_DIR_STR length (%lu) \"%*s\"[truncated] is greater than name space (%lu).\n", 
                ipc_prefix_dir_str_len, (int) MAX_IPC_SOCKET_NAME_SIZE, IPC_PREFIX_DIR_STR, MAX_IPC_SOCKET_NAME_SIZE);
            return false; 
        }

        ::memcpy((void*) &_ts->ipc_domain_id[0], (const void*) IPC_PREFIX_DIR_STR, ipc_prefix_dir_str_len);
        ipc_domain_str_len += ipc_prefix_dir_str_len;

        if (basename_str_len >= (MAX_IPC_SOCKET_NAME_SIZE - ipc_domain_str_len))
        {
            PATHOLOGICAL("basename length (%lu) \"%*s\"[truncated] is greater than remaining name space (%lu).\n", 
                ipc_name_suffix_str_len, (int) MAX_IPC_SOCKET_NAME_SIZE, ipc_name_suffix_str, 
                (MAX_IPC_SOCKET_NAME_SIZE - ipc_domain_str_len));
            return false; 
        }

        ::memcpy((void*) &_ts->ipc_domain_id[ipc_domain_str_len], (const void*) basename_str, basename_str_len);
        ipc_domain_str_len +=  basename_str_len;

        if (ipc_name_suffix_str_len >= (MAX_IPC_SOCKET_NAME_SIZE - ipc_domain_str_len))
        {
            PATHOLOGICAL("ipc_name_suffix_str length (%lu) \"%*s\"[truncated] is greater than remaining name space (%lu).\n", 
                ipc_name_suffix_str_len, (int) MAX_IPC_SOCKET_NAME_SIZE, ipc_name_suffix_str, MAX_IPC_SOCKET_NAME_SIZE);
            return false; 
        }

        ::memcpy((void*) &_ts->ipc_domain_id[ipc_domain_str_len], (const void*) ipc_name_suffix_str, ipc_name_suffix_str_len);
        ipc_domain_str_len +=  ipc_name_suffix_str_len;

        if (ipc_domain_str_len != ::strnlen(_ts->ipc_domain_id, MAX_IPC_SOCKET_NAME_SIZE))
        {
            PATHOLOGICAL("Unix Domaian Socket Filename length (%lu) doesn't match calculated length (%lu).\n", 
                ::strnlen(_ts->ipc_domain_id, MAX_IPC_SOCKET_NAME_SIZE), ipc_domain_str_len);
            return false; 
        }

        if (ipc_domain_str_len >= MAX_IPC_SOCKET_NAME_SIZE)
        {
            PATHOLOGICAL("Unix Domaian Socket Filename length (%lu) exceeds limit (%lu).\n", 
                ipc_domain_str_len, MAX_IPC_SOCKET_NAME_SIZE);
            return false; 
        }

        DIAGOUT(3, "Set IPC file name to \"%s\"\n", _ts->ipc_domain_id);

        return true;

    } // end of ipc_name_create().


    /**
     ** Creates the JVM-side IPC "server" receive socket.
     **/
    bool create_jvm_ipc_port()
    {

        return true;

    } // end of create_jvm_ipc_port().



//-------------------------------------------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------------------------------------
//
//
//-------------------------------------------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------------------------------------


    /**
     ** send_async_notify_fast() - Build an asynchronous notify message to send back to the Host Agent,
     ** but do so with a minimum of overhead.  This call should be used by any JVM thread that is critical
     ** path and may not block, or in an ISR.
     **/
    inline void send_async_notify_fast(SECTION_TYPE_E type, size_t index, SOCKET_FD_T fd, HA_HANDLE_T handle) 
    {
        JVM_NOTIFY_MSG_T    notify_msg;

        ::memset((void*) &notify_msg, 0, sizeof(JVM_NOTIFY_MSG_T));

        notify_msg.handle  = handle;
        notify_msg.type    = type;
        notify_msg.index   = index;
        notify_msg.cookie  = COOKIE_NOTIFY_MSG;

        write(fd, (const void*) &notify_msg, sizeof(JVM_NOTIFY_MSG_T));

    } // end of send_async_notify_fast()
 


    /**
     ** send_async_notify() - Build an asynchronous notify message to send back to the Host Agent, if an
     ** output_fd is supplied, then also write it to the local file system.
     **
     ** NOTE: There is no callback here that handles getting data into GPB form, as notification messages
     ** are generally expected to be high performance and can not afford the overhead of a second buffer
     ** copy or transform.  If you can afford the time to compress it, you should use the alternative
     ** mechanism to transfer.
     **/
    inline bool send_async_notify(SECTION_TYPE_E type, size_t index, SOCKET_FD_T fd, HA_HANDLE_T handle, SOCKET_FD_T output_fd = -1)
    {
        if (handle != _handle)
        {
            PATHOLOGICAL("handle (0x%lX) doesn't match _handle(0x%lX).\n", handle, _handle);
            return false; 
        }

        if (fd < 0)
        {
            PATHOLOGICAL("Unix Domain Socket (%d) invalid.\n", fd);
            return false;
        }

        JVM_NOTIFY_MSG_T    notify_msg;

        // Now build the notify message and send it to the Host Agent.
        ::memset((void*) &notify_msg, 0, sizeof(JVM_NOTIFY_MSG_T));

        notify_msg.handle  = handle;
        notify_msg.type    = type;
        notify_msg.index   = index;
        notify_msg.cookie  = COOKIE_NOTIFY_MSG;

        ERRNO_RESULT_T failure = 0;

        // Now that the parameters have been loaded into the shared memory, send only the notify message to the JVM.
        // It will read that, then process the shared memory file's response buffer for the actual data.
        ssize_t bytes_written = write_file(fd, (const void*) &notify_msg, sizeof(JVM_NOTIFY_MSG_T), &failure);

        if (bytes_written != sizeof(notify_msg)) 
        { 
            PATHOLOGICAL("send_async_notify(\"%s\", index %lu, fd %d, handle 0x%lX) wrote %lu of %lu bytes, failure was \"%s\".\n", 
                xlate_section_type(type), index, fd, handle, bytes_written, sizeof(notify_msg), ::strerror(failure)); 
            return false; 
        }

        DIAGOUT(3, "JVM has sent an asynchronous \"%s\" notify msg (index %lu) to the Host Agent (handle 0x%lX on fd %d).\n", 
            xlate_section_type(type), index, handle, fd);

        if (output_fd >= 0)
        {
            REQUEST_BUFFER_T* req  = (REQUEST_BUFFER_T*) _ts->sections[type_index(type)].file.jvm_mmap_addr;

            if (req == NULL)
            {
                PATHOLOGICAL("send_async_notify(request %p, fd %d) failed, jvm_mmap_addr was NULL.\n", req, output_fd);
                return false; 
            }

            ssize_t len = _ts->sections[type_index(type)].file.unit_size;

            if (len < 1)
            {
                PATHOLOGICAL("send_async_notify(request %p, fd %d) failed, unit_size was less than 1.\n", req, output_fd);
                return false; 
            }

            const char* addr = (const char*) req + (len * index);

            bytes_written = write_file(output_fd, (const void*) addr, len, &failure);

            if (bytes_written != len) 
            { 
                PATHOLOGICAL("JVM could not record index %lu (addr %p) of \"%s\" to output fd %d, wrote %lu bytes, len was %lu.\n", 
                    index, addr, xlate_section_type(type), fd, bytes_written, len); 
                return false; 
            }

            DIAGOUT(4, "JVM recorded index %lu (addr %p) of \"%s\" to output fd %d, wrote_%lu byets, len %lu.\n",
                    index, addr, xlate_section_type(type), fd, bytes_written, len); 
        }

        return true;

    } // end of send_async_notify()


    /**
     ** send_response - Send a response (to the Host Agent) that wasn't requested.  The presumption is that the request
     ** buffer has been updated with the response already.  The most likely use of this is when we are sending a shutdown
     ** notification asynchronously back to the Host Agent, to inform it that the JVM is voluntarily shutting down. If an
     ** output_fd is supplied, write the data to the local file system as well.
     **/
    bool inline send_response(HA_HANDLE_T handle, SECTION_TYPE_E type, SOCKET_FD_T output_fd = -1)
    {
        // TODO: add logic to fill in the associated request buffer section with the proper sequencing info.
        
        // We're building a temporary request_msg, filed in with the type & handle.
        HA_REQUEST_MSG_T request_msg;
       
        ::memset((void*) &request_msg, 0, sizeof(HA_REQUEST_MSG_T)); 
      
        request_msg.handle = handle;
        request_msg.type   = type;
        request_msg.cookie = COOKIE_REQUEST_MSG;

        return send_response(&request_msg, output_fd);
    }


    /**
     ** send_response - Send a response back from a Host Agent request.  The presumption is that the request
     ** buffer has been updated with the response already. If an output_fd is supplied, write the data to the
     ** local file system as well.
     **/
    bool inline send_response(HA_REQUEST_MSG_T* request, SOCKET_FD_T output_fd = -1)
    {
        SOCKET_FD_T         fd         = _ts->jvm_ipc_fd;
        ERRNO_RESULT_T      failure    = 0;

        IPC_STATE_E state = ipc_state(_ts);

        if (state != ipc_state_connection_est && state != ipc_state_got_start)
        {
            DIAGOUT(1, "IPC state \"%s\", notify msg for Host Agent (handle 0x%lX on fd %d) ignorned.\n", 
                xlate_ipc_state(state), _handle, fd);
            return true;
        }

        REQUEST_BUFFER_T* req = (REQUEST_BUFFER_T*) _ts->sections[type_index(request->type)].file.jvm_mmap_addr;

        if (req == NULL)
        {
            PATHOLOGICAL("send_response(request %p, fd %d) failed, jvm_mmap_addr was NULL.\n", request, output_fd);
            return false; 
        }

        req->hdr.cntl.response_token = req->hdr.cntl.request_token;

        // Build a message response to send back to the Host Agent.
        JVM_NOTIFY_MSG_T    notify_msg;

        ::memset((void*) &notify_msg, 0, sizeof(JVM_NOTIFY_MSG_T));

        notify_msg.handle  = _handle;
        notify_msg.type    = request->type;
        notify_msg.cookie  = COOKIE_NOTIFY_MSG;

        DIAGOUT(1, "Sending notify msg for type \"%s\" on fd %d to the Host Agent.\n", xlate_section_type(notify_msg.type), fd);
        DIAGBUFFEROUT(2, (volatile char*) &notify_msg, sizeof(JVM_NOTIFY_MSG_T));

//TODO: RKANE issue callback here that handles getting data into GPB form.

        // Now that the parameters have been loaded into the shared memory, send only the notify message to the JVM.
        // It will read that, then process the shared memory file's response buffer for the actual data.
        ssize_t bytes_written = write_file(fd, (const void*) &notify_msg, sizeof(notify_msg), &failure);

        if (bytes_written != sizeof(notify_msg)) 
        { 
            if (failure != EPIPE)
            {
              PATHOLOGICAL("Send to Host Agent on fd %d failed, wrote %lu bytes, size was %lu.\n", 
                fd, bytes_written, sizeof(notify_msg)); 
            }
            return false; 
        }

        DIAGOUT(1, "JVM has sent a \"%s\" notify msg (response token %lu) to the Host Agent (handle 0x%lX on fd %d).\n", 
            xlate_section_type(request->type), req->hdr.cntl.response_token, _handle, fd);

        if (output_fd >= 0 && req->hdr.cntl.response_size > 0)
        {
            ssize_t len = req->hdr.cntl.response_size;

            bytes_written = write_file(output_fd, (const void*) req->response_data_start, len, &failure);

            if (bytes_written != len) 
            { 
                PATHOLOGICAL("Send to Host Agent on fd %d failed, wrote %lu bytes, size was %lu.\n", fd, bytes_written, len); 
                return false; 
            }

            DIAGOUT(4, "JVM saved a \"%s\" response, token %lu, len %lu, to the output file fd %d.\n",
                xlate_section_type(request->type), req->hdr.cntl.response_token, req->hdr.cntl.response_size, fd);
        }

        return true;

    } // end of send_response()


    /**
     ** 
     **/
    bool process_host_agent_start_message(int timeout_ms)
    {
        SOCKET_FD_T            fd;
        HA_REQUEST_MSG_T       ha_request;
        struct pollfd          poll_info;
        ssize_t                bytes_read;
        int                    result;

        size_t                 eintr_cnt   = 0;
        ERRNO_RESULT_T         failure     = 0;

        if (type_stable_valid(_ts) != true) 
        { 
            PATHOLOGICAL("type_stable_valid() returns false.\n");
            goto exit;
        }

        if (_ts->jvm_ipc_fd <= 0)
        {
            PATHOLOGICAL("_ts->jvm_ipc_fd uninitialized.\n");
            goto exit;
        }

        fd = _ts->jvm_ipc_fd;

        DIAGOUT(3, "Polling %u milliseconds for Host Agent Start Request on fd %d...\n", timeout_ms, fd);

        poll_info.fd      = fd;
        poll_info.events  = (POLLIN | POLLPRI);
        poll_info.revents = 0;

        while (true)
        {
            result = ::poll(&poll_info, 1, timeout_ms);

            if (result >= 0                  ) { break; }
            if (result <  0 && errno != EINTR) { break; }
            if (10 < eintr_cnt++)              { break; }
        }

        if (result < 0)
        {
            failure = errno;
            PATHOLOGICAL("::poll(%p, 1, %u) on fd %d returned %d, errno is %s (%d).\n", 
                    &poll_info, timeout_ms, fd, result, ::strerror(errno), errno);
            goto exit;
        }

        if (result > 1)
        {
            // TODO: add OOB processing here.
            PATHOLOGICAL("Got %d file descriptors back from ::poll(%p, 1, %d) on fd %d, expecting only 1.\n", 
                result, &poll_info, timeout_ms, fd);
            goto exit;
        }

        if (result == 0)
        {
            DIAGOUT(3, "poll_info fd = %d, events = 0x%X, revents = 0x%X.\n", 
                poll_info.fd, poll_info.events, poll_info.revents);

            // If we timed out without processing any requests, that is still okay.
            if (timeout_ms >= 0) 
            { 
                DIAGOUT(3, "::poll(%p, 1, %d) on fd %d returned zero, timed out after %d milliseconds.\n",
                    &poll_info, timeout_ms, fd, timeout_ms);
                return true;
            }

            PATHOLOGICAL("Got zero back from ::poll(%p, 1, %d) call on fd %d when timeout should have been indefinite.\n", 
                    &poll_info, timeout_ms, fd);
            goto exit;
        }

        DIAGOUT(3, "::poll(%p, 1, %d) on fd %d returned %d file descriptor.\n",
            &poll_info, timeout_ms, fd, result);

        // Read the Host Agent request message.
        bytes_read = read_file(fd, &ha_request, sizeof(HA_REQUEST_MSG_T), &failure);

        if (failure == ECONNRESET || failure == EINTR || failure == EAGAIN)
        {
            DIAGOUT(2, "Got \"%s\" (%d), shutting down connection to Host Agent.\n", ::strerror(failure), failure);
            lost_connection_to_host_agent();
            return false;
        }

        if (bytes_read < 0)
        {
            PATHOLOGICAL("read_file(fd %d) failed with errno \"%s\" (%d).\n",
                fd, ::strerror(failure), failure);
            goto exit;
        }
        
        if (bytes_read != sizeof(ha_request))
        {
            DIAGOUT(1, "Lost connection?  Partial read of Host Agent request.  Expected %lu bytes, got %lu.\n",
                    sizeof(ha_request), bytes_read);
            lost_connection_to_host_agent();
            return false;
        }

        if (ha_request.type != start_token)
        {
            PATHOLOGICAL("invalid type \"%s\" (%d), should be \"%s\" (%d).\n",
                xlate_section_type(ha_request.type), ha_request.type, xlate_section_type(start_token), start_token);
            goto exit;
        }

        // Make sure that the message isn't corrupted.
        if (ha_request.cookie != COOKIE_REQUEST_MSG)
        {
            PATHOLOGICAL("Got Host Agent \"%s\", expected cookie 0x%lX, got 0x%lX.\n", 
                xlate_section_type(ha_request.type), COOKIE_REQUEST_MSG, ha_request.cookie);
            goto exit;
        }

        // If we already have a handle, then this isn't the first start message, so something is wrong.
        if (_handle != 0)
        {
            PATHOLOGICAL("Got Host Agent request for type \"%s\", handle was 0x%lX, but already initialized (0x%lX).\n",
                xlate_section_type(ha_request.type), ha_request.handle, _handle);
            goto exit;
        }

        // Record the handle passed to us by the Host Agent.
        _handle = ha_request.handle;

        DIAGOUT(1, "Received \"%s\" request from Host Agent (handle 0x%lX on fd %d), sending response.\n", 
                xlate_section_type(ha_request.type), ha_request.handle, fd);

        set_ipc_state(ipc_state_got_start);

        if (send_response(&ha_request) != true)
        {
            PATHOLOGICAL("send_response(%p) failed. type: \"%s\", handle: 0x%lX.\n",
                &ha_request, xlate_section_type(ha_request.type), ha_request.handle);
            goto exit;
        }

        return true;
exit:
        return false;

    } // end of process_host_agent_start_message().



    /**
     ** Set up to recieve requests from the Host Agent, and create response buffers with the
     ** appropriate data inside them.  We signal the Host Agent that the data is available 
     ** by sending a JVM_IPC_MSG_T back on the socket.
     **
     ** The limit parameter dictates how many message will be processed before returning.
     ** Returns the number of messages actually processed.
     **
     ** If the caller used a value less than zero for the timeout_ms parameter, then
     ** pend indefinitely during the ::poll() call.  If they used a value of zero,
     ** then don't pend at all.  Otherwise pend for no more than <timeout_ms> seconds
     ** during the ::poll() call.
     **
     ** If the caller is concerned about blocking, they should set the timeout to a very small value,
     ** and the limit to 1.
     **/
    size_t process_host_agent_requests(int timeout_ms, int limit, bool* got_shutdown)
    {
        SOCKET_FD_T            fd;
        HA_REQUEST_MSG_T       ha_request;
        struct pollfd          poll_info;
        IPC_STATE_E            state;
        ssize_t                bytes_read;
        int                    result;
        int                    msgs_processed = 0;
        ERRNO_RESULT_T         failure        = 0;

        if (*got_shutdown != false)
        {
            PATHOLOGICAL("got_shutdown was not false.\n");
            goto exit;
        }

        if (type_stable_valid(_ts) != true) 
        { 
            PATHOLOGICAL("type_stable_valid() returns false.\n");
            goto exit;
        }

        if (_ts->jvm_ipc_fd <= 0)
        {
            PATHOLOGICAL("_ts->jvm_ipc_fd uninitialized.\n");
            goto exit;
        }

        fd = _ts->jvm_ipc_fd;

        DIAGOUT(3, "Polling %u milliseconds for at most %u Host Agent Requests on fd %d...\n", timeout_ms, limit, fd);

        // We stay in this polling loop unti we either timeout on the ::poll(), or reach the limit of processed messages.
        while (true)
        {
            // Check that this message didn't exceed the limit, providing
            // we are worried about the total number of messages processed.
            if (limit >= 0 && msgs_processed >= limit)
            {
                DIAGOUT(1, "process_host_agent_requests(timeout_ms=%ums, limit=%u) returns %u messages processed.\n", 
                    timeout_ms, limit, msgs_processed);
                goto exit;
            }

            nfds_t                 nfds          = 1;
       
            poll_info.fd      = fd;
            poll_info.events  = (POLLIN | POLLPRI);
            poll_info.revents = 0;

            while (true)
            {
                result = ::poll(&poll_info, nfds, timeout_ms);

                if (result >= 0                  ) { break; }
                if (result <  0 && errno != EINTR) { break; }
            }

            DIAGOUT(4, "poll_info fd = %d, events = 0x%X, revents = 0x%X.\n", poll_info.fd, poll_info.events, poll_info.revents);

            if (result < 0)
            {
                failure = errno;
                PATHOLOGICAL("::poll(%p, %lu, %u) on fd %d returned %d, errno is %s (%d).\n", 
                        &poll_info, nfds, timeout_ms, fd, result, ::strerror(errno), errno);
                goto exit;
            }

            if (result > 1)
            {
                // TODO: add OOB processing here.
                PATHOLOGICAL("Got %d file descriptors back from ::poll(%p, %lu, %d) call on fd %d, expecting only %lu.\n", 
                    result, &poll_info, nfds, timeout_ms, fd, nfds);
                goto exit;
            }

            if (result == 0)
            {
                // If we timed out without processing any requests, that is still okay.
                if (timeout_ms >= 0) 
                { 
                    DIAGOUT(3, "::poll(%p, %lu, %d) on fd %d returned zero, timed out after %d milliseconds.\n",
                        &poll_info, nfds, timeout_ms, fd, timeout_ms);
                    goto exit;
                }

                PATHOLOGICAL("Got zero back from ::poll(%p, %lu, %d) call on fd %d when timeout should have been indefinite.\n", 
                        &poll_info, nfds, timeout_ms, fd);
                goto exit;
            }

            DIAGOUT(3, "::poll(%p, %lu, %d) on fd %d returned %d file descriptor.\n",
                &poll_info, nfds, timeout_ms, fd, result);

            failure = 0;

            // Read the Host Agent request message.
            bytes_read = read_file(fd, &ha_request, sizeof(HA_REQUEST_MSG_T), &failure);


            if (bytes_read < 0)
            {
                if (failure == ECONNRESET || failure == EINTR || failure == EAGAIN)
                {
                    DIAGOUT(2, "Got \"%s\" (%d), shutting down connection to Host Agent.\n", ::strerror(failure), failure);
                    lost_connection_to_host_agent();
                }
                else
                {
                    PATHOLOGICAL("read_file(fd %d) failed with errno \"%s\" (%d).\n", fd, ::strerror(failure), failure);
                }
                goto exit;
            }
            
            if (bytes_read != sizeof(ha_request))
            {
                DIAGOUT(1, "Lost connection?  Partial read of Host Agent request.  Expected %lu bytes, got %lu.\n",
                    sizeof(ha_request), bytes_read);
                lost_connection_to_host_agent();
                goto exit;
            }

            if (ha_request.type < start_token || ha_request.type >= end_token)
            {
                PATHOLOGICAL("invalid type (%d), should be a value greater than or equal to %d and less than %d.\n",
                    ha_request.type, start_token, end_token);
                goto exit;
            }

            // Make sure that the message isn't corrupted.
            if (ha_request.cookie != COOKIE_REQUEST_MSG)
            {
                PATHOLOGICAL("Got Host Agent \"%s\" request message, expected cookie 0x%lX, got 0x%lX.\n", 
                    xlate_section_type(ha_request.type), COOKIE_REQUEST_MSG, ha_request.cookie);
                goto exit;
            }

            if (forgo_data(_ts, ha_request.type) == true)
            {
                DIAGOUT(1, "Forgoing request type \"%s\" (%d).\n",
                    xlate_section_type(ha_request.type), ha_request.type);
                continue;
            }

            // We will now count this as a "processed message".
            msgs_processed++;

            // Grab the state first.
            state = ipc_state(_ts);

            // Process the Host Agent requests.
            // TODO:  These tests should be placed in order of frequency. 
            if (ha_request.type == config_details   ||
                ha_request.type == environment_vars ||
                ha_request.type == hotspot_flags    ||
                ha_request.type == general_stats    ||
                ha_request.type == top_busy_methods)
            {
                // Make sure that the Host Agent object handle matches the handle
                // we got from the most recent start_token request.
                if (ha_request.handle != _handle)
                {
                    PATHOLOGICAL(
                        "Got Host Agent \"%s\" request on fd %d, but object handle (0x%lX) didn't match JVM value (0x%lX).\n",
                        xlate_section_type(ha_request.type), fd, ha_request.handle, _handle);
                    goto exit;
                }

                if (state != ipc_state_connection_est)
                {
                    PATHOLOGICAL("Got \"%s\" request (object 0x%lX) on fd %d but state invalid.\n",
                        xlate_section_type(ha_request.type), ha_request.handle, fd);
                    PATHOLOGICAL("State was \"%s\" (0x%lX) should be \"%s\" (0x%lX).\n",
                        xlate_ipc_state(state), state, xlate_ipc_state(ipc_state_connection_est), ipc_state_connection_est);
                    goto exit;
                }

                DIAGOUT(1, "Received \"%s\" request from Host Agent (handle 0x%lX on fd %d), sending response.\n", 
                   xlate_section_type(ha_request.type), ha_request.handle, fd);

                if (send_response(&ha_request) != true)
                {
                    PATHOLOGICAL("send_response(%p) failed. type: \"%s\", handle: 0x%lX.\n",
                        &ha_request, xlate_section_type(ha_request.type), ha_request.handle);
                    goto exit;
                }
            }
            else if (ha_request.type == heartbeat)
            {
                // Make sure that the Host Agent object handle matches the handle
                // we got from the prior start_token request.
                if (ha_request.handle != _handle)
                {
                    PATHOLOGICAL(
                        "Got Host Agent \"%s\" request on fd %d, but object handle (0x%lX) didn't match JVM value (0x%lX).\n",
                        xlate_section_type(ha_request.type), fd, ha_request.handle, _handle);
                    goto exit;
                }

                DIAGOUT(1, "Received \"%s\" request from Host Agent (handle 0x%lX on fd %d), sending response.\n", 
                        xlate_section_type(ha_request.type), ha_request.handle, fd);
                
                FILE_DATA_T* file_data = file_data_by_type(_ts, heartbeat);

                if (file_data == NULL)
                {
                    PATHOLOGICAL("file_data_by_type(section = \"%s\" (%d)) function returned NULL.\n", 
                        xlate_section_type(heartbeat), heartbeat);
                    goto exit;
                }
                
                REQUEST_BUFFER_T* req = (REQUEST_BUFFER_T*) file_data->jvm_mmap_addr;

                if (req == NULL)
                {
                    PATHOLOGICAL("heartbeat jvm_mmap_addr is NULL.\n");
                    goto exit;
                }

                if (ARTA::heartbeat((void*) &req->params,              MAX_REQUEST_PARAMS_SIZE, 
                                    (void*) &req->response_data_start, MAX_RESPONSE_SIZE) != true)
                {
                    PATHOLOGICAL("ARTA::heartbeat() failed.\n");
                    goto exit;
                }
                
                if (send_response(&ha_request) != true)
                {
                    // This may have failed because the Host Agent exited and the fd or 
                    // pipe was closed, which isn't a pathological condition, but means
                    // we need to close the connection.
                    DIAGOUT(1, "send_response(%p) failed. type: \"%s\", handle: 0x%lX.\n",
                        &ha_request, xlate_section_type(ha_request.type), ha_request.handle);
                    goto exit;
                }
            }
            else if (ha_request.type == shutdown)
            {
                if (ha_request.handle != _handle)
                {
                    PATHOLOGICAL(
                        "Got Host Agent request for type \"%s\", but object handle (0x%lX) didn't match JVM value (0x%lX).\n",
                        xlate_section_type(ha_request.type), ha_request.handle, _handle);
                    goto exit;
                }

                DIAGOUT(1, "Received \"%s\" request from Host Agent handle 0x%lX on fd %d, sending response and exiting.\n", 
                    xlate_section_type(ha_request.type), ha_request.handle, fd);

                *got_shutdown = true;
                send_response(&ha_request);
                set_ipc_state(ipc_state_exiting);

                goto exit;
            }
            else
            {
                PATHOLOGICAL("illegal request type (%s) received, request ignored.\n", xlate_section_type(ha_request.type));
                goto exit;
            }
            
        }  // end of while (true)

        PATHOLOGICAL("Unexpectedly exited while loop.\n");

exit:
        DIAGOUT(3, "process_host_agent_requests(timeout_ms=%ums, limit=%u) returns %u messages processed.\n", 
            timeout_ms, limit, msgs_processed);

        return msgs_processed;

    } // end of process_host_agent_requests().


    /**
     ** Build the unit test Output Results file pathname in the "output_filename_str" buffer supplied
     ** by the caller.
     **/
    bool build_output_pathname(char*       pathname_buf, 
                               const char* zvision_dir_str,
                               const char* basename_str,
                               const char* output_filename_str) 
    { 
        // If we didn't get one, then presume we can use basename (prefix + date + time + procid) we've 
        // already created during initialization.  If that string is NULL, then it is an error.
        if (basename_str == NULL)
        {
            basename_str = &_basename[0];
        }

        size_t basename_str_len        = ::strnlen(basename_str,        MAX_ZVISION_PATHNAME_LEN);
        
        if (basename_str_len <= 0 || basename_str_len >= MAX_ZVISION_PATHNAME_LEN)
        {
            PATHOLOGICAL("basename_str_len (%lu) exceeds limit (%lu).\n",
                basename_str_len, MAX_ZVISION_PATHNAME_LEN);
            return false; 
        }

        DIAGOUT(3, "Using \"%s\" as basename in output pathname string.\n", basename_str);

        return build_pathname(pathname_buf, zvision_dir_str, basename_str, output_filename_str);

    }  // end of build_output_pathname().


    /**
     ** Build the basename prefix.
     **/
    bool create_basename_prefix(const char* basename_prefix_str)
    {
        size_t basename_prefix_str_len;
        
        // If the user supplied a prefix string, make that the default value, otherwise
        // use the default string (e.g. "azul_").  Note: if the user also specified their
        // own basename string, then it will override the prefix setting.
        if (basename_prefix_str != NULL)
        {
            basename_prefix_str_len = ::strnlen(basename_prefix_str, MAX_ZVISION_PATHNAME_LEN);
            ::memcpy((void*) _basename_prefix, (const void*) basename_prefix_str, basename_prefix_str_len);
        }
        else
        {
            basename_prefix_str_len = ::strnlen(BASENAME_PREFIX_STR, MAX_ZVISION_PATHNAME_LEN);
            ::memcpy((void*) _basename_prefix, (const void*) BASENAME_PREFIX_STR, basename_prefix_str_len);
        }

        // If the prefix won't fit in the buffer, something is wrong.
        if (basename_prefix_str_len >= MAX_ZVISION_PATHNAME_LEN)
        {
            PATHOLOGICAL("basename prefix len (%ld \"%s\") exceeds limit (%ld).\n", 
                basename_prefix_str_len, basename_prefix_str, MAX_ZVISION_PATHNAME_LEN); 
            return false;
        }
    
        // Ensure that there are no slashes within the prefix.
        if (::strchr(&_basename_prefix[0], '/') != NULL)
        {
            PATHOLOGICAL("Basename prefix string contains an illegal slash.\n");
            return false;
        }

        return true;

    }  // end of create_basename_prefix().


    /**
     ** ZVision::create_date_time_procid() - 
     **
     ** Build the (initialization) _date_time_procid string.  This will be used if the user didn't
     ** supply a specific name.  It is composed of the date & time stamps, and the process ID, which
     ** provides a unique identifier when creating the shared memory files.  We do this once and
     ** reuse it so that it doesn't vary with time during initialization.  Note: We do this regardless
     ** of whether the user specified their own basename, as this string contains a start date & time,
     ** as well as the process ID, which turn out to be useful in debugging.
     **
     ** RETURNS: true if successful, false on pathological error. 
     ** 
     ** NOTES: 
     **   - May be executed only once.
     **   - Called by JVM only.
     **
     **/
    bool create_date_time_procid()
    {
        PID_T       proc_id          = ::getpid();
        time_t      time_data        = ::time(NULL);
        struct tm*  time_struct_data = ::localtime(&time_data);
        size_t      buf_len          = MAX_ZVISION_PATHNAME_LEN;

        if (time_struct_data == NULL)
        {
            PATHOLOGICAL("localtime() returned NULL.\n");
            return false;
        }

        ::memset((void*) &_date_time_procid[0], 0, buf_len);

        // Fill in the date & time stamp in the _date_time_procid buffer.
        size_t ftime_len = ::strftime(&_date_time_procid[0], buf_len, ZVISION_DATESTAMP_FORMAT, time_struct_data);

        if (ftime_len != ZVISION_DATESTAMP_LEN)
        {
            PATHOLOGICAL("::strftime() returned %lu characters but expected %lu instead.\n", ftime_len, ZVISION_DATESTAMP_LEN); 
            return false;
        }

        // Now build the process id into the _date_time_procid, and a null terminator.
        ::sprintf(&_date_time_procid[ZVISION_DATESTAMP_LEN], "%d%c", proc_id, 0);

        // Ensure that there are no slashes within the _date_time_procid string.
        if (::strchr(&_date_time_procid[0], '/') != NULL)
        {
            PATHOLOGICAL("_date_time_procid string contains an illegal slash.\n");
            return false;
        }

        return true;

    }  // end of create_date_time_procid().

    
    /**
     ** ZVision::create_basename() - Builds the fully qualified filename name of the type_stable file.
     ** This file will be created by the JVM, and the contents memory mapped by both the JVM & Host Agent.
     **
     ** RETURNS: true if successful, false on pathological error. 
     ** 
     ** NOTES: 
     **   - May be executed only once.
     **   - Called by JVM only.
     **   - Presumes that the prefix and date/time/procid strings are valid.
     **
     **/
    bool create_basename(const char* basename_str)
    {
        size_t basename_str_len  = 0;
        size_t prefix_str_len    = 0;
        size_t component_str_len = 0;
        
        memset((void*) &_basename[0], 0, MAX_ZVISION_PATHNAME_LEN);

        // If the user provided one, set the default basename to it.
        if (basename_str != NULL) 
        { 
            basename_str_len = ::strnlen(basename_str, MAX_ZVISION_PATHNAME_LEN);

            if (basename_str_len <= 0 || basename_str_len >= MAX_ZVISION_PATHNAME_LEN)
            {
                PATHOLOGICAL("basename length (%ld) \"%*s\"[truncated] exceeds available basename size (%ld).\n", 
                    basename_str_len, (int) MAX_ZVISION_PATHNAME_LEN, basename_str, MAX_ZVISION_PATHNAME_LEN); 
                return false;
            }

            ::memcpy((void*) &_basename[0], (const void*) basename_str, basename_str_len);

            // Ensure that there are no slashes within the basename.
            if (::strchr(&_basename[0], '/') != NULL)
            {
                PATHOLOGICAL("Basename string contains an illegal slash.\n");
                return false;
            }

            return true;
        }

        // We assume that the _basename_prefix value is either the default, or if the user has
        // already overridden the value. We use whatever value it is currently.
        prefix_str_len = ::strnlen(_basename_prefix, MAX_ZVISION_PATHNAME_LEN);

        if (prefix_str_len <= 0 || prefix_str_len >= MAX_ZVISION_PATHNAME_LEN)
        {
            PATHOLOGICAL("basename prefix length (%ld) \"%*s\"[truncated] exceeds basename size (%ld).\n", 
                prefix_str_len, (int) MAX_ZVISION_PATHNAME_LEN, _basename_prefix, MAX_ZVISION_PATHNAME_LEN); 
            return false;
        }

        // Copy the prefix into the basename.
        ::memcpy((void*) &_basename[0], (const void*) _basename_prefix, prefix_str_len);
        basename_str_len += prefix_str_len;

        // We assume that the _date_time_procid component has already been set.
        component_str_len = ::strnlen(_date_time_procid, MAX_ZVISION_PATHNAME_LEN);

        if (component_str_len >= (MAX_ZVISION_PATHNAME_LEN - basename_str_len))
        {
            PATHOLOGICAL("date_time_procid prefix length (%ld) \"%*s\"[truncated] exceeds remaining available space (%ld).\n", 
                component_str_len, (int) MAX_ZVISION_PATHNAME_LEN, _date_time_procid, (MAX_ZVISION_PATHNAME_LEN - basename_str_len)); 
            return false;
        }

        // Copy the date/time stamp & proc ID into the basename.
        ::memcpy((void*) &_basename[basename_str_len], (const void*) &_date_time_procid[0], component_str_len);
        basename_str_len += component_str_len;

        // One final check that the basename string isn't too large.
        if (basename_str_len <= 0 || basename_str_len >= MAX_ZVISION_PATHNAME_LEN)
        {
            PATHOLOGICAL("basename length (%ld) \"%*s\"[truncated] exceeds available space (%lu).\n", 
                basename_str_len, (int) MAX_ZVISION_PATHNAME_LEN, basename_str, MAX_ZVISION_PATHNAME_LEN); 
            return false;
        }

        DIAGOUT(3, "Created basename string \"%s\" (from _prefix & _date_time_procid).\n", _basename);

        return true;

    } // end of create_basename().
    

    /** 
     **
     **/
    bool wait_for_host_agent_clear(int wait_secs)
    {
        int          timeout  = wait_secs;

        // wait for the Host Agent to set the PID field, which is the final step.
        while (host_agent_is_ready(_ts) == true)
        {
            if (timeout == 0)
            {
                DIAGOUT(1, "Host Agent state failed to shutdown connection in %d seconds.\n", wait_secs);
                return false;
            }
    
            DIAGOUTNOPREFIX(4, ".");
            ::sleep(1);
            if (timeout > 0) { timeout--; }
        }

        return true;
    }

    /** 
     **
     **/
    bool wait_for_host_agent_ready(int wait_secs)
    {
        int          timeout  = wait_secs;
        IPC_STATE_E  state    = ipc_state(_ts);

        while (state == ipc_state_got_connect)
        {
            if (timeout == 0)
            {
                DIAGOUT(1, "Host Agent state failed to come ready in %d seconds.\n", wait_secs);
                return false;
            }
    
            DIAGOUTNOPREFIX(4, ".");
            ::sleep(1);
        }

        if (state != ipc_state_got_start)
        {
            PATHOLOGICAL("Invalid JVM state \"%s\" (0x%lX) encountered waiting for Host Agent to come ready.\n",
                xlate_ipc_state(state), state); 
            PATHOLOGICAL("State should be \"%s\" (0x%lX).\n",xlate_ipc_state(ipc_state_got_start), ipc_state_got_start);
            return false;
        }

        DIAGOUT(2, "JVM State is now \"%s\" (0x%lX), waiting for Host Agent to come ready.\n",
            xlate_ipc_state(state), state);

        // wait for the Host Agent to set the PID field, which is the final step.
        while (host_agent_is_ready(_ts) != true)
        {
            if (timeout == 0)
            {
                DIAGOUT(1, "Host Agent state failed to come ready in %d seconds.\n", wait_secs);
                return false;
            }
    
            DIAGOUTNOPREFIX(4, ".");
            ::sleep(1);
            if (timeout > 0) { timeout--; }
        }
    
        set_ipc_state(ipc_state_connection_est);
    
        DIAGOUT(1, "Host Agent is now ready, Unix process id is %d, JVM state \"%s\" (0x%lX).\n", 
            host_agent_pid(_ts), ZVision::xlate_ipc_state(state), state); 
    
        return true;
    
    } // end of wait_for_host_agent_ready()


};  // end of class ZVisionJVM

#endif // ZVISION_JVM_HPP

// end of file zvision_jvm.hpp.

