package os::SafePushDir;

use Cwd;
use strict;

sub new {
    my $class = shift;
    my $dirName = shift;
    die if !defined $dirName;

    my $origDirName = cwd();
    my $this = {
                "origDirName" => $origDirName,
               };
    bless $this, $class;

    os::safeCd ($dirName);

    return $this;
}

sub DESTROY {
    my $this = shift;

    my $origDirName = $this->{"origDirName"};
    os::safeCd ($origDirName);
}


package os::TmpDir;

use strict;

my $counter = 0;

sub new {
    my $class = shift;
    die if !defined $class;

    $counter++;
    my $dirName = "/tmp/$$.${counter}";
    my $this = {
                "dirName" => $dirName,
               };
    bless $this, $class;

    os::safeMkdir ($dirName);

    return $this;
}

sub getDirName {
    my $this = shift;
    die if !defined $this;
    die if defined shift;

    return $this->{"dirName"};
}

sub DESTROY {
    my $this = shift;

    if (os::getKeepTmp()) {
        return;
    }

    my $dirName = $this->getDirName();
    my $origChildError = $?;
    os::safeRmDashRF ($dirName);
    $? = $origChildError;
}


package os;

use Cwd;
use Carp;
use File::Basename;
use File::Copy;
use File::stat;
use File::Spec;
use FileHandle;

no strict;
use Sys::Syslog qw(:DEFAULT);
use Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(true false success failure);
use strict;

use constant true => 1;
use constant false => 0;

use constant success => 1;
use constant failure => 0;

# Package globals.
my $printOnly = false;
my $verbose = false;
my $stackTraceOnFailure = false;
my $keepTmp = false;
my $logFh = *STDERR;

# Package initialization.
autoflush STDOUT 1;
autoflush STDERR 1;
$main::SIG{"QUIT"} = "IGNORE";          # Don't coredump on Ctrl-\


sub printIfVerbose {
    if ($verbose) {
        print @_;
    }
}



#
# Error printing helper functions.
#

my $logOpened = false;

sub openLogIfNecessary {
    if (!$logOpened) {
        my $programName = basename $main::0;
        openlog ($programName, 'pid', 'syslog');
        $logOpened = true;
    }
}

sub mysyslog {
    openLogIfNecessary();
    my $level = shift;
    my $message = shift;
    syslog ($level, '%s', $message);
}

sub info {
    my $message = shift;
    confess if !defined $message;

    print $logFh "Info:  ${message}.\n";
    mysyslog ("info", "LOG_INFO: " . $message);
}

sub warning {
    my $message = shift;
    confess if !defined $message;

    print $logFh "Warning:  ${message}.\n";
    mysyslog ("warning", "LOG_WARN: " . $message);
}

sub error {
    my $message = shift;
    confess if !defined $message;

    print $logFh "Error:  ${message}.\n";
    mysyslog ("err", "LOG_ERR: " . $message);
}

sub fatal {
    # Purpose:  Print an error message, and exit.

    my $message = shift;
    $message = "" if !defined $message;
    print $logFh "Fatal error:  ${message}.\n";
    mysyslog ("err", "LOG_ERR: " . $message);

    if ($stackTraceOnFailure) {
        print $logFh "\n";
        confess;
    }
    else {
        exit 1;
    }
}

sub fatalOptionIsRequired {
    my $name = shift;
    die if !defined $name;
    die if defined shift;

    os::fatal "Option ${name} is required";
    exit 1;
}


#
# All "unsafe" commands require the caller to check $? status to know
# if the command succeeded.
#

sub unsafeSystem {
    my @toexec = @_;
    printIfVerbose "+ CMD: @toexec\n";

    return if $printOnly;

    system (@toexec);
}


#
# All "safe" commands check to make sure the return code succeeded.
# If it did not, fatal is called.
#

sub safeCd {
    my $dir = shift;
    confess if !defined $dir;
    confess if defined shift;

    printIfVerbose "+ CD:  $dir\n";

    return if ($printOnly);

    chdir ($dir) || fatal "os::safeCd:  can't chdir to '$dir'";
}

sub safeSystem {
    my @toexec = @_;
    printIfVerbose "+ CMD: @toexec\n";

    return if ($printOnly);

    system (@toexec);
    my $exit_value = $? >> 8;
    my $signal_num = $? & 127;
    my $dumped_core = $? & 128;
    if ($exit_value || $signal_num || $dumped_core) {
        fatal "os::safeSystem:  command '@toexec' failed";
    }
}

sub safeExec {
    my @toexec = @_;
    printIfVerbose "+ CMD: @toexec\n";

    return if ($printOnly);

    exec (@toexec);
}

sub safeChmod {
    my $permissions = shift;
    my $fileOrDir = shift;
    confess if !defined $fileOrDir;
    confess if defined shift;

    printIfVerbose ("+ CMD: chmod ($permissions, $fileOrDir)\n");
    if (!$printOnly) {
        my $octPerm = oct ($permissions);
        if (!$octPerm) {
            fatal "Illegal permissions";
        }
        my @cannot = grep {not chmod $octPerm, $_} $fileOrDir;
        if (@cannot) {
            fatal "os::safeChmod:  chmod ($permissions, $fileOrDir) failed";
        }
    }
}

sub safeChown {
    my $owner = shift;
    my $fileOrDir = shift;
    confess if !defined $fileOrDir;
    confess if defined shift;

    os::safeSystem ("/bin/chown", $owner, $fileOrDir);
}

sub safeChgrp {
    my $owner = shift;
    my $fileOrDir = shift;
    confess if !defined $fileOrDir;
    confess if defined shift;

    os::safeSystem ("/bin/chgrp", $owner, $fileOrDir);
}

sub safeMv {
    my $from = shift;
    my $to = shift;
    confess if !defined $to;
    confess if defined shift;

    safeSystem ("/bin/mv", $from, $to);
}

sub safeCpDashP {
    my $from = shift;
    my $to = shift;
    confess if !defined $to;
    confess if defined shift;

    safeSystem ("/bin/cp", "-p", $from, $to);
}

sub safeMkdir {
    my $dirName = shift;
    confess if !defined $dirName;
    confess if defined shift;

    if (! mkdir ($dirName)) {
        my $s = "mkdir $dirName failed ($!)";
        os::fatal ($s);
    }
}

sub safeMkdirDashP {
    my $dirName = shift;
    confess if !defined $dirName;
    confess if defined shift;

    safeSystem ("/bin/mkdir", "-p", $dirName);
}

sub safeRmDashRF {
    my $tree = shift;
    confess if !defined $tree;
    confess if defined shift;

    safeSystem ("/bin/rm", "-rf", $tree);
}

sub safeRmDashF {
    my $tree = shift;
    confess if !defined $tree;
    confess if defined shift;

    safeSystem ("/bin/rm", "-f", $tree);
}

sub safeSymlink {
    my $lhs = shift;
    my $rhs = shift;
    confess if !defined $rhs;
    confess if defined shift;

    safeSystem ("/bin/ln", "-s", $lhs, $rhs);
}


#
# Support functions.
#

sub printFileToStdout {
    my $fileName = shift;
    confess if !defined $fileName;
    confess if defined shift;

    my $fh = new FileHandle ($fileName);
    if (!defined $fh) {
        os::error "File ${fileName} not found";
        return;
    }

    while (<$fh>) {
        chomp;
        print $_, "\n";
    }
}

sub parseCommonOptions {
    my $opt = shift;
    confess if !defined $opt;
    confess if defined shift;

    if (defined $opt->{"verbose"}) {
        os::setVerbose (true);
    }
    else {
        os::setVerbose (false);
    }

    if (defined $opt->{"n"}) {
        os::setPrintOnly (true);
    }
    else {
        os::setPrintOnly (false);
    }

    if (defined $opt->{"keeptmp"}) {
        os::setKeepTmp (true);
    }
    else {
        os::setKeepTmp (false);
    }
}


#
# Getters and setters.
#

sub setPrintOnly {
    $printOnly = shift;
    confess if !defined $printOnly;
    confess if defined shift;
}

sub getPrintOnly {
    confess if defined shift;
    return $printOnly;
}

sub setVerbose {
    $verbose = shift;
    confess if !defined $verbose;
    confess if defined shift;
}

sub getVerbose {
    confess if defined shift;
    return $verbose;
}

sub setStackTraceOnFailure {
    $stackTraceOnFailure = shift;
    confess if !defined $stackTraceOnFailure;
    confess if defined shift;
}

sub setKeepTmp {
    $keepTmp = shift;
    confess if !defined $keepTmp;
    confess if defined shift;
}

sub getKeepTmp {
    confess if defined shift;
    return $keepTmp;
}

sub setLogFh {
    $logFh = shift;
    confess if !defined $logFh;
}

sub setLogFhStdout {
    setLogFh (*STDOUT);
}

sub setLogFhStderr {
    setLogFh (*STDERR);
}

1;
