package zing_bundler_lib;
# Copyright 2013 Azul Systems, Inc.  All Rights Reserved.
# This perl module is used by zing-core-pattern-hook and zing-core-bundler
use File::Temp qw/tempfile tmpnam/;
use File::Basename;
use File::Spec;
use File::Compare;
use Sys::Syslog;

my $global_invoked_byhook = 0;
my $append_str = undef;
# The below strings are used to match output from 'file' command
my $str_sharedobject = "shared object";
my $str_executable = "LSB\[/ /\]+executable";
my $str_characterspecial = "character special";
my $str_directory = "directory";
# global variables
my $OPENSSL_PATH = "/usr/bin/openssl";
my $GPG_PATH = "/usr/bin/gpg";
my $GPGAGENT_PATH = "/usr/bin/gpg-agent";
# debug mode
my $debug_mode = $ENV{"DEBUG_ZING_BUNDLER"};
if (!defined($debug_mode)) {
  $debug_mode = 0;
}

sub msg_printer {
  my $tmp = $_[0];
  if ($debug_mode) {
    $tmp = "[$_[0]]\n";
  }
  if ($global_invoked_byhook == 0) {
    print $tmp;
  }
  else {
    syslog("info","$tmp");
  }
}

sub debug_printer {
  if ($debug_mode) {
    &msg_printer("$_[0]][" . $_[1]);
  }
}

sub print_die {
  &msg_printer($_[0]);
  &msg_printer("Bundler failed. Exiting");
  exit 1;
}

sub extract_profilelog {
  my @brk = split($_[0], $_[1]);
  if (defined($brk[1])) {
    my @path = split(/\s+/, $brk[1]);
    if (defined($path[0])) {
      if (-f $path[0]) {
        return $path[0];
      }
    }
  }
  return " ";
}


# find if file ($_[0]) is of type ($_[1]) based on ELF's magic number
# TODO: re-write this function
sub is_filetype_byarg {
  if (length($_[0]) > 0 && -e $_[0]) {
    my $file_cmd = 'file -L ' . "$_[0]";
    my $cmd_output = `$file_cmd`;
    if ($?) {
      syslog("info", "ERROR: $?");
    }
    else {
      # if file command's output contains string given by $_[1] (eg. "shared object") return TRUE
      if ($cmd_output =~ /\s$_[1]/) {
        # syslog("info","is_filetype_byarg: $_[0]");
        return 0;
      }
    }
  }
  return 1;
}

# must have executables for encrypt()
sub check_enc_executables {
  if (!-X $OPENSSL_PATH) { &msg_printer("ERROR: Executable not found $OPENSSL_PATH\n"); return 1;}
  if (!-X $GPG_PATH) { &msg_printer("ERROR: Executable not found $GPG_PATH\n"); return 1;}
  return 0;
}

=head1 METHOD
 'encrypt' is a function which uses Azul's public key
 found in /etc/zing/ to encrypt the core bundle:
 $_[0]:      path to the core bundle
Output:
 $_[0].gpg:  encrypted file
=cut
sub encrypt {
  my $homedir = tmpnam();
  my $gpgagent_pid = undef;
  my $fingerprint = "AA22AC698ADA2A6A7ADD0F2D394026C2F659E4E5";
  my $password_fname = $_[0] . "_password";
  if (&check_enc_executables()) {
    goto enc_abort;
  }
  debug_printer(__LINE__,"ENC: EXECUTING CMD = $OPENSSL_PATH rand -base64 30 > $password_fname");
  system("$OPENSSL_PATH rand -base64 30 > $password_fname");
  if ($?) {
    &msg_printer("ERROR: Failed to create passphrase\n");
    unlink($password_fname);
    goto enc_abort;
  }
  debug_printer(__LINE__,"ENC: EXECUTING CMD = /bin/rm -rf $homedir");
  debug_printer(__LINE__,"ENC: EXECUTING CMD = mkdir -p $homedir");
  system("/bin/rm -rf $homedir");
  system("mkdir -p $homedir");
  if ($?) {
    &msg_printer("ERROR: Failed to create tmp directory\n");
    goto enc_abort;
  }
  debug_printer(__LINE__,"ENC: EXECUTING CMD = chmod 0777 $homedir");
  system("chmod 0777 $homedir");
  if ($?) {
    &msg_printer("ERROR: Failed to set permissions to $homedir");
    goto enc_abort;
  }
  debug_printer(__LINE__,"ENC: EXECUTING CMD = /bin/cp /etc/zing/pubring.gpg $homedir");
  system("/bin/cp /etc/zing/pubring.gpg $homedir");
  if ($?) {
    &msg_printer("ERROR: Failed to copy pubring.gpg\n");
    goto enc_abort;
  }
  debug_printer(__LINE__,"ENC: INVOKED BY CORE PATTERN HOOK = $global_invoked_byhook");
  if (!$global_invoked_byhook) {
    if (-e $GPGAGENT_PATH) { 
      debug_printer(__LINE__,"ENC: EXECUTING CMD = $GPGAGENT_PATH --homedir=$homedir/ --quiet --use-standard-socket --write-env-file $homedir/gpg-agent.env --daemon");
      system("$GPGAGENT_PATH --homedir=$homedir/ --quiet --use-standard-socket --write-env-file $homedir/gpg-agent.env --daemon");
      if ($?) {
        # continue without gpg-agent
        &msg_printer("INFO: Continue without gpg-agent\n");
        goto skip_to_gpg;
      }
      else {
        my $env_handle;
        if (!open($env_handle, "$homedir/gpg-agent.env")) {
          &msg_printer("INFO: Continue without gpg-agent\n");
          goto skip_to_gpg;
        }
        my $line = <$env_handle>;
        close $env_handle;
        my @gpgagent_env = split(":", $line);
        $gpgagent_pid = $gpgagent_env[1];
        debug_printer(__LINE__,"ENC: GPGAGENT PID = $gpgagent_pid");
      }
    }
  }
skip_to_gpg:
  if (-e "$_[0].gpg") {
    &msg_printer("INFO: Target file $_[0].gpg already exists.\n");
    unlink($password_fname);
    goto enc_abort;
  }
  debug_printer(__LINE__,"ENC: EXECUTING CMD = $GPG_PATH --passphrase-file $password_fname --batch --no-tty --homedir=$homedir --no-permission-warning --no-armor --no-verbose --compress-algo Uncompressed --trust-model always -e -c -r $fingerprint $_[0]");
  my $ret = system("$GPG_PATH --passphrase-file $password_fname --batch --no-tty --homedir=$homedir --no-permission-warning --no-armor --no-verbose --compress-algo Uncompressed --trust-model always -e -c -r $fingerprint $_[0]");
  $ret = $ret >> 8;
  if ($ret) {
    # deciding GPG's execution status based on the output file
    if (-e "$_[0].gpg" && !-z "$_[0].gpg") {
      goto enc_success;
    }
    &msg_printer("ERROR: Encryption failed.\n");
    goto enc_abort;
  }
enc_success:
  # not using --trustdb-name and --no-random-seed-file (slower)
  &msg_printer("INFO: To decrypt use passphrase file $password_fname\n");
  &msg_printer("SUCCESS: Encrypted core bundle at $_[0].gpg\n");
  &msg_printer("*** PLEASE DO NOT UPLOAD THE PASSWORD FILE TO AZUL ***\n");
  &msg_printer("Azul already has a private key which can decrypt the core bundle.\n");
  &msg_printer("The passphrase is only for your use, in case you need to decrypt.\n");
  # removing plain-text core bundle
  unlink($_[0]);
  goto enc_end;
enc_abort:
  &msg_printer("WARNING: Core bundle is not encrypted\n");
  &msg_printer("INFO: Core bundle at $_[0]\n");
enc_end:
  if (defined($gpgagent_pid)) {
    debug_printer(__LINE__,"ENC: EXECUTING CWD = kill -TERM $gpgagent_pid");
    system("kill -TERM $gpgagent_pid");
  }
  debug_printer(__LINE__,"ENC: EXECUTING CWD = /bin/rm -rf $homedir");
  system("/bin/rm -rf $homedir");
}

=head1 METHOD
 'bundle' is a function with varying parameters based on the callee script:
   $o_core            : filepath of the core file
   $collect_data      : flag to collect data files
   $invoked_byhook    : is it invoked by zing-core-pattern-hook script
 Called by zing-core-pattern-hook:
   $libfiles          : array of libfiles to be collected
   $tarlist           : string with list of files to be collected
 Called by zing-core-bundler:
   $collect_executable: flag to collect executable (non-java) binary
   $o_exepath       : filepath of executable (non-java) binary
=cut
sub bundle {
  my $class = shift;

  my $GDB_PATH = 'gdb';
  my $stderr;
  my $stat_fname = tmpnam() . "_sysinfo";
  my $profilelog_fname = " ";
  my $tarlist;
  my $list_add_me = " ";
  my @libfiles;
  my %libshash;
  my $JAVA_PATH = '';
  my $EMPTY_BIN_PATH = undef;
  my $process_pwd = '';
  my $pid_from_hs_err = 0;
  my $pid_from_gdb = 0;
  my $gdb_found_pid = 0;
  my $cpy_list_exe;
  my $bundler_failed = 0;

  my $o_core = $_[0];
  my $o_hsspath = undef;
  my @temp_vals = split('#',$o_core);
  if (scalar(@temp_vals) > 1) {
    $o_core = $temp_vals[0];
    $o_hssfile = $temp_vals[1];
  }
  my $collect_data = $_[1];
  my $invoked_byhook = $_[2];
  my $bundle_on_current_host = $_[5];
  # docker path to be prefixed
  my $dockerpath = $_[6];
  my $outputdir= $_[7];

  $global_invoked_byhook = $invoked_byhook;

  my $hostname = `/bin/hostname`;
  chomp($hostname);
  my $append_str = $hostname;

  my $corefile;
  if (!open($corefile, $o_core)) {
    &print_die("Could not open corefile \'$o_core\'\n");
  }
  close($corefile);

  open my $fh, "-|", "strings", "$o_core";
  my $line;
  while ($line =  <$fh> ) {
    if (defined($dockerpath)) {
      $line = $dockerpath . "/" . $line;
    }
    if ($line =~ m#jre/lib/rt.jar#) {
      my @paths = split (/\:/, $line);
      for my $path (@paths) {
        chomp($path);
        if ($path =~ m#jre/lib/rt.jar#) {
          $path =~ s#lib/rt.jar##;
          if (defined($dockerpath)) {
            $path = $dockerpath . "/" . $path;
          }
          $JAVA_PATH = $path . "bin/java";
          debug_printer(__LINE__,"Extracted from core: JAVA_PATH = $JAVA_PATH");
          $path =~ s#/jre##;
          $EMPTY_BIN_PATH = $path . "bin/";
          debug_printer(__LINE__,"Extracted from core: JAVA_BIN_PATH = $EMPTY_BIN_PATH");
        }
      }
      last;
    }
  }
  close($fh);

  my $pid;
  my $logfile = "";

  if ($o_core =~ m/.*\.(\d+)/) {
    $pid = $1;
  }
  else {
    &msg_printer("WARNING: could not get PID from core file\n");
    $pid = $o_core;
  }
  debug_printer(__LINE__,"Pid from core file: PID = $pid");

  my $abs_path = File::Spec->rel2abs($o_core);
  my ($filename, $directory) = fileparse($abs_path);
  if (!defined($o_hssfile)) {    
    $logfile = $directory . "hs_err_pid$pid.log";
  }
  else {
    $logfile = $o_hssfile;
  }
  my $coredir = $directory;

  my $testfh;
  if (!open($testfh, ">libs.$pid.tgz")) {
    &msg_printer("ERROR: could not open libs.$pid.tgz for writing\n");
    return 1;
  }
  close($fh);
  unlink("libs.$pid.tgz");

  # if called from 'zing-core-pattern-hook', switch to TARring up files.
  if ($invoked_byhook == 1) {
    debug_printer(__LINE__,"INVOKED BY CORE PATTERN HOOK");
    @libfiles = @{$_[3]};
    $tarlist = $_[4];
  }
  else {
    # Starting with GDB
    my $gdb_cmd_fh;
    my $gdb_cmd_file = tmpnam();
    my $stderr_fname = tmpnam();
 
    open($gdb_cmd_fh, ">$gdb_cmd_file") or &msg_printer("ERROR: could not open GDB command file");
    print $gdb_cmd_fh "info auxv\n";
    print $gdb_cmd_fh "printf \"HOSTNAME:%s\\n\", _utsname.nodename\n";
    print $gdb_cmd_fh "info sharedlibrary\n";
    print $gdb_cmd_fh "info proc\n";
    print $gdb_cmd_fh "Quit\n";
    close($gdb_cmd_fh);

    my $cmdline = "$GDB_PATH -c " .  $o_core . "< $gdb_cmd_file 2> $stderr_fname";
    debug_printer(__LINE__,"GDB: STARTING WITH CMD = $cmdline");
    my $output = `$cmdline`;

    if ($?) {
      &msg_printer("ERROR: Failed to invoke GDB.\n");
      return 1;
    }

    my $exe_from_gdb = " ";
    my $exe_from_arg = undef;
    my $core_by_java_exe = 0;
    my $prog = " ";
    # preparing to read GDB's output for hostname
    {
      if (defined($_[4])) {
        my $is_exe = &is_filetype_byarg($_[4], $str_executable);
        if ($is_exe == 1) {
          # not exiting now.
        }
        $exe_from_arg = $_[4];
        $prog = $exe_from_arg;
        debug_printer(__LINE__,"GDB: USER SUPPLIED EXECUTABLE = $prog");
      }
      # keep a copy for AUXV output
      my $output_auxv = $output;
      my @lines_auxv = split("\n", $output_auxv);
      foreach my $line_auxv(@lines_auxv) {
        if ($line_auxv =~ m#AT_EXECFN#) {
          my @parsed_auxv = split(/\s+/, $line_auxv);
          if (defined($parsed_auxv[scalar(@parsed_auxv)-1])) {
            $exe_from_gdb = $parsed_auxv[scalar(@parsed_auxv)-1];
            $exe_from_gdb = substr($exe_from_gdb, 1);
            chop($exe_from_gdb);
            if (defined($dockerpath)) {
              $exe_from_gdb = $dockerpath . "/" . $exe_from_gdb;
            }
            debug_printer(__LINE__,"GDB: EXECUTABLE = $exe_from_gdb");
          }
        }
      }
      my $is_exe = &is_filetype_byarg($exe_from_gdb, $str_executable);
      if ($is_exe == 1) {
        open my $fh, "-|", "strings", "$o_core";
        my $line;
        while ($line =  <$fh> ) {
          if ($line =~ m#PWD=#) {
            if ($line !~ m#OLDPWD=#) {
              $process_pwd = substr ($line, 4);
              chop($process_pwd);
              debug_printer(__LINE__,"GDB: PWD = $process_pwd");
            }
          }
          if ($line =~ m#sun\.java\.launcher\.pid\=#) {
            $core_by_java_exe = 1;
            debug_printer(__LINE__,"GDB: CORE BY JAVA EXE = $core_by_java_exe");
          }
        }
        close($fh);
        if (-d $process_pwd) {
          my $tmp;
          if (defined($exe_from_gdb)) {
            $tmp = " ";
          } else {
            $tmp = substr($exe_from_gdb, 1);
          }
          $tmp = $process_pwd . "/" . $tmp;
          $exe_from_gdb = $tmp;
          debug_printer(__LINE__,"GDB: EXECUTABLE = $exe_from_gdb");
        }
      }
      $is_exe = &is_filetype_byarg($exe_from_gdb, $str_executable);
      if ($is_exe == 0) {
        $prog = $exe_from_gdb;
        debug_printer(__LINE__,"GDB: PROG EXTRACTED = $prog");
      }
    }

    # re-executing GDB with extracted $prog binary.
    $cmdline = "$GDB_PATH -c " .  $o_core . " " . $prog . " < $gdb_cmd_file 2> $stderr_fname";
    debug_printer(__LINE__,"GDB: STARTING AGAIN = $cmdline");
    $output = `$cmdline`;

    if ($?) {
      &msg_printer("ERROR: Failed to invoke GDB.\n");
      return 1;
    }

    unlink ($gdb_cmd_file);

    { 
      local $/;
      $/ = undef;
      my $gdb_log;
      open($stderr, "$stderr_fname") or &print_die("could not open stderr.text file for reading");
      $gdb_log = <$stderr>;

      if ($gdb_log =~ m/is not a core dump\: File format not recognized/ms) {
        &msg_printer("ERROR: Invalid core file format\n");
        return 1;
      }
      close($stderr);
      unlink ($stderr_fname);

      # Extract the inferior's PID from core.
      # It is based on the theory that the
      # core's PID is not being re-used by a different process
      # If so, $pid_from_gdb will remain 0. 
      my @arr_proc = split("\n", $gdb_log);
      foreach my $arr_line(@arr_proc) {
        if ($arr_line =~ m#No /proc directory:#) {
          my @arr = split(/proc\//, $arr_line);
          if (defined($arr[1])) {
            chop($arr[1]);
            $pid_from_gdb = $arr[1];
            $gdb_found_pid = 1;
            debug_printer(__LINE__,"GDB: PROC PID = $pid_from_gdb");
          }
        }
      }
    }

    # argument from 'zing-core-bundler'
    my $collect_executable = $_[3];
    debug_printer(__LINE__,"ARG: COLLECT_EXECUTABLE = $collect_executable");
    if (defined($_[4])) {
        my $is_exe = &is_filetype_byarg($_[4], $str_executable);
        if ($is_exe == 1) {
          &msg_printer("ERROR: File '$_[4]' is not an executable\n");
          return 1;
        }
        debug_printer(__LINE__,"ARG: USER SUPPLIED EXECUTABLE = $_[4]");
    }
    my @tmp_libfiles;
    # start collecting list of files from hs_error log file.
    if ( -f $logfile ) {
      if (!open($fh, $logfile)) {
        &msg_printer("ERROR: could not open hs_err file");
      }
      debug_printer(__LINE__,"HS_ERR: HS_ERR LOG FILE = $logfile");
      my $dynflg = 0;
      # This loop will collect the list of files in an array (tmp_libfiles)
      # and will ignore unwanted files like chr special file, directory or invalid files
      while (my $line = <$fh>) {
        # find PID from hs_err file. Present in head -n 4 of the file.
        if ($line =~ m#pid=#) {
          my @arr = split(/pid=/, $line);
          if (defined($arr[1])) {
            my @arr2 = split(/,/, $arr[1]);
            $pid_from_hs_err = $arr2[0];
            debug_printer(__LINE__,"HS_ERR: PID FROM HS_ERR FILE = $pid_from_hs_err");
            # if we have a valid hs_err file
            if ($gdb_found_pid == 1) {
              if ($pid_from_gdb ne $pid_from_hs_err) {
                &msg_printer("ERROR: File mismatch. Please make sure both hs_err and core file are from the same process\n");
                return 1;
              }
              debug_printer(__LINE__,"HS_ERR: PID FROM HS_ERR FILE = PID FROM GDB = $pid_from_hs_err");
            }
          }
        }
        if ($line =~ m/Host\:/) {
          my @arr = split(/:/, $line);
          if (defined($arr[1])) {
            my $hostname_from_hs_err = $arr[1];
            $hostname_from_hs_err =~ s/^\s+//;
            chomp($hostname_from_hs_err);
            debug_printer(__LINE__,"HS_ERR: HOSTNAME = $hostname_from_hs_err");
            debug_printer(__LINE__,"HOSTNAME = $hostname");
            debug_printer(__LINE__,"CHECK, HOST MATCHING = $bundle_on_current_host");
            if ($hostname_from_hs_err ne $hostname) { 
              $append_str = $hostname_from_hs_err . "_bundledon_" . $hostname;
              if ($bundle_on_current_host == 1) {
                next;
              }
              &msg_printer("ERROR: The bundler must run on host '$hostname_from_hs_err'\n");
              &msg_printer("ERROR: Use the argument '-bundle-on-current-host' to bundle on this host\n");
              exit 1;
            }
          }
          next;
        }
        if ($line =~ m/XX\:ProfileLog/) {
          $profilelog_fname = &extract_profilelog('ProfileLogIn=', $line);
          $profilelog_fname .= " " . &extract_profilelog('ProfileLogOut=', $line);
          debug_printer(__LINE__,"HS_ERR: ProfileLog File(s) = $profilelog_fname");
          next;
        }
        if ($line =~ m/^Dynamic libraries:/) {
          $dynflg = 1;
          debug_printer(__LINE__,"HS_ERR: Dynamic libraries found = $dynflg");
          next;
        }
        if ($dynflg) {
          if ($line =~ m/Cannot get library information/){
            last;
          }
          my @libs = split(/\s+/, $line);
          if (defined($libs[5])) {
            if ($libs[5] =~ m/^\[stack\]$/) {
              $dynflg = 0;
              next;
            }
          }
          if (defined($libs[3]) && $libs[3] ne 0 ) {
            if ($libs[5]) {
              if ($libs[5] ne '[heap]') {
                if (!defined($libshash{$libs[5]})) {
                  if (defined($dockerpath)) {
                    $libs[5] = $dockerpath . "/" . $libs[5];
                  }
                  unless (-e $libs[5] && -r $libs[5]) {
                    next;
                  }
                  my $is_dir = &is_filetype_byarg($libs[5], $str_directory);
                  if ($is_dir == 0) {
                    next;
                  }
                  my $is_chr_spl = &is_filetype_byarg($libs[5], $str_characterspecial);
                  if ($is_chr_spl == 0) {
                    next;
                  }
                  $libshash{$libs[5]} = 1;
                  push(@tmp_libfiles, $libs[5]);
                }
              }
            }
          }
        }
      }
      close($fh);
      if ($pid_from_hs_err == 0) {
        &msg_printer("ERROR: Please make sure hs_err file is valid.\n");
        return 1;
      }
      # Now, divide the files into groups of lists.
      my $list_so = " ";
      my $list_data = " ";
      my $list_exe = " ";
      my $flag_nonjava_exe = 0;
      foreach my $file(@tmp_libfiles) {
        if (length($file) > 0) {
          chomp($file);
          my $is_so = &is_filetype_byarg($file, $str_sharedobject);
          my $is_exe = &is_filetype_byarg($file, $str_executable);
          # check for SOs
          if ($is_so == 0) {
            debug_printer(__LINE__,"HS_ERR: SHARED OBJ = $file");
            $list_so .= $file . " ";
            next;
          }
          # check for data files
          if ($is_so == 1 && $is_exe == 1) {
            debug_printer(__LINE__,"HS_ERR: DATE FILE = $file");
            $list_data .= $file . " ";
            next;
          }
          # check for exe files
          if ($is_exe == 0) {
            # collect only non-java executable
            if (compare($file, $JAVA_PATH) != 0) {
              $flag_nonjava_exe = 1;
              debug_printer(__LINE__,"HS_ERR: NON-JAVA EXECUTABLE = $file");
              $list_exe = $file;
              next;
            }
          }
        }
      }
      if (defined($_[4])) {
        my $exe_from_arg = $_[4];
        if ($flag_nonjava_exe == 0) {
          $list_exe = $JAVA_PATH;
          debug_printer(__LINE__,"HS_ERR: NON-JAVA EXECUTABLE = JAVA_PATH = $list_exe");
        }
        my $ret = compare($exe_from_arg, $list_exe);
        if ($ret != 0) {
          &msg_printer("INFO: GDB will be used to compare '$exe_from_arg' and '$list_exe'\n");
        }
      }
      if ($collect_executable == 0 && $flag_nonjava_exe == 1) {
        &msg_printer("INFO: Collecting the executable used is required in order to debug the core.\n");
        &msg_printer("INFO: Please use '-getexe' to bundle the core\n");
        return 1;
      }
      if ($collect_data == 0) {
        $list_data = " ";
      }
      $tarlist = $list_so . " " . $list_data . " " . $list_exe . " ";
      $cpy_list_exe = $list_exe;
      if (defined $tarlist) {
        my @from_lines = split(" ", $tarlist);
        foreach my $line(@from_lines) {
          debug_printer(__LINE__,"HS_ERR: TARLIST FILES = $line");
          push(@libfiles, $line);
        }
      }
    }

    # argument from 'zing-core-bundler', using it again!!
    $collect_executable = $_[3];
    my $is_exe = &is_filetype_byarg($exe_from_gdb, $str_executable);
    if ($is_exe == 0) {
      if ($collect_executable == 0 && compare($JAVA_PATH, $exe_from_gdb) != 0) {
        &msg_printer("INFO: Executable $exe_from_gdb is required to debug the core\n"); 
        &msg_printer("INFO: Please use '-getexe' to bundle the core\n");
        return 1;
      }
      if ($collect_executable == 1 && defined($_[4])) {
        if (compare($exe_from_arg, $exe_from_gdb) != 0) {
          &msg_printer("ERROR: Executable '$exe_from_gdb' does not match the argument '$exe_from_arg'\n");
          return 1;
        }
      }
    }
    else {
      if (defined($_[4])) {
        &msg_printer("INFO: Failed to find executable from GDB: using file from argument -exepath '$exe_from_arg'\n");
        $exe_from_gdb = $exe_from_arg;
      }
      else {
        if ($core_by_java_exe == 0) {
          &msg_printer("ERROR: Failed to find executable from GDB\n");
          &msg_printer("ERROR: Please provide a valid executable using '-exepath' flag\n");
          return 1;
        }
        else {
          $exe_from_gdb = " ";
        }
      }
    }

    {
      # Extracting hostname from GDB's output
      my @lines= split("\n", $output);
      my $hostname_from_gdb = undef;
      foreach my $line(@lines) {
        if ($line=~ m#HOSTNAME#) {
          my @parsed= split(/\:/, $line);
          if (defined($parsed[1])) {
            $hostname_from_gdb = $parsed[1];
            chomp($hostname_from_gdb);
            debug_printer(__LINE__,"GDB: HOSTNAME = $hostname_from_gdb");
            if ($hostname_from_gdb ne $hostname) { 
              if (length($hostname_from_gdb) <= 1) {
                $hostname_from_gdb = undef;
                next;
              }
              $append_str = $hostname_from_gdb . "_bundledon_" . $hostname;
              if ($bundle_on_current_host == 1) {
                next;
              }
              &msg_printer("ERROR: The bundler must run on host '$hostname_from_gdb'\n");
              &msg_printer("ERROR: Use argument '-bundle-on-current-host' to bundle on this host\n");
              exit 1;
            }
          }
          last;
        }
      }
      # Need a check for core files without hostname (specially cores generated on RHEL5) and when hs_err file is not
      # available to supply the host name. In this case, user is asked to supply a hs_err file or use '-bundle-on-current-host'
      # argument. The bundle will be named as corelibs.<pid>.FOUND_AT_UNKNOWN_HOSTNAME_bundledon_$hostname
      if (!defined($hostname_from_gdb) && !-f $logfile) {
        if ($bundle_on_current_host == 0) {
          &msg_printer("ERROR: Failed to find hostname from core file.\n");
          &msg_printer("ERROR: Either provide a hs_err file using '-hsfile' or use argument '-bundle-on-current-host' to override.\n");
          exit 1;
        }
        else {
          $append_str = "FOUND_AT_UNKNOWN_HOSTNAME" . "_bundledon_" . $hostname;
        }
      }
    }
    debug_printer(__LINE__,"TAR FOLDER NAME = $append_str");

    # For bundling purposes keep only one valid copy of executable 
    if (defined($exe_from_gdb) && defined($cpy_list_exe)) {
      if (!compare($exe_from_gdb, $cpy_list_exe)) {
        $exe_from_gdb = " ";
      }
    }

    $tarlist .= $exe_from_gdb . " ";
    push(@libfiles, $exe_from_gdb);

    my $gdb_out = undef;
    # containing within dockerpath else can be replaced with the pattern.
    if (defined($dockerpath)) {
      $output =~ /[\(gdb\)]* From                To                  Syms Read   Shared Object Library(.*)/ms;
      $gdb_out = $1;
    } else {
      $output =~ /^\(gdb\) From                To                  Syms Read   Shared Object Library(.*)/ms;
      $gdb_out = $1;
    }
    if (!defined($gdb_out)) {
      $gdb_out = " ";
    }
   
    debug_printer(__LINE__,"GDB: CMD OUTPUT = $gdb_out");
    my @lines = split("\n", $gdb_out);
    foreach my $line(@lines) {
      @parsed = split(/\s+/, $line);
      my $parsed_line = $parsed[scalar(@parsed)-1];
      if (defined($parsed_line)) {
        chomp($parsed_line);
        if (defined($dockerpath)) {
          $parsed_line = $dockerpath . "/" . $parsed_line;
        }
        if (-e $parsed_line && !defined($libshash{$parsed_line})) {
          $tarlist .= $parsed_line . " ";
          $libshash{$parsed_line} = 1;
          debug_printer(__LINE__,"GDB: ACCEPT = $parsed_line");
          push(@libfiles, $parsed_line);
        }
        else {
          $parsed_line .= " ";
          debug_printer(__LINE__,"GDB: REJECTED = $parsed_line");
        }
      }
    }
pre_bundle:
  # Adding JAVA_PATH
  $tarlist .= $JAVA_PATH . " ";
  }
  # Add complete chain of symlinks to the tarlist
  # for example, if libaznix.so is required by GDB
  # then a complete symlink chain to the library is needed.
  foreach my $libfile(@libfiles) {
    $libfile =~ s/^\s+|\s+$//g;
    if (length($libfile) > 0) {
      my($filename, $directory) = fileparse($libfile);
      $filename = $directory . $filename;
      debug_printer(__LINE__,"BUNDLER, LOOK FOR A SYMLINK IN = $filename");
      while (-f $filename && -l $filename) {
        $filename = $directory . readlink($filename);
        $list_add_me .= $filename . " ";
        debug_printer(__LINE__,"BUNDLER, ADD SYMLINK = $filename");
      }
      if (compare($filename, $libfile) != 0) {
        $list_add_me .= $filename . " ";
        debug_printer(__LINE__,"BUNDLER, ADD FILE = $filename");
      }
    }
  }
  # copy symlinks of the libraries too
  # For instance, copy both the 'real name' of library /lib64/libc-2.12.so
  # and it's symlink name /lib64/libc.so.6
  foreach my $libfile(@libfiles) {
    if (length($libfile) > 0) {
      my($filename, $directory) = fileparse($libfile);
      $filename = quotemeta($filename);
      # The output of 'ls -l' is inconsistent.
      # making it consistent by using "ls -l --full-time"
      my $cmdout = `ls -l --full-time $directory`;
      my @lines = split("\n", $cmdout);
      foreach my $line(@lines) {
        if ($line =~ m/->/){
          # changing it from "\d\d:\d\d" to "\d\d -\d\d\d\d"
          if ($line =~ m/\d\d -\d\d\d\d (.*?) -> $filename$/) {
            $tarlist .= "$directory" . "$1" . " ";
            debug_printer(__LINE__,"BUNDLER, added to tarlist = $directory$1");
          }
        }
      }
    }
  }
  $tarlist .= $list_add_me;

  # Filter dups
  {
    my $tmp_tarlist = " ";
    my @files= split (" ", $tarlist);
    for my $file (@files) {
      my $is_dir = &is_filetype_byarg($file, $str_directory);
      if (-e $file && !exists($hash{$file}) && $is_dir != 0) {
        $tmp_tarlist .= $file . " ";
        debug_printer(__LINE__,"BUNDLER: ACCEPT = $file");
        $hash{$file} = 1;
      }
      else {
        debug_printer(__LINE__,"BUNDLER: FILTER = $file");
      }
    }
    $tarlist = $tmp_tarlist;
  }

  my $stat_fh;
  open($stat_fh, ">$stat_fname") or &print_die("could not open stat fh");

  foreach my $file ( '/proc/cpuinfo', '/proc/meminfo', '/proc/version', '/proc/modules' ) {  
    my $tmpfh;
    open($tmpfh, $file);

    print $stat_fh "BEGIN $file\n\n";
    while (my $line = <$tmpfh>) {
      print $stat_fh $line;
    }
    print $stat_fh "\n\nEND $file";
    close($tmpfh);
  }

  close($stat_fh);

  my $relocation_dir = dirname(File::Spec->rel2abs(__FILE__));
  if ($relocation_dir ne "/usr/sbin") {
    $relocation_dir =~ s!/usr/sbin!!;
  }
  else {
    $relocation_dir = "";
  }
  if ($relocation_dir eq "/sbin/") {
    $relocation_dir =~ s!/sbin/!!;
  }
 
  my $libs_path = undef;
  if (defined($outputdir)) {
    $libs_path = $outputdir . "libs.$pid.";
  } else {
    $libs_path = "libs.$pid.";
  }
  debug_printer(__LINE__,"BUNDLER, LIBS_PATH VAR = $libs_path");

  if ($invoked_byhook == 1) {
    $libs_path = "$coredir/$libs_path";
  }
  my $temp_tgz = $libs_path . "tgz";
  my $temp_tar = $libs_path . "tar";
   
  my $tar_prep_str = $relocation_dir . "/usr/lib/zing/bin/tar --ignore-failed-read --transform='s#^#corelibs" . ".$pid" . '_' . "$append_str" . '/#S\' ';
  my $tar_cmd_create = $tar_prep_str . " -vcf $temp_tar ";
  my $tar_cmd_append = $tar_prep_str . " -rf $temp_tar --no-rec ";

  my $gzip_cmd = $relocation_dir . "/usr/lib/zing/bin/pigz";
  system("$gzip_cmd -help 2>/dev/null");
  if ($? || !-x $gzip_cmd) {
    &msg_printer("WARNING: parallel gzip (pigz) does not seem to be installed, using regular gzip\n");
    $gzip_cmd = "gzip";
  }

  if (-e $stat_fname && -r $stat_fname) {
    debug_printer(__LINE__,"BUNDLER, STAT FILE = $stat_fname");
    $tarlist .= "$stat_fname" . " ";
  }
  if (-e $logfile && -r $logfile) {
    debug_printer(__LINE__,"BUNDLER, HS_ERR LOG FILE = $logfile");
    $tarlist .= "$logfile" . " ";
  }
 
  my $hotspot_logfile = "$coredir/hotspot.log";
  if (-e $hotspot_logfile && -r $hotspot_logfile) {
    goto tar_hotspot;
  }
  # extract path from hs_err file
  ($filename, $directory) = fileparse($logfile);
  $hotspot_logfile = $directory . "/hotspot.log";
  if (-e $hotspot_logfile && -r $hotspot_logfile) {
    goto tar_hotspot;
  }
tar_hotspot:
  debug_printer(__LINE__,"BUNDLER, HOTSPOT LOG FILE = $hotspot_logfile");
  $tarlist .= "$hotspot_logfile" . " ";

  debug_printer(__LINE__,"BUNDLER, PROFILE LOG FILE = $profilelog_fname");
  $tarlist .= $profilelog_fname . " ";

  debug_printer(__LINE__,"BUNDLER, EXECUTING CMD = $tar_cmd_create $tarlist $o_core");
  my $t1 = system("$tar_cmd_create $tarlist $o_core");
  if ($global_invoked_byhook) {
    # bitshift to get the return code
    $t1 = $t1>>10;
  }
  if ($t1) {
    $bundler_failed = 1;
    goto fail_and_clean;
  }
  if (defined($EMPTY_BIN_PATH)) {
    debug_printer(__LINE__,"BUNDLER, EXECUTING CMD = $tar_cmd_append $EMPTY_BIN_PATH");
    my $t2 = system("$tar_cmd_append $EMPTY_BIN_PATH");
    if ($global_invoked_byhook) {
      #bitshift by 10
      $t2 = $t2>>10;
    }
    if ($t2) {
      $bundler_failed = 1;
      goto fail_and_clean;
    }
  }
  debug_printer(__LINE__,"BUNDLER, EXECUTING CMD = $gzip_cmd -1 -b 512 -c $temp_tar > $temp_tgz");
  my $t3 = system("$gzip_cmd -1 -b 512 -c $temp_tar > $temp_tgz");
  if ($global_invoked_byhook) {
    $t3 = $t3>>8;
  }
  if ($t3 || -z $temp_tar || -z $temp_tgz) {
    $bundler_failed = 1;
    goto fail_and_clean;
  }

  debug_printer(__LINE__,"REMOVING FILE = $temp_tar");
  unlink("$temp_tar");
  &msg_printer("INFO: Encrypting core bundle $temp_tgz\n");
  &encrypt("$temp_tgz");

  debug_printer(__LINE__,"REMOVING HOTSPOT LOG FILE = $hotspot_logfile");
  unlink("$hotspot_logfile");
  debug_printer(__LINE__,"REMOVING HS_ERR LOG FILE = $logfile");
  unlink("$logfile");
  debug_printer(__LINE__,"REMOVING CORE FILE = $o_core");
  unlink("$o_core");

  closelog();
  my $exitValue = $?;

fail_and_clean:
  if ($bundler_failed == 1) {
    $exitValue = 1;
    &msg_printer("ERROR: Failed to bundle files\n");
    unlink($temp_tar);
    unlink($temp_tgz);
  }

  if (length ($stat_fname) > 0) {
    debug_printer(__LINE__,"REMOVING FILE = $stat_fname");
    unlink ($stat_fname);
  }

  $? = $exitValue;
  return $exitValue;

  bless {exitValue=>\$exitValue},$class;
}

1;

