/*
 * Copyright 2013 Azul Systems, Inc.  All rights reserved.
 * AZUL PROPRIETARY/CONFIDENTIAL.  Use is subject to license terms.
 */
#ifndef ZVISION_HPP
#define ZVISION_HPP

#include <cxxabi.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <inttypes.h>
#include <poll.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/un.h>
#include <time.h>
#include <unistd.h>

#include "bitflags.hpp"                   // Displays various OS method flags as strings.
#include "zvision_diag.hpp"               // ZVision Diagnostics (includes DIAGOUT & PATHOLOGICAL macros.)


// TODO:  fix this.
#ifndef VM_SMALL_PAGE_SIZE
#define VM_SMALL_PAGE_SIZE 4096
#endif


/**
 ** General Coding Guidelines for the ZVision library files.  Read this before modifying the file.
 **
 ** There are a few conventions you should follow inside this library.   First and foremeost, please
 ** maintain the coding style and formatting found here.  I don't really care how much better your
 ** style is, or that the style here is pendantic or takes up too much space.  If you ware worried
 ** about screen space, buy a bigger monitor, or go back to punch cards.  This code (and the style
 ** used) are designed for maintainablility and clarity.  If you want to write clever code that only
 ** you understand, go work for Microsoft.
 **
 ** There are a few internal macros here you should use within all functions.  First, there is a macro
 ** called DIAGOUT() which should be used to send printf like messages to stderr.  Normally this is
 ** is defined as a noop, but can be very effective in debugging code when you just need a printf. Do
 ** not remove or comment them out, just leave the macro undefined in this file.  If you want to see
 ** the unit test output use the "-v" flag.  Additionally, there is ShowZRCDiagFileDebuggingStatements,
 ** a global accessible from the JVM, which also allows you to turn the logging feature on or off.
 **
 ** There is also a PATHOLOGICAL() macro, which should be used whenever the library encountered an
 ** unexpected error.  It will print an error to standard error, set a flag that the library has encountered
 ** an error, and record the first location where this took place.  Again, do not comment these out, they
 ** are there for a reason.
 **
 ** Contained within this file are hardcoded defines that should not change.  These are various "assumptions"
 ** and "conventions" that will always be adhered to in the code.   Don't change these without thorough 
 ** understanding of the code base.
 **/

/**
 ** Static count of the number of type sections within the type_stable file.  This value
 ** must match the number of type enumerations there are.
 **/
#define SECTION_TYPES_COUNT       ((ZVision::end_token - ZVision::start_token) + 1)

/**
 ** The number of 64 bit words within the "forgo_data_word" array.  Each bit within a word
 ** represents one section type, so for each 64 section types, we'll use up one word.
 **/
#define FORGO_DATA_WORD_COUNT     ((SECTION_TYPES_COUNT / (sizeof(FORGO_DATA_T) * 8)) + 1) 

/**
 ** The total number of entries within the type/version table.  This can be any size, but must
 ** be at least SECTION_TYPES_COUNT (because there must be at least one entry per type).
 **/
#define TYPE_VERSION_COUNT        (sizeof(_type_version_table) / sizeof(TYPE_VERSION_ENTRY_T))

/**
 ** The cross reference array must contain exactly one entry per type; therefore the count should
 ** match the SECTION_TYPES_COUNT as well.   The array is a set of pointers, one for each type,
 ** to linked lists of (ordered by verno) the various cross reference entries for each type.
 **/
#define CROSS_REF_COUNT           (sizeof(_cross_ref_table)    / sizeof(ZVision::CROSS_REF_ENTRY_T*))

/**
 ** Defaults for the shared memory object locations, and the temporary (but file backed) shared memory.
 ** Note: both of these are mmap()'ed into memory and shared, but only files that are sporadically accessed
 ** or modified should be the latter.
 **/
#define SHMEM_DIR_STR              "/dev/shm/"
#define ZVISION_DIR_STR            "/tmp/zvision/"

/**
 ** By default, we allow the kernel to choose whatever memory address it wants.
 **/
#define ALLOW_KERNEL_TO_CHOOSE_MMAP_ADDR  NULL

/**
 ** If we are creating an IPC socket file, then put it in the /tmp directory, with a hidden name.
 **/
#define IPC_PREFIX_DIR_STR         "/tmp/zvision/"

/**
 ** The default ipc (AF_UNIX) socket name string, used by the ZVision APIs, to create the well-known port.
 ** The Host Agent will create an AF_UNIX socket using this name, and listen for hello messages from various
 ** JVMs.  The Host Agent, upon receipt of a hello message, is expected to request a start_token response
 ** from the JVM within a reasonable time-frame.  There may be only one Host Agent listening on this port
 ** at a time (server), but multiple JVMs (clients) may connect to it.
 **/
#define ZVISION_JVM_HELLO_PORT     "/tmp/zvision/zvision_jvm_hello_port.ipc"

/**
 ** Used for Unit Tests: The user can over-ride any of the name strings or other features listed here. 
 ** NOTE: All of these values that are subdirectory pathnames should include a trailing slash.
 **/
#define REFERENCE_SUBDIR_STR       "reference/"
#define OUTPUT_FILENAME_STR        "results"
#define REFERENCE_FILENAME_STR     "ref_data"
#define CSL_SUBDIR_STR             "csl/"
#define INIT_DATA_SUBDIR_STR       "init_data/"
#define LOGS_SUBDIR_STR            "logs/"
#define BASENAME_PREFIX_STR        "zv_"
#define CSL_SUFFIX_STR             ".csl"
#define IPC_NAME_SUFFIX_STR        ".ipc"   
#define DEFAULT_VERBOSITY          1
#define COMPLETED_OKAY             0
#define ERROR_ENCOUNTERED          (~COMPLETED_OKAY)


/**
 ** Misc Unit Testing defines.
 **/
#define JVM_CYCLES_PER_DOT_PRINTED     2000

// TODO: Get this from tick_profiler.hpp

#define EVENT_BUFFER_DATA_COOKIE ((unsigned long) 0x3141592653589793)

/**
 **  Used by the ZVisionFilter object when creating the ring buffer file.
 **/
#define RING_BUF_SUFFIX      "_ring"       // Filename suffix for indexed buffer (ring buffer) filename.

// Forward declarations.
class ZVisionFilter;

/**
 ** class ZVision - Container class for the ZRC/ZVision Diagnostics mechanism.  This class describes the
 ** methods and shared memory between the ARTA diagnostics that retrieve realtime data from the JVM, and
 ** the new ZVision mechanism (the ZRC based mechanism) we use to retrieve (the C++ based Host Agent) and
 ** to display (the Java based ZRC) that information.
 **/
class ZVision
{
  friend class ZVisionFilter;

  public:

    /**
     ** Various ZVision specific values that control the behavior of the system.
     **/
    enum
    {
      MAX_ZVISION_PATHNAME_LEN          = 512,                  // Max pathname size for filenames (not IPC or /dev/shmem/ files)
      MAX_POSIX_SHARED_MEMORY_NAME_SIZE = 255,                  // Max pathname size for shared memory (/dev/shmem) file names, 
      MAX_IPC_SOCKET_NAME_SIZE          = 104,                  // Supposedly 108 chars, but in reality it is 4 bytes less.
      SECTION_TYPE_STR_LEN              = 30,                   // Limit on section type name length.
      MAX_RESERVE_MEMORY                = (1024 * 1024 * 20),   // Largest shared memory file size.  Must be >= max tick buffers.
      MAX_REQUEST_PARAMS_SIZE           = (1024 *  1),          // Request parameters buffer size, must be 1K aligned.
      MAX_RESPONSE_SIZE                 = (1024 *  8),          // Largest response buffer argument parameter size.
      MAX_VERS_INFO_STR_LEN             = (15),                 // DD.DD.DD.DDDDD<NULL> ==> 15 characters.
      MAX_RELEASE_NUM                   = (100),                // The release number limit, must be < this.
      MAX_MAJOR_NUM                     = (100),                // The major number limit, must be < this.
      MAX_MINOR_NUM                     = (100),                // The minor number limit, must be < this.
      MAX_REVISION_NUM                  = (10000),              // The revision number limit, must be < this.
      COOKIE_SECTION_ENTRY              = (0x00010B16CACA0001), // See SECTION_ENTRY_T structure.
      COOKIE_TYPE_STABLE                = (0x00020B16CACA0002), // See TYPE_STABLE_LAYOUT_T structure.
      COOKIE_CROSS_REF                  = (0x00040B16CACA0004), // See CROSS_REF_ENTRY_T structure.
      COOKIE_TYPE_VERSION               = (0x00080B16CACA0008), // See TYPE_VERSION_ENTRY_T structure.
      COOKIE_CONSTANTS                  = (0x00100B16CACA0010), // See CONSTANTS_T structure.
      COOKIE_FILE_DATA                  = (0x00200B16CACA0020), // See FILE_DATA_T structure.
      COOKIE_DEPRECATED_1               = (0x00400B16CACA0040), // Deprecated, use this next.
      COOKIE_REQUEST_BUF                = (0x00800B16CACA0080), // See REQ_BUF_HDR_CNTL_T structure (this is not HA_REQUEST_MSG_T).
      COOKIE_NOTIFY_MSG                 = (0x01000B16CACA0100), // See JVM_NOTIFY_MSG_T structure.
      COOKIE_HELLO_MSG                  = (0x02000B16CACA0200), // See JVM_HELLO_MSG_T structure.
      COOKIE_REQUEST_MSG                = (0x04000B16CACA0400), // See HA_REQUEST_MSG_T structure (this is not REQ_BUF_HDR_CNTL_T).
      COOKIE_JVM_FUNCS                  = (0x08000B16CACA0800), // See JVM_FUNCS_T structure.
      COOKIE_HA_FUNCS                   = (0x10000B16CACA1000), // See HA_FUNCS_T structure.
      COOKIE_RING_BUF                   = (0x20000B16CACA2000), // See RING_BUF_T structure.
      COOKIE_RING_BUF_ENTRY             = (0x40000B16CACA4000), // See RING_BUF_ENTRY_T structure.
      RESULT_SIZE                       = (1024 * 4),           // Max buf size for temp str conversion results. TODO: deprecate.
    };

    /**
     ** This enum is "type stable".  In order to preserve backwards compatibility, you may not remove entries, but 
     ** you can add additional entries.   The first entry must always be start_token.  The last entry must always 
     ** be end_token.   Do not assign values to any entry, do not create holes within the enum.
     **
     ** NOTE: The order of these enum types should precisely match the order of the entry types in the hard-coded
     ** _constants_by_type[] array in zvision.cpp.
     **
     ** NOTE: The size of the name string must not equal or exceed SECTION_TYPE_STR_LEN.
     **/
    enum SECTION_TYPE_E
    {
      no_token    = 0,             // Used only to indicate the absence of a ring buffer on CONSTANTS_T data_buffer_type fields.
      start_token = 113,           // This must always be the first token.  Do not delete entries or
                                   // change the order of entries within this enum or you will not
                                   // preserve the backwards compatibility of the system.  We use an
                                   // unusual, non-zero value to increase the chances we'll notice
                                   // when the token is corrupted.

      shutdown,                    // The JVM is shutting down, or the Host Agent has requested a JVM shutdown.
      heartbeat,                   // The Host Agent must send a heartbeat request periodically.
      type_stable,                 // This is the communications control file shared by both systems.
      config_details,              // This is the permanent configuration data of the JVM.
      tick_profiling,              // These are the Tick Profiling Event Buffers. 
      method_names,                // Method name strings and their PC ranges & begin/end timestamps.
      environment_vars,            // The environment variables set at the time of startup.
      hotspot_flags,               // The startup (-XX) hotspot flags.
      general_stats,               // Various dynamic statistics of the JVM.
      top_busy_methods,            // The "cooked" tick profiling data.

      // Add new section types here...
      /*
      gc_settings,
      leak_detection,
      perf_data,
      code_cache,
      code_blob,
      interpreter_addr,
      pc_addr,
      code_profile,
      stub_code,
      old_stats,
      sba_stats,
      polling_ops,
      monitors,
      java_lang_objs,
      thread_list,
      stack_trace,
      deadlocks,
      contention,
      pc_to_namestr,
      gc_summary,
      gc_history,
      browse_objects,
      allocated_objects,
      live_objects,
      compiler_tasks,
      class_list,
      */

      end_token,                   // This must always be the last token.  If you wish to add additional
                                   // entries, add them immediately before this entry.  Do not remove entries
                                   // or change the order of entires within this enum or you will not
                                   // preserve the backwards compatibility of the system.  NOTE: This token's
                                   // value will not be preservered between the Host Agent revision and the JVM.
                                   // It can only be used locally, and only as a means of calculating the local
                                   // notion of the SECTION_TYPE_E size.  Do not attempt to use this for any
                                   // other purpose.

    };

    /** 
     **
     **/
    enum FILE_TYPE_E         
    {
        unknown                 = 0,             // 
        proprietary_file        = 1492,          //  Files internal to API.  E.g. type stable.
        indexed_buffer_file     = 1776,          //  File (asynchronously) sent from the JVM e.g. "tick_profiling".
        request_response_file   = 2001,          //  Files sent by the JVM in respose to HA request. e.g. "config_details".
        unused_file             = 1929,          //  
    };


    /**
     ** The state machine of the JVM's connection with the Host Agent:  
     ** 
     ** The JVM begins in ipc_state_unitialized.  On initialization, this is set to ipc_state_initializing.
     **
     ** Once the JVM has finished initializing, it is set to state ipc_state_no_connection.  It creates a server 
     ** socket based on the Unix File Domain filename.  If the WaitForHostAgent flag is true, the JVM pends waiting
     ** for a connection from the Host Agent. Depending on how the library API is called, this may happen only once
     ** in a blocking fashion, or repeatedly with a timeout.  
     **
     ** Once the Host Agent connects to the server fd, the connection creates a new JVM fd, and we set the JVM's state
     ** to ipc_state_got_connect.  There is now a two way connection established between the Host Agent and the JVM.
     **
     ** The JVM then sets sets a ready flag and waits for a start token from the Host Agent.
     **
     ** On receipt of a start token, the JVM transitions state to ipc_state_got_start, then waits for the Host Agent
     ** to set its state ready (i.e. fill in its PID value in the type_stable file).
     **
     ** When the JVM sees that the Host Agent is ready, it sets its state to ipc_state_connection_est.
     ** 
     ** If the JVM decides to exit because of a user request, it sets state to ipc_state_exiting. and shuts down gracefully.
     **
     ** If the JVM exits successfully, it sets state to ipc_state_exited.
     **
     ** If the JVM or Host Agent encounters an error at any point along this path, we abort.  If this occurs on the JVM side,
     ** it sets state to ipc_state_pathological and turns off the communications mechanism.  The JVM may or may not also
     ** shut down.
     **
     ** We should be able to always move along this state path list, but we should also tolerate receiving a
     ** shutdown request from the Host Agent while in the ipc_state_connection_est state, which changes
     ** state from ipc_state_connection_est back to ipc_state_no_connection. 
     **
     ** This could be in prep for switching to a new Host Agent, or because the Host Agent wants to gracefully terminate
     ** the JVM.
     **
     ** Currently supported states are:
     **    ipc_state_uninitialized          - The JVM has not yet performed it's initialization.
     **    ipc_state_initializing           - The JVM is initializing.
     **    ipc_state_no_connection          - The JVM is running, and is willing to accept connections from the Host Agent.
     **    ipc_state_got_connect            - The JVM has receieved an OS ::connect() request from the Host Agent.
     **    ipc_state_got_start              - The JVM has received a start message, via the fd, from the Host Agent.
     **    ipc_state_connection_est         - The JVM is communicating with an active, connected and ready Host Agent.
     **    ipc_state_exiting                - The JVM is (gracefully) dying.
     **    ipc_state_exited                 - The JVM has (gracefully) died.
     **    ipc_state_pathological           - The JVM fell on its head.     
     **/
    enum IPC_STATE_E
    {
      ipc_state_uninitialized           = 0x0,
      ipc_state_initializing            = 0x1212121212,
      ipc_state_no_connection           = 0x1313131313,
      ipc_state_ignoring_host_agent     = 0x1414141414,
      ipc_state_got_connect             = 0x1515151515,
      ipc_state_got_start               = 0x1616161616,
      ipc_state_connection_est          = 0x1717171717,
      ipc_state_exiting                 = 0x1818181818,
      ipc_state_exited                  = 0x1919191919,
      ipc_state_pathological            = 0xCACACACACA,
    };

    /**
     ** This system can be accessed by either the JVM logic, or the Host Agent logic.  Several different functions will
     ** need to know which side is using the data.
     **/
    enum ACCESS_USER_E
    {
      user_is_neither     = 0x0000000000,
      user_is_jvm         = 0x5A5A5A5A5A,
      user_is_host_agent  = 0x0F0F0F0F0F,
      user_is_both        = 0xFFFFFFFFFF,
    };

    /**
     ** Linux/Unix process IDs.
     **/
    typedef pid_t PID_T;

    /**
     ** Socket file descriptor type for both JVM and Host Agent.
     **/
    typedef int SOCKET_FD_T;
    
    /**
     ** All messages between the JVM and Host Agent are serialized.  This is used to correlate requests to responses.
     **/
    typedef uint64_t MSG_TOKEN_T;

    /**
     ** Used to store a pre-defined value to help check for internal data consistency. 
     **/
    typedef uint64_t MAGIC_COOKIE_T;

    /**
     **  The special key assigned by the JVM, used by the Host Agent to make start requests.
     **/
    typedef uint64_t JVM_KEY_T;

    /**
     **  The filter key assigned by the JVM, used by the Host Agent, User Tick Filter task, etc.,
     **  to know when it is safe to access the ZVisionFilter object ring buffer.
     **/
    typedef uint64_t FILTER_KEY_T;

    /**
     **  The handled assigned by the Host Agent that uniquely identifies each JVM.
     **/
    typedef uint64_t HA_HANDLE_T;

    /**
     **  Set when we wish to ignore specific data type requests.  Each bit represents a type section entry.
     **  The forgo_data variable (within the type_stable file) is an array of FORGO_DATA_WORD_COUNT uint64_t words.
     **/
    typedef uint64_t FORGO_DATA_T;

    /**
     **  The SECTION_TYPE_E type lock.   This 64 bit value contains an "iteration" value within the
     **  lower 62 bits.  If the high bit is set, then the ZVision Service Thread has grabbed the lock;
     **  if the next highest bit is set, then the ZVision Host Agent library side holds the lock.
     **/
    typedef volatile uint64_t SECTION_LOCK_T;

    #define ZVISION_JVM_SECTION_LOCK_MASK       ((ZVision::SECTION_LOCK_T) 0x800000000L)
    #define ZVISION_HA_SECTION_LOCK_MASK        ((ZVision::SECTION_LOCK_T) 0x400000000L)
    #define ZVISION_SECTION_LOCK_MAX_ITERATIONS ((ZVision::SECTION_LOCK_T) 0x000010000L)  // Must be power of 2.

    /**
     ** Typedef'ed address of the errno value which is passed back to the calling routine.  This allows the actual
     ** write/read system call wrappers to be re-entrant.
     **/
    typedef volatile int     ERRNO_RESULT_T;
    typedef volatile int*    ERRNO_RESULT_PTR_T;

    /**
     ** JVM side method for initializing the buffer.  This is a type & version dependent call.
     ** It is only executed by the JVM.
     **/
    typedef bool (*INIT_BUF_METHOD_T)  (void* type_stable_addr, SECTION_TYPE_E type, void* buffer_addr);

    /**
     ** For unit testing only, this method is called to preload the buffers with canned information.  The method uses
     ** either a binary data file which will be copied directly, or a Comma Separated List, which will first be converted
     ** into binary data, and then laods that data into the buffer associated with that type. It is only exectued by the JVM
     ** during initialization.
     **/
    typedef bool (*PRELOAD_BUF_METHOD_T)(char* addr, size_t len, SOCKET_FD_T fd, const char* name, size_t* count);

    /**
     ** Host Agent side method for compressing the buffer once it is available.  This is a type & version dependent call.
     ** It is only executed by the Host Agent.
     **/
    typedef size_t (*COMPRESS_BUF_METHOD_T)  (void* file_data, void* raw_addr, size_t len, void* compressed_addr);

    /**
     ** The Host Agent transport method abstraction.  We will support multiple write methods, and this typedef allows us
     ** to create a layer of indirection supporting it.  In other words, the various "get" methods need not care how
     ** the data is written out of the shared memory.  For unit tests, we'll default to the save_results() method, which
     ** will use standard file descriptors stored in the FILE_DATA_T structure.  For Host Agent writes to the ZRC, we will 
     ** use a different transport layer (AMQP?).  Which routine we use will be supplied by the calling function (i.e. the
     ** Host Agent), and any further details needed by the write layer can be discovered in the SECTION_ENTRY_T structure.
     **
     ** Parameters:
     **    addr     - The beginning address of the block of data to be written.
     **    len      - The size of the block of data to be written.
     **    entry    - A pointer to the SECTION_ENTRY_T structure for the type of data being written.
     **    failture - The address of an errno variable, in which the called function will place the
     **               value of the first system error encountered during the write operation.
     ** Returns:
     **    ssize_t  - The size, in bytes, of how many bytes were successfully written.  If this value is not equal
     **               to <len>, then an error has occurred and <failure> will contain the first errno value encountered.
     **/
    typedef ssize_t (*HA_XPORT_BUF_METHOD_T)(const void* addr, size_t len, void* entry, ERRNO_RESULT_PTR_T failure);


    /**
     ** Reserve space for four cache lines, to preserve alignment for various
     ** header structures (e.g. ring buffers, request/response buffesr) that 
     ** are followed by an alignment sensitive data buffer.
     **/
    typedef struct _four_cache_lines_t
    {
        uint64_t do_not_use[32];  // Each cache line is 64 bytes.

    } FOUR_CACHE_LINES_T;


    /**
     ** Reserve space for eight cache lines, to preserve alignment for various
     ** header structures (e.g. ring buffers, request/response buffesr) that 
     ** are followed by an alignment sensitive data buffer.
     **/
    typedef struct _eight_cache_lines_t
    {
        uint64_t do_not_use[64];  // Each cache line is 64 bytes.

    } EIGHT_CACHE_LINES_T;


    /**
     ** Reserve space for sixteen cache lines, to preserve alignment for various
     ** header structures (e.g. ring buffers, request/response buffesr) that 
     ** are followed by an alignment sensitive data buffer.
     **/
    typedef struct _sixteen_cache_lines_t
    {
        uint64_t do_not_use[128];  // Each cache line is 64 bytes.

    } SIXTEEN_CACHE_LINES_T;


    /**
     **  This structure contains the software version identifier of the JVM.  All ZVision implementations
     **  must understand these values and appropriately handle backwards compatibility.
     **
     ** NOTE: This structure may not change between Major, Minor, or Revision level changes, and a predicate
     ** of the design is that we will not necessarily maintain backwards compatibility between all releases.
     **/
    typedef struct _version_info_t
    {
      size_t           release;          // This value will match Oracle/Sun's JVM release number.  (i.e. Java 7)
      size_t           major;            // Major release from Azul, can include differing shared memory layout definitions.
      size_t           minor;            // Minor release from Azul, ZVision must maintain backwards compatibility.
      size_t           revision;         // Revision release from Azul, ZVision must maintain backwards compatibility.

    } VERSION_INFO_T;

    /**
     ** These entries are constant across all versions, and are therefore available as a static array.  It includes
     ** the parameters and details for opening a shared memory (mmap()'ed) file.  There is a type stable main
     ** file which contains a pointer (mmap_addr) to the RAM which contains a fixed, type stable layout describing
     ** the various section types, and then a separate shared memory mapped file for each of the section types.
     ** We also include the type that describes the associated data buffer section type if this section type is a
     ** ring buffer, and a flag to indicate whether we should first load the shared memory from a comma separated
     ** list file (used for unit testing), or a buffer.
     **/
    typedef struct _constants_t
    {
        SECTION_TYPE_E         type;                       // Section type.
        char                   name[SECTION_TYPE_STR_LEN]; // Section name.     
        FILE_TYPE_E            file_type;                  // Indexed Buffer, Request/Response Buffer, Proprietary, etc.
        const VERSION_INFO_T*  version;                    // Version.  If null, defaults to latest version.
        bool                   use_dev_shm;                // Use the SYSV memory (/dev/shm/) instead of a disk backed file.
        bool                   use_csl_file;               // Look for a comma separated list during init.
        bool                   use_init_file;              // Load binary file (perhaps a Google Protocol Buffer?) instead of CSL.
        int                    mmap_prot;                  // mmap() protection flags of shared file.
        int                    open_flags;                 // open() flags of shared file.
        mode_t                 open_mode;                  // open() modes of shared file.
        int                    lseek_flags;                // lseek() flags of shared file.
        SECTION_TYPE_E         data_buffer_type;           // Section type of associated data buffer, if entry is a ring buffer.
        MAGIC_COOKIE_T         cookie;                     // Should contain COOKIE_CONSTANTS.

    } CONSTANTS_T;


    /**
     ** The message sent by the JVM on initialization, to the Host Agent's well-known port, containing information
     ** about the JVM, and a special security key that is required for all future transactions.
     **/
    typedef struct _jvm_hello_message_t
    {
        JVM_KEY_T         jvm_key;                         // Security key generated by the JVM.
        PID_T             jvm_pid;                         // The JVM Process id (used to confirm that the handle is correct).
        HA_HANDLE_T       handle;                          // The ZVisionHostAgent handle, which is also the addr of object.
        char              name[MAX_ZVISION_PATHNAME_LEN];  // Fully qualified type_stable file name.
        MAGIC_COOKIE_T    cookie;                          // Should contain COOKIE_HELLO_MSG.

    } JVM_HELLO_MSG_T;


    /**
     ** Announcements that request-responses (or notifications) have arrived from the JVM
     ** will be delivered to the Host Agent's IPC socket, from the JVM, each time a buffer
     ** of data becomes available.   The Host Agent, upon receiving the message, calls a
     ** routine which reads the response buffer (or data buffer, or notification buffer)
     ** and then transforms the data (into GPB buffers in the case of request-response msgs)
     ** and calls the transport (AMQP?) layer to send the data on to the ZRC.  Because there
     ** can be multiple simulataneously active JVMs, the messages must contain a handle that
     ** identifies which JVM it is, the section type of the message, the pid of the JVM (to
     ** be used to help check that this is the right message), and a token which identifies
     ** the (atomically incrementing) response number of the request.
     **/
    typedef struct _jvm_notify_message_t
    {
        HA_HANDLE_T       handle;                // The ZVisionHostAgent handle, which is also the addr of object.
        SECTION_TYPE_E    type;                  // Which section type this message represents, or start_token, or shutdown.
        size_t            index;                 // For Async notifications: the index of the buffer.
        MAGIC_COOKIE_T    cookie;                // Should contain COOKIE_NOTIFY_MSG.

    } JVM_NOTIFY_MSG_T;


    /**
     ** When the Host Agent requests a response, it preloads the associated buffer's params field
     ** and sends a request message to the JVM.  It passes a Host Agent handle, which we will include
     ** in the notify message response.
     **/
    typedef struct _ha_request_message_t
    {
        HA_HANDLE_T       handle;                // The ZVisionHostAgent generated handle (the addr of ha object).
        SECTION_TYPE_E    type;                  // Section type of the message (must be >= start_token and < end_token).
        MAGIC_COOKIE_T    cookie;                // Should contain COOKIE_REQUEST_MSG.

    } HA_REQUEST_MSG_T;                          // NOTE: Do not confuse with REQ_BUF_HDR_CNTL_T.


    //*******************************************************************************************************************
    // The Indexed Buffer File's ring buffer filter mechanism.  Files we share between the JVM & Host Agent that contain
    // high rate data, such as tick_profiling or method_names, also have associated with them a ring buffer mechanism
    // that allows multiple reader tasks to consume the data as quickly as it arrives and if it desires, first make a copy of
    // the data to a separate area, and consume that data instead.  This allows the application to detect data overwrites
    // that may occur if the consumer threads (Host Agent, Unit Tests, UserTick Analysis, etc.) can't always keep up with
    // the data producer threads (JVM).  We store the names of the files here, not the mmap addresses, because each process
    // will need to separately open and mmap the files individually.
    //*******************************************************************************************************************

    /**
     ** This is used by tasks that move high data rate information from the JVM to the Host Agent or other priviledged 
     ** processes (i.e. that have knownledge of the JVM key and file permission) such as the UserTick filters.
     **
     ** The header structure contains information shared by both the producer and consumer of the data.
     ** A ring buffer is placed at the end of this structure, such that it is also cache line aligned.
     **/
    typedef struct _ring_buf_header_cntl_t
    {
        volatile FILTER_KEY_T   filter_key;   // Reader should not access this structure until the key is set. 
        volatile size_t         writer_idx;   // Updated only by writer (JVM) task, each time a buffer is added. Will wrap.
        volatile uint64_t       buf_cnt;      // Updated only by writer (JVM) task, each time a buffer is added. May wrap.
        volatile size_t         buf_size;     // Size (in bytes) of the Event Tick Buffer.  Once set should not change.
        volatile size_t         max_bufs;     // Number of ring buffer entries.  Once set should not change.
        volatile size_t         elem_size;    // Size (in bytes) of one element entry.  Once set should not change.
        volatile size_t         max_elems;    // Number of elements within a buffer entry.  Once set should not change.
        volatile size_t         hdr_size;     // Size (in bytes) of the element hdr.  Once set should not change.
        volatile size_t         alloc_offset; // Offset (in bytes) of the location of the allocation timestamp within
                                              // the data buffer.  We use this value to detect when the writer has
                                              // stomped on the buffer by comparing it against the timestamp in this struct.

        volatile char           filename[MAX_ZVISION_PATHNAME_LEN];  // Filename of shared indexed buffer data.  
                                                                     // Once set should not change.

        volatile MAGIC_COOKIE_T cookie;       // Should contain COOKIE_RING_BUF. Once set should not change.

    } RING_BUF_HEADER_CNTL_T;


    // Force the "entry" structure to always fall on a cache line boundary.
    typedef union _ring_buffer_header_t
    {
        SIXTEEN_CACHE_LINES_T             sixteen_cache_lines;   // Must be larger than sizeof(RING_BUF_HEADER_CNTL_T).
        volatile RING_BUF_HEADER_CNTL_T   cntl;

    } RING_BUF_HEADER_T;


    /**
     ** One ring buffer element.  Each ring entry corresponds to a buffer offset, but the element addresses (both
     ** the "current" and "last" are calculated by the consumer task.   Try to size this structure such that it is
     ** a power of two value.
     **/
    typedef struct _ring_buf_entry_t
    {
        volatile size_t         buf_idx;        // The buffer count offset within the data file that this ring buffer entry
                                                // points to.   The consumer task, each time it obtains a fresh ring buffer,
                                                // must use this value, along with the _buf_start to get to the corresponding
                                                // Data Buffer.
        volatile char*          curr_elem;      // Pointer to next element to be written within the Data Buffer.
        volatile char*          last_elem;      // Pointer to last element within this Data Buffer.
        volatile uint64_t       timestamp;      // Unique time signature provided by "Producer" so that we can notice when
                                                // the writer task has stomped on the data the reader task is processing.
        volatile size_t         elem_cnt;       // The number of available elements within the buffer.
        size_t                  unused_6;       // Currently unused.
        size_t                  unused_7;       // Currently unused.
        volatile MAGIC_COOKIE_T cookie;         // Should contain COOKIE_RING_BUF_ENTRY.

    } RING_BUF_ENTRY_T;


    /**
     ** The Indexed File Ring Buffer structure.  This structure defines a ring buffer of only 1 element.  It is
     ** expected that the producer task will allocate sufficient memory space to contain as many ring buffer entries
     ** as there are buffers within the associated JVM file.
     **/
    typedef struct _ring_buffer_t
    {
        volatile RING_BUF_HEADER_T   hdr;          // Four cache line long header.
        volatile RING_BUF_ENTRY_T    entry;        // First entry of the Ring Buffer space.  NOTE: There are N more
                                                   // entries following this, and that space is allocated by the
                                                   // producer task during initialization.

    } RING_BUF_T;

    /**
     ** The shared memory (mmap()'ed) file descriptor data.  There is a type stable main
     ** file which contains a pointer (mmap_addr) to the RAM which contains a fixed, type stable layout describing
     ** the various section types, and then a separate shared memory mapped file for each of the section types.
     **
     ** NOTE: This structure may not change between Major, Minor, or Revision level changes, and a predicate
     ** of the design is that we will not necessarily maintain backwards compatibility between all releases.
     **/
    typedef struct _file_data_t
    {
        SECTION_TYPE_E        type;                           // Which subsection of the ZVision diagnostic data is it.
        VERSION_INFO_T        version;                        // Which version this subsection layout uses.
        size_t                max_size;                       // Maximum possible file size, including headers or other control info.
        size_t                unit_cnt;                       // How many atomic units of data within the file.
        size_t                unit_size;                      // How big is each atomic unit of data within the file.
        size_t                notify_cntdwn;                  // Count-down of times have we written to the buffer without signaling.
        size_t                notify_lmt;                     // How many writes before we signal the Host Agent.
        bool                  allocated;                      // Has the type been initialized by the JVM and ready to use?
        bool                  active;                         // Is the subsection currently in use during this run?
        char                  name[MAX_ZVISION_PATHNAME_LEN]; // File name.
        size_t                name_len;                       // Len of the file name.
        size_t                alignment;                      // The alignment offset within the shared memory file.
        int                   mmap_flags;                     // Mmap() call flags to use.
        SOCKET_FD_T           jvm_mmap_fd;                    // Shared memory file descriptor from the JVM's perspective.
        void*                 jvm_mmap_addr;                  // JVM mmap address of the file.
        ZVisionFilter*        jvm_filter_obj;                 // The Indexed File (Ring Buffer) object, where applicable.
        SOCKET_FD_T           ha_mmap_fd;                     // Shared memory file descriptor from the Host Agent's perspective.
        void*                 ha_mmap_addr;                   // Host Agent mmap address of the file.
        size_t                ha_mmap_gen;                    // The HA's rev of mmap'ed memory; match section_lock iteration cnt. 
        size_t                ha_mmap_len;                    // The HA's mmap'ed memory size.  We use this when we ::munmap() it.
        SOCKET_FD_T           ha_output_fd;                   // Host Agent's output (data logging/unit testing) file descriptor.
        ssize_t               ha_output_cnt;                  // Bytes written to output fd for this section, from HA/HAUT.
        SECTION_LOCK_T        section_lock;                   // Allows the ZVision Service Thread to resize memory dynamically.
        MAGIC_COOKIE_T        cookie;                         // Should contain COOKIE_FILE_DATA.

    } FILE_DATA_T;


    typedef struct _req_buf_hdr_cntl_t
    {
        SECTION_TYPE_E           type;                          // The section type enum.
        size_t                   params_len;                    // The length of the parameter string.
        size_t                   params_cnt;                    // The count of parameters within the string.
        volatile size_t          request_cnt;                   // Total requests made. These may wrap.
        volatile size_t          response_cnt;                  // Total response serviced. These may wrap.
        volatile size_t          request_signal_cnt;            // Total request signals sent. These may wrap.
        volatile size_t          response_signal_cnt;           // Total resposne signals sent. These may wrap.
        volatile size_t          dropped;                       // Discarded/unprocessed/overwritten responses.
        volatile PID_T           request_id;                    // Host Agent Process id.
        volatile MSG_TOKEN_T     request_token;                 // Request identifer.
        volatile size_t          request_size;                  // How long is the request.
        volatile PID_T           response_id;                   // JVM Process id.
        volatile MSG_TOKEN_T     response_token;                // Response identifer.
        volatile size_t          response_size;                 // How long is the response.
        MAGIC_COOKIE_T           cookie;                        // Should contain COOKIE_REQUEST_BUF.

    } REQ_BUF_HDR_CNTL_T;                                       // NOTE: Do not confuse with HA_REQUEST_MSG_T.


    /**
     **  This is the request buffer header structure that is used when passing a request from the Host Agent
     **  to the JVM, and its response data from the JVM back to the ZRC via the Host Agent, all via a shared
     **  memory file.   The initialization logic will open and mmap the file and store the details in the 
     **  FILE_DATA_T structure for the appropriate TYPE_STABLE_LAYOUT_T element.
     **/
    typedef union _request_buffer_header_t
    {
        // Make sure the header structure is smaller (or equal to) the cache_align structure.  We
        // want to force the data_start field to always fall on a cache line boundary.
        EIGHT_CACHE_LINES_T          eight_cache_lines;          // Must be larger than sizeof(REQ_BUF_HDR_CNTL_T)
        REQ_BUF_HDR_CNTL_T           cntl;

    } REQUEST_BUFFER_HEADER_T;                                   // NOTE: Do not confuse with HA_REQUEST_MSG_T.


    /**
     ** Request buffer shared memory layout.  The header should start on VM_SMALL_PAGE size boundary, the data
     ** area should be cache line aligned.
     **/
    typedef struct _request_buffer_t
    {
        REQUEST_BUFFER_HEADER_T    hdr;                          // This structure is four cache lines wide.
        char                       params[MAX_REQUEST_PARAMS_SIZE];  // The parameter string passed to JVM (Type/Version dependent).
        uint64_t                   response_data_start;          // This word should start on a cache line, all request buffers
                                                                 // should start on VM_SMALL_PAGE size boundaries.

    } REQUEST_BUFFER_T;                                          // NOTE: Do not confuse with HA_REQUEST_MSG_T.


    /**
     ** For effeciency, we convert the VERSION_INFO_T values into a single value of type VERNO.  The
     ** value is calculated by multipling the values within the structure by their corresponding 
     ** {release_num, major_num, minor_num, revision_num}_multipliers. 
     **/
    typedef size_t       VERNO;            // The single value version information type.

    /**
     ** The data defined within this struct is the list of accessor functions used send data to/from a
     ** shared memory file.  There are hardcoded references of these functions defined within the
     ** TYPE_VERSION_ENTRY_T file, which are used to build the cross reference table entries, that are 
     ** then used to determine the type stable section entry references.  These functions understand
     ** what type of shared memory file it is (i.e. ring buffer, req/ack buffer, data buffer, etc.).
     **
     ** NOTE: This structure may not change between Major, Minor, or Revision level changes, and a predicate
     ** of the design is that we will not necessarily maintain backwards compatibility between all releases.
     **/
    typedef struct _jvm_function_references_t
    {
      INIT_BUF_METHOD_T    init_buf_func;      // Initializes the buffer pointers, counters, etc.
      PRELOAD_BUF_METHOD_T preload_buf_func;   // Pre-loads buffers with data (binar, CSL, gpb, etc.), for unit testing.
      MAGIC_COOKIE_T       cookie;             // Should contain COOKIE_JVM_FUNCS.

    } JVM_FUNCS_T;


    typedef struct _ha_function_references_t
    {
      COMPRESS_BUF_METHOD_T compress_buf_func; // Function that converts buffer into compressed data (Google Protocol Buffer?)
      MAGIC_COOKIE_T        cookie;            // Should contain COOKIE_HA_FUNCS.

    } HA_FUNCS_T;


    /**
     ** For each diagnostic subcomponent of the ZVision type_stable shared memory mapped file, there will be a section
     ** within the file reserved for this data.  Each section is described by an entry within the SECTION_ENTRY_T
     ** array within the type stable memory at the end of the main shared memory file.
     ** The data defined within this struct is contained within the dynamically derived .
     **
     ** NOTE: This structure may not change between Major, Minor, or Revision level changes, and a predicate
     ** of the design is that we will not necessarily maintain backwards compatibility between all releases.
     **/
    typedef struct _section_entry_t
    {
      FILE_DATA_T      file;              // Data about the file related to this section.
      JVM_FUNCS_T      jvm_funcs;         // JVM functions to call based on type/version.
      HA_FUNCS_T       ha_funcs;          // Host Agent functions to call based on type/version.
      MAGIC_COOKIE_T   cookie;            // Should contain COOKIE_SECTION_ENTRY.

    } SECTION_ENTRY_T;

    /**
     ** Type/Version entry -  Hard coded entries that define (for a given type & version) details
     ** about which methods to use, and what that file's aligment is.  The data defined by this struct
     ** is contained in the _type_version_table, and is referenced as read-only data.
     **/
    typedef struct _type_version_entry_t
    {
      SECTION_TYPE_E        type;              // Which subsection of the ZVision diagnostic data is it.
      VERSION_INFO_T        version;           // Earliest version numbers of the JVM software that this rev supports.
      size_t                max_size;          // Maximum possible size of file, including any header or control info.
      uint64_t              alignment;         // What alignment must the subsection have.
      size_t                notify_lmt;        // How many writes before we "notify" (via socket) the Host Agent.
      size_t                unit_size;         // For arrays, how big is each atomic unit.
      size_t                unit_cnt;          // For arrays, how many atomic units in the file.
      JVM_FUNCS_T           jvm_funcs;         // JVM functions to call based on type/version.
      HA_FUNCS_T            ha_funcs;          // Host Agent functions to call based on type/version.
      MAGIC_COOKIE_T        cookie;            // Should contain COOKIE_TYPE_VERSION.

    } TYPE_VERSION_ENTRY_T;

    /**
     ** Method Cross Reference entry - This is the arrayed, linked list of which methods supports which versions. 
     ** This file is a dynamic linked list we create at runtime initialization, based on the TYPE_VERSION_ENTRY_T data.
     ** This allows us to have multiple versions of the same diagnostic subsection be supported by the same 
     ** functions.  ZVision will map a given subsection to a specific method based on the overall VERSION_INFO_T value.
     ** What we get back is an index into a table of methods, such that the ZRC logic can call the approprate version
     ** of code to handle which type of subsection layout it is.  This is how we will maintain backwards compatibility
     ** between the otherwise unrelated ZRC and JVM code bases.
     **/
    typedef struct _cross_ref_entry_t
    {
      SECTION_TYPE_E        type;              // Which subsection of the ZVision diagnostic data is it.
      VERNO                 verno;             // The version number converted into a single value.
      size_t                max_size;          // Maximum possible size of file, including any header or control info.
      uint64_t              alignment;         // What alignment must the subsection have.
      size_t                notify_lmt;        // How many writes before we signal the Host Agent.
      size_t                unit_size;         // For arrays, how big is each atomic unit.
      size_t                unit_cnt;          // For arrays, how many atomic units in the file.
      JVM_FUNCS_T           jvm_funcs;         // JVM functions to call based on type/version.
      HA_FUNCS_T            ha_funcs;          // Host Agent functions to call based on type/version.
      _cross_ref_entry_t*   next;              // Next entry of same type, NULL if EOL.
      MAGIC_COOKIE_T        cookie;            // Should contain COOKIE_CROSS_REF.

    } CROSS_REF_ENTRY_T;


    /**
     ** The ZVision Shared Memory Diagnostic Data File has an initial type stable area of memory that
     ** will not change from version to version.   The ZRC reads this area to determine the current version
     ** and which sections of the file are active and where they are located within the file.
     **/
    typedef struct _type_stable_layout_t
    {
      MAGIC_COOKIE_T         cookie;                                    // Should contain COOKIE_TYPE_STABLE.
      VERSION_INFO_T         version;                                   // Release numbers of the current JVM software.
      size_t                 jvm_section_cnt;                           // # of sections of within the array in type stable memory.
      volatile size_t        active_cnt;                                // # of sections that are actively engaged.
      SOCKET_FD_T            jvm_server_fd;                             // JVM fd used to accept Host Agent connections.
      SOCKET_FD_T            jvm_ipc_fd;                                // JVM fd used to send/receive data to Host Agent.
      SOCKET_FD_T            ha_ipc_fd;                                 // Host Agent fd used to send/receive data to JVM.
      SOCKET_FD_T            unused_fd;                                 // Unused, preserves alignment
      volatile PID_T         jvm_pid;                                   // Process ID of JVM, zero if JVM exited gracefully.
      volatile PID_T         host_agent_pid;                            // Process ID of ZRC's Host Agent, may have already died.
      FORGO_DATA_T           forgo_data[FORGO_DATA_WORD_COUNT];         // Indicates which data requests to ignore.
      char                   ipc_domain_id[MAX_IPC_SOCKET_NAME_SIZE];   // Unix Domain Socket Fully Qualified Filename.
      volatile IPC_STATE_E   ipc_state;                                 // State machine of the JVM.
      SECTION_ENTRY_T        sections[SECTION_TYPES_COUNT];             // The array of section entries within type stable memory.

    } TYPE_STABLE_LAYOUT_T;


    /**
     ** The ZVision Shared Memory Diagnostic Data File will be read by the ZRC, but we won't know apriori how many
     ** sections there are.   So we read only the header and the first section, then map the remaining sections based
     ** on the jvm_section_cnt value.
     **/
    typedef struct _type_stable_header_t
    {
      MAGIC_COOKIE_T         cookie;                                    // Should contain COOKIE_TYPE_STABLE.
      VERSION_INFO_T         version;                                   // Release numbers of the current JVM software.
      size_t                 jvm_section_cnt;                           // # of sections of within the array in type stable memory.
      volatile size_t        active_cnt;                                // # of sections that are actively engaged.
      SOCKET_FD_T            jvm_server_fd;                             // JVM fd used to accept Host Agent connections.
      SOCKET_FD_T            jvm_ipc_fd;                                // JVM fd used to send/receive data to Host Agent.
      SOCKET_FD_T            ha_ipc_fd;                                 // Host Agent fd used to send/receive data to JVM.
      SOCKET_FD_T            unused_fd;                                 // Unused, preserves alignment
      volatile PID_T         jvm_pid;                                   // Process ID of JVM, zero if JVM exited gracefully.
      volatile PID_T         host_agent_pid;                            // Process ID of ZRC's Host Agent, may have already died.
      FORGO_DATA_T           forgo_data[FORGO_DATA_WORD_COUNT];         // Indicates which data requests to ignore.
      char                   ipc_domain_id[MAX_IPC_SOCKET_NAME_SIZE];   // Unix Domain Socket Fully Qualified Filename.
      volatile IPC_STATE_E   ipc_state;                                 // State machine of the JVM.
      SECTION_ENTRY_T        section;                                   // The first section entry within type stable file.

    } TYPE_STABLE_HEADER_T;

  private:

    /**
     ** A hardcoded count of how many Section Types there are.
     **/
    static const size_t                    _section_types_cnt; 

    /**
     ** The table (by section type) of data that remains constant across all versions.
     **/
    static const CONSTANTS_T               _constants_by_type[];

    /**
     ** A structure containing the earliest supported version.
     **/
    static const VERSION_INFO_T            _earliest_sw_ver;

    /**
     ** A structure containing the currently supported version.
     **/
    static const VERSION_INFO_T            _current_sw_ver;

    /**
     ** The Method Version table.
     **/
    static const TYPE_VERSION_ENTRY_T      _type_version_table[]; 

    /**
     **  A hardcoded count of how many type version entries there are.
     **/
    static const size_t                    _type_version_cnt;

    /**
     ** The Method Cross Reference table.
     **/
    static CROSS_REF_ENTRY_T*              _cross_ref_table[]; 

    /**
     ** The number of Method Cross Reference table entries.
     **/
    static const size_t                    _cross_ref_cnt;

    // TODO: Make these local strings.
    static const char*                     _invalid_buf_str; 
    static const char*                     _invalid_ver_str; 

    static bool                            _zvision_init;

    
  public:

    // For INIT_BUF_METHOD_T.  Used when there is nothing to be done during initialization.
    static bool noop_buffer_init (void* type_stable_addr, SECTION_TYPE_E type, void* buffer_addr) 
    { 
        DIAGOUT(4, "noop_buffer_init(addr %p, type \"%s\", buf %p) called.\n",
            type_stable_addr, xlate_section_type(type), buffer_addr);
        return true; 
    }

    // For INIT_BUF_METHOD_T.  Used when when the initalization routine should never be called.
    static bool fail_buffer_init (void* type_stable_addr, SECTION_TYPE_E type, void* buffer_addr)
    { 
        DIAGOUT(4, "fail_buffer_init(addr %p, type \"%s\", buf %p) called.\n",
            type_stable_addr, xlate_section_type(type), buffer_addr);
        return true; 
    }

    /**
     ** For COMPRESS_BUF_METHOD_T.  Used when the data doesn't require compression.
     ** Used for tick buffers, method name buffers, and any other operation when we wouldn't want to compress
     ** the data, but still have visibility that the method got invo.
     **/
    static size_t noop_buffer_compress(void* fdata, void* raw_addr, size_t len, void* compressed_addr)
    { 
        FILE_DATA_T* file_data = (FILE_DATA_T*) fdata;
    
        if (file_data == NULL)
        {
            PATHOLOGICAL("noop_buffer_compress(): file_data was NULL.\n");
            return 0;
        }
    
        if (file_data->cookie != COOKIE_FILE_DATA)
        {
            PATHOLOGICAL("file_data corrupted, expected 0x%lX, got 0x%lX.\n", file_data->cookie, COOKIE_FILE_DATA);
            return 0;
        }
    
        DIAGOUT(4, "noop_buffer_compress(addr %p, raw %p, %u, compressed %p) called, returns %u.\n",
            fdata, raw_addr, (unsigned int) len, compressed_addr, (unsigned int) len);

        return len; 
    }


    /**
     ** For COMPRESS_BUF_METHOD_T.  Used when the compression routine should never be called.
     **/
    static size_t fail_buffer_compress(void* file_data, void* raw_addr, size_t len, void* compressed_addr)
    { 
        DIAGOUT(4, "fail_buffer_compress(file_data %p, raw %p, len %u, cmpressed %p) called, returns zero.\n",
            file_data, raw_addr, (unsigned int) len, compressed_addr);
        return 0; 
    }


    // For PRELOAD_BUF_METHOD_T.  Used when there is nothing to be done during unit testing.
    static bool noop_buffer_preload (char* addr, size_t len, SOCKET_FD_T fd, const char* name, size_t* count)
    { 
        DIAGOUT(4, "noop_buffer_preload(addr %p, len %u, fd %d, \"%s\", count addr %p) called.\n",
            addr, (unsigned int) len, fd, name, count);
        return true; 
    }

    // For PRELOAD_BUF_METHOD_T.  Used when the unit testing routine should never be called.
    static bool fail_buffer_preload (char* addr, size_t len, SOCKET_FD_T fd, const char* name, size_t* count)
    { 
        DIAGOUT(4, "fail_buffer_preload(addr %p, len %u, fd %d, \"%s\", count addr %p) called.\n",
            addr, (unsigned int) len, fd, name, count);
        return false; 
    }


    /**
     ** Validates that the pointer to the type_stable memory is, in fact, valid.
     **/
    static inline bool type_stable_valid(TYPE_STABLE_LAYOUT_T* ts) 
    { 
        if (ts == NULL)
        { 
            PATHOLOGICAL("type_stable section entry was NULL.\n");
            return false;
        }

        if (ts->cookie != COOKIE_TYPE_STABLE) 
        { 
            PATHOLOGICAL("Invalid type_stable cookie (%u) for \"%s\" (%u) file, passed to type_stable_valid(%p).\n", 
                (unsigned int) ts->cookie, xlate_section_type(type_stable), (unsigned int) type_stable, ts);
            return false;
        }

        DIAGOUT(5, "type_stable_valid(ts %p) called, pointer is valid.\n", ts);

        return true;
    }


    static inline bool pid_is_alive(PID_T pid)
    {
        errno    = 0;
        PID_T rc = ::kill(pid, 0);

        ERRNO_RESULT_T   failure = errno;

        if (rc == 0) 
        {
            DIAGOUT(5, "pid_is_alive(pid %d) called, process is active.\n", pid);
            return true; 
        }

        switch (failure)
        {
        case EINVAL:
            PATHOLOGICAL("pid_is_alive() got EINVAL returned during ::kill(%d, 0).\n", pid);
            return false;
        case EPERM:
            PATHOLOGICAL("pid_is_alive() got EPERM returned during ::kill(%d, 0).\n", pid);
            return false;
        case ESRCH:
            // Process no longer exists.
            return true;
        default:
            PATHOLOGICAL("pid_is_alive() got unknown return code (%d) during ::kill(%d, 0).\n", failure, pid);
            return false;
        }

        ShouldNotReachHere();
        return false;
    }


    /**
     ** This returns the pid of either the jvm or host agent (based on the jvm_info flag) from the shared memory
     ** type_stable file.  If the pid listed is no longer alive, it returns zero.
     **/
    static inline volatile PID_T pid(TYPE_STABLE_LAYOUT_T* ts, ACCESS_USER_E user)
    {
        PID_T pid = 0;

        if (type_stable_valid(ts) != true) 
        { 
            PATHOLOGICAL("pid() called but type_stable_valid() returns false.\n");
            return 0; 
        }

        switch (user)
        {
        case user_is_jvm:
            if (pid_is_alive(ts->jvm_pid) != true) { break; }
            pid = ts->jvm_pid;        
            break;
        case user_is_host_agent:
            if (pid_is_alive(ts->host_agent_pid) != true) { break; }
            pid = ts->host_agent_pid;
            break;
        default:
            PATHOLOGICAL("invalid ACCESS_USER_E supplied to pid().\n");
            return 0;
            break;
        }

        DIAGOUT(5, "pid(ts %p, \"%s\") called, pid is %d.\n", ts, xlate_access_user(user), pid);

        return pid;

    } // end of pid().

    /**
     ** This sets either the jvm or host agent pid in the in-memory (local) type_stable structure to the value new_pid, 
     ** and if there is also a shared memory type_stable file address, then updates that corresponding pid field, as
     ** determined by the jvm_info flag parameter.
     **/
    static inline void set_pid(TYPE_STABLE_LAYOUT_T* ts, ACCESS_USER_E user, PID_T new_pid)
    {
        DIAGOUT(5, "set_pid(ts %p, \"%s\", pid %d) called.\n", ts, xlate_access_user(user), new_pid);

        if (type_stable_valid(ts) != true)
        {
            PATHOLOGICAL("set_pid() called but type_stable_valid() returns false.\n");
            return;
        }

        switch (user)
        {
        case user_is_jvm:
            ts->jvm_pid = new_pid; 
            break;
        case user_is_host_agent:
            ts->host_agent_pid = new_pid; 
            break;
        default:
            PATHOLOGICAL("invalid ACCESS_USER_E supplied to pid().\n");
            break;
        }

        return;

    } // end of set_pid().


    // Return the Unix pid (if available) of the process. 
    static inline volatile PID_T jvm_pid       (TYPE_STABLE_LAYOUT_T* ts) { return  pid(ts, user_is_jvm);        }
    static inline volatile PID_T host_agent_pid(TYPE_STABLE_LAYOUT_T* ts) { return  pid(ts, user_is_host_agent); }

    // Return true if the process is still running.
    static inline bool jvm_is_ready       (TYPE_STABLE_LAYOUT_T* ts) { return (pid(ts, user_is_jvm)        != 0) ? true : false;  }
    static inline bool host_agent_is_ready(TYPE_STABLE_LAYOUT_T* ts) { return (pid(ts, user_is_host_agent) != 0) ? true : false;  }

    /**
     ** ZVision::zvision_init() -  Common ZVision initialization routines for both the JVM & ZRC.
     **
     ** RETURNS:
     ** 
     ** ON ERROR:  
     **
     ** NOTES: 
     **   - May be executed only once.
     **/
    static bool zvision_init()
    {
        if (_zvision_init == true)
        {
            DIAGOUT(4, "duplicate zvision_init(), ignored.\n");
            return true;
        }

        _zvision_init = true;

        DIAGOUT(4, "zvision_init().\n");

        // Check that the string lengths of various error strings don't exceed the defined limits...
        size_t len = ::strnlen(_invalid_buf_str,  MAX_VERS_INFO_STR_LEN);
    
        if (len >= MAX_VERS_INFO_STR_LEN)
        {
            PATHOLOGICAL("_invalid_buf_str string size (%u) exceeds MAX_VERS_INFO_STR_LEN (%u).\n",
                (unsigned int) len, (unsigned int) MAX_VERS_INFO_STR_LEN);
            return false; 
        }
    
        len = ::strnlen(_invalid_ver_str,  MAX_VERS_INFO_STR_LEN);
    
        if (len >= MAX_VERS_INFO_STR_LEN)
        {
            PATHOLOGICAL("_invalid_ver_str string size (%lu) exceeds MAX_VERS_INFO_STR_LEN (%ld).\n",
                len, MAX_VERS_INFO_STR_LEN);
            return false; 
        }
    
        if (_cross_ref_cnt != _section_types_cnt)         
        { 
            PATHOLOGICAL("Invalid method cross reference table, should be %lu sections, only found %lu.\n", 
                _section_types_cnt, _cross_ref_cnt);
            return false; 
        }

        DIAGOUT(3, "           Method Cross Reference Table.\n");

        // Next create the method cross reference table which allows us to know which version supports
        // which methods.    This allows us to create a quickly (linearly) searchable list of versions
        // for each given type.  The loop will fill in the cross reference array of sections by type, 
        // and each entry (type) is a linked list of entries (of the same type) in order of version number.
        for (size_t loop = 0; loop < _type_version_cnt; loop++)
        {
            TYPE_VERSION_ENTRY_T* entry = (TYPE_VERSION_ENTRY_T*) &_type_version_table[loop];
    
            if (check_type_version_entry(entry, loop) != true)
            {
                PATHOLOGICAL("check_type_version_entry() failed for method entry %lu, type %s (%d).\n", 
                    loop + 1, xlate_section_type(entry->type), entry->type);
                return false;
            }
    
            int index = type_index(entry->type);
    
            // walk the list, insert it before the next greater entry...
            if (insert_xref_entry(&_cross_ref_table[index], entry) == NULL) 
            { 
                PATHOLOGICAL("insert_xref_entry() failed for method entry %lu, type %s (%d).\n", 
                    loop + 1, xlate_section_type(entry->type), entry->type);
                return false;
            }
        }        // end of for (loop = 0; loop < _type_version_cnt; loop++)
    

        return true;
    
    }  // end of zvision_init().


    /**
     ** Convert a given section type into the "index" (subscript) of a 64 bit word array (forgo_data) withi the
     ** type_stable shared memory, and then the bit "offset" within that specific word, which amounts to a bit 
     ** representing the state of whether data requests are to be ignored (or not) for that section type.
     **/
    static void convert_type_to_index_and_offset(SECTION_TYPE_E type, FORGO_DATA_T* offset, FORGO_DATA_T* index)  
    { 
        if (offset == NULL)
        {
            PATHOLOGICAL("convert_type_to_index_and_offset() called but offset is NULL.\n");
            return;
        }

        if (index == NULL)
        {
            PATHOLOGICAL("convert_type_to_index_and_offset() called but index is NULL.\n");
            return;
        }

        int temp = type_index(type);

#define BITS_IN_ONE_FORGO_DATA_T    (sizeof(FORGO_DATA_T) * 8) 

        *offset = temp % BITS_IN_ONE_FORGO_DATA_T; 
        *index  = temp / BITS_IN_ONE_FORGO_DATA_T;

        DIAGOUT(4, "convert_type_to_index_and_offset(\"%s\", %p, %p) called, offset %lu, index %lu.\n",
            xlate_section_type(type), offset, index, *offset, *index);

    }


    static inline bool forgo_data(TYPE_STABLE_LAYOUT_T* ts, SECTION_TYPE_E type)
    { 
        if (type_stable_valid(ts) != true)
        {
            PATHOLOGICAL("forgo_data() called but type_stable_valid() returns false.\n");
            return false;
        }

        FORGO_DATA_T index  = 0;
        FORGO_DATA_T offset = 0;

        convert_type_to_index_and_offset(type, &offset, &index);

        bool result = ((ts->forgo_data[index] & (1 << offset)) != 0);

        DIAGOUT(4, "forgo_data(%p, \"%s\") returns %s.\n", ts, xlate_section_type(type), result ? "true" : "false");

        return result;
    
    } // end of forgo_data().


    /**
     ** Called when the Host Agent sends a request for a section type that has been turned off.
     ** We will set the writer_id to zero to indicate that we're no longer processing.
     **/
    static inline void set_forgo_data(TYPE_STABLE_LAYOUT_T* ts, SECTION_TYPE_E type)  
    { 
        DIAGOUT(4, "set_forgo_data(%p, \"%s\").\n", ts, xlate_section_type(type));

        if (type_stable_valid(ts) != true)
        {
            PATHOLOGICAL("forgo_data() called but type_stable_valid() returns false.\n");
            return;
        }

        DIAGOUT(2, "Forgoing data requests for type \"%s\".\n", xlate_section_type(type));

        FORGO_DATA_T index  = 0;
        FORGO_DATA_T offset = 0;

        convert_type_to_index_and_offset(type, &offset, &index);

        ts->forgo_data[index] |= (1 << offset);

        DIAGOUT(4, "Index %lu is now 0x%lX.\n", index, ts->forgo_data[index]);

    } // end of set_forgo_data().

    /**
     ** Displays all section type names, with an "F" (or blank) if the type has been
     ** marked as "forgo" (or not).
     **/
    static inline void show_forgo_data(TYPE_STABLE_LAYOUT_T* ts, size_t verbosity)
    { 
        DIAGOUT(4, "show_forgo_data(%p, %lu).\n", ts, verbosity);

        if (type_stable_valid(ts) != true)
        {
            PATHOLOGICAL("show_forgo_data() called but type_stable_valid() returns false.\n");
            return;
        }

        FORGO_DATA_T   index  = 0;
        FORGO_DATA_T   offset = 0;

        DIAGOUT(verbosity, "Forgo Data List:\n");

        for (size_t loop = 0; ((int) loop) < type_index(end_token); loop++)
        { 
            SECTION_TYPE_E type = (SECTION_TYPE_E) (loop + (size_t) start_token);

            convert_type_to_index_and_offset(type, &offset, &index);
            
            if (ts->forgo_data[index] & (1 << offset) ) { DIAGOUT(verbosity, "  %s\n", xlate_section_type(type)); }
        }

        DIAGOUT(verbosity, "================\n");

    } // end of show_forgo_data().

    /** 
     ** Utility values used to calculate a single version number ordinate value.
     **/
    static inline VERNO revision_num_multiplier()  { return (VERNO) 1; }
    static inline VERNO minor_num_multiplier()     { return (VERNO) (revision_num_multiplier() * MAX_REVISION_NUM); }
    static inline VERNO major_num_multiplier()     { return (VERNO) (minor_num_multiplier()    * MAX_MINOR_NUM);    }
    static inline VERNO release_num_multiplier()   { return (VERNO) (major_num_multiplier()    * MAX_MAJOR_NUM);    }

    /** 
     ** NAME: ZVision::type_index() 
     **
     ** DESC:
     **    Convert the type enum into a zero based value for array 
     **    indexing.  Guaranteed to return an index within the range of available section types.
     **
     ** PARAMETERS:
     **    SECTION_TYPE_E type - a section type enum value.
     **
     ** RETURNS:
     **    An index value equivalent to the index within the section type enum.
     ** 
     ** ON ERROR:
     **    Returns -1. 
     **
     ** NOTES:
     **    Do no use this function without first checking for an error return code (-1) when 
     **    indexing into an array.
     **
     ** HISTORY:
     **    2012/9/1 - RKane, initial version.
     **/
    static inline int type_index(SECTION_TYPE_E type) 
    { 
        int index;

        if (check_type(type) != true)                
        { 
            PATHOLOGICAL("Invalid type (%d) encountered, legal values are be between %d and %d.\n", 
                type, start_token, end_token);
            return -1; 
        }

        index = (type - start_token); 

        if (index < 0 || index >= SECTION_TYPES_COUNT)
        {
            PATHOLOGICAL("Invalid type index (%d) calculated, legal values are be between 0 and %d.\n", 
                type, SECTION_TYPES_COUNT);
            return -1; 
        }

        return index; 
    }

    /**
     ** This returns the JVM ZVision machine state from the type_stable shared memory file.
     **/
    static inline volatile IPC_STATE_E ipc_state(TYPE_STABLE_LAYOUT_T* ts)
    {
        if (type_stable_valid(ts) != true)
        {
            PATHOLOGICAL("ipc_state() called but type_stable_valid() returns false.\n");
            return ipc_state_pathological; 
        }

        return ts->ipc_state;        

    } // end of ipc_state().


    /**
     ** 
     **/
    static const char* xlate_ipc_state(IPC_STATE_E   state)
    {
        static const char* error_str = "ERROR:UNKNOWN_ZVISION_IPC_STATE";

        switch (state)
        {
        case ipc_state_uninitialized:
            return "JVM Uninitialized";
            break;
        case ipc_state_initializing:
            return "JVM Initializing";
            break;
        case ipc_state_no_connection:
            return "JVM running without connection to Host Agent";
            break;
        case ipc_state_ignoring_host_agent:
            return "JVM proceeding without Host Agent";
            break;
        case ipc_state_got_connect:
            return "JVM received connection from Host Agent";
            break;
        case ipc_state_got_start:
            return "JVM got start_token from Host Agent";
            break;
        case ipc_state_connection_est:
            return "JVM communicating with Host Agent";
            break;
        case ipc_state_exiting:
            return "JVM Exiting";
            break;
        case ipc_state_exited:
            return "JVM Exited";
            break;
        case ipc_state_pathological:
            return "JVM Pathological";
            break;
        default:
            PATHOLOGICAL("Invalid IPC_STATE_E enum (0x%lX) encountered.\n", state);
            return error_str;
            break;
        }
    
        ShouldNotReachHere();
        return "JVM Pathological";
    
    } // end of xlate_ipc_state().
    

    /**
     ** 
     **/
    static const char* xlate_access_user(ACCESS_USER_E   user)
    {
        static const char* error_str = "ERROR:UNKNOWN_ZVISION_ACCESS_USER";

        switch (user)
        {
        case user_is_jvm:
            return "JVM";
            break;
        case user_is_host_agent:
            return "HA";
            break;
        default:
            PATHOLOGICAL("Invalid ACCESS_USER_E enum (%ld) encountered.\n", user);
            return error_str;
            break;
        }
    
        ShouldNotReachHere();
    
    } // end of xlate_access_user().
    

    /**
     ** 
     **/
    static inline const char* xlate_section_type(SECTION_TYPE_E   type)
    {
        static const char* error_str = "ERROR:UNKNOWN_ZVISION_SECTION_TYPE";
        const CONSTANTS_T* constants = constants_by_type(type);

        if (constants == NULL) { return error_str; }
        if (::strnlen(constants->name, SECTION_TYPE_STR_LEN) >= SECTION_TYPE_STR_LEN)
        { 
            PATHOLOGICAL("Section Type (%d) name string \"%*s\"[truncated] exceeds name length limit (%ld).\n",
                type, (int) SECTION_TYPE_STR_LEN, constants->name, SECTION_TYPE_STR_LEN);
            return error_str; 
        }

        return constants->name;
    
    } // end of xlate_section_type().
    
    
    /**
     **  Called by the JVM or Host Agent unit tests, when using the "-f (forgo unit test)" parameter.
     **/
    static SECTION_TYPE_E  xlate_type_string_to_type(const char* type_string)
    {
        // Do a linear search of all section type strings for a match.
        for (size_t index = 0; ((int) index) <= type_index(end_token); index++)
        {
            if (::strcmp(type_string, _constants_by_type[index].name) == 0) 
            { 
                return (SECTION_TYPE_E) (index + start_token); 
            }
        }
        
        return no_token;
    }
    
    /**
     ** 
     **/
    static const char* xlate_file_type(FILE_TYPE_E   file_type)
    {
        static const char* error_str = "ERROR:UNKNOWN_ZVISION_FILE_TYPE";

        switch(file_type)
        {
        case proprietary_file:
            return "Proprietary File Type";
            break;
        case request_response_file:
            return "Request Response File Type";
            break;
        case indexed_buffer_file:
            return "Indexed Buffer File Type";
            break;
        default:
            PATHOLOGICAL("Invalid file type enum (%d) encountered.\n", file_type);
            return error_str;
            break;
        }
    
        ShouldNotReachHere();
    
    } // end of xlate_file_type().
    
    

    /**
     ** ZVision::check_type - Check that the type supplied is within the enum range.
     **
     ** RETURNS: true, if the type is within bounds, false otherwise.
     ** 
     ** ON ERROR: 
     **
     ** NOTES: 
     **/
    static inline bool check_type(SECTION_TYPE_E type)
    {
        if (type < start_token || type > end_token)
        {
            PATHOLOGICAL("Illegal type %d.\n", type);
            return false;
        }

        return true;

    } // end of check_type().


    /**
     ** Return the address of the current and earliest supported version structures.
     **/
    static const VERSION_INFO_T* earliest_sw_ver() { return &_earliest_sw_ver; }
    static const VERSION_INFO_T* current_sw_ver()  { return &_current_sw_ver;  }

    /**
     ** Covnert the hard coded current and earliest supported version information into single number values.
     **/
    static VERNO earliest_sw_verno() { return verno((VERSION_INFO_T*) earliest_sw_ver()); }
    static VERNO current_sw_verno()  { return verno((VERSION_INFO_T*) current_sw_ver());  }

    /**
     ** The type_stable file's version number.
     **/
    static VERNO type_stable_verno(TYPE_STABLE_LAYOUT_T* ts)        
    { 
        DIAGOUT(4, "type_stable_verno(%p) called.\n", ts);

        if (type_stable_valid(ts) != true) { return 0; } else { return verno((VERSION_INFO_T*) &ts->version);  } 
    }

    /**
     ** verno - Convert a VERSION_INFO_T entry into a single value, which can be used
     ** for comparision against other version numbers.
     **
     ** RETURNS: 
     ** 
     ** ON ERROR:  Returns zero.
     **
     ** NOTES: 
     **/
    static VERNO  verno(VERSION_INFO_T* version)
    {
        VERNO  num = 0;

        if (version == NULL) 
        { 
            PATHOLOGICAL("Illegal (NULL) version pointer.\n");  
            return 0; 
        }

        num += (version->revision * revision_num_multiplier());
        num += (version->minor    * minor_num_multiplier());
        num += (version->major    * major_num_multiplier());
        num += (version->release  * release_num_multiplier());

        DIAGOUT(5, "verno(%p) called, result was %lu.\n", version, num);

        return num;

    } // end of verno().


    /**
     ** convert_verno - Convert a VERNO entry into a VERSION_INFO_T struct.
     **
     ** RETURNS: true.
     ** 
     ** ON ERROR:  Returns false.
     **
     ** NOTES: 
     **/
    static inline bool convert_verno(VERNO verno, VERSION_INFO_T* version)
    {
        DIAGOUT(4, "convert_verno(%lu, %p) called.\n", verno, version);

        size_t num = (size_t) verno;

        version->release  = num / release_num_multiplier();
        num              -= (version->release * release_num_multiplier());

        version->major    = num / major_num_multiplier();
        num              -= (version->major * major_num_multiplier());

        version->minor    = num / minor_num_multiplier();
        num              -= (version->minor * minor_num_multiplier()); 

        version->revision = num;

        if (check_version(version) != true)
        { 
            char buf[RESULT_SIZE] = { 0 };

            xlate_version_struct(version, buf);
            PATHOLOGICAL("convert_verno(%lu) failed, version is \"%s\".\n", verno, buf);
            return false; 
        }
        
        return true;

    } // end of convert_verno().


    /**
     ** ZVision::check_version - Check that the version is within the supported range of releases.
     **
     ** RETURNS: true, if the version is equal to or later than the earliest release, and
     ** equal to or sooner than the current release, false otherwise.
     ** 
     ** ON ERROR: 
     **
     ** NOTES: 
     **/
    static bool check_version(VERSION_INFO_T* version)
    {
        DIAGOUT(4, "check_version(%p) called.\n", version);

        if (version == NULL) 
        { 
            PATHOLOGICAL("Illegal (NULL) version pointer.\n");  
            return false; 
        }

        ssize_t release  = (ssize_t) version->release;
        ssize_t major    = (ssize_t) version->major;
        ssize_t minor    = (ssize_t) version->minor;
        ssize_t revision = (ssize_t) version->revision;

        ZV_ASSERT(release  >= 0);  
        ZV_ASSERT(major    >= 0);
        ZV_ASSERT(minor    >= 0);  
        ZV_ASSERT(revision >= 0);

        if (release >= MAX_RELEASE_NUM)
        { 
            DIAGOUT(2, "Illegal release value %lu, exceeds limit of %lu.\n", release, MAX_RELEASE_NUM);  
            return false; 
        }

        if (major >= MAX_MAJOR_NUM)
        { 
            DIAGOUT(2, "Illegal major value %lu, exceeds limit of %lu.\n", major, MAX_MAJOR_NUM);    
            return false; 
        }

        if (minor >= MAX_MINOR_NUM)
        { 
            DIAGOUT(2, "Illegal minor value %lu, exceeds limit of %lu.\n", minor, MAX_MINOR_NUM);    
            return false; 
        }

        if (revision >= MAX_REVISION_NUM)
        { 
            DIAGOUT(2, "Illegal revision value %lu, exceeds limit of %lu.\n", revision, MAX_REVISION_NUM); 
            return false; 
        }

        static VERNO earliest = earliest_sw_verno();
        static VERNO current  = current_sw_verno();

        VERNO this_verno = verno(version);
    
        if (this_verno > current)
        {
            DIAGOUT(2, "The version number is later than the currently compiled version number.\n"); 
            return false; 
        }

        if (this_verno < earliest)
        {
            DIAGOUT(2, "The version number is newer than the earliest supported version number.\n");
            return false; 
        }

        return true;

    } // end of check_version().


    /**
     ** xlate_version_struct() - convert a VERSION_INFO_T* into a user readable string supplied by the caller.
     ** The caller must supply a pointer to a null terminated character buffer of at least <MAX_VERS_INFO_STR_LEN>
     ** characters that has a zero string length (there must be a zero/null-terminator), along with the version
     ** information (release, major, minor, and revision numbers) to be translated.
     **
     ** No result will include a carriage return or new line character within the string.
     **
     ** RETURNS: The routine will return the user supplied pointer value of the buffer passed to it, with the 
     ** translated (human readable), null terminated string value of the current version within it.
     ** 
     ** ON ERROR: In the case of errors, a pointer to a static, null-terminated error string will be returned containing
     ** a brief description of the reason for failure.
     **
     ** NOTES:  May not be called by time critical logic paths.
     **/
    static char* xlate_version_struct(VERSION_INFO_T* version, char* buf)
    {
        if (version == NULL)
        {
            PATHOLOGICAL("Null version struct pointer.\n");
            return (char*) _invalid_buf_str; 
        }

        if (version->release > MAX_RELEASE_NUM)
        {
            PATHOLOGICAL("Version struct (%p) release value (%lu) exceeds limit (%lu).\n", 
                version, version->release, MAX_RELEASE_NUM);
            return (char*) _invalid_ver_str; 
        }
         
        if (version->major > MAX_MAJOR_NUM)
        {
            PATHOLOGICAL("Version struct (%p) major value (%lu) exceeds limit (%lu).\n",
                version, version->release, MAX_MAJOR_NUM);
            return (char*) _invalid_ver_str; 
        }
         
        if (version->minor > MAX_MINOR_NUM)
        {
            PATHOLOGICAL("Version struct (%p) minor value (%lu) exceeds limit (%lu).\n",
                version, version->release, MAX_MINOR_NUM);
            return (char*) _invalid_ver_str; 
        }
         
        if (version->release > MAX_REVISION_NUM)
        {
            PATHOLOGICAL("Version struct (%p) revision value (%lu) exceeds limit (%lu).\n",
                version, version->revision, MAX_REVISION_NUM);
            return (char*) _invalid_ver_str; 
        }
         
        if (buf == NULL)
        {
            PATHOLOGICAL("Null buf pointer.\n");
            return (char*) _invalid_buf_str; 
        }

        sprintf(buf, "%lu.%lu.%lu.%lu%c", version->release, version->major, version->minor, version->revision, 0);
    
        return buf; 

    } // end of xlate_version_struct().


    /**
     ** ZVision::xlate_version_data - Convert the version info into a user readable string supplied by the caller.
     ** The caller must supply a pointer to a null terminated character buffer of at least <MAX_VERS_INFO_STR_LEN> characters
     ** that has a zero string length (there must be a zero/null-terminator), along with the version
     ** information (release, major, minor, and revision numbers) to be translated.
     **
     ** No result will include a carriage return or new line character within the string.
     **
     ** RETURNS: The routine will return the user supplied pointer value of the buffer passed to it, with the 
     ** translated (human readable), null terminated string value of the current version within it.
     ** 
     ** ON ERROR: In the case of errors, a pointer to a static, null-terminated error string will be returned containing
     ** a brief description of the reason for failure.
     **
     ** NOTES:  May not be called by time critical logic paths.
     **/
    static inline char* xlate_version_data(size_t release, size_t major, size_t minor, size_t revision, char* buf)
    {
        VERSION_INFO_T   version;

        if (buf == NULL)
        {
            PATHOLOGICAL("Null buf pointer.\n");
            return (char*) _invalid_buf_str; 
        }

        version.release  = release;
        version.major    = major;
        version.minor    = minor;
        version.revision = revision;

        char* result = xlate_version_struct(&version, buf);

        return result;

    } // end of xlate_version_data().
    
    
    /**
     ** ZVision::xlate_verno - Convert a VERNO into a printable version string.
     **
     ** RETURNS: 
     ** 
     ** ON ERROR:  Returns zero.
     **
     ** NOTES: 
     **/
    static inline char* xlate_verno(VERNO verno, char* buf)
    {
        size_t release;
        size_t major;
        size_t minor;
        size_t revision;
        size_t num;

        release   = verno / release_num_multiplier();
        num       = release * release_num_multiplier();
        verno    -= num;

        major     = verno / major_num_multiplier();
        num       = major * major_num_multiplier();
        verno    -= num;

        minor     = verno / minor_num_multiplier();
        num       = minor * minor_num_multiplier();
        verno    -= num;

        revision  = verno;

        return xlate_version_data(release, major, minor, revision, buf);

    } // end of xlate_verno().


    /**
     ** ZVision::section_init_check  - Check that the secion is uninitialized.
     **
     ** RETURNS: true, if the section is available but unitialized, false otherwise.
     ** 
     ** ON ERROR: 
     **
     ** NOTES: 
     **/
    static bool section_init_check(TYPE_STABLE_LAYOUT_T* ts, SECTION_TYPE_E type)
    {
        if (type_stable_valid(ts) != true)
        {
            PATHOLOGICAL("type_stable pointer was NULL.\n");
            return false;
        }

        DIAGOUT(4, "section_init_check(%p, \"%s\") called.\n", ts, xlate_section_type(type));

        int            index   = type_index(type);

        if (index < 0) { return false; }

        SECTION_ENTRY_T*  section = section_entry_by_type(ts, type);

        if (section->file.max_size != 0)     
        { 
            PATHOLOGICAL("Non-zero (%lu) max_size in \"%s\" section[%d].\n", 
                section->file.max_size, xlate_section_type(type_stable), index);
            return false;
        }

        if (section->file.active != false) 
        { 
            PATHOLOGICAL("Active flag (%d) in \"%s\" section[%d] not false.\n", 
                section->file.active, xlate_section_type(type_stable), index);
            return false;
        }

        if (section->file.allocated != false) 
        { 
            PATHOLOGICAL("Allocated flag (%d) in \"%s\" section[%d].\n",
                section->file.allocated, xlate_section_type(type_stable), index);
            return false;
        }

        return true;

    } // end of section_init_check().

    /**
     **  Count of how my type/version entries there are in the array.
     **/
    static const size_t  type_version_cnt() { return _type_version_cnt; }
    
#define WRONG_ELEM(entry) { PATHOLOGICAL("Error at entry %lu in TYPE_VERSION_ENTRY_T table.\n", entry); }

    /**
     ** ZVision::check_type_version_entry - Check that the _type_version_table entry is sensible.
     **
     ** RETURNS: 
     ** 
     ** ON ERROR: 
     **
     ** NOTES: 
     **/
    static bool check_type_version_entry(TYPE_VERSION_ENTRY_T* ver, size_t element)
    {
        DIAGOUT(4, "check_type_version_entry(%p, %lu) called.\n", ver, element);

        char            result[RESULT_SIZE] = { 0 };         // Automatiic stack variable for flagbit translation.
  
        VERSION_INFO_T*          version           = &ver->version;
        SECTION_TYPE_E           type              = ver->type;
        size_t                   max_size          = ver->max_size;
        uint64_t                 alignment         = ver->alignment;
        MAGIC_COOKIE_T           version_cookie    = ver->cookie;
        INIT_BUF_METHOD_T        init_buf_func     = ver->jvm_funcs.init_buf_func;
        PRELOAD_BUF_METHOD_T     preload_buf_func  = ver->jvm_funcs.preload_buf_func;
        MAGIC_COOKIE_T           jvm_funcs_cookie  = ver->jvm_funcs.cookie;
        COMPRESS_BUF_METHOD_T    compress_buf_func = ver->ha_funcs.compress_buf_func;
        MAGIC_COOKIE_T           ha_funcs_cookie   = ver->ha_funcs.cookie;

        xlate_version_struct(version, result);
 
        DIAGOUT(3, "Entry %lu (%p)\n",               element, ver);
        DIAGOUT(3, "           type = %s\n",           xlate_section_type(type));
        DIAGOUT(3, "        version = %s\n",           result);
        DIAGOUT(3, "       max size = %ld (0x%0lX)\n", max_size, max_size);
        DIAGOUT(3, "      alignment = %ld (0x%0lX)\n", alignment, alignment);
        DIAGOUT(3, " version cookie = 0x%0lX)\n",      (uint64_t) version_cookie);
        DIAGOUT(3, "      init func = 0x%0lX)\n",      (uint64_t) init_buf_func);
        DIAGOUT(3, "  compress func = 0x%0lX)\n",      (uint64_t) compress_buf_func);
        DIAGOUT(3, "jvm func cookie = 0x%0lX)\n",      (uint64_t) jvm_funcs_cookie);
        DIAGOUT(3, "   preload func = 0x%0lX)\n",      (uint64_t) preload_buf_func);
        DIAGOUT(3, " ha func cookie = 0x%0lX)\n",      (uint64_t) ha_funcs_cookie);

        if (version_cookie != COOKIE_TYPE_VERSION)
        {
            WRONG_ELEM(element);
            PATHOLOGICAL("Illegal type version cookie.\n");
            return false;
        }
        
        if (jvm_funcs_cookie != COOKIE_JVM_FUNCS)
        {
            WRONG_ELEM(element);
            PATHOLOGICAL("Illegal type version JVM function refs cookie.\n");
            return false;
        }
        
        if (ha_funcs_cookie != COOKIE_HA_FUNCS)
        {
            WRONG_ELEM(element);
            PATHOLOGICAL("Illegal type version Host Agent function refs cookie.\n");
            return false;
        }
        
        if (check_version(version) != true)
        {
            WRONG_ELEM(element);
            PATHOLOGICAL("Illegal version.\n");
            return false;
        }

        if (check_type(type) != true) 
        {
            WRONG_ELEM(element);
            PATHOLOGICAL("Illegal type.\n");
            return false;
        }

        int index = type_index(type);

        if (index < 0 || index >= ((int) _section_types_cnt))
        {
            WRONG_ELEM(element);
            PATHOLOGICAL("Illegal index.\n");
            return false;
        }

        if (max_size <= 0)
        {
            const CONSTANTS_T* constants = constants_by_type(type);

            if (constants == NULL)
            {
                WRONG_ELEM(element);
                PATHOLOGICAL("NULL constants pointer for type \"%s\".\n", xlate_section_type(type));
                return false;
            }

            if (constants->file_type != indexed_buffer_file)
            {
                WRONG_ELEM(element);
                PATHOLOGICAL("Illegal size %lu.\n", max_size);
                return false;
            }
        }

        if (init_buf_func == NULL)
        {
            WRONG_ELEM(element);
            PATHOLOGICAL("Illegal initialiazation function.\n");
            return false;
        }

        if (preload_buf_func == NULL)
        {
            WRONG_ELEM(element);
            PATHOLOGICAL("Illegal preload function.\n");
            return false;
        }

        if (compress_buf_func == NULL)
        {
            WRONG_ELEM(element);
            PATHOLOGICAL("Illegal compress function.\n");
            return false;
        }

        return true;

    } // end of check_type_version_entry().


    /**
     **  Count of how my entries there are in the cross reference array.
     **/
    static const size_t  cross_ref_cnt() { return _cross_ref_cnt; }


    /**
     ** ZVision::create_xref_entry - Allocate memory for a CROSS_REF_ENTRY_T object, and fill it in with
     ** the information supplie by the "entry" parameter.
     **
     ** RETURNS: the address of the new object, or NULL if it encountered an error.
     ** 
     ** ON ERROR: 
     **
     ** NOTES: 
     **
     **/
    static CROSS_REF_ENTRY_T*   create_xref_entry(CROSS_REF_ENTRY_T** xref_addr, TYPE_VERSION_ENTRY_T* entry)
    {
        DIAGOUT(4, "create_xref_entry(%p, %p) called.\n", xref_addr, entry);

        if (entry == NULL)
        {
            char version_str[MAX_VERS_INFO_STR_LEN] = { 0 };

            xlate_version_struct(&entry->version, &version_str[0]);
            PATHOLOGICAL("Null entry given in create_xref_entry().\n");
            return NULL;
        }

        if (xref_addr == NULL)
        {
            char version_str[MAX_VERS_INFO_STR_LEN] = { 0 };

            xlate_version_struct(&entry->version, &version_str[0]);
            PATHOLOGICAL("Null cross reference pointer address given in create_xref_entry().\n");
            return NULL;
        }

        if (*xref_addr != NULL)
        {
            char version_str[MAX_VERS_INFO_STR_LEN] = { 0 };

            xlate_version_struct(&entry->version, &version_str[0]);
            PATHOLOGICAL("Cross reference pointer is non-NULL (0x%lX) in create_xref_entry().\n", (uint64_t) *xref_addr);
            return NULL;
        }

        CROSS_REF_ENTRY_T* new_xref = (CROSS_REF_ENTRY_T*) malloc(sizeof(CROSS_REF_ENTRY_T));

        if (new_xref == NULL)
        {
            char version_str[MAX_VERS_INFO_STR_LEN] = { 0 };

            xlate_version_struct(&entry->version, &version_str[0]);
            PATHOLOGICAL("malloc(%lu) failed processing entry type \"%s\" (%d), version \"%s\" in the TYPE_VERSION_ENTRY_T table.\n",
                sizeof(CROSS_REF_ENTRY_T), xlate_section_type(entry->type), entry->type, version_str);   
            return NULL;
        }

        new_xref->verno                      = verno(&entry->version);
        new_xref->type                       = entry->type;
        new_xref->max_size                   = entry->max_size;
        new_xref->alignment                  = entry->alignment;
        new_xref->notify_lmt                 = entry->notify_lmt;
        new_xref->unit_size                  = entry->unit_size;
        new_xref->unit_cnt                   = entry->unit_cnt;
        new_xref->jvm_funcs.init_buf_func    = entry->jvm_funcs.init_buf_func;
        new_xref->jvm_funcs.preload_buf_func = entry->jvm_funcs.preload_buf_func;
        new_xref->jvm_funcs.cookie           = entry->jvm_funcs.cookie;
        new_xref->ha_funcs.compress_buf_func = entry->ha_funcs.compress_buf_func;
        new_xref->ha_funcs.cookie            = entry->ha_funcs.cookie;
        new_xref->next                       = NULL;
        new_xref->cookie                     = COOKIE_CROSS_REF;

        *xref_addr = new_xref;   // Now update the orginal (NULL) pointer with the new entry.

        DIAGOUT(3, "malloc'ed %lu bytes for CROSS_REF_ENTRY_T.\n", sizeof(CROSS_REF_ENTRY_T));

        return new_xref;

    }  // end of create_xref_entry().


    /**
     ** ZVision::insert_xref_entry - Add an entry into the ascending version number cross referenced linked list (for that 
     ** SECTION_TYPE_E) before the next greater entry.
     **
     ** It is passed a pointer to the linked list CROSS_REF_ENTRY_T* for the appropriate SECTION_TYPE_E, and a pointer
     ** to the TYPE_VERSION_ENTRY_T entry.
     **
     ** RETURNS: the address of the new object, or NULL if it encountered an error.
     ** 
     ** ON ERROR: 
     **
     ** NOTES: 
     **
     **/
    static CROSS_REF_ENTRY_T*   insert_xref_entry(CROSS_REF_ENTRY_T** xref_addr, TYPE_VERSION_ENTRY_T* entry)
    {
        DIAGOUT(4, "insert_xref_entry(%p, %p) called.\n", xref_addr, entry);

        CROSS_REF_ENTRY_T* new_xref = NULL; 

        if (entry == NULL)
        { 
            PATHOLOGICAL("Error encountered building CROSS_REF_ENTRY_T table, NULL method version entry.\n"); 
            return NULL; 
        }

        if (xref_addr == NULL)
        { 
            PATHOLOGICAL("Error encountered building CROSS_REF_ENTRY_T table, NULL cross reference entry.\n"); 
            return NULL; 
        }

        // Calculate the index into the cross reference table based on segment type.
        int index = type_index(entry->type);

        if (index < 0)
        {
            char version_str[MAX_VERS_INFO_STR_LEN] = { 0 };

            xlate_version_struct(&entry->version, &version_str[0]);
            PATHOLOGICAL("Error while building CROSS_REF_ENTRY_T table, type \"%s\"(%d), version \"%s\".\n", 
                    xlate_section_type(entry->type), entry->type, version_str);   
            return NULL; 
        }

        // Make sure the method version makes sense...
        // TODO: Make sure this call is right.
        if (check_version(&entry->version) != true)
        {
            char version_str[MAX_VERS_INFO_STR_LEN] = { 0 };

            xlate_version_struct(&entry->version, &version_str[0]);
            PATHOLOGICAL("Error while building CROSS_REF_ENTRY_T table, type \"%s\"(%d), version \"%s\".\n", 
                    xlate_section_type(entry->type), entry->type, version_str);   
            return NULL; 
        }

        // If there are no previous entries for this, then create the first one.
        if (*xref_addr == NULL) 
        {
            new_xref = create_xref_entry(xref_addr, entry);

            if (new_xref == NULL) 
            { 
                char version_str[MAX_VERS_INFO_STR_LEN] = { 0 };

                xlate_version_struct(&entry->version, &version_str[0]);
                PATHOLOGICAL("Error while building CROSS_REF_ENTRY_T table, type \"%s\"(%d), version \"%s\".\n", 
                    xlate_section_type(entry->type), entry->type, version_str);   
                return NULL; 
            }

            // TODO:  Figure out how to add this in...
            // Sinced this is the first entry in the linked list for this type, we can
            // bump the _active_cnt variable in the type_stable header.
            // _ts->_active_cnt++;

            if (new_xref != *xref_addr)
            {
                PATHOLOGICAL("new_xref != *xref_addr.\n");
                return NULL;
            }

            return new_xref;
        }

        // We'll need to compare this to other cross reference entries, so calculate the entry version number.
        VERNO  entry_verno = verno(&entry->version);

        // Search the version info ordered linked list of entries for this particular segment type, looking
        // for the appropriate place to insert the entry.
        while (*xref_addr != NULL)
        {
            if (entry->type != (*xref_addr)->type) 
            {
                char version_str[MAX_VERS_INFO_STR_LEN] = { 0 };

                xlate_version_struct(&entry->version, &version_str[0]);
                PATHOLOGICAL("Error encountered building CROSS_REF_ENTRY_T table, type \"%s\"(%d), version \"%s\".\n", 
                    xlate_section_type(entry->type), entry->type, version_str);   
                return NULL;
            }

            if ((*xref_addr)->cookie != COOKIE_CROSS_REF)
            {
                char version_str[MAX_VERS_INFO_STR_LEN] = { 0 };

                xlate_version_struct(&entry->version, &version_str[0]);
                PATHOLOGICAL("Invalid token 0x%lX, found building CROSS_REF_ENTRY_T table, type \"%s\"(%d), version \"%s\".\n", 
                    (*xref_addr)->cookie, xlate_section_type(entry->type), entry->type, version_str);   
                return NULL;
            }

            // Is the version of the new entry less than the current xref linked list version...
            if (entry_verno <= (*xref_addr)->verno)
            {
                // We can't allow duplicate entries.
                if (entry_verno == (*xref_addr)->verno)
                {
                    char version_str[MAX_VERS_INFO_STR_LEN] = { 0 };

                    xlate_version_struct(&entry->version, &version_str[0]);
                    PATHOLOGICAL("Duplicate entry for type \"%s\"(%d), version \"%s\" in TYPE_VERSION_ENTRY_T table.\n",
                        xlate_section_type(entry->type), entry->type, version_str);   
                    return NULL;
                }
                  
                break;  // Drop out of the search, we've found our insertion point.
            }

            xref_addr = &(*xref_addr)->next;

        } // end of while (*xref_addr != NULL)
        
        // If we've gotten here it is either because we are at the end of the linked list, of at the point in the list
        // where we want to insert the object.  In either case, the operation is the same, but we need to first capture
        // the next pointer of the insertion point entry.  The create_xref_entry() call will update the xref_addr entry
        // with the new addres, but we also need to set the next pointer to the original entry.

        CROSS_REF_ENTRY_T* temp = *xref_addr;

        new_xref = create_xref_entry(xref_addr, entry);

        if (new_xref == NULL)
        {
            char version_str[MAX_VERS_INFO_STR_LEN] = { 0 };

            xlate_version_struct(&entry->version, &version_str[0]);
            PATHOLOGICAL("create_xref_entry failed for type \"%s\"(%d), version \"%s\" in TYPE_VERSION_ENTRY_T table.\n",
                xlate_section_type(entry->type), entry->type, version_str);   
            return NULL; 
        }

        new_xref->next = temp;
        
        if (*xref_addr != new_xref)
        {
            PATHOLOGICAL("new_xref != *xref_addr.\n");
            return NULL;
        }

        return new_xref;

    } // end of insert_xref_entry().



    /**
     ** ZVision::get_method_info - Check that the type & version supplied exists within the method cross reference
     ** table and points to a supported entry.
     **
     ** RETURNS: The address of the entry if there is a matching entry within the cross reference table, or NULL otherwise.
     ** 
     ** ON ERROR: 
     **
     ** NOTES: Linear search, sort of. We dont want to do this very often, it isn't very fast.
     **/
    const static CROSS_REF_ENTRY_T*   get_method_info(SECTION_TYPE_E type, VERSION_INFO_T* version)
    {
        DIAGOUT(4, "get_method_info(\"%s\", %p) called.\n", xlate_section_type(type), version);

        if (check_type(type) != true || check_version(version) != true) 
        { 
            PATHOLOGICAL("get_method_info() inputs invalid.\n");
            return NULL; 
        }
 
        // Calculate the index into the cross reference table based on segment type.
        CROSS_REF_ENTRY_T* xref = xref_data_by_type(type); 

        if (xref == NULL)
        {
            PATHOLOGICAL("xref_data_by_type(\"%s\" %d) returned NULL.\n", xlate_section_type(type), type);
            return NULL;
        }

        // Calculate the version number of the version to compare against.
        VERNO  this_verno = verno(version);

        // Search the linked list (of this particular segment type) for an entry
        // whose version information is supported by the specified version. (i.e the
        // verno value is equal to or greater than the entry, and less than the next
        // entry in the linked list.)
        while (xref != NULL)
        {
            if (xref->cookie != COOKIE_CROSS_REF)
            {
                PATHOLOGICAL("get_method_info() failed, ecountered  corrupted cross-reference entry.\n");
                return NULL;
            }

            if (type != xref->type)
            { 
                PATHOLOGICAL("get_method_info() failed, ecountered type \"%s\" (%d) when expecting type \"%s\" (%d).\n",
                    xlate_section_type(type), type, xlate_section_type(xref->type), xref->type);
                return NULL; 
            }

            // Is the version in question greater than or equal to this entry...
            if (this_verno >= xref->verno)
            {
                // If there wasn't another entry of the same type (ie. we are EOL), then use this entry.
                if (xref->next == NULL) { return xref; }

                // And if this version is less than then next entry's version (i.e. in between
                // the two cross reference entries) then use the current cross reference entry.
                if (this_verno < xref->next->verno) { return xref; }
            } 
            
            // If there is another cross reference entry within the linked list,
            // it should be larger than the current one.
            if (xref->next && xref->next->verno <= xref->verno)
            {
                PATHOLOGICAL("failed for type \"%s\", next verno (%lu) less than or equal to current verno (%lu).\n",
                    xlate_section_type(type), xref->next->verno, xref->verno);
                return NULL; 
            }

            xref = xref->next;
        }

        char version_str[MAX_VERS_INFO_STR_LEN] = { 0 };
        
        xlate_version_struct(version, &version_str[0]);

        DIAGOUT(0, "Could not find appropriate method supporting type \"%s\", version \"%s\".\n",
            xlate_section_type(type), version_str);
            
        return NULL; 

    } // end of get_method_info().
    
    
  
    /**
     ** lseek_file() - 
     **
     ** RETURNS: true if successful, false otherwise.
     **
     ** ON ERROR: returns false, PATHOLOGICAL error is set.
     **
     ** NOTES: 
     **
     **/
    static bool lseek_file(int fd, off_t offset, int whence)
    {
        DIAGOUT(4, "Issuing lseek(fd=%d, offset=%lu, flags=0x%X)...\n", fd, offset, whence);

        // Force errno to zero, in case the lseek() returns something we weren't expecting.
        errno    = 0;

        off_t rc = lseek(fd, offset, whence);

        if (rc == (off_t) -1)
        {
            ERRNO_RESULT_T   failure = errno;
    
            DIAGOUT(0, "lseek(fd=%d, 0, SEEK_SET) failed, errno set to %d (%s).\n", 
              fd, failure, ::strerror(failure)); 

            switch (failure)
            {
            case EBADF:
                DIAGOUT(0, "Invalid File Descriptor (%d) specified.\n", fd);
                break;
            case EINVAL:
                DIAGOUT(0, "Invalid \"whence\" parameter (0x%X) specified.\n", whence);
                break;
            case EOVERFLOW:
                DIAGOUT(0, "Invalid \"offset\" parameter (%lu) specified.\n", offset);
                break;
            case ESPIPE:
                DIAGOUT(0, "File descriptor (%d) is associated with a pipe, socket, or FIFO.\n", fd);
                break;
            default:
                DIAGOUT(0, "Unexpected error \"%s\" (%d) returned.\n", ::strerror(failure), failure);
                break;
            }

            return false;
        }

        return true;

    }  // end of lseek_file().

  
    /**
     ** read_file() - 
     **
     ** RETURNS: number of bytes read, or -1, and supplied errno is set.
     **
     ** ON ERROR:  
     **
     ** NOTES: 
     **
     **/
    static inline ssize_t read_file(SOCKET_FD_T fd, void* buffer, ssize_t len, ERRNO_RESULT_PTR_T failure)
    {
#define READ_RETRY_SLEEP             (1000 * 1000)
#define LIMIT_OF_EINTR_READ_ATTEMPTS  100
#define LIMIT_OF_EAGAIN_READ_ATTEMPTS 100

        char*   volatile addr            = (char*) buffer;
        ssize_t volatile expected        = len;
        ssize_t volatile total_read      = 0;
        ssize_t volatile bytes_read      = 0;
        size_t  volatile eintr_attempts  = 0;
        size_t  volatile eagain_attempts = 0;

        struct timespec time_spec;

        DIAGOUT(4, "read_file(fd %d, addr %p, len %lu, failure ptr %p)...\n", fd, addr, len, failure);

        if (fd < 0) 
        { 
            PATHOLOGICAL("fd parameter (%d) invalid.\n", fd);
            return -1; 
        }

        if (addr == NULL) 
        { 
            PATHOLOGICAL("addr parameter was NULL.\n");
            return -1; 
        }

        if (len < 1) 
        { 
            PATHOLOGICAL("length parameter (%lu) is less than 1.\n", len);
            return -1; 
        }

        if (failure == NULL) 
        { 
            PATHOLOGICAL("failure parameter was NULL.\n");
            return -1; 
        }

        // This loop is only so that we can retry certain error conditions.
        while (true)
        {
            errno      = 0;
            bytes_read = ::read(fd, addr, expected);
            *failure   = errno;
    
            if (bytes_read > expected) 
            {
                PATHOLOGICAL("::read(fd %d, addr %p, expected %ld) returned %ld bytes!\n", fd, addr, expected, bytes_read);
                return -1;
            }
            
            // It is possible for some versions of Linux to get an EINTR and still read some 
            // partial number of bytes?  That seems to be what the POSIX spec suggests.  This
            // should cope with that.
            if (bytes_read >= 0) 
            {
                total_read += bytes_read;
                addr       += bytes_read;
                expected   -= bytes_read;

                // We read less bytes than we were expecting, but there was no error.
                if (*failure == 0)
                {
                    if (total_read < len)
                    {
                        DIAGOUT(0, "read %lu bytes on fd %d but expected %lu.\n", total_read, fd, len);
                    }

                    return total_read;
                }
            }

            // Even if we we're interrupted, if we've gotten everything
            // we think we should have read, just return.
            if (total_read >= len) { return total_read; }

            if (*failure == EAGAIN || *failure == EWOULDBLOCK)
            {
                if (++eagain_attempts < LIMIT_OF_EAGAIN_READ_ATTEMPTS)
                {
                    DIAGOUT(0, "Got EAGAIN on ::read() of fd %d.\n", fd);
                    time_spec.tv_sec   = 0;
                    time_spec.tv_nsec  = READ_RETRY_SLEEP;
                    ::nanosleep(&time_spec, NULL);
                    continue;
                }

                DIAGOUT(2, "Exceeded EAGAIN limit (%lu) on ::read() of fd %d.\n", eagain_attempts, fd);
                break;
            }

            // The system interrupted this call in mid-stream.  POSIX allows a read() that can be interrupted
            // after reading some data.   We need to keep trying, but at some point, we should give up.
            if (*failure == EINTR)
            {
                if (++eintr_attempts < LIMIT_OF_EINTR_READ_ATTEMPTS)
                {
                    DIAGOUT(2, "Got EINTR on ::read() of fd %d.\n", fd);
                    time_spec.tv_sec   = 0;
                    time_spec.tv_nsec  = READ_RETRY_SLEEP;
                    ::nanosleep(&time_spec, NULL);
                    continue;
                }

                DIAGOUT(2, "Exceeded EINTR limit (%lu) on ::read() of fd %d.\n", eintr_attempts, fd);
                break;
            }

        }  // end of while(true)

        DIAGOUT(3, "read_file(fd %d, addr %p, len %lu) returned %lu bytes after %lu attempts, last errno was \"%s\" (%d).\n", 
            fd, addr, len, total_read, eintr_attempts + eagain_attempts, ::strerror(*failure), *failure);

        switch (*failure)
        {
        case EAGAIN:
            DIAGOUT(0, "Read from fd (%d) would have blocked, read only %lu bytes, expected %lu.\n", fd, total_read, len);
            return total_read;
            break;
        case EINTR:
            DIAGOUT(0, "Read from fd (%d) was interrupted, read only %lu bytes, expected %lu.\n", fd, total_read, len);
            return total_read;
            break;
        case ECONNRESET:
            DIAGOUT(0, "Connection reset by peer on fd %d.\n", fd);
            return total_read;
            break;
        case EBADF:
            DIAGOUT(0, "File Descriptor (%d) specified is not open for reading.\n", fd);
            break;
        case EDESTADDRREQ:
            DIAGOUT(0, "fd (%d) refers to a socket for which peer address has not been set using connect(2).\n", fd);
            break;
        case EFAULT:
            DIAGOUT(0, "Buffer address parameter (%p) specified is outside accessible address space.\n", addr);
            break;
        case EINVAL:
            DIAGOUT(0, "fd (%d) attached to an object which is unsuitable for reading.\n", fd);
            break;
        case EIO:
            DIAGOUT(0, "Low leve I/O error occurred.\n");
            break;
        case EISDIR:
            DIAGOUT(0, "The file descriptor (%d) refers to a directory.\n", fd);
            break;
        default:
            PATHOLOGICAL("Unexpected error (0x%X) returned from ::read.\n", *failure);
            break;

        } // end of switch (*failure)

        return -1;
     
    } // end of read_file().
    
  
    /**
     ** write_file() - write to the file descriptor <fd>, from the buffer pointed to by <addr>, the number 
     ** of bytes given by <len>.  If an error occurs, place the first unix errno created in the variable
     ** pointed to by <failure>.
     **
     ** RETURNS: number of bytes written, or -1, and supplied errno is set.
     **
     ** ON ERROR: Write the first errno encountered into the address of <failure>.
     **
     ** NOTES: 
     **
     **/
    static inline ssize_t write_file(SOCKET_FD_T fd, const void* addr, ssize_t len, ERRNO_RESULT_PTR_T failure)
    {
        DIAGOUT(3, "write_file(fd %d, addr %p, len %lu, fail %p) called.\n", fd, addr, len, failure);

#define WRITE_RETRY_SLEEP             (100 * 1000)
#define LIMIT_OF_EINTR_WRITE_ATTEMPTS  100
#define LIMIT_OF_EAGAIN_WRITE_ATTEMPTS 100
        
        if (fd < 0)
        {
            PATHOLOGICAL("fd (%d) invalid.\n", fd);
            return -1;
        }

        ssize_t volatile total_written   = 0;       // Total bytes we managed to write out to the fd.
        ssize_t volatile bytes_written   = 0;       // Bytes we wrote to the fd on each write attempt.
        ssize_t volatile bytes_remaining = len;     // Bytes remaining yet to be written.
        size_t  volatile eintr_attempts  = 0;       // Total number of EINTR errors we got trying to complete write.
        size_t  volatile eagain_attempts = 0;       // Total number of EAGAIN errors we got trying to complete write.

        struct timespec time_spec;

        while (bytes_remaining > 0)
        {
            errno         = 0;
            bytes_written = ::write(fd, addr, bytes_remaining);
            *failure      = errno;
    
            if (bytes_written == bytes_remaining) { break; }

            // We wrote less bytes than we were expecting, but there was no error.
            if (bytes_written >= 0 && bytes_written < bytes_remaining)
            { 
                total_written   += bytes_written;
                bytes_remaining -= bytes_written;

                // If we've read some bytes, we'll want to move the addr pointer as well.
                addr = (void*) (((char*) addr) + bytes_written);

                if (*failure == 0) 
                {
                    DIAGOUT(4, "Write to fd (%d) was interrupted, wrote only %lu of %lu bytes.\n", fd, total_written, len);
                    break; 
                }
            }

            if (*failure == EAGAIN || *failure == EWOULDBLOCK)
            {
                if (++eagain_attempts < LIMIT_OF_EAGAIN_WRITE_ATTEMPTS) 
                {
                    DIAGOUT(0, "Got EAGAIN on write().\n");
                    time_spec.tv_sec   = 0;
                    time_spec.tv_nsec  = READ_RETRY_SLEEP;
                    ::nanosleep(&time_spec, NULL);
                    continue;
                }

                DIAGOUT(0, "Write to fd (%d) might block, wrote %lu bytes.\n", fd, bytes_written);
                return bytes_written;
             }

            if (*failure == EINTR)
            {
                if (++eintr_attempts < LIMIT_OF_EINTR_WRITE_ATTEMPTS)
                {
                    DIAGOUT(0, "Got EINTR on write().\n");
                    time_spec.tv_sec   = 0;
                    time_spec.tv_nsec  = READ_RETRY_SLEEP;
                    ::nanosleep(&time_spec, NULL);
                    continue; 
                }
                DIAGOUT(0, "Write to fd (%d) was interrupted, wrote %lu bytes.\n", fd, bytes_written);
                return bytes_written;
            }
            
            // This isn't necessarily a pathological error.  The other side may have exited.
            if (*failure == EPIPE)
            {
                DIAGOUT(0, "write to fd (%d) is connected to a pipe or socket whose reading end is closed.\n", fd);
                return 0;
            }

            // This isn't necessarily a pathological error.  The other side may have exited.
            if (*failure == ECONNRESET)
            {
                DIAGOUT(0, "write to fd (%d) failed, connection reset by peer.\n", fd);
                return 0;
            }

            switch (*failure)
            {
            case EBADF:
                DIAGOUT(0, "File Descriptor (%d) specified is not open for writing.\n", fd);
                break;
            case EDESTADDRREQ:
                DIAGOUT(0, "fd (%d) refers to a socket for which peer address has not been set using connect(2).\n", fd);
                break;
            case EFAULT:
                DIAGOUT(0, "Buffer address parameter (%p) specified is outside accessible address space.\n", addr);
                break;
            case EFBIG:
                DIAGOUT(0, "Write exceeds implementation-defined maximum file size or the process's file size limit\n\
                         or to write at a position past the maximum allowed offset.\n");
                break;
            case EINVAL:
                DIAGOUT(0, "fd (%d) attached to an object which is unsuitable for writing.\n", fd);
                break;
            case EIO:
                DIAGOUT(0, "Low leve I/O error occurred while modifying the inode.\n");
                break;
            case ENOSPC:
                DIAGOUT(0, "The device containing the file referred to by the fd (%d) has no room for the data.\n", fd);
                break;
            default:
                DIAGOUT(0, "Unexpected error \"%s\" (0x%X) returned.\n", ::strerror(*failure), *failure);
                break;

            } // end of switch (*failure)

            PATHOLOGICAL("write_file(fd %d, addr %p, len %lu): returned %lu bytes after %lu attempts, errno was \"%s\" (%d).\n", 
                fd, addr, len, bytes_written, eintr_attempts + eagain_attempts, ::strerror(*failure), *failure);
            return -1;

        } // end of while (count > 0)

        return bytes_written;
     
    } // end of write_file().
    
  
    /**
     ** save_results() - Writes to the outupt results file the data that would have been sent by the
     ** host agent to the ZRC.  This method is used as an alternative transport layer when unit testing.
     **
     ** RETURNS: number of bytes written, or -1, and supplied errno is set.
     **
     ** ON ERROR: returns -1. 
     **
     ** NOTES: Called only by the Host Agent.
     **
     ** Matches the HA_XPORT_BUF_METHOD_T typedef:
     **
     **   i.e. ssize_t func(const void* addr, size_t len, SECTION_ENTRY_T* entry, ERRNO_RESULT_PTR_T failure)
     **/
    static inline ssize_t save_results(const void* addr, ssize_t len, void* fdata, ERRNO_RESULT_PTR_T failure)
    {
        DIAGOUT(4, "save_results(addr %p, len %lu, file_data %p, failure %p)...\n", addr, len, fdata, failure);

#define LIMIT_OF_EINTR_SAVE_ATTEMPTS 100
#define LIMIT_OF_EAGAIN_SAVE_ATTEMPTS 100
        
        FILE_DATA_T* file_data = (FILE_DATA_T*) fdata;

        if (file_data == NULL)
        {
            PATHOLOGICAL("file_data is NULL.\n");
            return -1;
        }

        SOCKET_FD_T fd = file_data->ha_output_fd;

        if (fd < 0)
        {
            PATHOLOGICAL("ha_output_fd (%d) invalid for type \"%s\" (%d).\n", 
                fd, xlate_section_type(file_data->type), file_data->type);
            return -1;
        }

        return write_file(fd, addr, len, failure);
     
    } // end of save_results().
    
  
    /**
     ** ZVision::open_file() - 
     **
     ** RETURNS: 
     **
     ** ON ERROR:  
     **
     ** NOTES: 
     **
     **/
    static SOCKET_FD_T open_file(char* name, SECTION_TYPE_E type)
    {
        SOCKET_FD_T fd = -1;

        return  fd;

    } // end of open_file().

  
    /**
     ** ZVision::mmap_file() - 
     **
     ** RETURNS: pointer to memory mapped file, or error. 
     **
     ** ON ERROR:  
     **
     ** NOTES: 
     **
     **/
    static void* mmap_file(SOCKET_FD_T fd, size_t len, char* name)
    {
        char  result1[RESULT_SIZE] = { 0 };         // Automatiic stack variable for flagbit translation.
        char  result2[RESULT_SIZE] = { 0 };         // Automatiic stack variable for flagbit translation.

        int   prot  = PROT_READ | PROT_WRITE;
        int   flags = MAP_SHARED | MAP_NORESERVE;

        void*        mmap_addr;

        ERRNO_RESULT_T failure;

        DIAGOUT(3, "Issuing mmap(addr=0x%lX, len=%ld(~%ldK), mmap_prot=0x%X \"%s\", flags=0x%X \"%s\", fd=%d, offset=0)...\n",
            (uint64_t) ALLOW_KERNEL_TO_CHOOSE_MMAP_ADDR, 
            len, (len > 1024) ? (len/1024) : 1, 
            prot,  XlateProt::mmap(prot,  result1, sizeof(result1)/sizeof(char)), 
            flags, XlateFlag::mmap(flags, result2, sizeof(result2)/sizeof(char)),
            fd);
    
        ZV_ASSERT(fd >= 0);

        errno     = 0;
        mmap_addr = ::mmap(ALLOW_KERNEL_TO_CHOOSE_MMAP_ADDR, len, prot, flags, fd, 0);
        failure   = errno;

        if (mmap_addr == MAP_FAILED)
        {
            switch (failure)
            {
            case EACCES:
                DIAGOUT(0, "The file file descriptor refers to a non-regular file.\n");
                break;
    
            case EAGAIN:   // same as case SIGSEGV: ??
                DIAGOUT(0, "The file \"%s\" has been locked, or too much memory has been locked,\n", 
                    name == NULL ? "Unavailable" : name); 
                DIAGOUT(0, "or there was an attempted write into a region mapped as read-only.\n");
                break;
    
            case EBADF:
                DIAGOUT(0, "The file descriptor(%d) is not a valid file descriptor.\n", fd);
                break;
    
            case EINVAL:
                DIAGOUT(0, "The file size (%ld) must be greater than zero, and/or\n", len);
                DIAGOUT(0, "the mmap() flags (0x%X \"%s\") contained neither MAP_PRIVATE or MAP_SHARED,\n",
                    flags, XlateFlag::mmap(flags, result2, sizeof(result2)/sizeof(char)));
                DIAGOUT(0, "or contained both of these values.\n");
                break;
    
            case ENFILE:
                DIAGOUT(0, "The system limit on the total number of open files has been reached.\n");
                break;
    
            case ENODEV:
                DIAGOUT(0, "The underlying file system of the specified file does not support memory mapping.\n");
                break;
    
            case ENOMEM:
                DIAGOUT(0, "No memory is available, or the process's maximum number of mappings would have been exceeded.\n");
                DIAGOUT(0, "Attempted to mmap() %ld bytes (~%ldK) for ZVision shared memory file.\n",
                     len, (len / 1024));
                break;
    
            case EPERM:
                DIAGOUT(0, "The mmap_prot argument asks for PROT_EXEC but the mapped area belongs to a file\n");
                DIAGOUT(0, "on a file system that was mounted no-exec.\n");
                break;
    
            case ETXTBSY:
                DIAGOUT(0, "MAP_DENYWRITE was set but the object specified by fd is open for writing.\n");
                break;
    
            case SIGBUS:
                DIAGOUT(0, "Attempted access to a portion of the buffer that does not correspond to the file.\n");
                break;
    
            default:
                DIAGOUT(0, "Unexpected error \"%s\" (%d) returned.\n", ::strerror(failure), failure);
                break;
            }
    
            if (::close(fd) != 0)
            {
                DIAGOUT(0, "::close(%d) failed.\n", fd);
            }

            PATHOLOGICAL("mmap() failed with errno set to %d (%s).\n", failure, ::strerror(failure)); 
            return MAP_FAILED;
    
        } // end of if (ts == MAP_FAILED)
    
        DIAGOUT(3, "finished mmap()'ing file, addr is %p\n", mmap_addr);

        return  mmap_addr;

    } // end of mmap_file().


    static bool    mark_file_blocking(SOCKET_FD_T fd) { return set_file_access(fd, true);  }
    static bool mark_file_nonblocking(SOCKET_FD_T fd) { return set_file_access(fd, false); }
    
    /**
     ** ZVision::set_file_acess() -  Sets a file's status to blocking, or non-blocking, depending on "block" parameter.
     **
     ** Sometimes it is necessary to make a socket NON-BLOCKING because any write that would pend could 
     ** inappropriately hang up the caller.
     **
     **/
    static inline bool set_file_access(SOCKET_FD_T fd, bool block)
    {
        ERRNO_RESULT_T failure     = 0;
        int            access_mode = ::fcntl(fd, F_GETFL, 0);

         DIAGOUT(4, "access mode for fd %d was 0%o.\n", fd, access_mode);

        if (access_mode < 0)
        {
            failure = errno;
            PATHOLOGICAL("Could not get fcntl flags for fd %d, Error is %d \"%s\".\n", fd, failure, ::strerror(failure));
            return false;
        }

        if (block == true)
        {
            access_mode &= O_NONBLOCK;
        }
        else
        {
            access_mode |= O_NONBLOCK;
        }

        if (::fcntl(fd, F_SETFL, access_mode) < 0)
        {
            failure = errno;
            PATHOLOGICAL("Could not set fcntl flags to 0x%X: for fd %d, error is %d \"%s\".\n", 
                access_mode, fd, failure, ::strerror(failure));
            return false;
        }

        access_mode = ::fcntl(fd, F_GETFL, 0);

        if (access_mode < 0)
        {
            failure = errno;
            PATHOLOGICAL("Could not get fcntl flags for fd %d, Error is %d \"%s\".\n", fd, failure, ::strerror(failure));
            return false;
        }

        DIAGOUT(3, "access mode for fd %d, mode is now 0%o.\n", fd, access_mode);

        return true;

    } // end of set_file_access().
    

//-------------------------------------------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------------------------------------
//
//             Request Response Functions.
//
//-------------------------------------------------------------------------------------------------------------------------------
//-------------------------------------------------------------------------------------------------------------------------------


    /**
     ** Request Buffer functions...
     **
     ** Called during JVM initialization.  This sets up the initial counters for the request buffer.
     **/
    static inline bool request_buffer_init(void*  mmap_addr)
    { 
        DIAGOUT(4, "request_buffer_init(addr %p)\n", mmap_addr);

        if (mmap_addr == NULL) 
        { 
            PATHOLOGICAL("mmap_addr is NULL.\n");
            return false; 
        }

        REQUEST_BUFFER_T* request = (REQUEST_BUFFER_T*) mmap_addr;

        request->hdr.cntl.request_id             = 0;
        request->hdr.cntl.request_token          = 0;
        request->hdr.cntl.request_size           = 0;
        request->hdr.cntl.request_signal_cnt     = 0;
        request->hdr.cntl.request_cnt            = 0;

        request->hdr.cntl.response_id            = ::getpid();
        request->hdr.cntl.response_token         = 0;
        request->hdr.cntl.response_size          = 0;
        request->hdr.cntl.response_signal_cnt    = 0;
        request->hdr.cntl.response_cnt           = 0;

        request->hdr.cntl.cookie                 = COOKIE_REQUEST_BUF;
        request->hdr.cntl.dropped                = 0;

        ZV_ASSERT(MAX_REQUEST_PARAMS_SIZE == sizeof(request->params));
        //ZV_ASSERT(RESPONSE_SIZE == sizeof());

        ::memset((void*) request->params, 0, sizeof(request->params));

        return true;

    } // end of request_buffer_init().
  

    /**
     ** Close an existing request buffer.
     **/
    static bool request_buffer_close()
    { 
        return false;

    }  // end of request_buffer_close().

  
//-----------------------------------------------------------------------------------------------------------------------------//
//-----------------------------------------------------------------------------------------------------------------------------//
//                                                                                                                             //
//                                                                                                                             //
//                                                                                                                             //
//-----------------------------------------------------------------------------------------------------------------------------//
//-----------------------------------------------------------------------------------------------------------------------------//


    /**
     ** 
     **/
    static inline const CONSTANTS_T* constants_by_type(SECTION_TYPE_E type)
    {
        int index = type_index(type);
        return (index < 0) ? NULL : &_constants_by_type[index];
    }


    /**
     ** 
     **/
    static inline SECTION_ENTRY_T* section_entry_by_type(TYPE_STABLE_LAYOUT_T* ts, SECTION_TYPE_E type)
    {
        if (type_stable_valid(ts) != true) 
        { 
            PATHOLOGICAL("section_entry_by_type(%p, %d) called but type_stable_valid() returned false.\n", ts, type);
            return NULL; 
        }

        int index = type_index(type);
        return (index < 0) ? NULL : &ts->sections[index];
    }


    /**
     ** 
     **/
    static inline FILE_DATA_T* file_data_by_type(TYPE_STABLE_LAYOUT_T* ts, SECTION_TYPE_E type)
    {
        SECTION_ENTRY_T* section = section_entry_by_type(ts, type);
        return (section == NULL) ? NULL : &section->file;
    }


    /**
     ** 
     **/
    static inline CROSS_REF_ENTRY_T* xref_data_by_type(SECTION_TYPE_E type)
    {
        int index = type_index(type);
        return (index < 0) ? NULL : _cross_ref_table[index];
    }


//-----------------------------------------------------------------------------------------------------------------------------//
//-----------------------------------------------------------------------------------------------------------------------------//
//                                                                                                                             //
//                                                                                                                             //
//                                                                                                                             //
//-----------------------------------------------------------------------------------------------------------------------------//
//-----------------------------------------------------------------------------------------------------------------------------//

    /**
     **
     **/
    static void print_section_entries(TYPE_STABLE_LAYOUT_T* ts, size_t verbosity, ACCESS_USER_E access = user_is_neither)
    {
        char        result[RESULT_SIZE] = { 0 };         // Automatiic stack variable for flagbit translation.

        DIAGOUT(verbosity, "SECTION ENTRIES: collated data (based on available type & current version) for each section type.\n\n");

        for (size_t loop = 0; ((int) loop) < type_index(end_token); loop++)
        {
            SECTION_TYPE_E   type           = (SECTION_TYPE_E) (loop + start_token);
            SECTION_ENTRY_T* section        = section_entry_by_type(ts, type);
            FILE_DATA_T*     file_data      = file_data_by_type(ts, type);
    
            xlate_version_struct(&file_data->version, result);

            DIAGOUT(verbosity, "     \"%s\"\n",                       xlate_section_type(type));
            DIAGOUT(verbosity, "       version:           \"%s\"\n",  result);
            DIAGOUT(verbosity, "       max size:          %lu\n",     file_data->max_size);
            DIAGOUT(verbosity, "       unit cnt:          %lu\n",     file_data->unit_cnt);
            DIAGOUT(verbosity, "       unit size:         %lu\n",     file_data->unit_size);
            DIAGOUT(verbosity, "       jvm mmap fd:       %d\n",      file_data->jvm_mmap_fd);
            DIAGOUT(verbosity, "       ha  mmap fd:       %d\n",      file_data->ha_mmap_fd);
            DIAGOUT(verbosity, "       output fd:         %d\n",      file_data->ha_output_fd);
            DIAGOUT(verbosity, "       output cnt:        %lu\n",     file_data->ha_output_cnt);
            DIAGOUT(verbosity, "       name:              \"%s\"\n",  file_data->name);
            DIAGOUT(verbosity, "       name len:          %lu\n",     file_data->name_len);
            DIAGOUT(verbosity, "       jvm mmap:          %p\n",      file_data->jvm_mmap_addr);
            DIAGOUT(verbosity, "       ha mmap:           %p\n",      file_data->ha_mmap_addr);
            DIAGOUT(verbosity, "       ha mmap len:       %lu\n",     file_data->ha_mmap_len);
            DIAGOUT(verbosity, "       alignment:         %lu\n",     file_data->alignment);

            XlateFlag::mmap(file_data->mmap_flags, result, sizeof(result)/sizeof(char));

            DIAGOUT(verbosity, "       mmap flags         0x%X \"%s\"\n", file_data->mmap_flags, result);
            DIAGOUT(verbosity, "       allocated:         %s\n",      file_data->allocated ? "true" : "false");
            DIAGOUT(verbosity, "       active:            %s\n",      file_data->active    ? "true" : "false");
            DIAGOUT(verbosity, "       file cookie:       0x%lX\n",   file_data->cookie);
            DIAGOUT(verbosity, "       -- jvm --\n");
            DIAGOUT(verbosity, "       init_buf_func:     %p\n",      section->jvm_funcs.init_buf_func);
            DIAGOUT(verbosity, "       preload_buf_func:  %p\n",      section->jvm_funcs.preload_buf_func);
            DIAGOUT(verbosity, "       func cookie:       0x%lX\n",   section->jvm_funcs.cookie);
            DIAGOUT(verbosity, "       -- host agent --\n");
            DIAGOUT(verbosity, "       compress_buf_func: %p\n",      section->ha_funcs.compress_buf_func);
            DIAGOUT(verbosity, "       func cookie:       0x%lX\n",   section->ha_funcs.cookie);

            if (access == user_is_jvm)
            {
                DIAGOUT(verbosity, "JVM's Buffer Contents:\n");
                DIAGBUFFEROUT(verbosity, (volatile char*) file_data->jvm_mmap_addr, file_data->max_size);
                DIAGOUT(verbosity, "\n");
            }
            else if (access == user_is_host_agent)
            {
                DIAGOUT(verbosity, "Host Agent's Buffer Contents:\n");
                DIAGBUFFEROUT(verbosity, (volatile char*) file_data->ha_mmap_addr, file_data->ha_mmap_len);
                DIAGOUT(verbosity, "\n");
            }
            else
            {
                DIAGOUT(verbosity, "       Buffer contents omitted.\n");
            }

            DIAGOUT(verbosity, "\n");
        }

        DIAGOUT(verbosity, "\n");

    } // end of print_section_entries();


    /**
     **
     **/
    static void print_cross_ref_entries(size_t verbosity)
    {
        
        char        result[RESULT_SIZE] = { 0 };         // Automatiic stack variable for flagbit translation.

        DIAGOUT(verbosity, "CROSS REFERENCE DATA: These are the cross reference arrays, ordered by version within type.\n\n");

        for (size_t loop = 0; ((int) loop) < type_index(end_token); loop++)
        {
            SECTION_TYPE_E          type   = (SECTION_TYPE_E) (loop + start_token);
            CROSS_REF_ENTRY_T*      xref   = xref_data_by_type(type);

            DIAGOUT(verbosity, "     \"%s\"\n", xlate_section_type(type));

            while (xref != NULL)
            {
                size_t cnt = 1;

                if (xref->type != type)
                {
                    PATHOLOGICAL("Non-matching type \"%s\" (%d) encountered in xref linked list.\n", \
                        xlate_section_type(xref->type), xref->type);
                    break;
                }

                xlate_verno(xref->verno, result);

                while (xref != NULL)
                {
                    DIAGOUT(verbosity, "       Entry %lu: %p\n",                   cnt++, xref);
                    DIAGOUT(verbosity, "         version:           \"%s\" (%ld)\n", result, xref->verno);
                    DIAGOUT(verbosity, "         max size:          %lu\n",          xref->max_size);
                    DIAGOUT(verbosity, "         alignment:         %lu\n",          xref->alignment);
                    DIAGOUT(verbosity, "         notify_lmt:        %lu\n",          xref->notify_lmt);
                    DIAGOUT(verbosity, "         unit_cnt:          %lu\n",          xref->unit_cnt);
                    DIAGOUT(verbosity, "         unit_size:         %lu\n",          xref->unit_size);
                    DIAGOUT(verbosity, "         init_buf_func:     %p\n",           xref->jvm_funcs.init_buf_func);
                    DIAGOUT(verbosity, "         preload_buf_func:  %p\n",           xref->jvm_funcs.preload_buf_func);
                    DIAGOUT(verbosity, "         compress_buf_func: %p\n",           xref->ha_funcs.compress_buf_func);
                    DIAGOUT(verbosity, "         next:              %p\n",           xref->next);
                    DIAGOUT(verbosity, "         cookie:            0x%lX\n\n",      xref->cookie);

                    xref = xref->next;
                }

            }       // end of while (xref != NULL);
        }       // end of for (size_t loop = 0; loop < type_index(end_token); loop++)

        DIAGOUT(verbosity, "\n");

    }   // end of print_cross_ref_entries().



    /**
     **
     **/
    static void print_type_version_entries(size_t verbosity)
    {
        char        result[RESULT_SIZE] = { 0 };         // Automatiic stack variable for flagbit translation.

        DIAGOUT(verbosity, "TYPE/VERSION DATA: These are hardcoded data entries for each type/version combination.\n\n");

        for (size_t index = 0; index < _type_version_cnt; index++)
        {
            const TYPE_VERSION_ENTRY_T* ver = &_type_version_table[index];

            xlate_version_struct((VERSION_INFO_T*) &ver->version, result);

            DIAGOUT(verbosity, "     \"%s\"\n",                       xlate_section_type(ver->type));
            DIAGOUT(verbosity, "       version:           %s\n",      result);
            DIAGOUT(verbosity, "       max size:          %lu\n",     ver->max_size);
            DIAGOUT(verbosity, "       alignment:         %lu\n",     ver->alignment);
            DIAGOUT(verbosity, "       notify_lmt:        %lu\n",     ver->notify_lmt);
            DIAGOUT(verbosity, "       unit_cnt:          %lu\n",     ver->unit_cnt);
            DIAGOUT(verbosity, "       unit_size:         %lu\n",     ver->unit_size);
            DIAGOUT(verbosity, "       init_buf_func:     %p\n",      ver->jvm_funcs.init_buf_func);
            DIAGOUT(verbosity, "       preload_buf_func:  %p\n",      ver->jvm_funcs.preload_buf_func);
            DIAGOUT(verbosity, "       compress_buf_func: %p\n",      ver->ha_funcs.compress_buf_func);
            DIAGOUT(verbosity, "       cookie:            0x%lX\n\n", ver->cookie);

        }       // end of for (size_t loop = 0; loop < _type_version_cnt; loop++)

        DIAGOUT(verbosity, "\n");

    }   // end of print_type_version_entries().


    /**
     **
     **/
    static void print_status(TYPE_STABLE_LAYOUT_T* ts, size_t verbosity)
    {
        char        result[RESULT_SIZE] = { 0 };         // Automatiic stack variable for flagbit translation.

        if (type_stable_valid(ts) != true)
        {
            PATHOLOGICAL("print_status() called but type_stable_valid() returned false.\n");
            return;
        }

        DIAGOUT(verbosity, "\n");
        DIAGOUT(verbosity, "ZVision Build Internals:\n");
        DIAGOUT(verbosity, "  File Basename:      \"%s\"\n",    ts->sections[type_index(type_stable)].file.name);

        xlate_version_struct((VERSION_INFO_T*) current_sw_ver(), result);

        DIAGOUT(verbosity, "  Build Version:      \"%s\"\n",    result);

        xlate_version_struct((VERSION_INFO_T*) earliest_sw_ver(), result);

        DIAGOUT(verbosity, "  Earliest Version:   \"%s\"\n",    result);
        DIAGOUT(verbosity, "  Type Sections:      %lu\n",       ts->jvm_section_cnt);
        DIAGOUT(verbosity, "  Total File Size:    %lu\n",       sizeof(TYPE_STABLE_LAYOUT_T));
        DIAGOUT(verbosity, "\n");
        DIAGOUT(verbosity, "         cookie:  0x%lX\n",           ts->cookie);

        xlate_version_struct((VERSION_INFO_T*) &ts->version, result);

        DIAGOUT(verbosity, "        version:  \"%s\"\n",          result);
        DIAGOUT(verbosity, "jvm section_cnt:  %lu\n",             ts->jvm_section_cnt);
        DIAGOUT(verbosity, "     active_cnt:  %lu\n",             ts->active_cnt);
        DIAGOUT(verbosity, "      ipc_state:  \"%s\" (0x%lX)\n",  xlate_ipc_state(ts->ipc_state), ts->ipc_state);
        DIAGOUT(verbosity, "        jvm_pid:  %d\n",              ts->jvm_pid);
        DIAGOUT(verbosity, "  hostagent_pid:  %d\n",              ts->host_agent_pid);
        DIAGOUT(verbosity, "  jvm server fd:  %d\n",              ts->jvm_server_fd);
        DIAGOUT(verbosity, "    jvm send fd:  %d\n",              ts->jvm_ipc_fd);
        DIAGOUT(verbosity, "     ha send fd:  %d\n",              ts->ha_ipc_fd);
        DIAGOUT(verbosity, "  ipc_domain_id:  \"%s\"\n",          ts->ipc_domain_id);
        show_forgo_data(ts, verbosity);
        DIAGOUT(verbosity, "\n");

        print_type_version_entries(verbosity + 1);
        print_cross_ref_entries(verbosity + 1);
        print_section_entries(ts, verbosity + 1);

    } // end of print_status().


//-----------------------------------------------------------------------------------------------------------------------------//
//-----------------------------------------------------------------------------------------------------------------------------//
//                                                                                                                             //
//                                                                                                                             //
//                                                                                                                             //
//-----------------------------------------------------------------------------------------------------------------------------//
//-----------------------------------------------------------------------------------------------------------------------------//

   /**
    ** Build a pathname that includes a zvision_dir_str, a subdir_str & a filename_str
    **/
    static bool build_pathname(char*       pathname_buf, 
                               const char* zvision_dir_str, 
                               const char* subdir_str, 
                               const char* filename_str) 
    { 
        if (pathname_buf == NULL) 
        { 
            PATHOLOGICAL("pathname_buf is NULL.\n");
            return false; 
        }

        if (zvision_dir_str == NULL)
        { 
            PATHOLOGICAL("zvision_dir_str is NULL.\n");
            return false; 
        }

        if (subdir_str == NULL)
        { 
            PATHOLOGICAL("subdir is NULL.\n");
            return false; 
        }

        if (filename_str == NULL)
        { 
            PATHOLOGICAL("filename is NULL.\n");
            return false; 
        }

        DIAGOUT(4, "build_pathname(\"%s\", \"%s\", \"%s\", \"%s\", ).\n", pathname_buf, zvision_dir_str, subdir_str, filename_str);

        size_t zvision_dir_str_len = ::strnlen(zvision_dir_str, MAX_ZVISION_PATHNAME_LEN);
        size_t subdir_str_len      = ::strnlen(subdir_str,      MAX_ZVISION_PATHNAME_LEN);
        size_t filename_str_len    = ::strnlen(filename_str,    MAX_ZVISION_PATHNAME_LEN);

        if (zvision_dir_str_len <= 0 || zvision_dir_str_len >= MAX_ZVISION_PATHNAME_LEN)
        {
            PATHOLOGICAL("zvision_subdir_len (%lu) exceeds limit (%ld).\n",
                zvision_dir_str_len, MAX_ZVISION_PATHNAME_LEN);
            return false; 
        }

        if (subdir_str_len <= 0 || subdir_str_len >= MAX_ZVISION_PATHNAME_LEN)
        {
            PATHOLOGICAL("subdir_len (%lu) exceeds limit (%ld).\n",
                subdir_str_len, MAX_ZVISION_PATHNAME_LEN);
            return false; 
        }

        if (filename_str_len <= 0 || filename_str_len >= MAX_ZVISION_PATHNAME_LEN)
        {
            PATHOLOGICAL("filename_len (%lu) exceeds limit (%ld).\n",
                filename_str_len, MAX_ZVISION_PATHNAME_LEN);
            return false; 
        }

        // If there isn't enough room in the buffer to build the reference filename string, complaine and die.
        if (zvision_dir_str_len + subdir_str_len + filename_str_len + 1 >= MAX_ZVISION_PATHNAME_LEN)
        {
            PATHOLOGICAL("pathname length (%lu) exceeds buffer maximum (%ld).\n", 
                zvision_dir_str_len + subdir_str_len + filename_str_len + 1, MAX_ZVISION_PATHNAME_LEN);
            return false;
        }

        char* pathname_str = pathname_buf;

        ::memcpy((void*) pathname_str, (const void*) zvision_dir_str, zvision_dir_str_len);
        pathname_str += zvision_dir_str_len;

        ::memcpy((void*) pathname_str, (const void*) subdir_str, subdir_str_len);
        pathname_str += subdir_str_len;

        *pathname_str = '/';
        pathname_str++;

        ::memcpy((void*) pathname_str, (const void*) filename_str, filename_str_len);
        pathname_str += filename_str_len;

        DIAGOUT(3, "Constructed pathname: \"%s\".\n", pathname_buf);
        return true;
    
    }  // end of build_pathname().


    /**
     ** Compare the expected (golden data) file contents against the generated output file.
     ** 
     ** Returns true if the files compare identically; false otherwise.
     **
     ** Both filenames must be fully qualified.  If not, they must be accessible from the 
     ** current working directory of whatever application is invoking it; this is discouraged.
     **/
    static bool compare_results(const char* output_pathname, const char* reference_pathname) 
    { 
        size_t           output_len         = 0;     // size of file's fully qualified pathname.
        size_t           reference_len      = 0;     // size of file's fully qualified pathname.
    
        ERRNO_RESULT_T   failure            = 0;     // errno variable for calls that make system requests.
    
        SOCKET_FD_T output_fd;                       // file descriptor for the unit test results output 
        SOCKET_FD_T reference_fd;                    // file descriptor for the unit test "golden data" reference file.
    
        // Get the fully qualified pathname's string length.
        output_len    = ::strnlen(output_pathname,    MAX_ZVISION_PATHNAME_LEN);
        reference_len = ::strnlen(reference_pathname, MAX_ZVISION_PATHNAME_LEN);
    
        // Make sure that the final namestring (including the trailing NULL) fits in the buffer.
        if (output_len >= MAX_ZVISION_PATHNAME_LEN)
        {
            PATHOLOGICAL("could not create output filename, total string length (%lu) exceeds buffer (%ld).\n",
                output_len, MAX_ZVISION_PATHNAME_LEN);
            return false;
        }
    
        // Make sure it makes sense.
        if (output_len <= 1)
        {
            PATHOLOGICAL("could not create output filename, total string length less than 1.\n");
            return false;
        }
    
        DIAGOUT(1, "Opening the Output Results from \"%s\".\n", output_pathname);
    
        output_fd = ::open(output_pathname, O_RDONLY, 0777);
    
        if (output_fd < 0)
        {
            failure = errno;
            
            PATHOLOGICAL("open(name=\"%s\") failed with errno set to %d (%s).\n", 
                output_pathname, failure, ::strerror(failure)); 
    
            return false; 
        }
    
        // Make sure that the final namestring (including the trailing NULL) fits in the buffer.
        if (reference_len >= MAX_ZVISION_PATHNAME_LEN)
        {
            PATHOLOGICAL("could not create reference filename, total string length (%lu) exceeds buffer (%ld).\n",
                reference_len, MAX_ZVISION_PATHNAME_LEN);
            return false;
        }
    
        // Make sure it makes sense.
        if (reference_len < 1)
        {
            PATHOLOGICAL("could not create reference filename, total string length less than 1.\n");
            return false;
        }
    
        DIAGOUT(1, "Opening the Reference file in \"%s\".\n", reference_pathname);
    
        reference_fd = ::open(reference_pathname, O_RDONLY, 0777);
    
        if (reference_fd < 0) 
        {
            failure = errno;
    
            PATHOLOGICAL("open(name=\"%s\") failed with errno set to %d (%s).\n", 
                reference_pathname, failure, ::strerror(failure)); 
            
            return false;
        }
    
        return compare_results(output_fd, reference_fd);
    
    }  // end of compare_results(). 
    
    
    /**
     ** Compare the expected (golden data) file contents against the generated output file.
     ** 
     ** Returns true if the files compare identically; false otherwise.
     **
     ** Both fds must be open and available for reading.
     **/
    static bool compare_results(SOCKET_FD_T output_fd, SOCKET_FD_T reference_fd)
    { 
        ssize_t  bytes_read = 0;                              // total bytes actually read.
        const ssize_t block_size = VM_SMALL_PAGE_SIZE;        // default block read size.
    
        char output_buf   [block_size];                       // buffer into which we read blocks of output file data.
        char reference_buf[block_size];                       // buffer into which we read blocks of reference file data.
    
        ERRNO_RESULT_T   failure   = 0;                       // errno variable for calls that make system requests.
    
        ssize_t output_read;                                  // bytes returned from read().
        ssize_t reference_read;                               // bytes returned from read().
    
        DIAGOUT(2, "Comparing the output (fd %d) and reference (fd %d) files, byte for byte...\n", output_fd, reference_fd);
    
        while (true)
        {
            output_read    = read_file(output_fd,    &output_buf,    block_size, &failure);
    
            if (output_read != block_size && failure != 0)
            {
                DIAGOUT(0, "Got error \"%s\" (%d) while reading output file fd %d.\n", 
                    ::strerror(failure), failure, output_fd);
                return false;
            }
    
            reference_read = read_file(reference_fd, &reference_buf, block_size, &failure);
    
            if (reference_read != block_size && failure != 0)
            {
                DIAGOUT(0, "Got error \"%s\" (%d) while reading reference file fd %d.\n", 
                    ::strerror(failure), failure, reference_fd);
                return false;
            }
    
            if (output_read != block_size || reference_read != block_size) { break; }
    
            if (::memcmp(output_buf, reference_buf, block_size) != 0) 
            { 
                break;
            }
    
            bytes_read += block_size;
        }
    
        // If there was a partial read, but both blocks contain the same amount read, and their contents
        // are the same, then this is a match.
        if (output_read == reference_read && ::memcmp(output_buf, reference_buf, reference_read) == 0) 
        {
            bytes_read += reference_read;
    
            DIAGOUT(1, "Output Results file matches Reference file (length %lu).\n", bytes_read);
    
            return true; 
        }
    
        // Determine the maximum number of bytes in common between the two buffers.
        size_t bytes_avail = (output_read > reference_read) ? reference_read : output_read;
        size_t index;
    
        // Point out the first character where they the files differ.
        for (index = 0; index < bytes_avail; index++)
        {
            if ((output_buf[index] & 0xFF) != (reference_buf[index] & 0xFF)) { break; }
        }
        
        DIAGOUT(0, "First discrepancy at byte %lu in output file. Expecting byte 0x%X, got byte 0x%X.\n", 
            bytes_read + index, (reference_buf[index] & 0xFF), (output_buf[index] & 0xFF));
    
        if (output_read != reference_read) 
        { 
            DIAGOUT(0, "Output Results bytes read %lu\n", bytes_read + output_read);
            DIAGOUT(0, "Reference File bytes read %lu\n", bytes_read + reference_read);
            DIAGOUT(0, "Files differ!\n"); 
        }
    
        return false;
    
    }  // end of compare_results(). 

};  // end of class ZVision


/**
 ** class ZVisionFilter - Container class for the ZRC/ZVision Filter Ticks mechanism.  This class is used
 ** by the JVM (in this context it would be considered the producer of tick buffers, in that it updates the writer_idx
 ** field within the tick buffer ring buffer header) to process Tick information through a ring buffer as buffers
 ** are placed on the recycle list within the JVM.
 **
 ** This is useful for UserTicks, but can also be used for filtering standard ticks to reduce the bandwidth
 ** of tick data being sent to the ZRC Broker task or Host Agent.
 **
 ** The class is designed to allow any number of consumer threads to process either the tick buffers directly, or read a saved
 ** copy of them, filtering on various aspects, through the ring buffer shared memory file.
 **
 ** It was designed with the intention that it can also serve as the basis of a Java API to allow tracking of UserTicks
 ** directly.  See Gil or Robb for more details on this.
 **/
class ZVisionFilter
{
  friend class ZVisionHostAgent;
  friend class ZVisionFilterTicks;

  public:

    enum FILTER_STATE_E
    {
      uninitialized = 0,
      initializing  = 0x12341355,
      active        = 0x13425123,
      blocked       = 0x93236161,
      exited        = 0xF0F0F0F0,
      pathological  = ~0,

    };

  private:

    // Each process that interacts with a ring buffer will have its own notion of where its
    // files are memory mapped to, and what the file descriptors are.

    ZVision::SOCKET_FD_T  _rng_fd;         // File Descriptor of the ring buffer header file.
    ZVision::SOCKET_FD_T  _buf_fd;         // File Descriptor of the Event Tick Buffer file, if applicable.
    ZVision::RING_BUF_T*  _rng_start;      // Start of the ring buffer file's memory mapped address.
    char*                 _buf_start;      // The start address of the indexed buffer file.
    ZVision::FILTER_KEY_T _filter_key;     // The key valve placed in the shared memory ring buffer header to
                                           // indicate that it is now safe for the reader task to access the buffer.

    // These items are only set at initialization by the consumer applications (not the JVM), and then
    // referenced onward.  They are ignored by the producer (JVM) task.

    char*                 _buf_name;       // The name of the data buffer file.
    size_t                _buf_size;       // Size (in bytes) of the Event Tick Buffer.
    size_t                _max_bufs;       // Number of ring buffer entries.
    size_t                _elem_size;      // Size (in bytes) of a tick entry.
    size_t                _max_elems;      // Number of elements within a buffer entry.
    size_t                _hdr_size;       // Size (in bytes) of the element header structure inside buffer.
    size_t                _alloc_offset;   // Offset (in bytes) of the element header structure inside buffer.

    // These items are usually only set (and used) by the consumer application (not the JVM) when filtering.

    volatile uint64_t     _timestamp;      // We record the buffer header timestamp to detect when we've been stomped.
    volatile size_t       _buffer_idx;     // Current data buffer index the ring buffer element refers to.
    volatile size_t       _reader_idx;     // Current buffer reader is accessing.
    size_t                _last_writer_idx;// Last write buffer index the reader accessed before blocking.
    uint64_t              _last_buf_cnt;   // Last write buffer count before the reader blocked.
    char*                 _next_element;   // Next readable element within buffer.
    char*                 _last_element;   // Last readable element within buffer.
    uint64_t              _elem_tot;       // Elements consumed to date.
    char*                 _buf_hdr;        // The buffer header address of the current buffer.
    char*                 _elem_buf;       // A buffer address provided by caller into which we copy the element.
    FILTER_STATE_E        _state;          // Object's health.

    size_t                _nth_element;    // Skip N unfiltered elements first, each time an element is requested from
                                           // a get_next_element() method.
    size_t                _nth_match;      // Skip N matching elements, each time an element is requested from
                                           // a get_next_element() method.

    char _rng_filename[ZVision::MAX_POSIX_SHARED_MEMORY_NAME_SIZE]; // The fully qualified name of the ring buffer file.

  public:


//--------------------------------------------------------------------------------------------------------//
//   The following are called by the ring buffer "producer" (i.e JVM's ZVision Service Thread).           //
//--------------------------------------------------------------------------------------------------------//


    /**
     ** Called by the JVM.  This creates a shared, (/dev/shm/) memory-mapped (not file backed) file that contains a ring buffer of
     ** pointers to tick buffers.  
     ** 
     ** Parameters:
     **
     ** file_data    - pointer to the "tick_profiling" FILE_DATA_T section of the (memory mapped area of the) type_stable file.
     ** filter_key   - the unique key provided by the Producer when creating the ring buffer file to indicate that it is
     **                now fully initialized and safe to access the ring buffer. 
     ** elem_size    - the size, in bytes, of one element within the buffer.
     ** hdr_size     - the size, in bytes, of the header structure within the data buffer.  Currently it is assumed to equal 
     **                elem_size.
     ** alloc_offset - the offset, in bytes, to the "allocation time" within the header structure of the data buffer.
     **
     ** The timestamp and alloc_offset fields are used to detect when the writer task has overwritten (i.e. "stomped on") a
     ** buffer being processed by the reader task.  The timestamp should be a (very unlikely to wrap) value of when the
     ** buffer was taken off the recycle list and put into use.  The alloc_offset field is the offset, in bytes, to the
     ** timestamp value within the data buffer.  This offset value should always be the same for every entry within the data
     ** buffer, so we can store it in the ring buffer header, but the timestamp should always be unique to each data buffer and
     ** instead is stored in the ring buffer entry structure.
     **
     ** [This is the Producer task's object initializaiton call]
     **
     **/
    ZVisionFilter(ZVision::FILE_DATA_T*  file_data, 
                  ZVision::FILTER_KEY_T  filter_key, 
                  size_t                 elem_size, 
                  size_t                 hdr_size,
                  size_t                 alloc_offset)
    {
      DIAGOUT(1, "ZVisionFilter([Producer]) called, file_data %p, filter key 0x%lX, basename \"%s\".\n", 
          file_data, filter_key, file_data->name);

      _state      = initializing;
      _filter_key = filter_key;

      _rng_fd     = -1;
      _buf_fd     = file_data->jvm_mmap_fd;   // This should have already been opened by the Host Agent.

      _rng_start  = NULL;
      _buf_start  = NULL;

      if (file_data == NULL)
      {
          _state = pathological; 
          PATHOLOGICAL("NULL file_data pointer.\n");
          return; 
      }

      if (file_data->cookie != ZVision::COOKIE_FILE_DATA)
      {
          _state = pathological; 
          PATHOLOGICAL("Corrupted file_data pointer %p, expected 0x%lX, found 0x%lX.\n", 
              file_data, ZVision::COOKIE_FILE_DATA, file_data->cookie);
          return; 
      }

      // We shouldn't get here if there is already a ring buffer in place.
      if (file_data->jvm_filter_obj != NULL)
      {
          _state = pathological; 
          PATHOLOGICAL("Ring Buffer Object (jvm_filter_obj = %p) in FILE_DATA_T* (%p) not NULL but should have been.\n",
              file_data->jvm_filter_obj, file_data);
          return; 
      }
        
      //------------------------------------- Original Data Buffer File ----------------------------------------------//

      // This file should have already been open and mmap()'ed in by the JVM.
      _buf_start = (char*) file_data->jvm_mmap_addr;

      if (_buf_start == NULL)
      {
          _state = pathological; 
          PATHOLOGICAL("Data buffer pointer NULL.\n");
          return; 
      }

      _buf_name     = file_data->name;
      _buf_size     = file_data->unit_size;
      _max_bufs     = file_data->unit_cnt;
      _elem_size    = elem_size;
      _max_elems    = (_buf_size / _elem_size) - 1;
      _hdr_size     = hdr_size;
      _alloc_offset = alloc_offset;
      _timestamp    = 0;

      // First check that the various paramters make sense.  
      if (_max_bufs < 2) 
      { 
          _state = pathological; 
          PATHOLOGICAL("Max buffers is less than 2.\n");
          return; 
      }

      if (_buf_size < 4096) 
      { 
          _state = pathological; 
          PATHOLOGICAL("buffer size (%lu) is less than one small kernel page (4096 bytes).\n", _buf_size);
          return; 
      }

      // This is a copy of the buf_cnt fields within the shared memory ring buffer header "buf_cnt" field.
      // TODO:  This field may wrap, if given enough time.  We need to be able to cope with that gracefully.   
      _last_buf_cnt    = 0;

      // These fileds are not referenced/used by the producer task and can be ignored here.
      _reader_idx      = ~0;
      _last_writer_idx = 0;

      //-------------------------------------------- Ring Buffer File ----------------------------------------------//

      char* rng_name = create_filename_str(_rng_filename, sizeof(_rng_filename), file_data->name);

      if (rng_name == NULL)
      {
          _state = pathological; 
          PATHOLOGICAL("Could not create ring buffer filename.\n");
          return; 
      }

      DIAGOUT(2, "Fully qualified ring buffer filename is \"%s\".\n", _rng_filename);
      DIAGOUT(1, "Creating shared memory ring buffer file \"%s\".\n", rng_name);

      ZVision::SOCKET_FD_T rng_fd = ::shm_open(rng_name, O_CREAT | O_EXCL | O_RDWR, 0700);

      if (rng_fd < 0)
      {
          ZVision::ERRNO_RESULT_T failure = errno;

          _state = pathological; 

          PATHOLOGICAL("UserTick's ::shm_open(\"%s\") failed, error was %d \"%s\".\n", _rng_filename, failure, ::strerror(failure));
          return;
      }

      // Calculate the size of the shared memory ring buffer header, including the first "shared" ring buffer entry.
      size_t len = sizeof(ZVision::RING_BUF_T);

      // Increase the total size of the file to include all shared ring buffer entries.  We allocate
      // as many ring buffer entries as there are data buffers.  The RING_BUF_T structure already accounts for
      // the first ring buffer entry.
      len += ((_max_bufs - 1) * sizeof(ZVision::RING_BUF_ENTRY_T));

      // Force errno to zero, in case the truncate gives back something we weren't expecting.
      errno = 0;

      // Set the file size by positioning the write pointer at the end of the file.
      if (::ftruncate(rng_fd, len) != 0)
      {
          ZVision::ERRNO_RESULT_T failure = errno;

          _state = pathological; 

          PATHOLOGICAL("::ftruncate(fd=%d, len=%ld) of ring buffer file \"%s\" failed with errno set to %d (%s).\n", 
              rng_fd, len, _rng_filename, failure, ::strerror(failure));

          if (::close(rng_fd) != 0) { DIAGOUT(0, "::close(fd=%d) failed.\n", rng_fd); }

          return;
      }

      // Force errno to zero, in case the lseek gives back something we weren't expecting.
      errno = 0;

      // Set the write pointer back to the beginning of the file.
      if (::lseek(rng_fd, 0, SEEK_SET) != 0)
      {
          ZVision::ERRNO_RESULT_T failure = errno;

          _state = pathological; 

          PATHOLOGICAL("::lseek(fd=%d, 0, SEEK_SET) of file \"%s\" failed with errno set to %d (%s).\n", 
              rng_fd, _rng_filename, failure, ::strerror(failure));

          if (::close(rng_fd) != 0) { DIAGOUT(0, "::close(fd=%d) failed.\n", rng_fd); }

          return;
      }

      // Map in the file, then fill in the ring buffer header info.
      _rng_start = (ZVision::RING_BUF_T*) ZVision::mmap_file(rng_fd, len, _rng_filename);

      if (_rng_start == NULL)
      {
          ZVision::ERRNO_RESULT_T failure = errno;

          _state = pathological; 
          errno  = 0;

          if (::close(rng_fd) != 0) { DIAGOUT(0, "::close(%d) failed.\n", rng_fd); }

          PATHOLOGICAL("::mmap_file(fd=%d, len=%ld, \"%s\") failed, error set to %d (%s).\n", 
              rng_fd, len, _rng_filename, failure, ::strerror(failure)); 
          return;
      }

      _rng_fd = rng_fd;
      
      // Now write into the shared memory ring buffer header the sizes of various items.
      _rng_start->hdr.cntl.buf_size     = _buf_size;
      _rng_start->hdr.cntl.max_bufs     = _max_bufs;
      _rng_start->hdr.cntl.elem_size    = _elem_size;
      _rng_start->hdr.cntl.max_elems    = _max_elems;
      _rng_start->hdr.cntl.hdr_size     = _hdr_size;
      _rng_start->hdr.cntl.alloc_offset = _alloc_offset;
      _rng_start->hdr.cntl.writer_idx   = 0;
      _rng_start->hdr.cntl.cookie       = ZVision::COOKIE_RING_BUF;


      // These values get filled in when the first indexed buffer notify message arrives, but we initialize
      // them to something reasonable to avoid confusion during startup.
      volatile ZVision::RING_BUF_ENTRY_T* rbe = &_rng_start->entry;

      for (size_t loop = 0; loop < _max_bufs; loop++)
      {
          rbe->buf_idx    = ~0;
          rbe->elem_cnt   = 0;
          rbe->curr_elem  = NULL;
          rbe->last_elem  = NULL;
      }

      // Copy the data buffers filename into the ring buffer header.  We need to open the data buffer file separately,
      // from within the consumer thread later on, and then mmap it. Note: the "consumer" tasks can't use the "producer" task's 
      // (e.g. JVM's) notion of the mmap addr.
      memcpy((void*) &_rng_start->hdr.cntl.filename[0], (const void*) _buf_name, ZVision::MAX_ZVISION_PATHNAME_LEN);

      // Okay, we've now created a shared memory ring buffer that the Host Agent (or others) can use to drive
      // UserTick data out to the application.  The caller is responsible for deleting the files when it exits.
      file_data->jvm_filter_obj = this;

      // Gratuitous flag, used only within Producer task to determine if this routine completed successfully.
      _state  = active;

      gratutious_log();

      // Once this field is set, the reader task can safely begin accessing the data.
      _rng_start->hdr.cntl.filter_key = _filter_key;

      return;

    } // end of ZVisionFilter().  [Producer]


//---------------------------------------------------------------------------------------------------------------------------//
//   The following are called by the "consumer" thread (i.e. Host Agent, Host Agent Unit Test, UserTick Unit Test, etc.)     //
//---------------------------------------------------------------------------------------------------------------------------//

    /**
     ** Establish the Tick consumer object.  It opens the already established ring buffer shared memory file
     ** and mmaps it in.  It then allows the caller to read continuously from the ring buffer.  As notifications
     ** arrive, the "producer task" that feeds the ring buffer pushes a new value into the shared header file.
     **
     ** NOTE: This presumes that the Host Agent (or some other producer task) has, on the consumer's behalf, 
     ** created the appropriate ring buffer, and will update the ring buffer whenever an Event Tick Buffer 
     ** notification arrives.
     **
     ** [This is the Consumer task object intialization call]
     **
     **/
    ZVisionFilter(int64_t max_tries, const char* basename_str, ZVision::FILTER_KEY_T filter_key, char* elem_buf)
    {
      DIAGOUT(1, "ZVisionFilter([Consumer]) called, max_tries %ld, basename \"%s\", element copy address is %p.\n", 
          max_tries, basename_str, elem_buf);

      _state      = initializing; 
      _filter_key = filter_key; 
      _elem_tot   = 0;
      _elem_buf   = elem_buf;

      _rng_fd     = -1;
      _buf_fd     = -1;

      _rng_start  = NULL;
      _buf_start  = NULL;

      size_t attempts = 0;

      ZVision::SOCKET_FD_T rng_fd = -1;

      if (elem_buf == NULL)
      {
          _state = pathological; 
          PATHOLOGICAL("NULL element buffer.\n");
          return; 
      }

      if (basename_str == NULL)
      {
          _state = pathological; 
          PATHOLOGICAL("NULL basename pointer.\n");
          return; 
      }

      //-------------------------------------------- Ring Buffer File ----------------------------------------------//

      char* rng_name = create_filename_str(_rng_filename, sizeof(_rng_filename), basename_str);

      if (rng_name == NULL)
      {
          _state = pathological; 

          PATHOLOGICAL("Could not create ring buffer filename.\n");
          return; 
      }

      if (max_tries >= 0)
      {
          DIAGOUT(0, "Attemping (no more than %ld times) to open the (already created) shared-memory ring-buffer file \"%s\".\n", 
              max_tries, _rng_filename);
      }
      else
      {
          DIAGOUT(0, "Attemping to open the (already created) shared-memory ring-buffer file \"%s\".\n", _rng_filename);
      }

      while (true) 
      {
          // Force errno to zero, in case the shm_open() gives back something we weren't expecting.
          errno  = 0;
          rng_fd = ::shm_open(rng_name, O_RDWR, 0700);

          ZVision::ERRNO_RESULT_T failure = errno;
              
          if (rng_fd >= 0) 
          { 
              DIAGOUT(2, "Opened file \"%s\", fd=%d.\n", _rng_filename, rng_fd);
              break; 
          }

          if (failure == ENOENT)
          {
              if (max_tries >= 0) 
              { 
                  if (++attempts >= ((size_t) max_tries)) 
                  { 
                      DIAGOUT(1, "Could not open file \"%s\" after %ld attempts, giving up.\n", _rng_filename, max_tries);
                      _state = pathological;
                      return; 
                  }
              }

              DIAGOUT(2, "Could not open ring file \"%s\". Pausing 1 second before retrying.\n", _rng_filename);
              ::sleep(1);
              continue;
          }

          _state = pathological;
          DIAGOUT(1, "Shared File ::shm_open(\"%s\") failed, error was \"%s\".\n", _rng_filename, ::strerror(failure));
          return;

      }

      // Read the ring buffer header structure into a temporary structure, we'll use the information in it
      // to map in the entire ring buffer file.
      ZVision::RING_BUF_HEADER_T temp;

      while (true)
      {
          // Force errno to zero, in case the lseek() gives back something we weren't expecting.
          errno  = 0;
           
          // Set the write pointer to the beginning of the file.
          if (::lseek(rng_fd, 0, SEEK_SET) != 0)
          {
              ZVision::ERRNO_RESULT_T failure = errno;

              _state = pathological; 

              PATHOLOGICAL("::lseek(fd=%d, 0, SEEK_SET) to start of file \"%s\" failed with errno set to %d (%s).\n", 
                  rng_fd, _rng_filename, failure, ::strerror(failure));

              if (::close(rng_fd) != 0) { DIAGOUT(0, "::close(fd=%d) failed.\n", rng_fd); }

              return;
          }

          // Force errno to zero, in case the read() gives back something we weren't expecting.
          errno  = 0;

          // Read the header, checking that the filter key is set appropriately.
          ssize_t bytes_read = ::read(rng_fd, (void*) &temp, sizeof(ZVision::RING_BUF_HEADER_T));

          if (bytes_read != sizeof(ZVision::RING_BUF_HEADER_T))
          {
              ZVision::ERRNO_RESULT_T failure = errno;

              _state = pathological; 

              PATHOLOGICAL("File name \"%s\" ::read() failed, expected %ld bytes, read %lu, error was \"%s\".\n", 
                  _rng_filename, sizeof(ZVision::RING_BUF_HEADER_T), bytes_read, ::strerror(failure));
              return;
          }

          // Don't go any further until the value matches the supplied filter key.
          if (temp.cntl.filter_key == _filter_key) 
          { 
              _max_bufs = temp.cntl.max_bufs;
              DIAGOUT(3, "Opened ring buffer file \"%s\", max_bufs %lu.\n", _rng_filename, _max_bufs);
              break; 
          }


          if (max_tries >= 0)
          { 
              if (++attempts >= ((size_t) max_tries)) 
              { 
                  DIAGOUT(1, "Ring buffer file \"%s\" key match failed after %ld attempts, giving up.\n", 
                      _rng_filename, max_tries);
                  _state = pathological;
                  return; 
              }
          }

          DIAGOUT(2, "Filter key 0x%lX did not match expected value 0x%lX. Pausing 1 second before retrying.\n", 
              temp.cntl.filter_key, _filter_key);
          ::sleep(1);
      }

      // First calculate the size of the shared memory ring buffer header, including the first "shared" 
      // ring buffer entry.  
      size_t len = sizeof(ZVision::RING_BUF_T);
      
      // Then increase the total size of the file to include all shared ring buffer entries.  We allocate
      // as many ring buffer entries as there are buffers in the data file.  The RING_BUF_T structure already
      // accounts for the first ring buffer entry.
      len += ((_max_bufs - 1) * sizeof(ZVision::RING_BUF_ENTRY_T));

      // Force errno to zero, in case the ftruncate() gives back something we weren't expecting.
      errno  = 0;

      // Set the file size by positioning the write pointer at the end of the file.
      if (::ftruncate(rng_fd, len) != 0)
      {
          ZVision::ERRNO_RESULT_T failure = errno;

          _state = pathological; 

          PATHOLOGICAL("::ftruncate(fd=%d, len=%ld) of file \"%s\" failed with errno set to %d (%s).\n", 
              rng_fd, len, _rng_filename, failure, ::strerror(failure));

          if (::close(rng_fd) != 0) { DIAGOUT(0, "::close(fd=%d) failed.\n", rng_fd); }

          return;
      }

      // Force errno to zero, in case the lseek() gives back something we weren't expecting.
      errno  = 0;

      // Set the file pointer back to the beginning of the file.
      if (::lseek(rng_fd, 0, SEEK_SET) != 0)
      {
          ZVision::ERRNO_RESULT_T failure = errno;

          _state = pathological; 

          PATHOLOGICAL("::lseek(fd=%d, 0, SEEK_SET) back to beginning of file \"%s\" failed with errno set to %d (%s).\n", 
              rng_fd, _rng_filename, failure, ::strerror(failure));

          if (::close(rng_fd) != 0) { DIAGOUT(0, "::close(fd=%d) failed.\n", rng_fd); }

          return;
      }

      // Map in the entire ring buffer file, including the header.
      _rng_start = (ZVision::RING_BUF_T*) ZVision::mmap_file(rng_fd, len, _rng_filename);

      if (_rng_start == NULL)
      {
          if (::close(rng_fd) != 0) { DIAGOUT(0, "::close(%d) failed.\n", rng_fd); }

          _state = pathological; 

          PATHOLOGICAL("::mmap_file(fd=%d, len=%ld, \"%s\") failed.\n", 
              rng_fd, len, _rng_filename); 
          return;
      }
      
      _rng_fd          = rng_fd;

      ZV_ASSERT(_rng_start->hdr.cntl.buf_size   == temp.cntl.buf_size);
      ZV_ASSERT(_rng_start->hdr.cntl.max_bufs   == temp.cntl.max_bufs);
      ZV_ASSERT(_rng_start->hdr.cntl.elem_size  == temp.cntl.elem_size);
      ZV_ASSERT(_rng_start->hdr.cntl.max_elems  == temp.cntl.max_elems);
      ZV_ASSERT(_rng_start->hdr.cntl.filter_key == temp.cntl.filter_key);
      ZV_ASSERT(_rng_start->hdr.cntl.cookie     == ZVision::COOKIE_RING_BUF);

      // The buffer name was already filled in by the Host Agent process.
      _buf_name        = (char*) &_rng_start->hdr.cntl.filename[0];

      // Get the rest of these value from the file.
      _buf_size        = _rng_start->hdr.cntl.buf_size;
      _max_bufs        = _rng_start->hdr.cntl.max_bufs;
      _elem_size       = _rng_start->hdr.cntl.elem_size;
      _max_elems       = _rng_start->hdr.cntl.max_elems;
      _hdr_size        = _rng_start->hdr.cntl.hdr_size;
      _alloc_offset    = _rng_start->hdr.cntl.alloc_offset;
      _last_writer_idx = _rng_start->hdr.cntl.writer_idx;
      _last_buf_cnt    = _rng_start->hdr.cntl.buf_cnt;
      _reader_idx      = _last_buf_cnt;

      // Set the reader to be one buffer behind the current writer.
      if (_reader_idx == 0) { _reader_idx = _max_bufs - 1; } else { _reader_idx--; }


      //-------------------------------------------- Event Tick Buffer Data File ----------------------------------------------//


      // Open the data buffer file & map it in.  This file may be the actual data or a copy of the data.
      DIAGOUT(0, "Opening the shared-memory Event Tick Buffers file \"%s\".\n", _rng_start->hdr.cntl.filename);

      attempts = 0;

      while (true)
      {
          // Force errno to zero, in case the shm_open() gives back something we weren't expecting.
          errno  = 0;

          _buf_fd = ::shm_open(_buf_name, O_RDWR, 0777);

          ZVision::ERRNO_RESULT_T failure = errno;
              
          if (_buf_fd >= 0) 
          { 
              DIAGOUT(2, "Opened file \"%s\", fd=%d.\n", _buf_name, _buf_fd);
              break; 
          }

          if (failure == ENOENT)
          {
              ::sleep(1);

              if (max_tries >= 0) 
              { 
                  if ((++attempts) >= ((size_t) max_tries)) 
                  { 
                      DIAGOUT(1, "Could not open file \"%s\" after %ld attempts, giving up.\n", _buf_name, max_tries);
                      _state = pathological;
                      return; 
                  }
              }

              continue;
          }

          _state = pathological;
          DIAGOUT(1, "Shared File ::shm_open(\"%s\") failed, error was \"%s\".\n", _buf_name, ::strerror(failure));
          return;
      }

      size_t buf_len = _rng_start->hdr.cntl.buf_size * _rng_start->hdr.cntl.max_bufs;

      // Map in the file, then fill in the data buffer header info.
      _buf_start = (char*) ZVision::mmap_file(_buf_fd, buf_len, _buf_name);

      if (_buf_start == NULL)
      {
          _state = pathological; 

          if (::close(_buf_fd) != 0) { DIAGOUT(0, "::close(%d) failed.\n", _buf_fd); }

          PATHOLOGICAL("::mmap_file(fd=%d, len=%ld, \"%s\") failed.\n", 
              _buf_fd, buf_len, _buf_name); 
          return;
      }
      
      // Start out as blocked, so that we force the first get_next_element() call to bump
      // to the next buffer.
      _state = blocked;

      gratutious_log();

    } // end of ZVisionFilter().  [Consumer]


//---------------------------------------------------------------------------------------------------------------------------//
//   The following are shared routines between the "consumer" and "producer" threads.                                        //
//---------------------------------------------------------------------------------------------------------------------------//

    /**
     ** Create a fully qualified ring buffer filename based on the JVM supplied name, in the buffer provided
     ** by the caller.  NOTE: remember that we don't need the SHMEM_DIR_STR ("/dev/shm/") prefix, as it is
     ** assumed by the shared memory commands (i.e. shm_open) that the name is passed without the prefix.  
     **
     ** Instead we build a fully qualified name (including the prefix) but return a pointer to the string
     ** following the SHMEM_DIR_STR within that buffer, or NULL if an error was encountered, and change the
     ** _state variable to reflect the problem.
     **
     **/
    inline char* create_filename_str(char* name, size_t len, const char* basename_str)
    {
      if (name == NULL)
      {
          _state = pathological; 
          PATHOLOGICAL("NULL name buffer pointer supplied to create_filename_str() method.\n");
          return NULL;
      }

      if (basename_str == NULL)
      {
          _state = pathological; 
          PATHOLOGICAL("NULL basename pointer supplied to create_filename_str() method.\n");
          return NULL;
      }

      char* basename = (char*) basename_str;

      // Skip past any leading slashes.
      while (*basename == '/') { basename++; }

      size_t dev_shm_dir_len   = ::strlen(SHMEM_DIR_STR); 
      size_t ring_suffix_len   = ::strlen(RING_BUF_SUFFIX); 
      size_t basename_len      = ::strlen(basename); 
      char*  offset            = name;              

      // Make sure we can build the full pathname within the area allotted.
      if ((dev_shm_dir_len + ring_suffix_len + basename_len) >= len)
      {
          _state = pathological; 
          PATHOLOGICAL("Filename size (%lu) lacks room for prefix (\"%s\"), basename \"%s\", and %ld bytes for the suffix.\n", 
            len, SHMEM_DIR_STR, basename, ring_suffix_len);
          return NULL;
      }

      ::memcpy((void*) offset, (const void*) SHMEM_DIR_STR, dev_shm_dir_len);
      offset += dev_shm_dir_len;

      // We return the shm_open name string, not the fully qualified name.
      char* shm_name = offset - 1;

      ::memcpy((void*) offset, (const void*) basename, basename_len);
      offset += basename_len;

      ::memcpy((void*) offset, (const void*) RING_BUF_SUFFIX, ring_suffix_len);
      offset += ring_suffix_len;

      ::sprintf(offset++, "%c", 0);

      DIAGOUT(3, "Fully qualified indexed buffer shared ring filename is \"%s\".\n", name);
      DIAGOUT(2, "shm_open name string is \"%s\".\n", shm_name);

      return shm_name;

    }  // end of create_filename_str().


    /**
     ** Return the state of the object.
     **/
    FILTER_STATE_E  state() { return _state; }


    /**
     ** Return this object's notion of the start of the ring buffer file's memory mapped address.
     **/
    ZVision::RING_BUF_T*  ring() { return _rng_start; }


    /**
     ** Return the File Descriptor of the Ring Buffer header file.
     **/
    ZVision::SOCKET_FD_T  rng_fd()   { return _rng_fd; }


    /**
     ** Return the fully qualified ZVisionFilter Ring Buffer filename in use.
     **/
    char* rng_filename()   { return _rng_filename; }


    /**
     ** Return the fully qualified data buffer filename in use.
     **/
    char* buf_filename()   { return _buf_name; }


    /**
     ** Return the data buffer provided by the consumer where the current element will
     ** be copied to.
     **/
    char* elem_buf()   { return _elem_buf; }


    /**
     ** Return the Event Tick Buffer File Descriptor.
     **/
    ZVision::SOCKET_FD_T  buf_fd()   { return _buf_fd; }


    /**
     ** Return this object's notion of the Shared Event Tick Buffer file memory mapped address.
     **/
    void*                 buf_start()  { return _buf_start; }


    /**
     ** Return this object's notion of the current writer_idx field..
     **/
    size_t                writer_idx()  { return _rng_start->hdr.cntl.writer_idx; }

    
    /**
     ** Return this object's notion of the current reader_idx field..
     **/
    size_t                reader_idx()  { return _reader_idx; }

    
    /** 
     ** Gratutious log output.
     **/
    void gratutious_log()
    {
        DIAGOUT(2, "=================================================\n");
        DIAGOUT(2, " Shared Ring Buffer Filename: %s.\n",  _rng_filename);
        DIAGOUT(2, "                  Filter Key: %lu.\n", _filter_key);
        DIAGOUT(2, "              Shared Ring Fd: %d.\n",  _rng_fd);
        DIAGOUT(2, "                  Ring Start: %p.\n",  _rng_start);
        DIAGOUT(2, "        Data Buffer Filename: %s.\n",  _buf_name);
        DIAGOUT(2, "                 Data Buf Fd: %d.\n",  _buf_fd);
        DIAGOUT(2, "   Data Buffer Start Address: %p.\n",  _buf_start);
        DIAGOUT(2, "Caller's Element Buf Address: %p.\n",  _elem_buf);
        DIAGOUT(2, "                 Buffer Size: %ld.\n", _buf_size);
        DIAGOUT(2, "                 Max Buffers: %ld.\n", _max_bufs);
        DIAGOUT(2, "                Element Size: %ld.\n", _elem_size);
        DIAGOUT(2, "                Max Elements: %ld.\n", _max_elems);
        DIAGOUT(2, "                 Header Size: %ld.\n", _hdr_size);
        DIAGOUT(2, " Allocation Timestamp Offset: %ld.\n", _alloc_offset);
        DIAGOUT(2, "                  Reader Idx: %ld.\n", _reader_idx);
        DIAGOUT(2, "             Last Writer Idx: %ld.\n", _last_writer_idx);
        DIAGOUT(2, "      Published Buffer Count: %lu.\n", _last_buf_cnt);
    
    } // end of gratutious_log().


//---------------------------------------------------------------------------------------------------------------------------//
//   The following are routines called only by the "producer" threads.                                                       //
//---------------------------------------------------------------------------------------------------------------------------//

    /**
     ** Add another Data Buffer entry to the ring buffer.  Add the buffer to the next entry within the ring buffer.
     ** Writes never block and if the write index has caught up with itself then we simply overwrite what is there.
     ** NOTE: THis most likely is called from within an ISR in the JVM, do not use any blocking (i.e. printf) calls.
     **
     ** The timestamp and alloc_offset fields are used to detect when the writer task has overwritten (i.e. "stomped on") a
     ** buffer being processed by the reader task.  The timestamp should be a (very unlikely to wrap) value of when the
     ** buffer was taken off the recycle list and put into use.  The alloc_offset field is the offset, in bytes, to the
     ** timestamp value within the data buffer.  This value should always be the same for every entry within the data
     ** buffer, so we can store it in the ring buffer header, but the timestamp should always be unique to each data buffer.
     **/
    void put_buffer(size_t buf_index, uint64_t timestamp)
    {
        if (_rng_start == NULL)
        { 
            _state = pathological; 
            return; 
        }

        // Gratuitous sanity checking.
        if (_rng_start->hdr.cntl.cookie != ZVision::COOKIE_RING_BUF)
        {
            _state = pathological; 
            return;
        }

        ZV_ASSERT(buf_index      < _rng_start->hdr.cntl.max_bufs);
        ZV_ASSERT(_buf_size     == _rng_start->hdr.cntl.buf_size);
        ZV_ASSERT(_max_bufs     == _rng_start->hdr.cntl.max_bufs);
        ZV_ASSERT(_elem_size    == _rng_start->hdr.cntl.elem_size);
        ZV_ASSERT(_max_elems    == _rng_start->hdr.cntl.max_elems);
        ZV_ASSERT(_hdr_size     == _rng_start->hdr.cntl.hdr_size);
        ZV_ASSERT(_alloc_offset == _rng_start->hdr.cntl.alloc_offset);
        ZV_ASSERT(_filter_key   == _rng_start->hdr.cntl.filter_key);

        size_t offset    = _buf_size * buf_index;
        char*  buf_addr  = (char*) (_buf_start + offset);

        // Make sure the first two words of the buffer header are the data buffer cookie, and the element count.
        volatile uint64_t* temp       = (uint64_t*) buf_addr;
        volatile uint64_t  cookie     = (uint64_t) *temp++;
        volatile size_t    elem_cnt   = (size_t)   *temp;
        volatile size_t    writer_idx = _rng_start->hdr.cntl.writer_idx;

        if (cookie != EVENT_BUFFER_DATA_COOKIE)
        {
            _state = pathological; 
            return; 
        }

        // If there are no elements in this entry then discard it.
        if (elem_cnt == 0)
        {
            return;
        }

        ZV_ASSERT(elem_cnt < _max_elems);

        // If there are too many then complain & discard it.
        if (elem_cnt > _max_elems) 
        { 
            return; 
        }


// TODO: NEED ARBITRATION LOGIC HERE FOR RING BUFFER SO WE AVOID COLLISIONS+++++++++


        // Point at the next entry within the ring buffer.
        volatile ZVision::RING_BUF_ENTRY_T* rbe = &_rng_start->entry + writer_idx;

        // Establish the reader's initial (element) pointer context for this new ring buffer entry.  
        // NOTE: We are implicitly assuming that the header will always be less than one element in size.
        // TODO: Add checks for this during initialization.
        rbe->buf_idx      =  buf_index;
        rbe->elem_cnt     =  elem_cnt;
        rbe->curr_elem    = (char*)  (buf_addr + _hdr_size);
        rbe->last_elem    = (char*) ((buf_addr + _hdr_size) + (_elem_size * (elem_cnt)));
        rbe->timestamp    = timestamp;
        rbe->cookie       = ZVision::COOKIE_RING_BUF_ENTRY;
              
        // Bump the ring buffer index, then wrap if we've hit the end of the buffer space.
        ++writer_idx; if (writer_idx >= _max_bufs) { writer_idx = 0; }

        // Once we bump the pointer, the reader can access the ring buffer entry.
        _rng_start->hdr.cntl.buf_cnt++;
        _rng_start->hdr.cntl.writer_idx = writer_idx;

        return;

    } // end of put_buffer().


//---------------------------------------------------------------------------------------------------------------------------//
//   The following are routines called only by the "consumer" threads.                                                       //
//---------------------------------------------------------------------------------------------------------------------------//

    /**
     ** Return either the next available entry, or NULL if blocked or pathological condition has occurred.  
     **/
    inline bool get_next_element(int64_t timeout)
    {
        if (_rng_start == NULL) 
        { 
            _state = pathological; 
            PATHOLOGICAL("Ring obj %p with NULL ring pointer.\n", this);
            return false;
        }

        while (true)
        {
            if (_state == blocked || _next_element > _last_element) 
            { 
                volatile size_t   cur_writer_idx = _rng_start->hdr.cntl.writer_idx;
                volatile uint64_t cur_buf_cnt    = _rng_start->hdr.cntl.buf_cnt;
                int64_t           sleep_us_cnt   = 0;

                // If nothing has changed since the last time we looked, we're still blocked.
                while (_state == blocked && cur_writer_idx == _last_writer_idx && cur_buf_cnt == _last_buf_cnt) 
                {
                    if (timeout >= 0 && sleep_us_cnt++ >= timeout) 
                    { 
                        _state = blocked; 
                        DIAGOUT(4, "Ring object %p timed out after %ldus (writer %lu, count %lu) state is \"%s\".\n",
                            this, sleep_us_cnt, cur_writer_idx, cur_buf_cnt, xlate_state(blocked));
                        return false;
                    }

                    ::usleep(1);
                    cur_buf_cnt    = _rng_start->hdr.cntl.buf_cnt;
                    cur_writer_idx = _rng_start->hdr.cntl.writer_idx;
                }

                uint64_t next_reader_idx = _reader_idx + 1;

                if (next_reader_idx >= _max_bufs) { next_reader_idx = 0; }

                while (next_reader_idx == _rng_start->hdr.cntl.writer_idx)
                {
                    if (timeout >= 0 && sleep_us_cnt++ >= timeout) 
                    { 
                        _state = blocked; 
                        DIAGOUT(4, "Ring object %p timed out after %ldus (writer %lu, count %lu, reader %lu) state is \"%s\".\n",
                            this, sleep_us_cnt, cur_writer_idx, cur_buf_cnt, _reader_idx, xlate_state(blocked));
                        return false;
                    }

                    ::usleep(1);
                }

                _reader_idx = next_reader_idx;
                _state      = active; 

                DIAGOUT(3, "Advancing ring object %p reader index to %lu, state was \"%s\", now \"%s\".\n", 
                    this, _reader_idx, xlate_state(blocked), xlate_state(active));

                volatile ZVision::RING_BUF_ENTRY_T* entry = (ZVision::RING_BUF_ENTRY_T*) &_rng_start->entry + _reader_idx;

                // Record the time created (provided by the Producer).  This is used by the reader ("Consumer") tasks to
                // determine if the data has been stomped on before we had a chance to copy it into the caller's buffer.
                volatile uint64_t timestamp     = entry->timestamp;
                volatile size_t   buffer_idx    = entry->buf_idx;
                volatile size_t   element_count = entry->elem_cnt;

                // Set pointers to the header structure, as well as the first and last buffer element.  If we get 
                // stomped here, we'll notice it after the copy to the buffer occurs.
                _buf_hdr      = _buf_start + (buffer_idx * _buf_size);
                _next_element = _buf_hdr + _hdr_size;
                _last_element = _next_element + (_elem_size * element_count);
                _timestamp    = timestamp;
                _buffer_idx   = buffer_idx;

                DIAGOUT(4, "Setting next element to %p, last last element to %p, timestamp 0x%lX\n", 
                    _next_element, _last_element, timestamp);
            }

            if (_next_element > _last_element)
            {
                PATHOLOGICAL("_next_element %p is greater than _last_element %p.\n", _next_element, _last_element);
                _state = pathological;
                return false;
            }

            if (_elem_buf == NULL)
            {
                PATHOLOGICAL("_elem_buf is NULL.\n");
                _state = pathological;
                return false;
            }

            // If the data buffer's "modified" timestamp (in the data buffer's header struct) matches our
            // notion of the fields value when we were first told about it, then the element we've copied
            // into the caller-supplied buffer is still intact (i.e. not stomped on by the writer task) and we can
            // return "true", which informs the caller that that the data is valid and can be further qualified.
            uint64_t* data_buffer_allocated_timestamp = (uint64_t*) ((char*) (_buf_hdr + _alloc_offset));
           
            // Copy the data into the buffer provided by the caller.
            ::memcpy((void*) _elem_buf, (const void*) _next_element, _elem_size);

            if (_timestamp == *data_buffer_allocated_timestamp) { break; }

            // If the buffer was stomped on, then throw away all the remaining elements in this buffer
            // and force the logic to move to the next available ring buffer entry.
            DIAGOUT(3, "STOMP: element %p (last=%p) ring %p reader idx %lu, buf idx %lu, expected timestamp 0x%lX, got 0x%lX.\n", 
                _next_element, _last_element, this, _reader_idx, _buffer_idx, _timestamp, *data_buffer_allocated_timestamp);

            _next_element = _last_element + _elem_size;
            _state        = blocked;

        } // end of while (true)
        
        _elem_tot++;

        DIAGOUT(2, "returning element (%p) from ring object %p, reader index %lu, total element count %lu\n", 
            _next_element, this, _reader_idx, _elem_tot);

        _next_element += _elem_size;

        return true;

    } // end of get_next_element().


    /**
     ** Skip to the Nth Buffer Element, prior to any other filters being applied.
     **/
    void filterEveryNthElement(size_t nth_element)
    {
      _nth_element = nth_element;
      return;
    }

    /**
     ** Return every Nth entry that matches, after all the other filter criteria have been applied.
     **/
    void filterEveryNthMatch(size_t nth_match)
    {
      _nth_match = nth_match;
      return;
    }

//-------------------------------------------------------------------------------------------------------//
//                               static methods only                                                     //
//-------------------------------------------------------------------------------------------------------//

    static const char* xlate_state(FILTER_STATE_E state)
    {
        switch(state)
        {
        case uninitialized:
            return "uninitialized";
            break;
        case active:
            return "active";
            break;
        case blocked:
            return "blocked";
            break;
        case initializing:
            return "initializing";
            break;
        case pathological:
            return "pathological";
            break;
        case exited:
            return "exited";
            break;
        default:
            break;
        }

        return "Error";
    }

    static const char* xlate_type(uint64_t type)
    {
        switch(type)
        {
        case 0:
            return "Unknown";
            break;
        case 1:
            return "Timer";
            break;
        case 2:
            return "TTSP";
            break;
        case 3:
            return "TISP";
            break;
        case 4:
            return "Counter";
            break;
        case 5:
            return "Meta";
            break;
        case 6:
            return "User";
            break;
        default:
            break;
        }

        return "Error";
    }

    static const char* xlate_cat(uint64_t cat)
    {
        switch(cat)
        {
        case 0:
            return "Uninitialized";
            break;
        case 1:
            return "Application";
            break;
        case 2:
            return "Compiler";
            break;
        case 3:
            return "GC";
            break;
        case 4:
            return "I/O";
            break;
        case 5:
            return "Unused";
            break;
        case 6:
            return "Unused";
            break;
        case 7:
            return "None";
            break;
        default:
            break;
        }

        return "Error";
    }

};  // end of class ZVisionFilter


/**
 ** class ZVisionFilterTicks - 
 **/
class ZVisionFilterTicks : public ZVisionFilter
{
  friend class ZVisionHostAgent;

  public:

    // anonymous size info.
    enum
    {
      max_tags           = 10,         // Max tag ids we'll allow for a filter.
      max_tids           = 10,         // Max task ids we'll allow for a filter.
      max_cpus           = 10,         // Max CPU ids we'll allow for a filter.
      tick_element_size  = 128,        // Size in bytes of one tick profiling entry (i.e. 2 cache lines).
      tick_header_size   = 128,        // Size in bytes of the header (i.e. 2 cache lines, must match element size).
      tick_alloc_offset  = (8 * 14),   // Offset, in bytes, of the allocation timestamp wthin the data buffer.
    };

    // This maps to the tickProfiler.hpp tick type enum, but do we want to expose all tick types?
    enum TICK_FILTER_TYPE_E
    {
      filter_no_ticks      = 0x0,
      filter_timer_ticks   = 0x1,
      filter_ttsp_ticks    = 0x2,
      filter_tisp_ticks    = 0x4,
      filter_counter_ticks = 0x8,
      filter_meta_ticks    = 0x10,
      filter_user_ticks    = 0x20,
      filter_all_ticks     = 0x3F,
      filter_tick_error    = ~0,
    };

    enum TICK_FILTER_CATEGORY_E
    {
      filter_no_cats        = 0x0,
      filter_app_cats       = 0x1,
      filter_comp_cats      = 0x2,
      filter_gc_cats        = 0x4,
      filter_io_cats        = 0x8,
      filter_unassigned1    = 0x10,
      filter_unassigned2    = 0x20,
      filter_all_cats       = 0xF,
      filter_category_error = ~0,
    };

    typedef uint64_t TICK_PC_T;
    typedef uint64_t TICK_TID_T;
    typedef uint64_t TICK_TAG_T;
    typedef uint64_t TICK_CPU_T;
    typedef uint64_t TICK_CAT_T;
    typedef uint64_t TICK_TIMESTAMP_T;
 
    // This maps to the tickProfiler.hpp tick entry structure.
    typedef struct tick_profiling_entry_t
    {
        uint64_t entry;
        uint64_t u1;
        uint64_t u2;
        uint64_t u3;
        uint64_t rpcs[12];
        
    } TICK_PROFILING_ENTRY_T;

    const static TICK_TIMESTAMP_T first_time = (TICK_TIMESTAMP_T) 0;
    const static TICK_TIMESTAMP_T last_time  = (TICK_TIMESTAMP_T) ~0;

    const static TICK_PC_T first_pc = (TICK_PC_T) 0;
    const static TICK_PC_T last_pc  = (TICK_PC_T) ~0;

  private:

    enum
    {
      max_bufs           = 64 * 1024,  // Limit of how many tick buffers we'll allow.
      tick_preface_bytes = 32,         // First four words of a tick entry. (i.e. _entry + _u1 + _u2 + _u3) [see tickProfiler.hpp]
    };

    /**
     ** This is a hard-coded copy of the tickProfiler.hpp bit positions of the entry field.  It must
     ** match that defined in the Zing codebase.  TODO: Make these be shared between platforms.
     **/
    enum 
    {
      type_shift = 0,
      type_bits  = 3,
      type_mask  = 0x7,

      cpu_id_shift = type_shift + type_bits,
      cpu_id_bits  = 10,
      cpu_id_mask  = 0x3FF,

      category_shift = cpu_id_shift + cpu_id_bits,
      category_bits  = 3,
      category_mask  = 0x7,

      thread_id_shift = category_shift + category_bits,
      thread_id_bits  = 48,
      thread_id_mask  = 0xFFFFFFFFFFFF,
    };

    /**
     ** This is a hard-coded copy of the tickProfiler.hpp "type" subfield.  It must
     ** match that defined in the Zing codebase.  TODO: Make these be shared between platforms.
     **/
    enum TICK_TYPES_E
    {
      is_unknown_tick     = 0,     // Profile Entry is unitialized.
      is_timer_tick       = 1,     // Profiler ticks generated via timer (1/ms) which generates a signal.
      is_ttsp_tick        = 2,     // Identical to timer_ticks, used to measure time to safepoint of each thread.
      is_tisp_tick        = 3,     // Identical to timer_ticks, used to measure time to safepoint of each thread.
      is_counter_tick     = 4,     // As yet unimplemented, generated when a counter or limit is reached within OS.
      is_meta_tick        = 5,     // Metatick entry, not generated by timer event.
      is_user_tick        = 6,     // Usertick entry, not generated by timer event.
      is_ignore_tick_type = 7      // Used only for the ProfileFilter, NEVER set an entry to this value!
    };

    // This maps to the tickProfiler.hpp "category" field enum within the "entry" field.
    enum TICK_CATS_E
    {
      is_uninit_cat       = 0,
      is_app_cat          = 1,
      is_comp_cat         = 2,
      is_gc_cat           = 3,
      is_io_cat           = 4,
      is_unassigned1_cat  = 5,
      is_unassigned2_cat  = 6,
      is_no_cat           = 7,
    };


    // Object contents.
    TICK_FILTER_TYPE_E     _type;           // Mask of tick types to filter on.
    TICK_FILTER_CATEGORY_E _cat;            // Mask of tick functions (categories) to filter on.
    TICK_CPU_T             _cpu[max_cpus];  // A array of CPU IDs we will filter on.
    size_t                 _cpu_cnt;        // The count of active entries within the _cpu array.
    TICK_TAG_T             _tag[max_tags];  // A array of TIDs we will filter on.
    size_t                 _tag_cnt;        // The count of active entries within the _tag array.
    TICK_TID_T             _tid[max_tids];  // A array of TIDs we will filter on.
    size_t                 _tid_cnt;        // The count of active entries within the _tid array.
    TICK_TIMESTAMP_T       _begin_time;     // Earliest tick we'll allow.
    TICK_TIMESTAMP_T       _finish_time;       // Latest tick we'll allow.
    TICK_PC_T              _start_pc;       // Beginning of PC range we'll allow.
    TICK_PC_T              _end_pc;         // End of PC range we'll allow.


    size_t                 _return_len;     // How many bytes of each element we will return from a get_next_element() method.


  public:

    /**
     ** Returns the next valid (by whatever filters are currently in place) tick entry in the buffer originally provided 
     ** by the caller (during the constructor method).  If the timeout value is less than zero, it will spin indefinitely
     ** waiting for the writer to supply a new, valid tick_entry, otherwise it will pend for up to "timeout" microseconds 
     ** while waiting for the next available tick entry.
     **
     ** Returns true if a new tick was copied into the buffer, or false and state "blocked" if there was no tick available
     ** within the timeout given, or false and state "pathological" if an error has occurred. 
     **/
    inline bool get_next_filtered_tick(int64_t timeout, TICK_PROFILING_ENTRY_T* tick_entry)
    {
        static size_t nth_element = 0;
        static size_t nth_match   = 0;

        while (get_next_element(timeout) == true)
        {
            if (nth_element++ < _nth_element)
            { 
                DIAGOUT(3, "Skipping tick_entry %ld (%p) from ring object %p, reader index %lu, total entry count %lu.\n",
                    nth_element, tick_entry, this, _reader_idx, _elem_tot);
                continue; 
            }

            nth_element = 0;

            TICK_PC_T        pc        = 0;
            TICK_TAG_T       tag       = 0;
            TICK_TIMESTAMP_T timestamp = 0;

            // Assign the type, CPU ID, Thread ID, and function category from the subfields
            // within the type-stable entry field.
            TICK_TYPES_E type = (TICK_TYPES_E) ((tick_entry->entry >> type_shift)      & type_mask);
            TICK_CPU_T   cpu  = (TICK_CPU_T)   ((tick_entry->entry >> cpu_id_shift)    & cpu_id_mask);
            TICK_TID_T   tid  = (TICK_TID_T)   ((tick_entry->entry >> thread_id_shift) & thread_id_mask);
            TICK_CAT_T   cat  = (TICK_CAT_T)   ((tick_entry->entry >> category_shift)  & category_mask);

            // Assign the appropriate U1, U2, & U3 fields based on the tick type.
            switch (type)
            {
            case is_timer_tick:
              pc        = (TICK_PC_T)        tick_entry->u1;
              tag       = (TICK_TAG_T)       tick_entry->u2;
              timestamp = (TICK_TIMESTAMP_T) tick_entry->u3;
              break;

            case is_ttsp_tick:
              pc        = (TICK_PC_T)        tick_entry->u1;
              tag       = (TICK_TAG_T)       tick_entry->u2;
              timestamp = (TICK_TIMESTAMP_T) tick_entry->u3;
              break;

            case is_tisp_tick:
              pc        = (TICK_PC_T)        tick_entry->u1;
              tag       = (TICK_TAG_T)       tick_entry->u2;
              timestamp = (TICK_TIMESTAMP_T) tick_entry->u3;
              break;

            case is_counter_tick:
              pc        = 0;
              tag       = 0;
              timestamp = (TICK_TIMESTAMP_T) tick_entry->u3;
              break;

            case is_meta_tick:
              pc        = 0;
              tag       = (TICK_TAG_T)       tick_entry->u2;
              timestamp = (TICK_TIMESTAMP_T) tick_entry->u3;
              break;

            case is_user_tick:
              pc        = 0;
              tag       = (TICK_TAG_T)       tick_entry->u2;
              timestamp = (TICK_TIMESTAMP_T) tick_entry->u3;
              break;

            default:
              break;
            }

            // Apply filters...

            // Is it one of the tick types we've selected.
            if (
                ( _type == filter_all_ticks                                      ) ||
                ( ((_type & filter_timer_ticks   ) && (type == is_timer_tick)   )) ||
                ( ((_type & filter_ttsp_ticks    ) && (type == is_ttsp_tick)    )) ||
                ( ((_type & filter_tisp_ticks    ) && (type == is_tisp_tick)    )) ||
                ( ((_type & filter_counter_ticks ) && (type == is_counter_tick) )) ||
                ( ((_type & filter_meta_ticks    ) && (type == is_meta_tick)    )) ||
                ( ((_type & filter_user_ticks    ) && (type == is_user_tick)    ))
               )
            {
                DIAGOUT(4, "Matched Tick Type.\n");
            }
            else 
            { 
                DIAGOUT(4, "Failed to match Tick Type.\n");
                continue; 
            }


            // Is it one of the function categories we've selected.
            if (
                ( _cat == filter_all_cats                              ) ||
                ( ((_cat & filter_app_cats  ) && (cat == is_app_cat)  )) ||
                ( ((_cat & filter_gc_cats   ) && (cat == is_gc_cat)   )) ||
                ( ((_cat & filter_comp_cats ) && (cat == is_comp_cat) ))
               )
            {
                DIAGOUT(4, "Matched Category.\n");
            }
            else
            { 
                DIAGOUT(4, "Failed to match Category.\n");
                continue; 
            }

            // Is it after the start PC?
            if (_start_pc > first_pc && pc < _start_pc) 
            { 
                DIAGOUT(4, "Failed to match Start PC.\n");
                continue; 
            }
            else
            {
                DIAGOUT(4, "Matched Start PC.\n");
            }

            // Is it before the end PC?
            if (_end_pc < last_pc && pc > _end_pc)
            { 
                DIAGOUT(4, "Failed to match End PC.\n");
                continue; 
            }
            else
            {
                DIAGOUT(4, "Matched End PC.\n");
            }

            // Is it after the start time?
            if (_begin_time > first_time && timestamp < _begin_time)
            { 
                DIAGOUT(4, "Failed to match Begin Time.\n");
                continue; 
            }
            else
            {
                DIAGOUT(4, "Matched Begin Time.\n");
            }

            // Is it before the end time?
            if (_finish_time < last_time && timestamp > _finish_time)
            { 
                DIAGOUT(4, "Failed to match End Time.\n");
                continue; 
            }
            else
            {
                DIAGOUT(4, "Matched Finish Time.\n");
            }

            // Is it within the list of tags we've selected.
            if (_tag_cnt > 0)
            {
                size_t index;
                for (index = 0; index < _tag_cnt; index++) { if (_tag[index] == tag) { break; } }
                if (index >= _tag_cnt)
                { 
                    DIAGOUT(4, "Failed to match Tag.\n");
                    continue; 
                }
                else
                {
                    DIAGOUT(4, "Matched Tag.\n");
                }
            }
               
            // Is it within the list of CPU Ids we've selected.
            if (_cpu_cnt > 0)
            {
                size_t index;
                for (index = 0; index < _cpu_cnt; index++) { if (_cpu[index] == cpu) { break; } }
                if (index >= _cpu_cnt)
                { 
                    DIAGOUT(4, "Failed to match CPU.\n");
                    continue; 
                }
                else
                {
                    DIAGOUT(4, "Matched CPU.\n");
                }
            }
               
            // Is it within the list of tids we've selected.
            if (_tid_cnt > 0)
            {
                size_t index;
                for (index = 0; index < _tid_cnt; index++) { if (_tid[index] == tid) { break; } } 
                if (index >= _tid_cnt)
                { 
                    DIAGOUT(4, "Failed to match CPU.\n");
                    continue; 
                }
                else
                {
                    DIAGOUT(4, "Matched TID.\n");
                }
            }
               
            // Skip every N matches...
            if (nth_match++ < _nth_match) 
            { 
                DIAGOUT(3, "Skipping match %ld (%p) from ring object %p, reader index %lu, total entry count %lu.\n",
                    nth_match, tick_entry, this, _reader_idx, _elem_tot);
                continue; 
            }
            
            nth_match = 0;

            DIAGOUT(2, "Returning tick_entry (%p) from ring object %p, reader index %lu, total entry count %lu.\n",
                tick_entry, this, _reader_idx, _elem_tot);

            return true; 

        }  // end of while (true).

        if (_state != blocked) 
        {  
            PATHOLOGICAL("Illegal state \"%s\" at end of while loop.\n", ZVisionFilterTicks::xlate_state(_state));
            _state = pathological; 
        }

        return false;


    } // end of get_next_filtered_tick().


    /**
     ** Inherit the base class constructors for the "Producer" task.
     **/
    ZVisionFilterTicks(ZVision::FILE_DATA_T*  file_data, 
                       ZVision::FILTER_KEY_T  filter_key, 
                       size_t                 elem_size, 
                       size_t                 hdr_size,
                       size_t                 alloc_offset) 

       : ZVisionFilter(file_data, filter_key, elem_size, hdr_size, alloc_offset)
    {
        // Nothing to set for the producer task.
    }

    /**
     ** Inherit the base class constructors for the "Consumer" task.
     **/
    ZVisionFilterTicks(int64_t                   max_tries, 
                       const char*               basename_str, 
                       ZVision::FILTER_KEY_T     filter_key, 
                       TICK_PROFILING_ENTRY_T*   elem_buf)

        : ZVisionFilter(max_tries, basename_str, filter_key, (char*) elem_buf)
    {
        _type        = filter_all_ticks;
        _cat         = filter_all_cats;
        _tag_cnt     = 0;
        _cpu_cnt     = 0;
        _tid_cnt     = 0;
        _start_pc    = 0;
        _end_pc      = ~0;
        _begin_time  = 0;
        _finish_time = ~0;
    }


    /**
     ** Include [or not] the 12 words of call stack information when returning the next tick.
     ** Sets current filter parameter to only return the first 4 words of a tick.  For block 
     ** transfer requests, this feature will be ignored.
     **/
    void filterIncludingStackTrace(bool include_stack_trace)
    {

      if (include_stack_trace) 
      { 
        _return_len = tick_element_size; 
      }
      else
      { 
        _return_len = tick_preface_bytes;
      }

      return;
    }

    /**
     ** Only return ticks that are within the types described by the currently OR'ed set.
     **/
    bool filterByTickType(TICK_FILTER_TYPE_E tick_type)
    {
      if ((tick_type & filter_all_ticks) == 0)
      {
          DIAGOUT(0, "Illegal type filter requested (0x%X), possible bit settings are 0x%X.\n", tick_type, filter_all_ticks);
          return false;
      }

      _type = (TICK_FILTER_TYPE_E) (((unsigned int) _type) | ((unsigned int) tick_type));
      return true;
    }

    /**
     ** Reset filter type to none.
     **/
    void clearFilterTickType() { _type = filter_no_ticks; }

    /**
     ** Only return ticks that are within the function categories described by the currently OR'ed set.
     **/
    bool filterByTickCategory(TICK_FILTER_CATEGORY_E cat)
    {
      if ((cat & filter_all_cats) == 0)
      {
          DIAGOUT(0, "Illegal category filter requested (0x%X), possible bit settings are 0x%X.\n", cat, filter_all_cats);
          return false;
      }

      _cat = (TICK_FILTER_CATEGORY_E) (((unsigned int) _cat) | ((unsigned int) cat));
      return true;
    }

    /**
     ** Reset filter cat to none.
     **/
    void clearFilterCatType() { _cat = filter_no_cats; }

    /**
     ** Only return ticks that occurred after the time given.
     **/
    void filterByTickBeginTime(TICK_TIMESTAMP_T begin_time)
    {
      DIAGOUT(1, "Setting Start Time to 0x%lX.\n", begin_time);
      _begin_time = begin_time;
      return;
    }

    /**
     ** Only return ticks that occurred before the time given.
     **/
    void filterByTickFinishTime(TICK_TIMESTAMP_T finish_time)
    {
      DIAGOUT(1, "Setting End Time to 0x%lX.\n", finish_time);
      _finish_time = finish_time;
      return;
    }


    /**
     ** Only return ticks whose CPU IDs are in the TICK_CPU_T array.
     **/
    bool filterByTickCPU(TICK_CPU_T cpu)
    {
      if (_cpu_cnt >= max_cpus)
      {
          DIAGOUT(0, "Exceeded thread Id list limit (%d).\n", max_cpus);
          return false;
      }

      _cpu[_cpu_cnt++] = cpu;

      return true;
    }

    void resetTickCPU()
    {
        _cpu_cnt = 0;
        for (size_t index = 0; index < max_cpus; index++) { _cpu[index] = 0; }
    }


    /**
     ** Only return ticks whose Tag IDs are in the TICK_TAG_T array.
     **/
    bool filterByTickTag(TICK_TAG_T tag)
    {
      if (_tag_cnt >= max_tags)
      {
          DIAGOUT(0, "Exceeded thread Id list limit (%d).\n", max_tags);
          return false;
      }

      _tag[_tag_cnt++] = tag;

      return true;
    }

    void resetTickTag()
    {
        _tag_cnt = 0;
        for (size_t index = 0; index < max_tags; index++) { _tag[index] = 0; }
    }


    /**
     ** Only return ticks whose thread IDs are in the TICK_TID_T array.
     **/
    bool filterByTickTid(TICK_TID_T tid)
    {
      if (_tid_cnt >= max_tids)
      {
          DIAGOUT(0, "Exceeded thread Id list limit (%d).\n", max_tids);
          return false;
      }

      _tid[_tid_cnt++] = tid;

      return true;
    }

    void resetTickTid()
    {
        _tid_cnt = 0;
        for (size_t index = 0; index < max_tids; index++) { _tid[index] = 0; }
    }

    /**
     ** Only return ticks whose program counters are greater than or equal to start.
     **/
    void filterByTickStartPC(TICK_PC_T start_pc)
    {
      DIAGOUT(1, "Setting Start PC range to 0x%lX.\n", start_pc);
      _start_pc = start_pc;
      return;
    }

    /**
     ** Only return ticks whose program counters are less than or equal to end.
     **/
    void filterByTickEndPC(TICK_PC_T end_pc)
    {
      DIAGOUT(1, "Setting End PC range to 0x%lX.\n", end_pc);
      _end_pc = end_pc;
      return;
    }


};  // end of class ZVisionFilterTicks


/**
 ** class ZVisionFilterUserTicks - 
 **/
class ZVisionFilterUserTicks : public ZVisionFilterTicks
{
  friend class ZVisionHostAgent;

  public:

    ZVisionFilterUserTicks(ZVision::FILE_DATA_T*  file_data, 
                           ZVision::FILTER_KEY_T  filter_key, 
                           size_t                 elem_size, 
                           size_t                 hdr_size,
                           size_t                alloc_offset) 

       : ZVisionFilterTicks(file_data, filter_key, elem_size, hdr_size, alloc_offset)
    {
    }

    ZVisionFilterUserTicks(int64_t                  max_tries, 
                           const char*              basename_str, 
                           ZVision::FILTER_KEY_T    filter_key, 
                           TICK_PROFILING_ENTRY_T*  elem_buf) 

       : ZVisionFilterTicks(max_tries, basename_str, filter_key, elem_buf)
    {
    }


};  // end of class ZVisionFilterUserTicks


#endif // ZVISION_HPP

// end of file zvision.hpp.

